/*******************************************************************************
 * File:            shr_tick2port_top.e
 *
 * Written by:      Mike Pedneau  512-750-8062  pedneau@hotmail.com
 *  
 * Distribution:    For use as is. Freely distribute and modify. I will try
 *                  to enhance it's accuracy over time.
 *
 * Description:     Parse legal ecode and write out approximate port-compliant 
 *                  code. This is intended to accelerate the adoption of the
 *                  new specman port feature available in 4.2 and greater
 * 
 * Usage:           
 *     specman -p "load <user_code>; load shr_tick2port/e/shr_tick2port_top" -c tick2port
 *     NOTE: This program creates a file called "DEBUG_CHANGES" which shows all
 *           lines of e code that were changed.
 *
 * Limitations:     1) Some cases are not handled properly. For example if 
 *                     a user assigns to a computed name inside a method:
 *                       drive_signal(signal: string,  value: bit) is { ... };
 *                     Anything inside this method cannot be converted into a 
 *                     port access
 *                  2) Uncommon coding such as "unit <unit_name> on seperate
 *                     lines. Also things like "keep <field> == <string> on
 *                     seperate lines is also problematic in my simple parser
 *                  3) If a method receives a pointer to a unit which has a 
 *                     computed name, and that pointer is used inside the method
 *                     to reference the string, then these cases are not handled
 *                     properly.
 * 
 * Enhancements:    1) Automatic insert of "keep bind(...)" lines
 *                  2) Cases of:  "event is true('(signal)') @clk"
 *                  3) Cases of:  type cast on hdl read/write
 *                  4) Cases of:  "event is rise('(signal)'[0]) @clk"
 *                  5) Cases of:  "'(signal)@z' = 0xffff"
 *                  6) Cases of:  parent of port is struct
 *                  7) Cases of:  sig_node: string;  keep soft sig_node == "node" on same line
 *
 * 
 * Revision history:     
 *    0.1  Aug-11-2003   mikep  Initial release
 *    0.2  Aug-12-2003   mikep  Fixed some bug with various coding style
 *    0.3  Aug-13-2003   mikep  Invoke with macro. Fixed complicated vector 
 *                              computed name:  "sig_cmd$[(7+8*i):(8*i)]"
 *    0.3  Aug-13-2003   mikep  Added shr_debug flag
 *    0.4  Aug-14-2003   mikep  Fixed cases of sequences and sequence_drivers
 *    0.5  Aug-17-2003   mikep  Fixed cases of 
 *    0.6  Aug-20-2003   mikes  Changed to eRM package: shr_tick2port
 *    0.7  Aug-25-2003   mikep  Support sequence drivers, private, package types
 *    1.0  Aug-26-2003   mikep  First release
 *    1.1  Sep-30-2003   mikep  Fixes for removed comments and compiled support
 *    1.2  Nov-02-2003   mikep  Fixes specman versions with patches and support
 *                              for code with existing port code
 ******************************************************************************/
<'
package shr_tick2port;
//------------------------------------------------------------------------------
// Top level controls
//------------------------------------------------------------------------------
define <shr_tick2port_cmd'command> "tick2port" as {
    sys.convert_to_ports();
};
extend sys {
    !shr_outdir:              string;

    !shr_debug:               bool;
    !shr_debug_changelines:   list of shr_tick2port_loaded_code_s;

    !shr_specman_path:        list of string;
    !shr_loaded_modules:      list of shr_tick2port_loaded_module_s;
    !shr_loaded_topology:     list of shr_tick2port_unit_info_s;

    convert_to_ports() is {
        shr_debug  = TRUE;
        shr_outdir = "tick2port_out";
        shr_setup_check();
        shr_get_loaded_modules();
        shr_parse_code();
        shr_scan_topology();
        shr_tag_computed();
        shr_tag_port_fields();
        shr_tag_port_lines();
        shr_tag_port_hdlpaths();
        shr_replace_computed();
        shr_replace_port();
        shr_replace_hdlpath();
        shr_replace_port_strings();
        shr_write_files();
    };

    init() is also {
        set_config(print, line_size, 160);
    };
};


//------------------------------------------------------------------------------
// Structs
//------------------------------------------------------------------------------

// shr_tick2port_loaded_module_s
//    module_filename  => name of file on disk
//    module_contents  => initial copy of file from disk
//    module_ecode     => list of structs with info about each line of 'e' code
struct shr_tick2port_loaded_module_s {
    !module_filename:       string;
    !module_contents:       list of string;
    !module_ecode:          list of shr_tick2port_loaded_code_s;
};

// shr_tick2port_loaded_module_s
//     code_filename    => name of file on disk
//     code_linenum     => number of line in original file
//     code_unit        => text string name of unit 
//     code_str         => copy of line
//     code_computed    => lines to be re-written --> computed names
//     code_port        => lines to be re-written --> from string to port
//     code_hdlassign   => lines to be re-written --> keep hdl_path()
struct shr_tick2port_loaded_code_s {
    !code_filename:         string;
    !code_linenum:          index;
    !code_unit:             string;
    !code_str:              string;
    !code_computed:         list of string;
    !code_port:             list of string;
    !code_hdlassign:        list of string;
};

// shr_tick2port_unit_info_s
//     unit_name    => name of unit or struct
//     unit_fields  => list of info about each field
struct shr_tick2port_unit_info_s {
    !unit_name:             string;
    !unit_fields:           list of shr_tick2port_field_info_s;
};

// shr_tick2port_field_info_s
//      field_filename  => name of file on disk
//      field_linenum   => name of file on disk
//      field_name      => name of field
//      field_type      => user type of field
//      field_port      => is this field of string to be changed to simple_port
struct shr_tick2port_field_info_s {
    !field_filename:        string;
    !field_linenum:         index;
    !field_name:            string;
    !field_type:            string;
    !field_port:            bool;
};


//------------------------------------------------------------------------------
// Methods
//------------------------------------------------------------------------------
extend sys {
    // shr_setup_check() Test the SPECMAN_PATH  If this is not set then this
    // program will not work properly
    shr_setup_check() is {
        outf("*** Setup checks\n");
        var paths := str_split(get_symbol("SPECMAN_PATH"), ":");
        paths = paths.all(it != "");
//      paths = paths.all(it !~ "/sn_rel4.2/");
        shr_specman_path = paths;
        var status := system(append("rm -rf ", shr_outdir));
        assert(status == 0);
        status = system(append("mkdir ", shr_outdir));
        assert(status == 0);
    };

    // shr_get_loaded_modules looks inside specman internals (to bypass file load)
    // It builds the list of files in your eVC in the proper order. It stores
    // the full unix-path and filename.
    shr_get_loaded_modules() is {
        outf("*** Building file load list\n");
        var tmp_module:   shr_tick2port_loaded_module_s = new;

        shr_loaded_modules.clear();
        var modules := global.symtab.modules;
        modules = modules.all(.file_name !~ "/\/proj\/specman\/src/");
        modules = modules.all(.checksum > 0 or .loaded);
        modules = modules.all(.name !~ "/shr_tick2port_top/");
        modules = modules.all(.name !~ "/^evc_util.*/"      );
        modules = modules.all(.name !~ "/^erm_.*/"          );
        modules = modules.all(not .encrypted);
        for each (m) in modules do {
            tmp_module          = new;
            tmp_module.module_filename = append(m.relative_path,
                                                m.name, 
                                                ".e");
            shr_loaded_modules.add(tmp_module.copy());
        };
    };

    // shr_parse_code() scans each file in the eVC. It strips out multi and
    // single line comments. It then flags the lines which are using computed
    // signal names  .
    shr_parse_code() is {
        outf("*** Scanning files\n");
        for each (m) in shr_loaded_modules do {
            var in_code:      bool  = FALSE;
            var in_mline:     bool  = FALSE;
            var next_mline:   bool  = FALSE;
            var next_code:    bool  = FALSE;
            var linenum:      index = 0;

            outf("    *** Parsing file:   %s\n", m.module_filename);
            var path_match:    bool = FALSE;
            for each (p) in shr_specman_path do {
                var path := str_sub(m.module_filename, 0, str_len(p));
                if (path ~ p) {
                    path_match = TRUE;
                };
            };
            if (not path_match) {
//              outf("## WARNING :: specman_path:(%s)  not in file_path(%s)\n", 
//                   get_symbol("SPECMAN_PATH"),
//                   m.module_filename);
            };
            for each line (l) in file m.module_filename do {
                if (m.module_contents.size() == 0 and l ~ "/^Encrypted.*/") {
                    outf("## WARNING :: encrypted file:(%s)  is being skipped\n", 
                         m.module_filename);
                    break;
                };
                m.module_contents.add(l);
                if      (not in_code and l ~ "/^<'/")   { in_code    = TRUE;
                                                          next_code  = TRUE  }
                else if (    in_code and l ~ "/^'>/")   { next_code  = FALSE }
                else if (    in_code and l ~ "/\/\*/")  { in_mline   = TRUE  }
                else if (    in_code and l ~ "/\*\//")  { next_mline = FALSE };
                if (in_code and not in_mline) {
                    // remove single-line comments
                    var lnc : string;
                    lnc = str_replace(l, "/\-\-.*/", "");
                    lnc = str_replace(lnc, "/\/\/.*/", "");
                    // has a non-whitespace
                    if (lnc ~ "/.*\S.*/") {
                        var eline: shr_tick2port_loaded_code_s = new;
                        eline.code_filename     = m.module_filename;
                        eline.code_linenum      = linenum;
                        eline.code_str          = lnc;
                        m.module_ecode.add(eline);
                    };
                };
                if (in_code and not next_code) { in_code = FALSE };
                in_mline = next_mline;
                linenum += 1;
            };
            var breakpoint: uint;
        };
    };

    // shr_scan_topology() scans all real lines of non-comment code and tags
    // the lines which has computed names
    shr_scan_topology() is {
        outf("*** Scanning user code topology\n");
        // first find all unit/struct declarations
        outf("    *** Searching for unit/structs\n");
        var new_struct: shr_tick2port_unit_info_s = new;
        new_struct.unit_name = "sys";
        shr_loaded_topology.add(new_struct.copy());
        new_struct.unit_name = "global";
        shr_loaded_topology.add(new_struct.copy());
        new_struct.unit_name = "any_env";
        shr_loaded_topology.add(new_struct.copy());
        new_struct.unit_name = "dut_error_struct";
        shr_loaded_topology.add(new_struct.copy());
        for each (m) in shr_loaded_modules do {
            for each (l) in m.module_ecode do {
                //if ((l.code_str ~ "/^unit\s+(\S.*)\s.*{/"    ) or 
                    //(l.code_str ~ "/^\sunit\s+(\S.*)\s.*{/"  ) or 
                    //(l.code_str ~ "/^unit\s+(\S.*){/"        ) or 
                    //(l.code_str ~ "/^\sunit\s+(\S.*){/"      ) or 
                    //(l.code_str ~ "/^struct\s+(\S.*)\s.*{/"  ) or
                    //(l.code_str ~ "/^\sstruct\s+(\S.*)\s.*{/") or
                    //(l.code_str ~ "/^struct\s+(\S.*){/"      ) or
                    //(l.code_str ~ "/^\sstruct\s+(\S.*){/"    )) {
                if ((l.code_str ~ "/^unit\s+(\w+)/"    ) or 
                    (l.code_str ~ "/^\sunit\s+(\w+)/"  ) or 
                    (l.code_str ~ "/^struct\s+(\w+)/"  ) or
                    (l.code_str ~ "/^\sstruct\s+(\w+)/")
                    ){
                        var new_unit: string = $1;
                        new_unit = str_split(new_unit, "/\s/")[0];
                        if (not shr_loaded_topology.has(.unit_name == new_unit)) {
                            new_struct = new;
                            new_struct.unit_name = new_unit;
                            shr_loaded_topology.add(new_struct.copy());
                        };
                };
                if ((l.code_str ~  "/created_driver.*=(.*),/") or
                    (l.code_str ~  "/created_driver.*=(.*);/") and
                    (l.code_str !~ "/\sextend\s/"            ) and
                    (l.code_str !~ "/^extend\s/"             )) {
                        var new_unit: string = $1;
                        new_unit = str_replace(new_unit, "/\s/", "");
                        new_struct = new;
                        new_struct.unit_name = new_unit;
                        shr_loaded_topology.add(new_struct.copy());
                };
                if (((l.code_str ~  "/^\ssequence\s(.*)/")  or
                     (l.code_str ~  "/^sequence\s(.*)/"  )) and
                     (l.code_str !~ "/\sextend\s/"       )  and
                     (l.code_str !~ "/^extend\s/"        )) {
                        var new_unit: string = $1;
                        new_unit = str_split(new_unit, "/\s/")[0];
                        new_struct = new;
                        new_struct.unit_name = new_unit;
                        shr_loaded_topology.add(new_struct.copy());
                        if (l.code_str ~  "/sequence\s(.*)/") {
                            new_unit = append(new_unit, "_driver");
                            new_struct.unit_name = new_unit;
                            shr_loaded_topology.add(new_struct.copy());
                        };
                };
            };
        };
        // then tag each line of code to which unit it resides
        var this_unit:   string;
        outf("    *** Tagging unit/structs\n");
        for each (m) in shr_loaded_modules do {
            this_unit = "";
            for each (l) in m.module_ecode do {
                //if ((l.code_str ~ "/^unit\s+(\S.*)\s.*{/"    ) or -- u {
                    //(l.code_str ~ "/^\sunit\s+(\S.*)\s.*{/"  ) or -- u {
                    //(l.code_str ~ "/^unit\s+(\S.*){/"        ) or -- u{
                    //(l.code_str ~ "/^\sunit\s+(\S.*){/"      ) or -- u{
                    //(l.code_str ~ "/^struct\s+(\S.*)\s.*{/"  ) or -- s {
                    //(l.code_str ~ "/^\sstruct\s+(\S.*)\s.*{/") or -- s {
                    //(l.code_str ~ "/^struct\s+(\S.*){/"      ) or -- s{
                    //(l.code_str ~ "/^\sstruct\s+(\S.*){/"    ) or -- s{
                    //(l.code_str ~ "/^extend\s+(\S.*)\s.*{/"  ) or -- e {
                    //(l.code_str ~ "/^\sextend\s+(\S.*)\s.*{/") or -- e {
                    //(l.code_str ~ "/^extend\s+(\S.*){/"      ) or -- e{
                    //(l.code_str ~ "/^\sextend\s+(\S.*){/"    ) or -- e{
                    //(l.code_str ~ "/^\sextend\s+(\S.*){/"    )) {
                if ((l.code_str ~ "/^unit\s+(\w+)/"         ) or -- unit x
                    (l.code_str ~ "/^\sunit\s+(\w+)/"       ) or --  unit x
                    (l.code_str ~ "/^struct\s+(\w+)/"       ) or -- struct x
                    (l.code_str ~ "/^\sstruct\s+(\w+)/"     ) or --  struct x
                    (l.code_str ~ "/^extend\s+(\S.*)\s+{/"  ) or -- extend <subtype> x {
                    (l.code_str ~ "/^\sextend\s+(\S.*)\s+{/") or --  extend <subtype> x {
                    (l.code_str ~ "/^extend\s+(\S.*){/"      ) or -- extend <subtype> x{
                    (l.code_str ~ "/^\sextend\s+(\S.*){/"    ) or --  extend <subtype> x{
                    (l.code_str ~ "/^extend\s+(\S.*)\s*$/"   ) or -- extend <subtype> x\n
                    (l.code_str ~ "/^\sextend\s+(\S.*)s*$/"  )    --  extend <subtype> x\n
                    ){
                        this_unit = $1;
                        if (this_unit !~ "/:/") {
                            this_unit = str_replace(this_unit, "/\slike\s.*/", "");
                            this_unit = str_replace(this_unit, "/\slike$/", "");
                            var ee   := str_split(this_unit, "/\s/");
                            ee        = ee.all(it != "");
                            this_unit = ee[ee.size()-1];
                            if (this_unit ~ "/.*\s(\S.*)/") {
                                this_unit = $1;
                            };
                            assert(shr_loaded_topology.has(.unit_name ~ this_unit));
                        };
                    } else if (l.code_str ~ "/^sequence\s(\S.*)\s/") {
                        this_unit = $1;
                        this_unit = str_replace(this_unit, "/\susing.*/", "");
                };
                l.code_unit = this_unit;
            };
            m.module_ecode = m.module_ecode.all(.code_unit != "");
        };
        // next find all fields under each unit/struct
        outf("    *** Building fields\n");
        var lm := shr_loaded_modules;
        var lt := shr_loaded_topology;
        for each (s) in lt do {
            var s_ecode := lm.module_ecode.all(.code_unit == s.unit_name);
            for each (l) in s_ecode do {
                if ((l.code_str ~ "/(.*)(\/\/.*)/") or 
                    (l.code_str ~ "/(.*)(\-\-.*)/")) {
                        l.code_str = $1;
                };
                if ((l.code_str ~ "/\s(\S.*):(.*);/")    or 
                    (l.code_str ~ "/\s(\S.*)\s.*:(.*);/")) {
                        if ((l.code_str !~ "/.*var .*/"         ) and 
                            (l.code_str !~ "/.*item .*/"        ) and
                            (l.code_str !~ "/.* = .*/"          ) and
                            (l.code_str !~ "/.*\sevent\s.*is\s/") and 
                            (l.code_str !~ "/.*\);/"            ) and
                            (l.code_str !~ "/.*append.*/"       ) and
                            (l.code_str !~ "/.*out.*\(.*\)/"    ) and
                            (l.code_str !~ "/.*message.*\(.*/"  ) and
                            (l.code_str !~ "/.*error\(.*/"      )) {
                                var fname: string = str_split(l.code_str, ":")[0];
                                var ftype: string = str_split(l.code_str, ":")[1];
                                if (ftype ~ "/.*list.* of (.*)/") {
                                    ftype = $1;
                                };
                                if (ftype ~ "/\s(\S.*)\sis\sinstance/") {
                                    ftype = $1;
                                    ftype = str_replace(ftype, " ", "");
                                };
                                if (ftype ~ "/.*\s(\S.*);/") {
                                    ftype = $1;
                                };
                                var n := str_split(fname, " ").all(it != "");
                                if (n.size() > 1) {
                                    n  = n.all(it != "package");
                                    n  = n.all(it != "private");
                                };
                                fname  = n[0];
                                fname  = str_replace(fname, "!", "");
                                ftype  = str_split(ftype, " ").all(it != "")[0];
                                ftype  = str_split(ftype, "(").all(it != "")[0];
                                ftype  = str_replace(ftype, ";", "");
                                // keep select cases aren't handled properly for now...
                                if ((fname ~  "/[a-zA-Z].*/") and 
                                    (fname !~ "/\..*/"      ) and
                                    (ftype ~  "/[a-zA-Z].*/") and 
                                    (ftype !~ "/\..*/"      )) {
                                        var this_field: shr_tick2port_field_info_s = new;
                                        this_field.field_name     = fname;
                                        this_field.field_type     = ftype;
                                        this_field.field_linenum  = l.code_linenum;
                                        this_field.field_filename = l.code_filename;
                                        s.unit_fields.add(this_field);
                                };
                        };
                };
            };
        };
        // sequence driver have built-in fields
        outf("    *** Building sequence fields\n");
        var driver_ecode := lm.module_ecode.all(.code_str ~ "/created_driver/");
        for each (driver) in driver_ecode do {
            var unit_ptr := lt.first(.unit_name == driver.code_unit);
            var this_field: shr_tick2port_field_info_s = new;
            this_field.field_name     = "driver";
            if ((driver.code_str ~ "/created_driver\s=\s(.*),/") or
                (driver.code_str ~ "/created_driver\s=\s(.*);/")) {
                    this_field.field_type     = $1;
            };
            this_field.field_linenum  = driver.code_linenum;
            this_field.field_filename = driver.code_filename;
            unit_ptr.unit_fields.add(this_field);
            var breakpoint : uint;
        };
    };

    // shr_tag_computed() scans all real lines of non-comment code and tags
    // the lines which has computed names
    shr_tag_computed() is {
        outf("*** Tagging all lines with computed names for replacement\n");
        for each (m) in shr_loaded_modules do {
            for each (l) in m.module_ecode do {
                if ((l.code_str ~ "/.*\'\(.*\)\'.*/"       ) or
                    (l.code_str ~ "/.*\'\(.*\)\@x\'.*/"    ) or
                    (l.code_str ~ "/.*\'\(.*\)\@z\'.*/"    ) or
                    (l.code_str ~ "/.*\'\(.*\)\[.*\]\'.*/")) {
                        var clist := str_split(l.code_str, "/\'\(/").all(it ~ "/\).*\'/");
                        for each (c) in clist do {
                            var cc := str_split(c, "/\)\'/")[0];
                            if (c ~ "/\)@x/") {
                                cc = str_split(c, "/\)@x\'/")[0];
                                cc = append(cc, "@x");
                            } else if (c ~ "/\)@z/") {
                                cc = str_split(c, "/\)@z\'/")[0];
                                cc = append(cc, "@z");
                            } else if (c ~ "/(.*)\)(\[.*\])/") {
                                cc = append($1, $2);
                            };
                            if (cc !~ "/append.*(.*)/") {
                                l.code_computed.add(cc);
                            };
                        };
                };
            };
        };
    };

    // shr_scane_port_fields() tags all lines which need to be changed from 
    // "string" type to "simple_port" type.
    shr_tag_port_fields() is {
        outf("*** Searching for all computed accesses for replacement\n");
        var lm     := shr_loaded_modules;
        var lt     := shr_loaded_topology;
        var cnames := lm.module_ecode.all(.code_computed.size() > 0);
        for each (cline) in cnames do {
            for each (cname) in cline.code_computed do {
                var hierarchy := str_split(cname, ".");
                var port      := hierarchy.last(TRUE);
                    port       = str_replace(port, "/\@.*/", "");
                    port       = str_replace(port, "/\[.*/", "");
                var deep      := hierarchy.size() - 1;
                var unit_ptr  :  shr_tick2port_unit_info_s;
                var indirect  :  bool = FALSE;
                unit_ptr = lt.first(.unit_name == cline.code_unit);
                for each (ptr) in hierarchy.all(index < deep) do {
                    if (unit_ptr == NULL) {
                        indirect = TRUE;
                        break;
                    };
                    var field_ptr := unit_ptr.unit_fields.first(.field_name == ptr);
                    if (field_ptr == NULL) {
                        indirect = TRUE;
                        break;
                    } else {
                        unit_ptr = lt.first(.unit_name == field_ptr.field_type);
                        if (unit_ptr == NULL) {
                            indirect = TRUE;
                            break;
                        };
                    };
                };
                if (indirect) {
                    cline.code_computed.clear();
                } else {
                    var port_ptr := unit_ptr.unit_fields.first(.field_name == port);
                    if (port_ptr != NULL) {
                        port_ptr.field_port = TRUE;
                    };
                };
            };
        };
    };

    // shr_tag_port_lines() Mark the ecode lines which will eventually
    // be replaced from a string field type to a simple_port type
    shr_tag_port_lines() is {
        outf("*** Tagging all strings for port replacement\n");
        var lm     := shr_loaded_modules;
        var lt     := shr_loaded_topology;
        var pfields := lt.unit_fields.all(.field_port);
        for each (p) in pfields do {
            var mptr := lm.first(.module_filename == p.field_filename);
            var eptr := mptr.module_ecode.first(.code_linenum == p.field_linenum);
            eptr.code_port.add(p.field_name);
        };
    };

    // shr_tag_port_lines() Mark the ecode lines which will eventually
    // be replaced from a string field type to a simple_port type
    shr_tag_port_hdlpaths() is {
        outf("*** Tagging all hdl_path assignments\n");
        var lm      := shr_loaded_modules;
        var lt      := shr_loaded_topology;
        var pfields := lm.module_ecode.all(.code_port.size() > 0);
        for each (p) in pfields do {
            var module_e := lm.module_ecode.all(.code_unit == p.code_unit);
            var port_e   := module_e.all( str_split(.code_str, p.code_port[0]).size() > 1);
            for each (pe) in port_e do {
                if (pe.code_str ~ "/(keep.*)==.*\"(.*)\"/") {
                    pe.code_hdlassign.add(p.code_port);
                };
            };
        };
    };

    // shr_replace_computed() replace all computed name accesses
    shr_replace_computed() is {
        outf("*** Replacing string with simple_port\n");
        var lm      := shr_loaded_modules;
        var lt      := shr_loaded_topology;
        var pcode   := lm.module_ecode.all(.code_computed.size() > 0);
        for each (p) in pcode do {
            for each (pcomputed) in p.code_computed do {
                var hdl_computed: string = append("'(", pcomputed, ")'");
                var hdl_port:     string = append(pcomputed, "$");
                // special case for '(signal)[x:y]'
                var slist := str_split(hdl_port, "[");
                if (slist.size() > 1) {
                    var slist1_copy: string = slist[1];
                    slist[1] = str_replace(slist[1], "/\$/", "");
                    slist[1] = str_replace(slist[1], "(", "\(");
                    slist[1] = str_replace(slist[1], ")", "\)");
                    slist[1] = str_replace(slist[1], "/\*/", "\*");
                    slist[1] = str_replace(slist[1], "+", "\+");
                    hdl_computed = append("/'\\(", 
                                           slist[0], 
                                           "\)\\[", 
                                           str_replace(slist[1], "/\\]/", ""),
                                           "\\]\\'/");
                    slist1_copy = append("$[", slist1_copy);
                    slist1_copy = str_replace(slist1_copy, "/\$$/", "");
                    hdl_port    = append(slist[0], slist1_copy);
                };
                // special case for '(signal)@x'
                slist = str_split(hdl_port, "@");
                if (slist.size() > 1) {
                    if (slist[1] ~ "/(.)\$/") {
                        slist[1] = append("$@", $1);
                        hdl_computed = append("/'\\(",
                                              slist[0],
                                              "\\)@",
                                              $1,
                                              "'/");
                    };
//                  hdl_port = append(slist[0], "$", slist[1]);
                    hdl_port = append(slist[0], ".has_", str_sub(slist[1], 2, 1), "()");
                };
                //   hdl_computed = "/\'\(sig_cmd\)\[\(7\+8\*i\):\(8\*i\)\]\'/"
                //   hdl_port     = "sig_cmd$[(7+8*i):(8*i)]"
                p.code_str = str_replace(p.code_str, hdl_computed, hdl_port);
                if (p.code_str ~ "/event.*is.*true/") {
                    var event_str := str_split(p.code_str, "true")[1]; 
                    event_str = str_replace(event_str, "/^.*\(/", "");
                    if (event_str ~ "/^(.*)\).*/") {
                        event_str = $1;
                    };
                    if ((event_str !~ "/=/") and (event_str != "")) {
                        var event_bool := append(event_str, " == 1'b1");
                        p.code_str = str_replace(p.code_str, event_str, event_bool);
                    };
                };
                var fp := lm.first(.module_filename == p.code_filename);
                var cp := fp.module_contents[p.code_linenum];
                fp.module_contents[p.code_linenum] = str_replace(cp, hdl_computed, hdl_port);
                if (shr_debug) {
                    shr_debug_changelines.add(p);
                };
            };
        };
    };

    // shr_replace_port() replace string to simple_port
    shr_replace_port() is {
        outf("*** Replacing string with simple_port\n");
        var lm      := shr_loaded_modules;
        var lt      := shr_loaded_topology;
        var pcode   := lm.module_ecode.all(.code_port.size() > 0);
        for each (p) in pcode do {
            var old_type: string = "string";
            var new_type: string = "simple_port of uint is instance";
            p.code_str = str_replace(p.code_str, old_type, new_type);
            var fp := lm.first(.module_filename == p.code_filename);
            var cp := fp.module_contents[p.code_linenum];
            fp.module_contents[p.code_linenum] = str_replace(cp, old_type, new_type);
            if (shr_debug) {
                shr_debug_changelines.add(p);
            };
        };
    };

    // shr_replace_hdlpath() assign through hdl_path() method of port
    shr_replace_hdlpath() is {
        outf("*** Replacing keep with hdl_path()\n");
        var lm      := shr_loaded_modules;
        var lt      := shr_loaded_topology;
        var pcode   := lm.module_ecode.all(.code_hdlassign.size() > 0);
        pcode = pcode.all(.code_port.size() > 0);
        for each (p) in pcode do {
            var match_str := append("/", p.code_port[0], "\s/");
            var new_keep  := append(p.code_hdlassign[0], ".hdl_path()");
            p.code_str = str_replace(p.code_str, match_str, new_keep);
            var fp := lm.first(.module_filename == p.code_filename);
            var cp := fp.module_contents[p.code_linenum];
            fp.module_contents[p.code_linenum] = str_replace(cp, match_str, new_keep);
            if (shr_debug) {
                shr_debug_changelines.add(p);
            };
        };
    };

    // shr_replace_port_strings() 
    shr_replace_port_strings() is {
        outf("*** Replacing all hdl_path used as strings\n");
        var lm      := shr_loaded_modules;
        var lt      := shr_loaded_topology;
        var pfields := lm.module_ecode.all(.code_port.size() > 0);
        for each (p) in pfields do {
            var match_str := append("/", p.code_port[0], "/");
            var port_e := lm.module_ecode.all( str_split(.code_str, match_str).size() > 1);
            port_e = port_e.all(.code_port.size()      == 0);
            port_e = port_e.all(.code_str !~ "/message.*(.*)/");
            port_e = port_e.all(.code_str !~ "/append.*(.*)/");
            for each (pe) in port_e do {
                if (not lt.unit_fields.has(.field_filename == pe.code_filename and 
                                           .field_linenum == pe.code_linenum)) {
                    var fp       := lm.first(.module_filename == pe.code_filename);
                    var mlist:      list of string;
                    var rlist:      list of string;
                    var clist:      list of string;
                    var pelist:     list of string;
                    var new_str:    string;
                    var bgn_str:    string;
                    var end_str:    string = fp.module_contents[pe.code_linenum];
                    clist = str_split(pe.code_str, "/[(),;{} ]/");
                    clist = clist.all(it ~ (append("/", p.code_port[0], "/")));
                    clist = clist.all(it !~ "/\$/");
                    if (pe.code_str ~ "/keep.*bind.*(.*)/") {
                        clist.clear();
                    };
                    for each (c) in clist do {
                        var unit_ptr:    shr_tick2port_unit_info_s;
                        var field_ptr:   shr_tick2port_field_info_s;
                        var field_name:  string;
                        unit_ptr  = lt.first(.unit_name == pe.code_unit);
                        field_ptr = unit_ptr.unit_fields.first(.field_name == c);
                        if (c != p.code_port[0]) {
                            if ((c ~ "/\./") and (c !~ "/^\".*\"$/")) {
                                var hierarchy := str_split(c, ".");
                                field_name = hierarchy.last(TRUE);
                                hierarchy  = hierarchy.all(index < hierarchy.size()-1);
                                unit_ptr = lt.first(.unit_name == pe.code_unit);
                                for each (f) in hierarchy do {
                                    var ff   := str_replace(f, "/\s/", "");
                                    field_ptr = unit_ptr.unit_fields.first(.field_name == ff);
                                    if (index == 0 and ff == "sys") {
                                        unit_ptr  = lt.first(.unit_name == "sys");
                                        field_ptr = unit_ptr.unit_fields.first(.field_name == field_name);
                                    } else if (field_ptr != NULL and 
                                               lt.has(.unit_name == field_ptr.field_type)) {
                                        unit_ptr = lt.first(.unit_name == field_ptr.field_type);
                                    } else {
                                        break;
                                    };
                                };
                                field_ptr = unit_ptr.unit_fields.first(.field_name == field_name);
                            };
                        };
                        if (unit_ptr != NULL and field_ptr != NULL) {
                            var nbuff := str_split(pe.code_str, c).first(TRUE);
                            bgn_str = append(nbuff, c);
                            new_str = append(nbuff, c, ".hdl_path()");
                        } else {
                            var nbuff := str_split(end_str, c).first(TRUE);
                            bgn_str = append(bgn_str, nbuff, c);
                            new_str = append(new_str, nbuff, c);
                        };
                        if (str_len(bgn_str) < str_len(fp.module_contents[pe.code_linenum])) {
                            end_str = str_sub(fp.module_contents[pe.code_linenum], str_len(bgn_str), 1000);
                        };
                    };
                    new_str = append(new_str, end_str);
                    pe.code_str = new_str;
                    var cp := fp.module_contents[pe.code_linenum];
                    fp.module_contents[pe.code_linenum] = new_str;
                    if (shr_debug) {
                        shr_debug_changelines.add(pe);
                    };
                };
            };
        };
    };

    // shr_write_files()
    shr_write_files() is {
        outf("*** Writing port-compliant files\n");
        for each (lm) in shr_loaded_modules do {
            var fullpath:   string;
            var outdir:     string;
            var outfile:    string;
            var path_match: string = "";
            for each (p) in shr_specman_path do {
                var path := str_sub(lm.module_filename, 0, str_len(p));
                if (path ~ p) {
                    path_match = p;
                    break;
                };
            };
            if (path_match != "") {
                fullpath = str_replace(lm.module_filename, path_match, "");
            } else {
                fullpath = str_replace(lm.module_filename, "/^.*\//", "");
            };
            fullpath = str_replace(fullpath, "/^\//", "");
            fullpath = append(shr_outdir, "/", fullpath);
            if (fullpath ~ "/(.*\/)(.*)/") {
                outdir  = $1;
                outfile = $2;
                var cmd    := append("mkdir -p ", outdir);
                var status := system(cmd);
                if (status != 0) {
                    outf("## WARNING :: Cannot mkdir: %s\n", outdir);
                } else {
                    outf("    *** Writing file: %s\n", fullpath);
                    files.write_string_list(fullpath, lm.module_contents);
                };
            };
        };
        if (shr_debug) {
            var debug_filename : string = append(shr_outdir, "/DEBUG_CHANGES");
            outf("*** Writing debug file: %s\n", debug_filename);
            var change_lines: list of string;
            shr_debug_changelines = shr_debug_changelines.sort(.code_linenum);
            shr_debug_changelines = shr_debug_changelines.sort(.code_filename);
            shr_debug_changelines = shr_debug_changelines.unique(it);
            for each (l) in shr_debug_changelines do {
                change_lines.add(appendf("File:%-70s  Line:%5d   %s",
                                         l.code_filename,
                                         l.code_linenum,
                                         l.code_str));
            };
            files.write_string_list(debug_filename, change_lines);
        };
    };
};


'>
