/*******************************************************************************
 * File:        reset_driver.e
 * By:          Jerel Canales (Texas Instruments)
 * Company:     Texas Instruments, Inc.
 * Purpose:     Top-level reset driver
 ******************************************************************************/

<'

struct reset_s {

  //INTERFACE
  clk                 : clock_stub;
  sig_reset           : string;

  // used to select a initial non-random reset
  initial_reset_enable : enable_t;
  keep soft initial_reset_enable == DISABLED;

  // used to select width of initial non-random reset
  initial_reset_width : uint;
  keep soft initial_reset_width == 40;

  random_reset_enable : enable_t;
     keep soft random_reset_enable == DISABLED;

  random_stop_enable  : enable_t;
     keep soft random_stop_enable == ENABLED;

  random_reset_count  : uint;
     keep soft random_reset_count in [1..25];

  reset_interval_list[random_reset_count] : list of uint;
  keep for each in reset_interval_list {
     soft it in [1000..5000];
  };

  reset_width_list[random_reset_count] : list of uint;
  keep for each in reset_width_list {
     soft it in [1..40];
  };

  // used to keep reset assert/deassert off of clock edge.
  // this is used for both initial and random resets.
  reset_delay  : uint; keep soft reset_delay == 2; 

  // clock used for TCMs
  event rclk is cycle @clk.raw_rise;


  // new method: used by random generator
  Apply_Global_Reset(width : uint) @rclk is {

    '(sig_reset)' = 1;  // start with deassert
    wait[2];
    wait delay(reset_delay);
    '(sig_reset)' = 0;  // reset signal asserted
    wait[width];
    wait delay(reset_delay);
    '(sig_reset)' = 1;  // end with deassert

  }; // Apply_Global_Reset



  //-------- Initial & Random Global Reset control -- 
  Control_Global_Reset() @rclk is {

    var reset_interval : uint;
    var reset_width    : uint;
    var reset_index    : uint;
    var glob_reset_id  : uint;

    // allow this global reset controller to have a say in when the test is done
    if ( (random_reset_enable == ENABLED) && (random_stop_enable == ENABLED) ) then {
      glob_reset_id  = STOP_RUN.register_watch("global_reset_bfm",1,100);
    };

    // check if initial reset is desired
    if ( initial_reset_enable == ENABLED ) then {
      Apply_Global_Reset(initial_reset_width);
    };

    // first reset is mandatory
    reset_index = 0;
    // random resets are optional
    if (random_reset_enable == ENABLED) then {
      // will assert X number of random resets
      while ( reset_index < random_reset_count) do {
        // each random reset interval is a different number of clocks
        reset_interval = reset_interval_list[reset_index];
        reset_width    = reset_width_list[reset_index];
        while reset_interval>0 do {
           wait[1]*cycle;
           reset_interval -= 1;
        };
        // apply random reset now
        reset_index = reset_index + 1;
        Apply_Global_Reset(reset_width);
      }; // end while

      // allow test to stop - if feature enabled
      if ( random_stop_enable == ENABLED ) then {
        STOP_RUN.report_done(1,glob_reset_id);
      };

    }; // if random resets ENABLED

  }; // Control_Global_Reset

  run() is also {
     start Control_Global_Reset();
  };

};


'>


