//***********************************************************************
//*  FILE       : switch.e
//*  BY         : Jerel Canales
//*  PURPOSE    : The switch structure allows for easy configuration of
//*               heirachy modules, coverage monitors, and debug outputs.
//*               It also supports checker configuration, however this
//*               should be used ONLY for developmental purposes.
//*
//*  06/02/01     jerelc   Initial code
//*  08/06/01     jerelc   Add documentation (checker examaple)
//***********************************************************************


// CONTENTS (you can search for these words to jump to the appropriate section):
//     USAGE
//     INPUTS
//     OUTPUTS
//     EXAMPLES
//     DECLARATIONS
//     METHODS
    
// USAGE: Instantiate the switch_s structure as a field in the parent field 
//        a checker, coverage monitor or module declaration.

// INPUTS: None.

// OUTPUTS: 
//     module  : list of enable_t;  module switches which by default are enabled
//     monitor : list of enable_t;  monitor switches which by default are disabled
//     checker : list of enable_t;  checker switches which by default are enabled
//     debug   : list of enable_t;  debug switches which by default are disabled
    
// EXAMPLES:
//    
// Module example--while allowing for a unit declaration, extention of a unit via
//                 the switch reduces the memory usage when DISABLED.
//
//     unit GALAXY {
//        unit_switch  : enable_t;
//        debug_switch : enable_t;
//     };
//     unit DSPS {
//        s        : switch_s;
//        keep {
//           s.module.size() == 2;
//           s.debug.size()  == 2;
//           s.module_name[0] == "GALAXYI0";
//           s.module_name[1] == "GALAXYI1";
//        };
//        galaxyi0 : GALAXY is instance;
//           keep galaxyi0.unit_switch == s.module[0];
//           keep galaxyi0.debug_switch == s.debug[0];
//        galaxyi1 : GALAXY is instance;
//           keep galaxyi1.unit_switch == s.module[1];
//           keep galaxyi1.debug_switch == s.debug[1];
//     };
// 
//     The following conditional GALAXY extentions will either
//     declare the SUBSYSTEM unit or not declare SUBSYSTEM & freeze
//     the subsystem clock:
//
//     extend ENABLED'unit_switch GALAXY {
//        subsystemi0 : SUBSYSTEM is instantce;
//     };
//
//     extend DISABLED'unit_switch GALAXY {
//        freeze_clock() is {				
//           force '(galaxy_clock)' = 0;
//        };
//        run() is also {
//           me.freeze_clock();
//        };       
//     };
//
//     Finally, in a configuration file, you can program the switches.
//     For example, to enable Galaxy 0 and disable Galaxy 1, your
//     config file read as follows:
//
//     extend DSPS {
//        keep {
//           s.module[0] == ENABLED;
//           s.module[1] == DISABLED;
//           s.debug[0]  == ENABLED;
//           s.debug[1]  == DISABLED;
//        };
//     };
//        
//
// Checker example--allows for conditional extention of a checker; when DISABLED, all
//                  temporal expression, fields, and tcm's are not defined.
// 
// struct laijin_proto_chk_s {
// 
//    switch:       enable_t;   keep soft switch       == ENABLED;  // always keep soft to ENABLED,
// };                                                               // otherwise load errors will occur
//
//
// extend laijin_proto_chk_s {
//   when ENABLED'switch laijin_proto_chk_s {
//      
//    *** define protocol checks here ***
//
//    };
// };
//
// extend LEAD3_WRAPPER {
//      
//    switch : switch_s; 
//    keep switch.checker.size() == 1;
//    keep switch.checker_name[0] == append (kind, " ", checkpoint, " PBUS Protocol Checker");
// 
//    -- Program bus
//    pbus_proto_chk: laijin_proto_chk_s;
//    keep pbus_proto_chk.switch == switch.checker[0];
//
// Then, in a configuration file, program the checker:
//
// extend LEAD3_WRAPPER {
//    keep soft switch.checker[0] == DISABLED;
// };    
    
<'

type enable_t : [ DISABLED, ENABLED ];
    
struct switch_s {

////////////////////////////////////////////////////////////////////////
//  
//   ####   #    #     #     #####   ####   #    #           ####
//  #       #    #     #       #    #    #  #    #          #
//   ####   #    #     #       #    #       ######           ####
//       #  # ## #     #       #    #       #    #               #
//  #    #  ##  ##     #       #    #    #  #    #          #    #
//   ####   #    #     #       #     ####   #    # #######   ####
// 
// The switch struct allows for conditional enabling/disabling of units,
// checkers, monitors, and debug output.
   
// DECLARATIONS:   
   
   //Module Switch
   module        :list of enable_t;
   keep soft module.size() == 0;
   keep for each in module do {
      soft it == ENABLED; 
   };
   
   //Module & Checker Identifiers
   module_name   :list of string; 
   keep module_name.size() == module.size();
   keep for each in module_name do {
      soft it == "EMPTY"; 
   };   
   
   //Coverage Monitor Switch  
   monitor       :list of enable_t;
   keep soft monitor.size() == 0;
   keep for each in monitor do {
      soft it == DISABLED; 
   };   
   
   monitor_name  :list of string;
   keep monitor_name.size() == monitor.size();
   keep for each in monitor_name do {
      soft it == "EMPTY"; 
   };   
   
   //Checker Switches
   checker       :list of enable_t;
   keep soft checker.size() == 0;
   keep for each in checker do {
      soft it == ENABLED; 
   };
   
   checker_name  :list of string;
   keep checker_name.size() == checker.size();
   keep for each in checker_name do {
      soft it == "EMPTY"; 
   };

   //Debug Statement Switch 
   debug         :list of enable_t;
   keep soft debug.size() == 0;
   keep for each in debug do {
      soft it == DISABLED; 
   };
   
   debug_name  :list of string;
   keep debug_name.size() == debug.size();
   keep for each in debug_name do {
      soft it == "EMPTY"; 
   };
   
};//end struct switch_s
   

// METHODS:
    
extend switch_s {

   monitor_outf () is {
      
      for each in monitor do {
         if (it==ENABLED) then {      
            outf("%s monitor is enabled.\n", monitor_name[index] );
         };
      };
   };

   module_outf () is {
      
      for each in module do {
         if (it==DISABLED) then {            
            outf("%s is disabled and empty.\n", module_name[index] );
         };
      };
   };

   debug_outf () is {    

      for each in debug do {
         if (it==ENABLED) then {            
            outf("%s's debug statments are enabled.\n", debug_name[index] );
         };
      };
   };

   checker_outf () is {
         
      for each in checker do {
         if (it==DISABLED) then {
            outf("%s is disabled.\n", checker_name[index] ); 
         }; 
      };
   };
   
   run() is also {
      monitor_outf();
      module_outf();
      debug_outf();
      checker_outf();
   };
};//end extend switch_s

'>

/////////////////////////////////////////////////////////////////////////
//
//  #######
//     #     #    #  ######          ######  #    #  #####
//     #     #    #  #               #       ##   #  #    #
//     #     ######  #####           #####   # #  #  #    #
//     #     #    #  #               #       #  # #  #    #
//     #     #    #  #               #       #   ##  #    #
//     #     #    #  ######          ######  #    #  #####
//
/////////////////////////////////////////////////////////////////////////
   
