// $Header: $
//********************************************************************************
//
//  Filename: trace_logger.e
//
//  Description: provides a way (actually two) to log any system info into the
//               sim log on every rising edge of the system clock.
//
//               There are currently two types of these cycle loggers.
//               (either can be used and reused by various projects)
//               Each of these types has advantages and disadvantages.
//
//               trace_logger is designed to be used as a passive logger
//                 that provides a couple methods so that various modules
//                 can a) register their log entries (done once)
//                 and b) report their log info (done every clock)
//                 It is the responsibility of these various methods to construct
//                 and format the appropriate info into an ASCII log string.
//
//               string_logger is designed to be used as an active logger
//                 that can/will probe anywhere in the hierachy (using supplied
//                 signal name strings) and construct/format the ASCII log
//                 strings in one easy-to-find project-wide place (in this file).
//
//
//  Major Revision Descriptions:
//  Peter Spyra    11/02/01   original version
//  Peter Spyra    11/12/01   added column title lines for logging
//  Peter Spyra    11/15/01   changed name from debug logger to trace logger
//  Peter Spyra    11/26/01   created new methods for trace
//  Peter Spyra    12/06/01   added R, C, F, rom, rhea, & icache columns
//  Peter Spyra    12/24/01   modified trace_log methods to use row_id param
//
//
//
//*********************************************************************************
//
//                               : indicates request status (when address is displayed)
//      lead3/subsystem id       * indicates ready & not dismissed
//      (since more than 1)      # indicates ready & dismissed
//       |                       X indicates not ready & dismissed
//       |                       |                   
//   0   |  1     2        3     |   4      5       6        7        8        9       10  11   12     13
// clock:x: R   P bus    B bus   | C bus  D bus   E bus    F bus    saram    daram    rom rhea dport icache  
// -------- - -------- -------- -------- -------- -------- -------- -------- -------- --- ---- ---- --------
//    62:0:|.|P:......|B:......|C:......|D:000101|E*......|F:......|........|.......b|...|....|....|........|
//    62:1:|.|P:001ffc|B:000103|C:......|D:......|E:......|F:......|........|.......p|...|....|....|........|
//    62:2:|.|P:......|B:000103|C:......|D:......|E*......|F:......|........|.......b|...|....|....|........|
//    62:3:|.|P*......|B:......|C:......|D:000100|E:......|F:......|........|.......p|...|....|....|........|
//    62:4:|.|P*......|B:000102|C:......|D:000100|E:000ffe|F:......|........|.......p|...|....|....|........|
//    62:5:|.|P:......|B:000103|C:......|D*......|E:000ffe|F:......|........|.......p|...|....|....|........|
//    63:0:|.|P:......|B:......|C:......|D:......|E:......|F:......|........|.......p|...|....|....|........|
//    63:1:|.|P:......|B*000103|C:......|D:......|E:001000|F:......|........|......bp|...|....|....|........|
//    63:2:|.|P*......|B:......|C:......|D:......|E:......|F:......|........|.......p|...|....|....|........|
//    63:3:|.|P:......|B:000ffe|C:......|D*......|E:001000|F:......|........|......dp|...|....|....|........|
//    63:4:|.|P:......|B:000103|C:......|D*......|E*......|F:......|........|......dp|...|....|....|........|
//    63:5:|.|P:001ffc|B*000ffe|C:......|D:......|E:......|F:......|........|......d.|...|....|....|........|
//    64:0:|.|P:......|B:......|C:......|D*......|E:......|F:......|........|.......p|...|....|....|........|
//    64:1:|.|P:001ffc|B*......|C:......|D:000100|E:......|F:......|........|......bb|...|....|....|........|
//    64:2:|.|P:......|B*......|C:......|D:......|E:001000|F:......|........|......bp|...|....|....|........|
//    64:3:|.|P:000204|B*000fff|C:......|D:......|E:......|F:......|........|......bd|...|....|....|........|
//    64:4:|.|P:......|B*000103|C:......|D:......|E:......|F:......|........|......bd|...|....|....|........|
//    64:5:|.|P:......|B:000ffe|C:......|D:......|E*......|F:......|........|.......b|...|....|....|........|
//    65:0:|.|P*001ffc|B:000ffe|C:......|D:......|E:001000|F:......|........|......d.|...|....|....|........|
//    65:1:|.|P:......|B:......|C:......|D:000100|E*......|F:......|........|.......b|...|....|....|........|
//    65:2:|.|P*......|B:000ffe|C:......|D:000100|E:......|F:......|........|......pb|...|....|....|........|
//    65:3:|.|P:......|B:000fff|C:......|D:......|E*......|F:......|........|.......b|...|....|....|........|
//    65:4:|.|P:001ffc|B*......|C:......|D:......|E:001000|F:......|........|......bb|...|....|....|........|
//    65:5:|.|P*......|B:......|C:......|D:......|E:......|F:......|........|.......p|...|....|....|........|
//
// ----------------------------------------------------------------------------------------------------


<'
define  NUM_OF_TRACE_LOG_LINES    6;  // one for each galaxy in Janus chip, 0 thru 5
define  NUM_OF_TRACE_LOG_COLUMNS 14;  // must be = (last column id + 1) in subsystem_defines.e

// IMPORTANT: when adding trace log columns please follow these steps:
// -----------------------------------------------------------------------------------
// 1) reserve column id or ids for your particular module (now in subsystem_defines.e)
// 2) remember that each column can be any number of characters wide
// 3) keep NUM_OF_TRACE_LOG_COLUMNS up to date! (in this file)
// 4) add a call to method: Register_Trace_Group in your method (at time=0)
// 5) also add periodic calls (every clock) to method: Log_Trace_Group

struct log_column {
  row   : int;       // log line id 
  col   : int;       // log column id
  new   : bool;      // set to TRUE when info is logged
  width : int;       // specified column width
  title : string;    // string for column header
  info  : string;    // string that contains the updated log info
  space : string;    // string used when info is not updated
}; // end of log_column


//------------------------------------------------------------------------------
// The distributed trace logger (provides methods that other modules can call) 
//------------------------------------------------------------------------------

unit TRACE_LOGGER_UNIT {

   num_of_trace_log_columns  : uint;  keep soft num_of_trace_log_columns == 0;
   num_of_trace_log_lines    : uint;  keep soft num_of_trace_log_lines == 0;
   num_of_trace_groups       : uint;  keep soft num_of_trace_groups == (num_of_trace_log_columns * num_of_trace_log_lines);
   keep soft gen (num_of_trace_log_columns, num_of_trace_log_lines) before (num_of_trace_groups);

  // each trace log line needs to have a enable/disable switch
  // note that these switches should only prevent the print function,
  // not the actual logging or the counting of clocks 
   
  // default master switch to OFF
  master_trace_log_sw : bool;  keep soft master_trace_log_sw == FALSE;

  specific_trace_log_sw : list of bool;
  keep soft specific_trace_log_sw.size() == num_of_trace_log_lines;
  keep soft gen (num_of_trace_log_lines) before (specific_trace_log_sw.size());
   
  // default individual specific switches to ON
  keep for each (sw) in specific_trace_log_sw {
     soft sw == TRUE;
  };

  trace_title_printed : list of bool; 
  keep soft trace_title_printed.size() == num_of_trace_log_lines;
  keep soft gen (num_of_trace_log_lines) before (trace_title_printed.size());
  // init all print bools to false 
  keep for each (pr) in trace_title_printed {
     soft pr == FALSE;
  };

  // local trace variables
  trace_groups  : list of log_column;
  keep soft trace_groups.size() == num_of_trace_groups;
  keep soft gen (num_of_trace_groups) before (trace_groups.size());

  trace_clock_count : list of uint; 
  keep soft trace_clock_count.size() == num_of_trace_log_lines;
  keep soft gen (num_of_trace_log_lines) before (trace_clock_count.size());


  // global trace methods (used by all/any logging modules)

  Register_Trace_Group( row_id: int, col_id:int, no_of_chars:int, column_header:string ) is {

    var row_index : uint;

    // ignore this method if passed no_of_chars is zero
    if ( no_of_chars > 0 ) then {
      // treat trace_groups as a 2-dimensional array of log_column structs
      row_index = (row_id * num_of_trace_log_columns) + col_id;
      // now fill in specific group info 
      trace_groups[row_index].width = no_of_chars;
      // limit num of chars in column(s) to specified value
      trace_groups[row_index].title = str_exactly( column_header, no_of_chars );
      // create a string that contains exactly the specified number of spaces
      // this will be used to maintain column spacing when log info has not been updated.
      trace_groups[row_index].space = str_pad( " ", no_of_chars );
      // update booleans to show new specified column header
      trace_groups[row_index].new = FALSE;
      trace_title_printed[row_id] = FALSE;
    };

  }; // Register_Trace_Group


  Log_Trace_Group( row_id: int, col_id:int, log_info:string ) is {

    var row_index : uint;

    // treat trace_groups as a 2-dimensional array of log_column structs
    row_index = (row_id * num_of_trace_log_columns) + col_id;
    // now fill in specific group info with updated info
    trace_groups[row_index].info = log_info;
    trace_groups[row_index].new  = TRUE;

  }; // Log_Trace_Group


  // Init method should be called once (for each row-subystem)

  Init_Trace_Group( row_id : uint ) is {

    var first_row_index : uint;
    var final_row_index : uint;
    var row_index       : uint;
    var col_id          : uint;
 
    // initialize all trace_groups structs for the single specified row_id
    trace_clock_count[row_id] = 0;
    // treat trace_groups as a 2-dimensional array of log_column structs
    first_row_index = row_id * num_of_trace_log_columns;
    final_row_index = first_row_index + num_of_trace_log_columns - 1;
    // first column of each row is always reserved for clock count
    trace_groups[first_row_index].row = row_id;
    trace_groups[first_row_index].col = 0;
    trace_groups[first_row_index].new = FALSE;
    trace_groups[first_row_index].width = 10;
    trace_groups[first_row_index].title = appendf(" count :%1d:",row_id);
    trace_groups[first_row_index].info  = appendf("%7d:%1d:",0,row_id);
    trace_groups[first_row_index].space = "..........";
    // rest of columns in each row are user-configurable ...
    first_row_index = first_row_index + 1;
    col_id = 1;
    for row_index from first_row_index to final_row_index do {
       // init each field with a default value
       trace_groups[row_index].row = row_id;
       trace_groups[row_index].col = col_id;
       trace_groups[row_index].new = FALSE;
       trace_groups[row_index].width = 0;
       trace_groups[row_index].title = "+";
       trace_groups[row_index].info  = "+";
       trace_groups[row_index].space = "+";
       // update for next iteration
       col_id = col_id + 1;
    }; // for loop

  }; // Init_Trace_Group


  // Print method should be called once per clock period (for each row-subsystem)

  Print_Trace_Group( row_id : uint ) is {

    var first_row_index : uint;
    var final_row_index : uint;
    var row_index       : uint;

    // always increment count, regardless of enable config
    trace_clock_count[row_id] = trace_clock_count[row_id] + 1;

    // only print if global & specific trace logging is enabled
    if ( master_trace_log_sw and specific_trace_log_sw[row_id] ) then {
      // print the titles just before printing for the first time only
      if ( not trace_title_printed[row_id] ) then {
        Print_Trace_Header (row_id);
        trace_title_printed[row_id] = TRUE;
      }; 
      // this will print exactly the single row that is specified
      first_row_index = row_id * num_of_trace_log_columns;
      final_row_index = first_row_index + num_of_trace_log_columns - 1;
      // always print clock value & row_id in first column
      outf( "%7d:%1d:|", trace_clock_count[row_id], row_id );
      // now print the rest of the columns in this particular row
      first_row_index = first_row_index + 1;
      for row_index from first_row_index to final_row_index do {
         if ( trace_groups[row_index].width == 0 ) then {
           // skip print for all uninitialized columns
         } else if ( trace_groups[row_index].new ) then {
           // print actual info if updated since last clock
           outf( "%s|", trace_groups[row_index].info );
         } else {
           // print spaces only if not updated since last clock
           outf( "%s|", trace_groups[row_index].space );
         };
         // update for next print clock
         trace_groups[row_index].new = FALSE;
      }; // end for loop
      // finish up with a new line
      outf("\n");
    }; // trace log enabled

  }; // Print_Trace_Group


  Print_Trace_Header( row_id : uint ) is {

    // print all of the headers for the single specified row_id
    // note that this cannot be done immediately at start of run time because
    // we must wait for all modules to register their column titles
    for each in trace_groups do {
       // determine if next column is part of this row_id
       if ( ( row_id == it.row ) and ( it.width > 0 ) ) then {
         // print current column header block
         outf( "%s|", it.title );
       };
    }; // for loop
    // end last column with a new line
    outf("\n");

  }; // Print_Trace_Header


}; // TRACE_LOGGER_UNIT



//------------------------------------------------------------------------------
// The string trace logger (requires signal strings that will be probed)
//------------------------------------------------------------------------------

define  NUM_OF_SIGNALS  10;


unit STRING_LOGGER_UNIT {

//  columns: 0     1        2       3        4         5       6        7         8       9       10
//----------+-+--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+
//          |r|        |        |        |        |        |        |        |        |        |        |
//          |e|       * indicates when ready is asserted (in all cases)                                 |
//          |s|                                                                                         |
//  sim     |e|  smem  |  smem  |  smem  |  smem  |  smem  |  smem  |  smem  |  smem  |  emu   | emu    |
//  time    |t| icache0| icache1| icache2| icache3| icache4| icache5|  vbus0 |  vbus1 | write  | read   |
//----------+-+--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+
//1234567:T:| |0:F40000|1:FF8000|2:FF9000|3:FFA000|4:FFB000|5:FFC000|V:F40000|        |W:123456|R:123456|
//1021720:T:| |        |        |        |        |4*......|        |        |        |        |        |
//1021740:T:| |0*......|        |        |        |        |5*......|        |        |        |        |
//1021760:T:| |        |        |        |3*......|        |        |        |        |        |        |
//1021780:T:| |        |        |2*......|3:FF8000|        |        |        |        |        |        |
//1021800:T:| |        |1*......|        |3:......|        |        |        |        |        |        |
//1021820:T:| |        |        |        |3*......|        |        |V:FFB000|        |        |        |
//1021840:T:| |        |        |        |        |        |        |V:......|        |        |        |
//1021860:T:| |        |        |        |        |        |        |V*......|        |        |        |
//1021880:T:| |        |        |        |        |        |        |        |v:F50000|        |        |


  // revision history of String_Logger
  // v1.0  original version
  // v1.1  added reset column
  //

  string_log_version : string;   keep      string_log_version  == "v1.1";
  string_log_enable  : bool;     keep soft string_log_enable   == FALSE;
  string_display_time: bool;     keep soft string_display_time == TRUE;
  string_clock_count : uint;     keep soft string_clock_count  == 0;

  string_signal_port  : list of simple_port of uint is instance;
  string_request_port : list of simple_port of uint is instance;
  string_ready_port   : list of simple_port of uint is instance;
     keep soft string_signal_port.size()  == NUM_OF_SIGNALS;
     keep soft string_request_port.size() == NUM_OF_SIGNALS;
     keep soft string_ready_port.size()   == NUM_OF_SIGNALS;

  string_reset_port   : simple_port of uint is instance;


  // Instructions to add signal values and new columns
  // --------------------------------------------------------
  // a) modify the NUM_OF_SIGNALS define above
  // b) add signal names to list(s) in Init_String_Logger()
  // c) add logic to Print_String_Log to display the new signal values
  // d) add column titles to Init_String_Logger()
  // e) modify (increment) string_log_version
       


  // this method needs to be called at every rising clock edge
  Print_String_Log ( ) is {
 
    var address  : uint;
    var rdy_char : string;
    var bus_char : string;

    // always increment count, regardless of enable config
    string_clock_count = string_clock_count + 1;
    // perform rest of stuff only if enabled
    if ( string_log_enable ) then {

      // configurable to display sim time or clock cycle count
      if ( string_display_time ) then {
        outf( "%7d:T:|", sys.time );
      } else {
        outf( "%7d:C:|", string_clock_count );
      };

      // print 1 column reset status
      if ( string_reset_port$ == 0 ) then {
        outf( "R|" );
      } else {
        outf( ".|" );
      };

      // first 6 ports are icache ports (8 chars wide)
      for i from 0 to 5 do {
        // indicate when ready (active low) 
        if ( string_ready_port[i]$ == 0 ) then {
          rdy_char = "*";
        } else {
          rdy_char = ":";
        };
        // only log address when request is active (active low)
        if ( string_request_port[i]$ == 0 ) then {
          // fix smem addresses due to missing 2 bits 
          address = string_signal_port[i]$ * 4;
          outf("%d%s%6X|", i, rdy_char, address );
        } else {
          outf("%d%s......|", i, rdy_char);
        };
      }; // end for loop

      // next 2 ports are vbus ports
      for i from 6 to 7 do {
        // indicate when ready (active high)
        if ( string_ready_port[i]$ == 1 ) then {
          rdy_char = "*";
        } else {
          rdy_char = ":";
        };
        // only have 1 char to differentiate between two vbus ports
        if ( i == 6 ) then {
          bus_char = "V";
        } else {
          bus_char = "v";
        };
        // only log address when request is active (active high)
        if ( string_request_port[i]$ == 1 ) then { 
          address = string_signal_port[i]$;
          outf("%s%s%6X|", bus_char, rdy_char, address );
        } else {
          outf("%s%s......|", bus_char, rdy_char);
        };
      };// end for loop
   
      // last 2 ports are emu ports
      for i from 8 to 9 do {
        // indicate when ready (active low)
        if ( string_ready_port[i]$ == 0 ) then {
          rdy_char = "*";
        } else {
          rdy_char = ":";
        };
        // only have 1 char to differentiate between two emu ports
        if ( i == 8 ) then {
          bus_char = "W";
        } else {
          bus_char = "R";
        };
        // only log address when request is active (active low)
        if ( string_request_port[i]$ == 0 ) then { 
          address = string_signal_port[i]$;
          outf("%s%s%6X|", bus_char, rdy_char, address );
        } else {
          outf("%s%s......|", bus_char, rdy_char);
        };
      }; // end for loop
   
      // always finish the line
      outf("\n");

    }; // end if enabled   
 
  }; // Print_String_Log


  // this method needs to be called only once at init time
  Init_String_Logger ( ) is {

    var top_title : string;
    var mid_title : string;
    var bot_title : string;

   // print log version
   outf ("%10d SYS_LOG Gx: string logger version %s\n", sys.time, string_log_version );

   // print column titles
   if ( string_display_time ) then {
     top_title = "  sim  :T:";
     mid_title = "  time :T:";
     bot_title = "-------:T:";
   } else {
     top_title = " clock :C:";
     mid_title = " count :C:";
     bot_title = "-------:C:";
   };

   if ( string_log_enable ) then {
     //   these are the column headers that will appear in the sim log -------------------------------- 1 
     //   10            2         3         4         5         6         7         8         9         0
     //        123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123
     outf("%10s|r|  smem  |  smem  |  smem  |  smem  |  smem  |  smem  |  smem  |  smem  |  smem  |  smem  |\n",top_title);
     outf("%10s|s| icache0| icache1| icache2| icache3| icache4| icache5|  vbus0 |  vbus1 | emu wrt| emu rd |\n",mid_title);
     outf("%10s|-+--------+--------+--------+--------+--------+--------+--------+--------+--------+--------+\n",bot_title);
   };

   // initialize the signal names for the generic string logger
// string_reset_port.hdl_path()      = "topi0/corei0/smemi0/GLOBAL_RESET_N_PI";
  
// gen string_signal_port.hdl_path() keeping { .size() == NUM_OF_SIGNALS; };
// string_signal_port[0].hdl_path()  = "topi0/corei0/smemi0/CACHE_0_ADDRESS_PI";
// string_signal_port[1].hdl_path()  = "topi0/corei0/smemi0/CACHE_1_ADDRESS_PI";
// string_signal_port[2].hdl_path()  = "topi0/corei0/smemi0/CACHE_2_ADDRESS_PI";
// string_signal_port[3].hdl_path()  = "topi0/corei0/smemi0/CACHE_3_ADDRESS_PI";
// string_signal_port[4].hdl_path()  = "topi0/corei0/smemi0/CACHE_4_ADDRESS_PI";
// string_signal_port[5].hdl_path()  = "topi0/corei0/smemi0/CACHE_5_ADDRESS_PI";
// string_signal_port[6].hdl_path()  = "topi0/corei0/smemi0/VBUSP_0_ADDRESS_PI";
// string_signal_port[7].hdl_path()  = "topi0/corei0/smemi0/VBUSP_1_ADDRESS_PI";
// string_signal_port[8].hdl_path()  = "topi0/corei0/smemi0/EMU_WR_ADDR_PI";
// string_signal_port[9].hdl_path()  = "topi0/corei0/smemi0/EMU_RD_ADDR_PI";

// gen string_request_port.hdl_path() keeping { .size() == NUM_OF_SIGNALS; };
// string_request_port[0].hdl_path() = "topi0/corei0/smemi0/CACHE_REQ_N_PI[0]";
// string_request_port[1].hdl_path() = "topi0/corei0/smemi0/CACHE_REQ_N_PI[1]";
// string_request_port[2].hdl_path() = "topi0/corei0/smemi0/CACHE_REQ_N_PI[2]";
// string_request_port[3].hdl_path() = "topi0/corei0/smemi0/CACHE_REQ_N_PI[3]";
// string_request_port[4].hdl_path() = "topi0/corei0/smemi0/CACHE_REQ_N_PI[4]";
// string_request_port[5].hdl_path() = "topi0/corei0/smemi0/CACHE_REQ_N_PI[5]";
// string_request_port[6].hdl_path() = "topi0/corei0/smemi0/VBUSP_0_REQUEST_PI";
// string_request_port[7].hdl_path() = "topi0/corei0/smemi0/VBUSP_1_REQUEST_PI";
// string_request_port[8].hdl_path() = "topi0/corei0/smemi0/EMU_WR_REQ_N_PI";
// string_request_port[9].hdl_path() = "topi0/corei0/smemi0/EMU_RD_REQ_N_PI";

// gen string_ready_port.hdl_path() keeping { .size() == NUM_OF_SIGNALS; };
// string_ready_port[0].hdl_path()   = "topi0/corei0/smemi0/CACHE_RDY_N_POR[0]";
// string_ready_port[1].hdl_path()   = "topi0/corei0/smemi0/CACHE_RDY_N_POR[1]";
// string_ready_port[2].hdl_path()   = "topi0/corei0/smemi0/CACHE_RDY_N_POR[2]";
// string_ready_port[3].hdl_path()   = "topi0/corei0/smemi0/CACHE_RDY_N_POR[3]";
// string_ready_port[4].hdl_path()   = "topi0/corei0/smemi0/CACHE_RDY_N_POR[4]";
// string_ready_port[5].hdl_path()   = "topi0/corei0/smemi0/CACHE_RDY_N_POR[5]";
// string_ready_port[6].hdl_path()   = "topi0/corei0/smemi0/VBUSP_0_READY_PO";
// string_ready_port[7].hdl_path()   = "topi0/corei0/smemi0/VBUSP_1_READY_PO";
// string_ready_port[8].hdl_path()   = "topi0/corei0/smemi0/EMU_WR_RDY_N_POR";
// string_ready_port[9].hdl_path()   = "topi0/corei0/smemi0/EMU_RD_RDY_N_POR";

  } ; // Init_String_Logger

  // perform required initialization
  run() is first {
     Init_String_Logger();
  };

}; // STRING_LOGGER_UNIT




//------------------------------------------------------------------------------
// Both Loggers defined ... now create 1 each instances of each (Trace & String)
//------------------------------------------------------------------------------

extend sys {

  // instantiate string logger at top system level
  string_logger : STRING_LOGGER_UNIT is instance;

  // instantiate trace logger at top system level
  trace_logger  : TRACE_LOGGER_UNIT  is instance;
  keep soft trace_logger.num_of_trace_log_lines   == NUM_OF_TRACE_LOG_LINES;
  keep soft trace_logger.num_of_trace_log_columns == NUM_OF_TRACE_LOG_COLUMNS;

};
   

'>
