//***********************************************************************
//*  FILE       : watch.e
//*  BY         : Jerel Canales
//*  PURPOSE    : This watch file has 2 functions. Both of these
//*               functions "watch" for situations that could stop the test.
//*
//*               1) provide a mechanism for BFMs and test methods to
//*                  signal for proper test stoppage. This is implemented
//*                  by having all BFMs register with this module at the 
//*                  beginning of a test (promise) and then reporting done
//*                  when complete (fulfilling the promise). This module
//*                  will stop the test when all BFMs have reported done.    
//*               2) keep track of test clock cycles and optionally stop
//*                  the test when a watchdog limit has been exceeded.
//*         
//*
//*  06/10/01     jerelc   Initial code
//*  09/01/01     jerelc   added groups for multiple priorities 
//*                        and wait_for_master & report_done methods
//*  09/05/01     spyra    added a watchdog timer feature
//*  09/07/01     spyra    added bfm name to log messages
//*  10/25/01     spyra    changed log calls to outf so that system_logger 
//*                        is not required to be imported first.
//*
//*
//***********************************************************************

// CONTENTS (you can search for these words to jump to the appropriate section):
//     USAGE
//     EXAMPLE
//     METHODS
//     DECLARATIONS
//     ACTUAL_CODE


// USAGE: Instantiate the watch_s structure as a field in the top level
//        testbench.  Each module should register with watch_s during test
//        generation and report done when all of its stimulus are complete.
//        The term "module" can be any entity (clead BFM or stim TCM) that 
//        controls a test sequence and therefore knows when the test may be 
//        complete. Note that this module has been recently updated with a
//        a set of methods that allow for a more sophisticated manner of
//        stopping the test as well as better (more detailed) logging.
//        Therefore try to use the "new methods" as opposed to the "old methods".


// METHODS:
//   (old method)   register( delay:uint ) : uint
//                  inputs: delay specifies how many clocks to wait before
//                          ending the test after all methods have reported done.
//                  output: returns a uint that specifies the id of registered method.
//
//   (new method)   register_watch( myname:string, level:int, done_delay:uint) : uint
//                  inputs: myname is a string description used for logging 
//                          level is the dependancy level (priority)
//                          clead BFMs should use 0, test TCMs should use 1
//                          delay is same as old method
//                  output: returns a uint that specifies the id of registered method.
//
//   (old method)   done( id:uint )
//                  inputs: id that identifies which method is complete
//                  output: none
//
//   (new method)   report_done( level:uint, id:uint )
//                  inputs: level is the dependancy level (priority)
//                          id that identifies which method is complete
//                  output: none
//
//   (new only)     Set_Watchdog_Timer( value:uint )
//                  inputs: max number of clocks that the test is expected to take
//                  output: none
//
    

// EXAMPLE:
//     extend TESTBENCH {
//       stop_run_manager : watch_s;
//     };
// 
//     define   STOP_RUN   sys.f751710_tb.stop_run_manager;  
//
//     extend YOUR_TEST_METHOD {
//
//        id  : uint; keep id == STOP_RUN.register_watch( my_test_method_name, 1, 10);
//        
//        done_flag : bool; keep done_flag == FALSE;
//    
//        done_tcm()@sys.any is {
//          STOP_RUN.Set_Watchdog_Timer(500);
//          wait true (done_flag);
//          STOP_RUN.report_done(1,id);
//        };
//     };
    
<'

////////////////////////////////////////////////////////
//
// #    #    ##     #####   ####   #    #           ####
// #    #   #  #      #    #    #  #    #          #
// #    #  #    #     #    #       ######           ####
// # ## #  ######     #    #       #    #               #
// ##  ##  #    #     #    #    #  #    #          #    #
// #    #  #    #     #     ####   #    # #######   ####
//
/////////////////////////////////////////////////////////


// structure used for each BFM that registers
// every BFM that registers must report done at some point.
// a priority value was introduced so that all higher priority
// BFMs must finish before the lower priority ones can.
// In Janus, 2 priority levels will be used, and when the
// higher priority BFMs are done, a 1st handshake will take
// place. Then all lower priority BFMs can finish up and a
// 2nd handshake will take place (and the test will end)
    
struct group_s {

   !done_flag   : list of bool;
   !module_name : list of string;
   !delay       : list of uint;

};


define STOP_RUN   sys.stop_run_manager;

struct watch_s {

// DECLARATIONS:
   
   clk        : clock_stub;
   event fclk is @clk.raw_fall;
   event rclk is @clk.raw_rise;
   
   !groups    : list of group_s;
   max_delay  : uint; keep soft max_delay == 0;
   
   max_watchdog_count : uint; keep soft max_watchdog_count == 0;
   cur_watchdog_count : uint; keep soft cur_watchdog_count == 0;

   log_enable         : bool; keep soft log_enable         == TRUE;
   all_bfms_look_done : bool; keep soft all_bfms_look_done == FALSE;

// ACTUAL_CODE:
  
   //-------------------------------------
   // Used to set optional watchdog timer
   //-------------------------------------
   Set_Watchdog_Timer( value : uint ) is {
      if ( log_enable ) then {
        outf("%10d  NOTE  Watchdog timer's max count just set to %d \n", sys.time, value );
      }; 
      max_watchdog_count = value;     
   };


   //----------------------------------
   // Check Test Watchdog counter 
   //----------------------------------   
   
   Test_Watchdog_Counter() @rclk is {
      
      while TRUE {
        // always count clocks
        cur_watchdog_count = cur_watchdog_count + 1;
        // do not make any checks if watchdog mechansim is disabled      
        if ( max_watchdog_count == 0 ) {
          // 0 means disabled
          wait[1];
        } else {
          // otherwise enabled
          if ( cur_watchdog_count > max_watchdog_count ) {
            // watchdog exceeded, stop test now
            dut_error( "FATAL: ",appendf("Watchdog timer expired at count %d \n", cur_watchdog_count) );
            stop_run();
          } else {
            // watchdog not exceeded, keep waiting    
            wait[1];
          };
        }; 
      };
      
   }; // Test_Watchdog_Counter


   //----------------------------------
   // -----  Old Method  -----
   // BFMs register() during test gen;
   // pass in the # of clocks to delay
   //----------------------------------
   register( done_delay : uint ) : uint is {

     // any module that uses this old method gets group = 1 and a made up name
     return( register_watch( "Generic_BFM_Module", 1, done_delay ) );

   }; // register
   
   
   //----------------------------------
   // -----  New Method  -----
   // BFMs register() during test gen;
   // parameters passed in:
   //  1) module name string
   //  2) dependacy_level (priority)
   //  3) # of clocks to delay
   //----------------------------------
   register_watch(myname : string, level : int, done_delay : uint) : uint is {

     var max_group_index : int;
     var reg_id          : uint;
 
     // figure out how many groups currently exist
     max_group_index = groups.size() - 1;
     // must build enough groups to go from 0 to level
     while ( level > max_group_index ) do {
        // not enough groups yet
        var new_group : group_s;
        gen new_group;
        // create a new group
        groups.add(new_group);
        // keep track of how many so far 
        max_group_index = groups.size() - 1;
     }; 

     // 1. Register name string
     groups[level].module_name.push(myname);
     // 2. Register done flag
     groups[level].done_flag.push(FALSE);
     // 3. Register delay value
     groups[level].delay.push(done_delay);

     // log and return id 
     reg_id = groups[level].done_flag.size() - 1;
     if ( log_enable ) then {
       outf("%10d  NOTE  %s BFM %d of group %d just registered with Watch.e \n", sys.time, myname, reg_id, level );
     }; 
     return( reg_id );

   }; // register_watch


   //---------------------------------
   // -----  Old Method  -----
   // BFMs report done() when complete
   // only id must be passed in.
   //--------------------------------- 
   done( id : uint ) is {
     
     // any module that uses this old method assumes group = 1
     report_done( 1, id );
   }; // done

   //----------------------------------
   // -----  New Method  -----
   // BFMs report done() when complete
   // both group & id must be passed in
   //----------------------------------
   report_done( level : uint, id : uint ) is {

      var myname : string;

      // log which BFM reported done
      myname = groups[level].module_name[id];
      if ( log_enable ) then {
       outf("%10d  NOTE  %s BFM %d of group %d just reported that it is done \n", sys.time, myname, id, level );
      }; 
      // and simply set appropriate flag
      groups[level].done_flag[id] = TRUE;
   }; // report_done
   

   on fclk {
      start wait_for_bfms();
   };
   
   //----------------------------------
   // Stop sim when all BFMs are done 
   //----------------------------------   
   wait_for_bfms ()@fclk is {

      var found_bfm_not_done := FALSE;
      var num_of_groups      := groups.size();
      var num_of_bfms        : uint;

      // this TCM will run until all registered bfms are done

      if not all_bfms_look_done then {
   
         // must go thru all registered groups 
         for i from 1 to num_of_groups do {
            // must go thru all registered bfms 
            num_of_bfms = groups[i-1].done_flag.size(); 
            for j from 1 to num_of_bfms do {
               if ( groups[i-1].done_flag[j-1] == FALSE ) then {
                  // as soon as 1 bfm is not done, exit the loop
                  found_bfm_not_done = TRUE;
                  break;
               };
            }; // j loop for groups
         };   // i loop for bfms
      
         // check if all BFMs are done
         all_bfms_look_done = not found_bfm_not_done;

         if all_bfms_look_done then {

            // wait at least 1 clock (maybe more)
            max_delay = 1;
            for each in groups do {
               max_delay = ((max_delay < groups[index].delay.max_value(it)) ? groups[index].delay.max_value(it) : max_delay);
            };
         };
      };
   }; // wait_for_bfms
   
   event wait_e is true(all_bfms_look_done)@fclk;
   
   on wait_e {
      if max_delay==0 then {
         stop_run();  
      }
      else {
         max_delay -= 1; 
      };
   };


   //---------------------------------------------------------------------
   // used to wait until all BFMs in next higher priority group are done
   // this way, the lower priority BFMs will know when they should stop.
   //---------------------------------------------------------------------

   wait_for_master( level : uint) @fclk is {

      var master             : uint;
      var all_bfms_are_done  : bool;
      var found_bfm_not_done : bool;
      var num_of_bfms        : uint;

      // the higher the level, the higher the priority
      master = level + 1;
      // return immediately if no higher group exists 
      if ( groups.size() > master ) then {
         // wait until all bfms in the next higher pri group to complete
         all_bfms_are_done = FALSE;
         while not ( all_bfms_are_done ) do {
           // this TCM will run until all registered bfms are done
           found_bfm_not_done = FALSE;
           num_of_bfms = groups[master].done_flag.size(); 
           // determine if all BFMs are done
           for j from 1 to num_of_bfms do {
             if ( groups[master].done_flag[j-1] == FALSE ) then {
               // as soon as 1 bfm is not done, exit the loop
               found_bfm_not_done = TRUE;
               break;
             };
           }; 
           // check if all BFMs are done
           if ( found_bfm_not_done == FALSE ) then {
             // loop is done, set flag
             all_bfms_are_done = TRUE;
             if ( log_enable ) then {
               outf("%10d  NOTE  Level %d is complete due to master complete \n", sys.time, level );
             }; 
           } else {        
             // loop not done yet, try again
             wait[1];
           };
         }; // while loop
      } else {
        if ( log_enable ) then {
          outf("%10d  NOTE  Level %d is complete due to no master \n", sys.time, level );
        }; 
      };
      
   }; // wait_for_master



   //-------------------------------------
   // start up the TCMs in this module
   //-------------------------------------
   
   run() is also {
      start Test_Watchdog_Counter(); 
   };
   
};//end struct watch_s

extend sys {
   
   stop_run_manager : watch_s;  
    
};


'>

/////////////////////////////////////////////////////////////////////////
//
//  #######
//     #     #    #  ######          ######  #    #  #####
//     #     #    #  #               #       ##   #  #    #
//     #     ######  #####           #####   # #  #  #    #
//     #     #    #  #               #       #  # #  #    #
//     #     #    #  #               #       #   ##  #    #
//     #     #    #  ######          ######  #    #  #####
//
/////////////////////////////////////////////////////////////////////////

