//********************************************************************************
//
//  Filename: cpubus_protocol_checker.e
//
//  Description: cpu bus protocol checker
//
//  Major Revision Descriptions:
//  oishi          09/08/02         copy from UMA1.0 branch
//  oishi          09/09/02         removed bottom signals
//                                  removed mdp related functions
//                                  removed ack related functions
//                                  removed data dismiss related functions
//                                  removed stall signals
//  oishi          09/13/02         modified resetn to cpu_resetn and gl_resetn
//  oishi          10/04/02         added checking function of signal connection.
//  oishi          10/21/02         modified address extension checker
//                                  added simple lock checker
//  oishi          11/11/02         updated address bus checking
//  oishi          11/14/02         corrected address holding checker
//  oishi          11/15/02         added byte ready qualifier checking
//  oishi          12/05/02         enable qualifier checking for double/byte
//                                  added address checking for lock access
//  oishi          01/21/03         byte qualifier checking is changed 'warning' from 'error'.
//                                  added rule that byte access is not allowed when dual access.
//                                  clear the local valiable when abortready/reset is issued.
//  hiraoka        03/18/03         comment-out TCM_Error_CheckIcemakerReady0
//                                  On NON debug mode, IceMaker Memory Mapped area access may happen.
//  hiraoka        03/27/03         add dblcoef qualifier checking
//  hiraoka        05/21/03         add Debug access vs P-dismiss
//*********************************************************************************
<'

struct CPUBUS_PROTOCOL_CHECKER {
    -- enable/disable flag --
    disable  : bool; 
    keep soft disable == FALSE;
    
    -- timeout limitation --
    timeout  : uint;
    keep soft timeout == 200;
    
    -- global configurations --
    config   : CPUBUS_CONFIGURATIONS;
    paths    : CPU_SIGNAL_PATH;
    cpu      : CPU_BUSES;
    
    prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
    data_attributes : DATA_ACCESS_ATTRIBUTES;
    
    -- set clock event for state transitions
    event raw_rise_clock is rise(paths.clock$)    @sim;
    event rise_clock     is true( disable == FALSE ) @raw_rise_clock;
    
    post_generate() is also {
        
        -- check signal connection --
        paths.check_signal_connection( config.id );
        
    };
    
    ------------------------------
    -- cpu bus state transition --
    ------------------------------
    on rise_clock
    {
        -- keep current values
        cpu.keep_current_values();
        
        -- get current value
        cpu.get_current_value( config, paths );
        
        -- state transition
        cpu.state_transition();
        
        -- set current request attributes --
        if( cpu.ocuur_program_request() ) {
            gen prog_attributes;
            prog_attributes.set_current_attributes( cpu );
        };
        if( cpu.ocuur_data_write_requests() or cpu.ocuur_data_read_requests() ) {
            gen data_attributes;
            data_attributes.set_current_attributes( cpu );
        };

        -- for keeping pending requests --
        cpu.pending_prog_requests( prog_attributes );
        cpu.pending_data_requests( data_attributes, data_attributes );
        
        -- state initialize when reset 
        if( cpu.system.current.cpu_resetn == 0 ) {
            cpu.set_reset_values();
        };
    };
    
    ------------------------------
    -- check for protocol rules --
    ------------------------------
    
    -- ERROR : check state transition error
    TCM_Error_CheckForStateTransittion() @rise_clock is {
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {
                
                if( cpu.pbus.state.current == error ) {
                    dut_error( "Error at ",sys.time," : Pbus has state transition error.");
                    outf("Error Info : Previous state %s : Current request = %1d ready = %1d \n",
                          cpu.pbus.state.previous, cpu.pbus.current.req, cpu.pbus.current.ready );
                };
                if( cpu.bbus.state.current == error ) {
                    dut_error( "Error at ",sys.time," : Bbus has state transition error.");
                    outf("Error Info : Previous state %s : Current request = %1d ready = %1d \n",
                          cpu.bbus.state.previous, cpu.bbus.current.req, cpu.bbus.current.ready );
                };
                if( cpu.cbus.state.current == error ) {
                    dut_error( "Error at ",sys.time," : Cbus has state transition error.");
                    outf("Error Info : Previous state %s : Current request = %1d ready = %1d \n",
                          cpu.cbus.state.previous, cpu.cbus.current.req, cpu.cbus.current.ready );
                };
                if( cpu.dbus.state.current == error ) {
                    dut_error( "Error at ",sys.time," : Dbus has state transition error.");
                    outf("Error Info : Previous state %s : Current request = %1d ready = %1d \n",
                          cpu.dbus.state.previous, cpu.dbus.current.req, cpu.dbus.current.ready );
                };
                if( cpu.ebus.state.current == error ) {
                    dut_error( "Error at ",sys.time," : Ebus has state transition error.");
                    outf("Error Info : Previous state %s : Current request = %1d ready = %1d \n",
                          cpu.ebus.state.previous, cpu.ebus.current.req, cpu.ebus.current.ready );
                };
                if( cpu.fbus.state.current == error ) {
                    dut_error( "Error at ",sys.time," : Fbus has state transition error.");
                    outf("Error Info : Previous state %s : Current request = %1d ready = %1d \n",
                          cpu.fbus.state.previous, cpu.fbus.current.req, cpu.fbus.current.ready );
                };
            };
            
            wait cycle;
        };
    };
    
    -- ERROR : check for cpu_ioXreq_tr signals transitions
    -- The request signals as ioreq and req doesn't ocuur at same time in same bus as following example.
    --                   _______      _______       ____________________
    --     cpu_Xreq_tr          |____|                                  
    --                   ____________________   or  _______      _______
    --     cpu_ioXreq_nr                                   |____|       
    --                                                                  
    TCM_Error_CheckForSignal_ioreq0() @rise_clock is {
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {
                
                if( cpu.dbus.current.req == 0 and  cpu.dbus.current.ioreq == 0 ) {
                    dut_error( "Error at ",sys.time," : Both requests as iodreq and dreq are set '0', this is illegal protocol.");
                };
                if( cpu.ebus.current.req == 0 and  cpu.ebus.current.ioreq == 0 ) {
                    dut_error( "Error at ",sys.time," : Both requests as ioereq and ereq are set '0', this is illegal protocol.");
                };
            };
            
            wait cycle;
        };
    };
    
    -- ERROR : check for double qualifier transitions
    -- double qualifiers as cpu_doubler_tr/cpu_doublew_tr should be asserted 
    -- coincident with assertion of the request signals as cpu_dreq_nr/cpu_ereq_nr.
    -- and creq/freq should not be assert.
    TCM_Error_CheckForSignals_double_qualifier0() @rise_clock is {
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {
                
                if( cpu.qualifier.current.doubler == 1 ) {
                    check that ( cpu.dbus.current.req == 0 and cpu.dbus.current.ioreq == 1 )
                        else dut_error( "Error at ",sys.time," : cpu_doubler_tr should be asserted coincident with cpu_dreq_nr.");
                    check that ( cpu.cbus.current.req == 1 )
                        else dut_error( "Error at ",sys.time," : creq signal can't assert when double read request.");
                };
                if( cpu.qualifier.current.doublew == 1 ) {
                    check that ( cpu.ebus.current.req == 0 and cpu.ebus.current.ioreq == 1 )
                        else dut_error( "Error at ",sys.time," : cpu_doublew_tr should be asserted coincident with cpu_ereq_nr.");
                    check that ( cpu.fbus.current.req == 1 )
                        else dut_error( "Error at ",sys.time," : freq signal can't assert when double write request.");
                };

                if( cpu.qualifier.current.dblcoef == 1 ) {
                    check that ( cpu.bbus.current.req == 0 )
                        else dut_error( "Error at ",sys.time," : cpu_dblcoef_tr should be asserted coincident with cpu_breq_nr.");
                };                

            };
            wait cycle;
        };
    };
    
    -- Error : check for byte qualifier transitions
    -- when dual access, byte access is not allowed.
    TCM_Error_CheckForSignals_byte_qualifier0() @rise_clock is {
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {
                
                if( cpu.qualifier.current.wrbyte == 1 ) {
                    check that ( cpu.fbus.current.req == 1 )
                        else dut_error( "Error at ",sys.time," : When dual access, byte access does not allow.");
                };
                if( cpu.qualifier.current.rdbyte == 1 ) {
                    check that ( cpu.cbus.current.req == 1 )
                        else dut_error( "Error at ",sys.time," : When dual access, byte access does not allow.");
                };
                
            };
            
            wait cycle;
        };
    };
    
    -- WARNING : check for byte qualifier transitions
    -- byte qualifiers as cpu_wrbyte_tr/cpu_wrlsbe_tr(cpu_rdbyte_tr/cpu_rdlsbe_tr)
    -- should be asserted coincident with assertion of the request signals as 
    -- cpu_(io)ereq_nr(cpu_(io)dreq_nr).
    -- Note that it is allowed to access for mmr area with byte qualifier,
    -- in this cases, only byte qualifier is issued without request. 
    -- So this rule is categorized 'Warning'.
    TCM_Warning_CheckForSignals_byte_qualifier0() @rise_clock is {
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {
                
                if( cpu.qualifier.current.wrbyte == 1 ) {
                    check that ( cpu.ebus.current.req == 0 or cpu.ebus.current.ioreq == 0 )
                        else dut_error( "Warning at ",sys.time," : cpu_wrbyte_tr should be asserted coincident with cpu_ereq_nr.");
                };
                if( cpu.qualifier.current.wrlsbe == 1 ) {
                    check that ( cpu.ebus.current.req == 0 or cpu.ebus.current.ioreq == 0 )
                        else dut_error( "Warning at ",sys.time," : cpu_wrlsbe_tr should be asserted coincident with cpu_ereq_nr.");
                };
                
                if( cpu.qualifier.current.rdbyte == 1 ) {
                    check that ( cpu.dbus.current.req == 0 or cpu.dbus.current.ioreq == 0 )
                        else dut_error( "Warning at ",sys.time," : cpu_rdbyte_tr should be asserted coincident with cpu_dreq_nr.");
                };
                if( cpu.qualifier.current.rdlsbe == 1 ) {
                    check that ( cpu.dbus.current.req == 0 or cpu.dbus.current.ioreq == 0 )
                        else dut_error( "Warning at ",sys.time," : cpu_rdlsbe_tr should be asserted coincident with cpu_dreq_nr.");
                };
                
            };
            
            wait cycle;
        };
    };
    
    -- ERROR : check for address changing
    -- address bus should changed only when request signal is asserted.
    !keep_pbus_counter : uint;
    !keep_bbus_counter : uint;
    !keep_cbus_counter : uint;
    !keep_dbus_counter : uint;
    !keep_ebus_counter : uint;
    !keep_fbus_counter : uint;
    
    !keep_rslot0 : DATA_ACCESS_ATTRIBUTES;
    !keep_rslot1 : DATA_ACCESS_ATTRIBUTES;
    !keep_wslot0 : DATA_ACCESS_ATTRIBUTES;
    !keep_wslot1 : DATA_ACCESS_ATTRIBUTES;
    
    init() is also {
        keep_pbus_counter = 0;
        keep_bbus_counter = 0;
        keep_cbus_counter = 0;
        keep_dbus_counter = 0;
        keep_ebus_counter = 0;
        keep_fbus_counter = 0;
        
        keep_rslot0 = NULL;
        keep_rslot1 = NULL;
        keep_wslot0 = NULL;
        keep_wslot1 = NULL;
    };
    
    TCM_Error_CheckForBus_AddressChange0() @rise_clock is {
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {
                
                -- check for pabus --
                if( cpu.pbus.pending.counter > 0 and cpu.pbus.current.req != 0 ) {
                     check that ( cpu.pbus.previous.abus == cpu.pbus.current.abus )
                        else dut_error( "Error at ",sys.time," : pabus address should not be changed in this cycle.");
                };
                
                -- check for babus --
                if( keep_bbus_counter == 2 ) {
                    check that ( cpu.bbus.previous.abus == cpu.bbus.current.abus )
                        else dut_error( "Error at ",sys.time," : babus address should not be changed in this cycle.");
                };
                
                -- check for cabus --
                if( keep_cbus_counter == 2 ) {
                    check that ( cpu.cbus.previous.abus == cpu.cbus.current.abus )
                        else dut_error( "Error at ",sys.time," : cabus address should not be changed in this cycle.");
                };
                
                -- check for dabus --
                if( keep_dbus_counter == 2 ) {
                    check that ( cpu.dbus.previous.abus == cpu.dbus.current.abus )
                        else dut_error( "Error at ",sys.time," : dabus address should not be changed in this cycle.");
                };
                
                -- check for eabus --
                if( keep_ebus_counter == 2 ) {
                    check that ( cpu.ebus.previous.abus == cpu.ebus.current.abus )
                        else dut_error( "Error at ",sys.time," : eabus address should not be changed in this cycle.");
                };
                
                -- check for fabus --
                if( keep_fbus_counter == 2 ) {
                    check that ( cpu.fbus.previous.abus == cpu.fbus.current.abus )
                        else dut_error( "Error at ",sys.time," : fabus address should not be changed in this cycle.");
                };
                
                -- address extension check --
                if( keep_rslot0 != NULL and keep_rslot1 != NULL ) {
                    
                    if( keep_rslot0.bbus_active == active ) {
                        
                        if( keep_cbus_counter != 2 and 
                            keep_rslot1.cbus_active != no_active and 
                            keep_rslot1.doubler     == word          ) { 
                            check that ( cpu.cbus.previous.abus == cpu.cbus.current.abus )
                                else dut_error( "Error at ",sys.time," : cabus address should not be changed in this cycle.");
                        };
                        if( keep_dbus_counter != 2 and 
                            keep_rslot1.dbus_active != no_active     ) { 
                            check that ( cpu.dbus.previous.abus == cpu.dbus.current.abus )
                                else dut_error( "Error at ",sys.time," : dabus address should not be changed in this cycle."); 
                       };
                       
                    };
                    
                    if( keep_rslot0.cbus_active == active and keep_rslot0.doubler == word     ) {
                        
                        if( keep_bbus_counter != 2 and 
                            keep_rslot1.bbus_active != no_active     ) { 
                            check that ( cpu.bbus.previous.abus == cpu.bbus.current.abus )
                                else dut_error( "Error at ",sys.time," : babus address should not be changed in this cycle.");
                        };
                        if( keep_dbus_counter != 2 and 
                            keep_rslot1.dbus_active != no_active     ) { 
                            check that ( cpu.dbus.previous.abus == cpu.dbus.current.abus )
                                else dut_error( "Error at ",sys.time," : dabus address should not be changed in this cycle."); 
                        };
                        
                    };
                    
                    if( keep_rslot0.dbus_active == active ) {
                        
                        if( keep_dbus_counter != 2 and 
                            keep_rslot1.cbus_active != no_active and 
                            keep_rslot1.doubler     == word       ) { 
                            check that ( cpu.cbus.previous.abus == cpu.cbus.current.abus )
                                else dut_error( "Error at ",sys.time," : cabus address should not be changed in this cycle."); 
                        };
                        if( keep_bbus_counter != 2 and 
                            keep_rslot1.bbus_active != no_active     ) { 
                            check that ( cpu.bbus.previous.abus == cpu.bbus.current.abus )
                                else dut_error( "Error at ",sys.time," : babus address should not be changed in this cycle.");
                        };
                        
                    };
                    
                };
                
                if( keep_wslot0 != NULL and keep_wslot1 != NULL ) {
                    
                    if( keep_wslot0.ebus_active == active ) {
                        if( keep_fbus_counter != 2 and 
                            keep_wslot1.fbus_active == active and 
                            keep_wslot1.doublew     == word       ) { 
                            check that ( cpu.fbus.previous.abus == cpu.fbus.current.abus )
                                else dut_error( "Error at ",sys.time," : fabus address should not be changed in this cycle.");
                        };
                    };
                    
                    if( keep_wslot0.fbus_active == active and keep_wslot0.doublew == word ) {
                        if( keep_ebus_counter != 2 and 
                            keep_wslot1.ebus_active != no_active ) {
                            check that ( cpu.ebus.previous.abus == cpu.ebus.current.abus )
                                else dut_error( "Error at ",sys.time," : eabus address should not be changed in this cycle.");
                        };
                    };
                    
                };
                
            };
            
            keep_pbus_counter = cpu.pbus.pending.counter;
            keep_bbus_counter = cpu.bbus.pending.counter;
            keep_cbus_counter = cpu.cbus.pending.counter;
            keep_dbus_counter = cpu.dbus.pending.counter;
            keep_ebus_counter = cpu.ebus.pending.counter;
            keep_fbus_counter = cpu.fbus.pending.counter;
            
            keep_rslot0 = cpu.read_inst_slot[  cpu.read_inst_index ];
            keep_rslot1 = cpu.read_inst_slot[ ~cpu.read_inst_index ];
            keep_wslot0 = cpu.write_inst_slot[  cpu.write_inst_index ];
            keep_wslot1 = cpu.write_inst_slot[ ~cpu.write_inst_index ];
            
            wait cycle;
        };
      
    };
    
    -- ERROR : check for pbus address
    -- lower two bits are always zero in program address bus.
    TCM_Error_CheckForBus_Pabus0() @rise_clock is {
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {
                
                check that ( cpu.pbus.previous.abus[1:0] == 0 )
                    else dut_error( "Error at ",sys.time," : lower two bits in pabus should be always zero." );
            
            };
            
            wait cycle;
        };
    };

    -- ERROR : check for pdismiss
    -- Pdismiss should NOT occur when P bus DT-DMA access exists
    TCM_Error_CheckForPdissmiss() @rise_clock is {
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {

               if ( cpu.pbus.pending.dismiss_progress[0] != NULL and  cpu.pbus.pending.dismiss_progress[1] != NULL ) {

                  if( cpu.pbus.current.req == 0 and cpu.pbus.current.dismiss == 1 ) {

                      -- emulation qualifier check; there is this case on actual RTL.
                      -- check that ( cpu.emulation.current.dmapr == 0 ) 
                      -- else dut_error( "Error at ",sys.time," : Pdismiss and P emu qualifier should not occur simultaneously." );

                      -- Pending request check; the below code is related with pending_prog_requests() of cpubus_attributes.e
                      if ( cpu.pbus.pending.dismiss_counter == 2 ) {
                          check that ( ! ( ( cpu.pbus.pending.dismiss_progress[0].pbus_active == active and 
                                             cpu.pbus.pending.dismiss_progress[0].pbus_type   == debug ) or
                                           ( cpu.pbus.pending.dismiss_progress[1].pbus_active == active and
                                             cpu.pbus.pending.dismiss_progress[1].pbus_type   == debug )))
                                else dut_error( "Error at ",sys.time," : Pdismiss should not occur when P bus emulation access is not completed. case 1 \n");
                      };
                      if ( cpu.pbus.pending.dismiss_counter == 1 ) {
                          check that ( ! ( ( cpu.pbus.pending.dismiss_rdy_index == 0 and 
                                             cpu.pbus.pending.dismiss_progress[1].pbus_active == active and
                                             cpu.pbus.pending.dismiss_progress[1].pbus_type   == debug )  or
                                           ( cpu.pbus.pending.dismiss_rdy_index == 1 and 
                                             cpu.pbus.pending.dismiss_progress[0].pbus_active == active and
                                             cpu.pbus.pending.dismiss_progress[0].pbus_type   == debug )))
                                else dut_error( "Error at ",sys.time," : Pdismiss should not occur when P bus emulation access is not completed. case 2 \n");
                      };

                  };
               };
             };
            
            wait cycle;
        };
    };
    
    -- ERROR : check for read data bus
    TCM_Error_CheckForBus_ReadData0() @rise_clock is {
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {
                
                -- pdbus --
                if( cpu.pbus.previous.ready != 0 ) {
                    check that ( cpu.pbus.current.dbus == 0 )
                        else dut_error( "Error at ",sys.time," : pdbus should be zero in this phase." );
                };
                
                -- bdbus --
                if( cpu.bbus.previous.ready != 0 ) {
                    check that ( cpu.bbus.current.dbus == 0 )
                        else dut_error( "Error at ",sys.time," : bdbus should be zero in this phase." );
                };
                
                -- cdbus --
                if( cpu.cbus.previous.ready != 0 and 
                    !(cpu.dbus.pending.finish != NULL and cpu.dbus.previous.ready == 0 and cpu.dbus.pending.finish.doubler == double) ) {
                    check that ( cpu.cbus.current.dbus == 0 )
                        else dut_error( "Error at ",sys.time," : cdbus should be zero in this phase." );
                };
                
                -- ddbus --
                if( cpu.dbus.previous.ready != 0 ) {
                    check that ( cpu.dbus.current.dbus == 0 )
                        else dut_error( "Error at ",sys.time," : ddbus should be zero in this phase." );
                };
                
            };
            
            wait cycle;
        };
    };
    
    -- ERROR : check for write data bus
    previous_ebus_pending_counter : uint;
    previous_fbus_pending_counter : uint;
    current_ebus_double_flag      : bool;
    previous_ebus_double_flag     : bool;
    current_ebus_double_ready     : bool;
    previous_ebus_double_ready    : bool;
    ebus_cnt : uint;
    fbus_cnt : uint;
    TCM_Error_CheckForBus_WriteData0() @rise_clock is {
        previous_ebus_pending_counter = 0;
        previous_fbus_pending_counter = 0;
        ebus_cnt = 0;
        fbus_cnt = 0;
        while( TRUE ) {
            
            current_ebus_double_ready  = ( previous_ebus_double_flag and cpu.ebus.state.ready == ready );
            current_ebus_double_flag   = ( cpu.ebus.pending.progress[cpu.ebus.pending.rdy_index] != NULL and 
                                           cpu.ebus.pending.progress[cpu.ebus.pending.rdy_index].ebus_active == active and 
                                           cpu.ebus.pending.progress[cpu.ebus.pending.rdy_index].doublew == double );
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {
                
                -- edbus --
                if( cpu.ebus.x.dbus == 0 ) {
                    if( previous_ebus_pending_counter != 0 and 
                        !(cpu.ebus.state.previous_ready == ready) and 
                        !(previous_ebus_pending_counter == 1 and cpu.ebus.state.previous_request == request) ) {
                        ebus_cnt = ebus_cnt + 1;
                        
                        if( cpu.ebus.previous.dbus != cpu.ebus.current.dbus ) {
                            dut_error( "Error at ",sys.time," : edbus should not change in this cycle.");
                        };
                        
                    };
                };
                
                -- fdbus --
                if( cpu.fbus.x.dbus == 0 ) {
                    
                    if( (previous_fbus_pending_counter != 0 or previous_ebus_double_flag or current_ebus_double_ready) and
                        !(previous_ebus_double_flag and previous_ebus_pending_counter == 1 and cpu.ebus.state.previous_request == request) and
                        !(previous_ebus_double_flag and cpu.ebus.state.previous_ready == ready) and
                        !previous_ebus_double_ready and 
                        !(previous_fbus_pending_counter == 1 and cpu.fbus.state.previous_request == request and !previous_ebus_double_flag) and
                        !(cpu.fbus.state.previous_ready == ready) ) {
                        fbus_cnt = fbus_cnt + 1;
                        
                        if( cpu.fbus.previous.dbus != cpu.fbus.current.dbus ) { 
                            dut_error( "Error at ",sys.time," : fdbus should not change in this cycle.");
                        };
                        
                    };
                };
                
            };
            
            previous_ebus_pending_counter = cpu.ebus.pending.counter;
            previous_fbus_pending_counter = cpu.fbus.pending.counter;
            previous_ebus_double_flag     = current_ebus_double_flag;
            previous_ebus_double_ready    = current_ebus_double_ready;
            
            wait cycle;
            
        };
    };
    
    -- ERROR : ICEMAKER ready check
    TCM_Error_CheckIcemakerReady0() @rise_clock is {
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {
                
                -- check for dbus icemaker access --
                if( cpu.dbus.current.raw_ready == 0 ) {
                    check that ( cpu.dbus.current.ice_ready != 0 )
                        else dut_error( "Error at ",sys.time," : dready should not return in this cycle(overlap ICEMAKER's ready). ");
                };
                
                if( cpu.dbus.state.ready == ready and cpu.dbus.current.ice_ready == 0 ) {
                    
                    check that ( cpu.dbus.pending.detect_ready.rspace == io and 
                                 cpu.dbus.pending.detect_ready.dbus_active == active and
                                 cpu.dbus.pending.detect_ready.dabus is in ICEMAKER_IOMMR )
                        else dut_error( "Error at ",sys.time," : icemaker's dready is applied for normal request." );
                    
                };
                
                if( cpu.dbus.state.ready == ready and cpu.dbus.current.ice_ready != 0 ) {
                    
                    check that !( cpu.dbus.pending.detect_ready.rspace == io and 
                                  cpu.dbus.pending.detect_ready.dbus_active == active and
                                  cpu.dbus.pending.detect_ready.dabus is in ICEMAKER_IOMMR )
                        else dut_error( "Error at ",sys.time," : this dready is applied for icemaker's request." );
                        
                };
                
                -- check for ebus icemaker access --
                if( cpu.ebus.current.raw_ready == 0 ) {
                    check that ( cpu.ebus.current.ice_ready != 0 )
                        else dut_error( "Error at ",sys.time," : eready should not return in this cycle(overlap ICEMAKER's ready). ");
                };
                
                if( cpu.ebus.state.ready == ready and cpu.ebus.current.ice_ready == 0 ) {
                    
                    check that ( cpu.ebus.pending.detect_ready.wspace == io and 
                                 cpu.ebus.pending.detect_ready.ebus_active == active and
                                 cpu.ebus.pending.detect_ready.eabus is in ICEMAKER_IOMMR )
                        else dut_error( "Error at ",sys.time," : icemaker's eready is applied for normal request." );
                        
                 };
                 
                if( cpu.ebus.state.ready == ready and cpu.ebus.current.ice_ready != 0 ) {
                    
                    check that !( cpu.ebus.pending.detect_ready.wspace == io and 
                                  cpu.ebus.pending.detect_ready.ebus_active == active and
                                  cpu.ebus.pending.detect_ready.eabus is in ICEMAKER_IOMMR )
                        else dut_error( "Error at ",sys.time," : this eready is applied for icemaker's request." );
                        
                 };
                 
            };
            
            wait cycle;
        };
    };
    
    -- Qualifier assertion check in request phase --
    TCM_Error_CheckQualifierAssertion0() @rise_clock is {
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {
                
                -- checking for double and byte qualifier assertion --
                if( cpu.ebus.current.req == 0 ) {
                    check that ( !(cpu.qualifier.current.doublew == 1 and cpu.qualifier.current.wrbyte == 1) )
                        else dut_error( "Error at ",sys.time," : wrbyte and doublew can't assert same time in request phase.");
                };
                
                if( cpu.dbus.current.req == 0 ) {
                    check that ( !(cpu.qualifier.current.doubler == 1 and cpu.qualifier.current.rdbyte == 1) )
                        else dut_error( "Error at ",sys.time," : rdbyte and doubler can't assert same time in request phase.");
                };
                
                if( (cpu.ebus.current.req == 0 or cpu.ebus.current.ioreq == 0) and cpu.qualifier.current.wrlsbe == 1 ) {
                    check that ( cpu.qualifier.current.wrbyte == 1 )
                        else dut_error( "Error at ",sys.time," : wrlsbe is issued without wrbyte signal in request cycle.");
                };
                
                if( (cpu.dbus.current.req == 0 or cpu.dbus.current.ioreq == 0) and cpu.qualifier.current.rdlsbe == 1 ) {
                    check that ( cpu.qualifier.current.rdbyte == 1 )
                        else dut_error( "Error at ",sys.time," : rdlsbe is issued without rdbyte qualifier in request cycle.");
                };
                
                if( cpu.dbus.current.ioreq == 0 ) {
                    check that ( cpu.qualifier.current.doubler == 0 )
                        else dut_error( "Error at ",sys.time," : double qualifier don't allow to issue for io request.");
                };
                if( cpu.ebus.current.ioreq == 0 ) {
                    check that ( cpu.qualifier.current.doublew == 0 )
                        else dut_error( "Error at ",sys.time," : double qualifier don't allow to issue for io request.");
                };
                
                -- checking for io request assertion --
                if( cpu.dbus.current.ioreq == 0 and cpu.ebus.current.ioreq == 0 ) {
                    dut_error( "Error at ",sys.time," : ioereq and iodreq can issue at same time.");
                };
                
                -- checking for lock qualifier assertion --
                if( cpu.qualifier.current.lock == 1 ) {
                    
                    check that ( cpu.dbus.current.req == 0 )
                        else dut_error( "Error at ",sys.time," : lock qualifier should issue with dbus request.");
                    
                    check that ( cpu.dbus.current.ioreq == 1 and cpu.ebus.current.ioreq == 1 and
                                 cpu.bbus.current.req   == 1 and cpu.cbus.current.req   == 1 and
                                 cpu.ebus.current.req   == 1 and cpu.fbus.current.req   == 1     )
                        else dut_error( "Error at ",sys.time," : lock qualifier should be issued with dreq only.");
                    
                };
                
            };
            
            wait cycle;
        };
    };
    
    -- Lock assertion checker --
    !lock_access  : bool;
    !lock_address : uint (bits: 23);
    init() is also {
        lock_access  = FALSE;
        lock_address = 0;
    };
    
    TCM_Warning_CheckForLockAssertion0() @rise_clock is {
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {
                
                if( lock_access == TRUE and cpu.ebus.current.req == 0 ) {
                    lock_access  = FALSE;
                    
                    check that ( cpu.ebus.current.abus == lock_address )
                        else dut_error( "Warning at ",sys.time," : write address differs from read address when lock access.");
                    
                };
                
                if( lock_access ) {
                    if( cpu.bbus.current.req   == 0 or cpu.cbus.current.req   == 0 or 
                        cpu.dbus.current.req   == 0 or cpu.fbus.current.req   == 0 or
                        cpu.dbus.current.ioreq == 0 or cpu.ebus.current.ioreq == 0    ) {
                        dut_error( "Warning at ",sys.time," : request signals can not issue during lock access.");
                        lock_access = FALSE;
                    };
                };
                
                if( lock_access == FALSE and cpu.qualifier.current.lock == 1 ) {
                    lock_access  = TRUE;
                    lock_address = cpu.dbus.current.abus;
                };
                
            } else {
                lock_access = FALSE;
            };
            
            wait cycle;
        };
    };
    
    -- WARNING : timeout warning
    TCM_Warning_CheckForTimeout() @rise_clock is {
        var pbus_timer : uint; var pbus_flag : bool;
        var bbus_timer : uint; var bbus_flag : bool;
        var cbus_timer : uint; var cbus_flag : bool;
        var dbus_timer : uint; var dbus_flag : bool;
        var ebus_timer : uint; var ebus_flag : bool;
        var fbus_timer : uint; var fbus_flag : bool;
        
        pbus_timer = 0; pbus_flag = TRUE;
        bbus_timer = 0; bbus_flag = TRUE;
        cbus_timer = 0; cbus_flag = TRUE;
        dbus_timer = 0; dbus_flag = TRUE;
        ebus_timer = 0; ebus_flag = TRUE;
        fbus_timer = 0; fbus_flag = TRUE;
        
        while( TRUE ) {
            
            -- timeout check --
            if( pbus_flag and cpu.pbus.current.ready == 0 ) { pbus_timer = 0;  pbus_flag = TRUE; };
            if( bbus_flag and cpu.bbus.current.ready == 0 ) { bbus_timer = 0;  bbus_flag = TRUE; };
            if( cbus_flag and cpu.cbus.current.ready == 0 ) { cbus_timer = 0;  cbus_flag = TRUE; };
            if( dbus_flag and cpu.dbus.current.ready == 0 ) { dbus_timer = 0;  dbus_flag = TRUE; };
            if( ebus_flag and cpu.ebus.current.ready == 0 ) { ebus_timer = 0;  ebus_flag = TRUE; };
            if( fbus_flag and cpu.fbus.current.ready == 0 ) { fbus_timer = 0;  fbus_flag = TRUE; };
            
            if( pbus_flag and cpu.pbus.state.current != waiting_req ) { pbus_timer = pbus_timer + 1; };
            if( bbus_flag and cpu.bbus.state.current != waiting_req ) { bbus_timer = bbus_timer + 1; };
            if( cbus_flag and cpu.cbus.state.current != waiting_req ) { cbus_timer = cbus_timer + 1; };
            if( dbus_flag and cpu.dbus.state.current != waiting_req ) { dbus_timer = dbus_timer + 1; };
            if( ebus_flag and cpu.ebus.state.current != waiting_req ) { ebus_timer = ebus_timer + 1; };
            if( fbus_flag and cpu.fbus.state.current != waiting_req ) { fbus_timer = fbus_timer + 1; };
            
            if( pbus_flag and pbus_timer > timeout ) {
                dut_error( "Warning at ",sys.time," : timeout(",timeout," cycle) occur in pbus." );
                pbus_flag = FALSE;
            };
            if( bbus_flag and bbus_timer > timeout ) {
                dut_error( "Warning at ",sys.time," : timeout(",timeout," cycle) occur in bbus." );
                bbus_flag = FALSE;
            };
            if( cbus_flag and cbus_timer > timeout ) {
                dut_error( "Warning at ",sys.time," : timeout(",timeout," cycle) occur in cbus." );
                cbus_flag = FALSE;
            };
            if( dbus_flag and dbus_timer > timeout ) {
                dut_error( "Warning at ",sys.time," : timeout(",timeout," cycle) occur in dbus." );
                dbus_flag = FALSE;
            };
            if( ebus_flag and ebus_timer > timeout ) {
                dut_error( "Warning at ",sys.time," : timeout(",timeout," cycle) occur in ebus." );
                ebus_flag = FALSE;
            };
            if( fbus_flag and fbus_timer > timeout ) {
                dut_error( "Warning at ",sys.time," : timeout(",timeout," cycle) occur in fbus." );
                fbus_flag = FALSE;
            };
            
            wait cycle;
        };
    };
    
    -- WARNING : checking for unknown driving
    TCM_Warning_CheckForUnknownDriving() @rise_clock is {
        
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {
                
                if( !config.unknown_drive ) {
                    
                    check that ( cpu.pbus.x.abus    == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.pabus,"."   );
                    check that ( cpu.bbus.x.abus    == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.babus,"."   );
                    check that ( cpu.cbus.x.abus    == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.cabus,"."   );
                    check that ( cpu.dbus.x.abus    == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.dabus,"."   );
                    check that ( cpu.ebus.x.abus    == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.eabus,"."   );
                    check that ( cpu.ebus.x.dbus    == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.edbus,"."   );
                    check that ( cpu.fbus.x.abus    == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.fabus,"."   );
                    check that ( cpu.fbus.x.dbus    == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.fdbus,"."   );
                    check that ( cpu.qualifier.x.wrbyte  == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.wrbyte,"." );
                    check that ( cpu.qualifier.x.wrlsbe  == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.wrlsbe,"." );
                    check that ( cpu.qualifier.x.doubler == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.doubler,".");
                    check that ( cpu.qualifier.x.doublew == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.doublew,".");
                    check that ( cpu.emulation.x.dmapr == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.dmapr,"." );
                    check that ( cpu.emulation.x.dmapw == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.dmapw,"." );
                    check that ( cpu.emulation.x.dbugd == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.dbugd,"." );
                    check that ( cpu.emulation.x.dbuge == 0 ) 
                        else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.dbuge,"." );
                };
                
                check that ( cpu.pbus.x.req == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.preq,"."    );
                check that ( cpu.pbus.x.dbus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.pdbus,"."   );
                check that ( cpu.pbus.x.ready   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.pready,"."  );
                check that ( cpu.pbus.x.dismiss == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.pdismiss,".");
                
                check that ( cpu.bbus.x.req     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.breq,"."    );
                check that ( cpu.bbus.x.dbus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.bdbus,"."   );
                check that ( cpu.bbus.x.ready   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.bready,"."  );
                
                check that ( cpu.cbus.x.req     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.creq,"."    );
                check that ( cpu.cbus.x.dbus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.cdbus,"."   );
                check that ( cpu.cbus.x.ready   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.cready,"."  );
                
                check that ( cpu.dbus.x.req     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.dreq,"."    );
                check that ( cpu.dbus.x.ioreq   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.iodreq,"."  );
                check that ( cpu.dbus.x.dbus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.ddbus,"."   );
                check that ( cpu.dbus.x.ready   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.dready,"."  );
                
                check that ( cpu.ebus.x.req     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.ereq,"."    );
                check that ( cpu.ebus.x.ioreq   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.ioereq,"."  );
                check that ( cpu.ebus.x.ready   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.eready,"."  );
                
                check that ( cpu.fbus.x.req     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.freq,"."    );
                check that ( cpu.fbus.x.ready   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.fready,"."  );
                
                check that ( cpu.qualifier.x.lock    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.lock,"."   );
                
                check that ( cpu.emulation.x.debugenable == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.debugenable,"." );
                check that ( cpu.emulation.x.abortready  == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.abortready,"."  );
                
                check that ( cpu.cache.x.cachefreeze     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.cachefreeze,"." );
                check that ( cpu.cache.x.cacheenable     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.cacheenable,"." );
                check that ( cpu.cache.x.cacheclr        == 0 ) 
                else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.cacheclr,"."    );
                check that ( cpu.cache.x.cacheendclr     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.cacheendclr,"." );
                
                check that ( cpu.system.x.gl_resetn      == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.gl_resetn.hdl_path(),"."   );
                check that ( cpu.system.x.cpu_resetn     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.cpu_resetn.hdl_path(),"."  );
                check that ( cpu.system.x.progdismiss    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.progdismiss,"." );
                check that ( cpu.system.x.gl_mpnmc       == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.gl_mpnmc,"."    );
                check that ( cpu.system.x.cpu_mpnmc      == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.cpu_mpnmc,"."   );
                check that ( cpu.system.x.avis           == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.avis,"."        );
                check that ( cpu.system.x.idlemaster     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.idlemaster,"."  );
                check that ( cpu.system.x.idlewakeup     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.idlewakeup,"."  );
                check that ( cpu.system.x.int            == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.int,"."         );
                check that ( cpu.system.x.nmi            == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.nmi,"."         );
                check that ( cpu.system.x.buserror       == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.buserror,"."    );
                    check that ( cpu.system.x.debugerror     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Unknown is driven in ",paths.debugerror,"."  );
                
            };
        
            wait cycle;
        };
    };
    
    -- WARNING : checking for Hi-z driving
    TCM_Warning_CheckForHizDriving() @rise_clock is {
        
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn != 0 and cpu.emulation.current.abortready != 1 ) {
                
                check that ( cpu.pbus.z.req == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.preq,"."    );
                check that ( cpu.pbus.z.abus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.pabus,"."   );
                check that ( cpu.pbus.z.dbus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.pdbus,"."   );
                check that ( cpu.pbus.z.ready   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.pready,"."  );
                check that ( cpu.pbus.z.dismiss == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.pdismiss,".");
                
                check that ( cpu.bbus.z.req     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.breq,"."    );
                check that ( cpu.bbus.z.abus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.babus,"."   );
                check that ( cpu.bbus.z.dbus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.bdbus,"."   );
                check that ( cpu.bbus.z.ready   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.bready,"."  );
                
                check that ( cpu.cbus.z.req     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.creq,"."    );
                check that ( cpu.cbus.z.abus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.cabus,"."   );
                check that ( cpu.cbus.z.dbus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.cdbus,"."   );
                check that ( cpu.cbus.z.ready   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.cready,"."  );
                
                check that ( cpu.dbus.z.req     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.dreq,"."    );
                check that ( cpu.dbus.z.ioreq   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.iodreq,"."  );
                check that ( cpu.dbus.z.abus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.dabus,"."   );
                check that ( cpu.dbus.z.dbus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.ddbus,"."   );
                check that ( cpu.dbus.z.ready   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.dready,"."  );
                
                check that ( cpu.ebus.z.req     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.ereq,"."    );
                check that ( cpu.ebus.z.ioreq   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.ioereq,"."  );
                check that ( cpu.ebus.z.abus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.eabus,"."   );
                check that ( cpu.ebus.z.dbus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.edbus,"."   );
                check that ( cpu.ebus.z.ready   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.eready,"."  );
                
                check that ( cpu.fbus.z.req     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.freq,"."    );
                check that ( cpu.fbus.z.abus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.fabus,"."   );
                check that ( cpu.fbus.z.dbus    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.fdbus,"."   );
                check that ( cpu.fbus.z.ready   == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.fready,"."  );
                
                check that ( cpu.qualifier.z.wrbyte  == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.wrbyte,"." );
                check that ( cpu.qualifier.z.wrlsbe  == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.wrlsbe,"." );
                check that ( cpu.qualifier.z.doubler == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.doubler,".");
                check that ( cpu.qualifier.z.doublew == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.doublew,".");
                check that ( cpu.qualifier.z.lock    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.lock,"."   );
                
                check that ( cpu.emulation.z.dmapr == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.dmapr,"." );
                check that ( cpu.emulation.z.dmapw == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.dmapw,"." );
                check that ( cpu.emulation.z.dbugd == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.dbugd,"." );
                check that ( cpu.emulation.z.dbuge == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.dbuge,"." );
                
                check that ( cpu.emulation.z.debugenable == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.debugenable,"." );
                check that ( cpu.emulation.z.abortready  == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.abortready,"."  );
                
                check that ( cpu.cache.z.cachefreeze     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.cachefreeze,"." );
                check that ( cpu.cache.z.cacheenable     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.cacheenable,"." );
                check that ( cpu.cache.z.cacheclr        == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.cacheclr,"."    );
                check that ( cpu.cache.z.cacheendclr     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.cacheendclr,"." );
                
                check that ( cpu.system.z.gl_resetn      == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.gl_resetn.hdl_path(),"."   );
                check that ( cpu.system.z.cpu_resetn     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.cpu_resetn.hdl_path(),"."  );
                check that ( cpu.system.z.progdismiss    == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.progdismiss,"." );
                check that ( cpu.system.z.gl_mpnmc       == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.gl_mpnmc,"."    );
                check that ( cpu.system.z.cpu_mpnmc      == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.cpu_mpnmc,"."   );
                check that ( cpu.system.z.avis           == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.avis,"."        );
                check that ( cpu.system.z.idlemaster     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.idlemaster,"."  );
                check that ( cpu.system.z.idlewakeup     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.idlewakeup,"."  );
                check that ( cpu.system.z.int            == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.int,"."         );
                check that ( cpu.system.z.nmi            == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.nmi,"."         );
                check that ( cpu.system.z.buserror       == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.buserror,"."    );
                check that ( cpu.system.z.debugerror     == 0 ) 
                    else dut_error( "Warning at ",sys.time," : Hi-z is driven in ",paths.debugerror,"."  );
            };
            
            wait cycle;
        };
    };
    
    -- status clear when abort/reset --
    TCM_reset_check() @rise_clock is {
        
        while( TRUE ) {
            
            if( cpu.system.current.cpu_resetn == 0 or cpu.emulation.current.abortready == 1 ) {
                cpu.init();
                init();
            };
            
            wait cycle;
        };
    };

    ---------
    -- run --
    ---------
    run() is also {
        
        if (disable == FALSE) {
            
            start TCM_Error_CheckForStateTransittion();
            start TCM_Error_CheckForSignal_ioreq0();
            start TCM_Error_CheckForBus_AddressChange0();
            start TCM_Error_CheckForBus_Pabus0();
            start TCM_Error_CheckForBus_ReadData0();
            start TCM_Error_CheckForBus_WriteData0();
          --start TCM_Error_CheckIcemakerReady0();
            start TCM_Error_CheckQualifierAssertion0();
            start TCM_Error_CheckForSignals_byte_qualifier0();
            start TCM_Error_CheckForSignals_double_qualifier0();
            start TCM_Error_CheckForPdissmiss();

            start TCM_Warning_CheckForSignals_byte_qualifier0();
            start TCM_Warning_CheckForLockAssertion0();
            start TCM_Warning_CheckForTimeout();
          --start TCM_Warning_CheckForUnknownDriving();
          --start TCM_Warning_CheckForHizDriving();
            
            start TCM_reset_check();
        };
        
    };
    
};

'>
