//********************************************************************************
//
//  Filename: cpubus_state_transitions.e
//
//  Description: state transitions
//               note : program address : byte addressing
//                      data address    : word addressing
//
//  Major Revision Descriptions:
//  oishi          09/08/02         copied from UMA1.0 branch
//  oishi          09/09/02         removed bottom signals
//                                  removed mdp related functions
//                                  removed ack related functions
//                                  removed data dismiss functions
//                                  removed stall signals
//  oishi          09/11/02         changed bdbus width to 32bits
//                                  added following qualifiers;
//                                    cpu_dblcoef_tr/cpu_rdbyte_tr/cpu_rdlsbe_tr
//  oishi          09/12/02         added MemoryProtection related signals
//                                  added cpu_avis_tr siganl
//  oishi          09/13/02         modified resetn to cpu_resetn and gl_resetn
//                                  following signals are added;
//                                    gl_buserror_nr/gl_debugerror_nr
//                                    gl_nmi_nr/gl_int_nr/cpu_idlewakeup_ta
//  oishi          09/30/02         added iack/idlecpu/idlecpud signal
//  oishi          10/02/02         removed check_pending_lock_requests()
//  oishi          10/04/02         supports suspend signal
//  oishi          11/13/02         applied MikeP suggesion code for several functions
//  oishi          11/25/02         added memory protection violation handling
//  oishi          12/13/02         added perrack signal
//  hiraoka        03/25/03         add following signal
//                                    gl_homnsam_tr(3 DOWNTO 0)
//                                    cpu_homnsam_tr(3 DOWNTO 0)
//                                    gl_hompnsamp_init_tr
//                                    gl_homrnsamr_init_tr
//                                    gl_homxnsamx_init_tr
//                                    gl_homynsamy_init_tr
//  hiraoka        05/21/03         add dismiss_rdy_index/dismiss_progress to PENDING_PROG_REQUESTS
//                                  for saving status data from  dismiss-clear on pending_prog_requests()
//  hiraoka        06/03/03         correct hiz-get-info
//  hiraoka        07/08/03         speed improvement around get_current_value()
//*********************************************************************************
<'

struct CPUBUS_STATES {

    !current  : bus_state;
    !previous : bus_state;
    
    !stall   : stall_state;
    !ready   : ready_state;
    !request : request_state;
    
    !previous_ready   : ready_state;
    !previous_request : request_state;
    
    ---------------------------------
    -- state treatment for dismiss --
    ---------------------------------
    progdismiss( dismiss : uint ) is {
        if( dismiss == 1 ) {
            current = waiting_req;
        };
    };
    
    -----------------------
    -- state transitions --
    -----------------------
    transition( req : uint, rdy : uint ) is {
        
        previous = current.copy();
        
        case previous {
            [ waiting_req ] : {      -- error : any requests doesn't occur before this ready signals
                case {
                    req == 0 and rdy == 0 { current = error;         stall = no_stall;  ready = no_ready; request = no_request; };
                    req == 0 and rdy == 1 { current = rdy_stall_one; stall = no_stall;  ready = no_ready; request = request;    };
                    req == 1 and rdy == 0 { current = error;         stall = no_stall;  ready = no_ready; request = no_request; };
                    req == 1 and rdy == 1 { current = waiting_req;   stall = no_stall;  ready = no_ready; request = no_request; };
                    default               { current = error;         stall = no_stall;  ready = no_ready; request = no_request; };
                };
            };
            [ rdy_stall_one ] : {
                case {
                    req == 0 and rdy == 0 { current = rdy_stall_one; stall = no_stall;  ready = ready;    request = request;    };
                    req == 0 and rdy == 1 { current = rdy_stall_two; stall = no_stall;  ready = no_ready; request = request;    };
                    req == 1 and rdy == 0 { current = waiting_req;   stall = no_stall;  ready = ready;    request = no_request; };
                    req == 1 and rdy == 1 { current = rdy_stall_one; stall = rdy_stall; ready = no_ready; request = no_request; };
                    default               { current = error;         stall = no_stall;  ready = no_ready; request = no_request; };
                };
            };
            [ rdy_stall_two ] : {
                case {
                    req == 0 and rdy == 0 { current = error;         stall = no_stall;  ready = no_ready; request = no_request; };
                    req == 0 and rdy == 1 { current = error;         stall = no_stall;  ready = no_ready; request = no_request; };
                    req == 1 and rdy == 0 { current = rdy_stall_one; stall = no_stall;  ready = ready;    request = no_request; };
                    req == 1 and rdy == 1 { current = rdy_stall_two; stall = rdy_stall; ready = no_ready; request = no_request; };
                    default               { current = error;         stall = no_stall;  ready = no_ready; request = no_request; };
                };
            };
            [ error ] : 
                { current = waiting_req; stall = no_stall;  ready = no_ready; request = no_request };
            default : 
                { current = error; stall = no_stall;  ready = no_ready; request = no_request; };
        };
    };
    
    ------------------------
    -- set initial values --
    ------------------------
    init() is also {
        current          = waiting_req;
        previous         = waiting_req;
        stall            = no_stall;
        ready            = no_ready;
        request          = no_request;
        previous_ready   = no_ready;
        previous_request = no_request;
    };
};

struct PENDING_PROGRAM_REQUESTS {
    !rdy_index : uint (bits:  1);
    !req_index : uint (bits:  1);

    !counter   : uint (bits:  2);
    
    !progress[2]  : list of PROGRAM_ACCESS_ATTRIBUTES;
    !detect_ready : PROGRAM_ACCESS_ATTRIBUTES;
    !finish       : PROGRAM_ACCESS_ATTRIBUTES;

    -- For dismiss check
    !dismiss_rdy_index   : uint (bits:  1);    
    !dismiss_progress[2] : list of PROGRAM_ACCESS_ATTRIBUTES;
    !dismiss_counter     : uint (bits:  2);
    
    init() is also {
        rdy_index  = 0;
        req_index  = 0;
        counter    = 0;
        progress[0]  = NULL;
        progress[1]  = NULL;
        finish       = NULL;
        detect_ready = NULL;
        dismiss_rdy_index = 0;
        dismiss_counter   = 0;
        dismiss_progress[0]  = NULL;
        dismiss_progress[1]  = NULL;
    };
    
};

struct PENDING_DATA_REQUESTS {
    !rdy_index : uint (bits:  1);
    !req_index : uint (bits:  1);
    !counter   : uint (bits:  2);
    
    !progress[2]  : list of DATA_ACCESS_ATTRIBUTES;
    !detect_ready : DATA_ACCESS_ATTRIBUTES;
    !finish       : DATA_ACCESS_ATTRIBUTES;
    
    init() is also {
        rdy_index  = 0;
        req_index  = 0;
        counter    = 0;
        progress[0]  = NULL;
        progress[1]  = NULL;
        finish       = NULL;
        detect_ready = NULL;
    };
};

struct CPU_PBUS {
    current  : CPU_PBUS_SIGNALS;
    previous : CPU_PBUS_SIGNALS;
    state    : CPUBUS_STATES;
    pending  : PENDING_PROGRAM_REQUESTS;
    
    -- for detecting unknown/hiz --
    x       : CPU_PBUS_SIGNALS;
    z       : CPU_PBUS_SIGNALS;

    set_init_values() is {
        x.req     = 0; z.req     = 0;
        x.abus    = 0; z.abus    = 0;
        x.dbus    = 0; z.dbus    = 0;
        x.ready   = 0; z.ready   = 0;
        x.dismiss = 0; z.dismiss = 0;
    };
    
    post_generate() is also{
        set_init_values();
    };
    
};

struct CPU_BBUS {
    current  : CPU_BBUS_SIGNALS;
    previous : CPU_BBUS_SIGNALS;
    state    : CPUBUS_STATES;
    pending  : PENDING_DATA_REQUESTS;
    
    -- for detecting unknown/hiz --
    x        : CPU_BBUS_SIGNALS;
    z        : CPU_BBUS_SIGNALS;
    
    set_init_values() is {
        x.req     = 0; z.req     = 0;
        x.abus    = 0; z.abus    = 0;
        x.dbus    = 0; z.dbus    = 0;
        x.ready   = 0; z.ready   = 0;
    };
    
    post_generate() is also{
        set_init_values();
    };
};

struct CPU_CBUS {
    current  : CPU_CBUS_SIGNALS;
    previous : CPU_CBUS_SIGNALS;
    state    : CPUBUS_STATES;
    pending  : PENDING_DATA_REQUESTS;
    
    -- for detecting unknown/hiz --
    x        : CPU_CBUS_SIGNALS;
    z        : CPU_CBUS_SIGNALS;

    set_init_values() is {
        x.req     = 0; z.req     = 0;
        x.abus    = 0; z.abus    = 0;
        x.dbus    = 0; z.dbus    = 0;
        x.ready   = 0; z.ready   = 0;
    };
    
    post_generate() is also{
        set_init_values();
    };
};

struct CPU_DBUS {
    current  : CPU_DBUS_SIGNALS;
    previous : CPU_DBUS_SIGNALS;
    state    : CPUBUS_STATES;
    pending  : PENDING_DATA_REQUESTS;
    
    -- for detecting unknown/hiz --
    x        : CPU_DBUS_SIGNALS;
    z        : CPU_DBUS_SIGNALS;

    set_init_values() is {
        x.ioreq   = 0; z.ioreq   = 0;
        x.req     = 0; z.req     = 0;
        x.abus    = 0; z.abus    = 0;
        x.dbus    = 0; z.dbus    = 0;
        x.ready   = 0; z.ready   = 0;
    };
    
    post_generate() is also{
        set_init_values();
    };
};

struct CPU_EBUS {
    current  : CPU_EBUS_SIGNALS;
    previous : CPU_EBUS_SIGNALS;
    state    : CPUBUS_STATES;
    pending  : PENDING_DATA_REQUESTS;
    
    -- for detecting unknown/hiz --
    x        : CPU_EBUS_SIGNALS;
    z        : CPU_EBUS_SIGNALS;

    set_init_values() is {
        x.ioreq   = 0; z.ioreq   = 0;
        x.req     = 0; z.req     = 0;
        x.abus    = 0; z.abus    = 0;
        x.dbus    = 0; z.dbus    = 0;
        x.ready   = 0; z.ready   = 0;
    };

    post_generate() is also{
        set_init_values();
    };
};

struct CPU_FBUS {
    current  : CPU_FBUS_SIGNALS;
    previous : CPU_FBUS_SIGNALS;
    state    : CPUBUS_STATES;
    pending  : PENDING_DATA_REQUESTS;
    
    -- for detecting unknown/hiz --
    x        : CPU_FBUS_SIGNALS;
    z        : CPU_FBUS_SIGNALS;

    set_init_values() is {
        x.req     = 0; z.req     = 0;
        x.abus    = 0; z.abus    = 0;
        x.dbus    = 0; z.dbus    = 0;
        x.ready   = 0; z.ready   = 0;
    };
    
    post_generate() is also{
        set_init_values();
    };
};

struct CPU_QUALIFIERS {
    current  : CPU_QUALIFIERS_SIGNALS;
    previous : CPU_QUALIFIERS_SIGNALS;
    
    -- for detecting unknown/hiz --
    x        : CPU_QUALIFIERS_SIGNALS;
    z        : CPU_QUALIFIERS_SIGNALS;

    set_init_values() is {
        x.wrbyte  = 0; z.wrbyte  = 0;
        x.wrlsbe  = 0; z.wrlsbe  = 0;
        x.rdbyte  = 0; z.rdbyte  = 0;
        x.rdlsbe  = 0; z.rdlsbe  = 0;
        x.doubler = 0; z.doubler = 0;
        x.doublew = 0; z.doublew = 0;
        x.dblcoef = 0; z.dblcoef = 0;
        x.lock    = 0; z.lock    = 0;
    };
    
    post_generate() is also{
        set_init_values();
    };
};

struct CPU_EMULATIONS {
    current  : CPU_EMULATION_SIGNALS;
    previous : CPU_EMULATION_SIGNALS;
    
    -- for detecting unknown/hiz --
    x        : CPU_EMULATION_SIGNALS;
    z        : CPU_EMULATION_SIGNALS;

    set_init_values() is {
        x.dmapr       = 0; z.dmapr       = 0;
        x.dmapw       = 0; z.dmapw       = 0;
        x.dbugd       = 0; z.dbugd       = 0;
        x.dbuge       = 0; z.dbuge       = 0;
        x.debugenable = 0; z.debugenable = 0;
        x.abortready  = 0; z.abortready  = 0;
        x.suspend     = 0; z.suspend     = 0;
    };
    
    post_generate() is also{
        set_init_values();
    };
};

struct CPU_CACHE {
    current  : CPU_CACHE_SIGNALS;
    previous : CPU_CACHE_SIGNALS;
    
    -- for detecting unknown/hiz --
    x        : CPU_CACHE_SIGNALS;
    z        : CPU_CACHE_SIGNALS;

    set_init_values() is {
        x.cachefreeze = 0; z.cachefreeze = 0;
        x.cacheenable = 0; z.cacheenable = 0;
        x.cacheclr    = 0; z.cacheclr    = 0;
        x.cacheendclr = 0; z.cacheendclr = 0;
    };
    
    post_generate() is also{
        set_init_values();
    };
};

struct CPU_SYSTEM {
    current  : CPU_SYSTEM_SIGNALS;
    previous : CPU_SYSTEM_SIGNALS;
    
    -- for detecting unknown/hiz --
    x        : CPU_SYSTEM_SIGNALS;
    z        : CPU_SYSTEM_SIGNALS;

    set_init_values() is {
        x.gl_resetn   = 0; z.gl_resetn   = 0;
        x.cpu_resetn  = 0; z.cpu_resetn  = 0;
        x.progdismiss = 0; z.progdismiss = 0;
        x.gl_mpnmc    = 0; z.gl_mpnmc    = 0;
        x.cpu_mpnmc   = 0; z.cpu_mpnmc   = 0;
        x.gl_hompnsamp_init  = 1; z.gl_hompnsamp_init  = 1;
        x.gl_homrnsamr_init  = 1; z.gl_homrnsamr_init  = 1;
        x.gl_homxnsamx_init  = 1; z.gl_homxnsamx_init  = 1;
        x.gl_homynsamy_init  = 1; z.gl_homynsamy_init  = 1;
        x.gl_homnsam  = 1; z.gl_homnsam  = 1;
        x.cpu_homnsam = 1; z.cpu_homnsam = 1;
        x.avis        = 0; z.avis        = 0;
        x.idlemaster  = 0; z.idlemaster  = 0;
        x.idlewakeup  = 0; z.idlewakeup  = 0;
        x.idlecpu     = 0; z.idlecpu     = 0;
        x.idlecpud    = 0; z.idlecpud    = 0;
        x.iack        = 0; z.iack        = 0;
        x.int         = 0; z.int         = 0;
        x.nmi         = 0; z.nmi         = 0;
        x.buserror    = 0; z.buserror    = 0;
        x.debugerror  = 0; z.debugerror  = 0;
        x.perrack     = 0; z.perrack     = 0;
    };
    
    post_generate() is also {
        set_init_values();
    };
};

struct CPU_MP {
    current  : CPU_MP_SIGNALS;
    previous : CPU_MP_SIGNALS;
    
    -- for detecting unknown/hiz --
    x        : CPU_MP_SIGNALS;
    z        : CPU_MP_SIGNALS;

    set_init_values() is {
        x.gl_privenable  = 0; z.gl_privenable  = 0;
        x.mpcxviol       = 0; z.mpcxviol       = 0;
        x.mpcbviol       = 0; z.mpcbviol       = 0;
        x.mpccviol       = 0; z.mpccviol       = 0;
        x.mpcdviol       = 0; z.mpcdviol       = 0;
        x.mpceviol       = 0; z.mpceviol       = 0;
        x.mpcfviol       = 0; z.mpcfviol       = 0;
        x.cpu_privenable = 0; z.cpu_privenable = 0;
        x.ppm            = 0; z.ppm            = 0;
        x.bpm            = 0; z.bpm            = 0;
        x.cpm            = 0; z.cpm            = 0;
        x.dpm            = 0; z.dpm            = 0;
        x.epm            = 0; z.epm            = 0;
        x.fpm            = 0; z.fpm            = 0;
        x.rptpcvio       = 0; z.rptpcvio       = 0;
        x.rptintr        = 0; z.rptintr        = 0;
        x.rptpvio        = 0; z.rptpvio        = 0;
        x.rptbvio        = 0; z.rptbvio        = 0;
        x.rptcvio        = 0; z.rptcvio        = 0;
        x.rptdvio        = 0; z.rptdvio        = 0;
        x.rptiodvio      = 0; z.rptiodvio      = 0;
        x.rptevio        = 0; z.rptevio        = 0;
        x.rptioevio      = 0; z.rptioevio      = 0;
        x.rptfvio        = 0; z.rptfvio        = 0;
        x.rptpm          = 0; z.rptpm          = 0;
        x.dbgmpviol      = 0; z.dbgmpviol      = 0;
        x.xdmadmpovr     = 0; z.xdmadmpovr     = 0;
        x.xdmaempovr     = 0; z.xdmaempovr     = 0;
        x.emiodready     = 0; z.emiodready     = 0;
        x.emioeready     = 0; z.emioeready     = 0;
    };
    
    post_generate() is also {
        set_init_values();
    };
};

struct CPU_MMR {
    
    !reg[ 0x60 ] : list of uint(bits: 16);
    !rdata       : uint(bits: 16);
    !shadow_reg[ 0x1 ] : list of uint(bits: 16);

    set_init_values() is {
        var i : int;
        for{ i = 0; i < 0x60; i+=1 } do {
            reg[ i ] = 0;
        };
        shadow_reg[ 0 ] = 1;
    };
    
    post_generate() is also{
        set_init_values();
    };
    
};

unit CPU_BUSES {
    pbus      : CPU_PBUS;
    bbus      : CPU_BBUS;
    cbus      : CPU_CBUS;
    dbus      : CPU_DBUS;
    ebus      : CPU_EBUS;
    fbus      : CPU_FBUS;
    
    qualifier : CPU_QUALIFIERS;
    emulation : CPU_EMULATIONS;
    cache     : CPU_CACHE;
    mp        : CPU_MP;
    system    : CPU_SYSTEM;
    
    
    mmr       : CPU_MMR;
    
    state_transition() is {
        var masked_preq   : uint(bits : 1);
        var masked_breq   : uint(bits : 1);
        var masked_creq   : uint(bits : 1);
        var masked_dreq   : uint(bits : 1);
        var masked_ereq   : uint(bits : 1);
        var masked_freq   : uint(bits : 1);
        var masked_iodreq : uint(bits : 1);
        var masked_ioereq : uint(bits : 1);
        
        var masked_prdy   : uint(bits : 1);
        var masked_brdy   : uint(bits : 1);
        var masked_crdy   : uint(bits : 1);
        var masked_drdy   : uint(bits : 1);
        var masked_erdy   : uint(bits : 1);
        var masked_frdy   : uint(bits : 1);
        
        var masked_pdis   : uint(bits : 1);
        
        -- treatement for abortready signal --
        if( emulation.current.abortready == 1 ) {
            set_reset_values();
            
            masked_preq = 1;
            masked_breq = 1;
            masked_creq = 1;
            masked_dreq = 1; masked_iodreq = 1;
            masked_ereq = 1; masked_ioereq = 1;
            masked_freq = 1;
            
            masked_prdy = 1;
            masked_brdy = 1;
            masked_crdy = 1;
            masked_drdy = 1;
            masked_erdy = 1;
            masked_frdy = 1;
            
            masked_pdis = 0;
        } else {
            masked_preq = pbus.current.req;
            masked_breq = bbus.current.req;
            masked_creq = cbus.current.req;
            masked_dreq = dbus.current.req; masked_iodreq = dbus.current.ioreq; 
            masked_ereq = ebus.current.req; masked_ioereq = ebus.current.ioreq;
            masked_freq = fbus.current.req;
            
            masked_prdy = pbus.current.ready;
            masked_brdy = bbus.current.ready;
            masked_crdy = cbus.current.ready;
            masked_drdy = dbus.current.ready;
            masked_erdy = ebus.current.ready;
            masked_frdy = fbus.current.ready;
            
            masked_pdis = pbus.current.dismiss;
        };
        
        -- state treatment when program dismiss ocuur --
        pbus.state.progdismiss( masked_pdis );
        
        -- state transitions --
        pbus.state.transition( masked_preq,                 masked_prdy | masked_pdis );
        bbus.state.transition( masked_breq,                 masked_brdy );
        cbus.state.transition( masked_creq,                 masked_crdy );
        dbus.state.transition( masked_dreq & masked_iodreq, masked_drdy );
        ebus.state.transition( masked_ereq & masked_ioereq, masked_erdy );
        fbus.state.transition( masked_freq,                 masked_frdy );
        
        -- set mpnmc in ST3[6] during reset --
        if( system.current.cpu_resetn == 0 ) {
            
            if( system.current.gl_mpnmc == 0 ) {
                mmr.reg[MMR_ST3_55] = mmr.reg[MMR_ST3_55] & 0xffbf;
            } else {
                mmr.reg[MMR_ST3_55] = mmr.reg[MMR_ST3_55] | 0x0040;
            };
            
        };
        
    };
    
    keep_current_values() is {
        -- keep signals --
        pbus.previous      = pbus.current.copy();
        bbus.previous      = bbus.current.copy();
        cbus.previous      = cbus.current.copy();
        dbus.previous      = dbus.current.copy();
        ebus.previous      = ebus.current.copy();
        fbus.previous      = fbus.current.copy();
        qualifier.previous = qualifier.current.copy();
        emulation.previous = emulation.current.copy();
        cache.previous     = cache.current.copy();
        mp.previous        = mp.current.copy();
        system.previous    = system.current.copy();
        -- keep state --
        pbus.state.previous = pbus.state.current.copy();
        bbus.state.previous = bbus.state.current.copy();
        cbus.state.previous = cbus.state.current.copy();
        dbus.state.previous = dbus.state.current.copy();
        ebus.state.previous = ebus.state.current.copy();
        fbus.state.previous = fbus.state.current.copy();
        -- keep ready state --
        pbus.state.previous_ready = pbus.state.ready.copy();
        bbus.state.previous_ready = bbus.state.ready.copy();
        cbus.state.previous_ready = cbus.state.ready.copy();
        dbus.state.previous_ready = dbus.state.ready.copy();
        ebus.state.previous_ready = ebus.state.ready.copy();
        fbus.state.previous_ready = fbus.state.ready.copy();
        -- keep request state --
        pbus.state.previous_request = pbus.state.request.copy();
        bbus.state.previous_request = bbus.state.request.copy();
        cbus.state.previous_request = cbus.state.request.copy();
        dbus.state.previous_request = dbus.state.request.copy();
        ebus.state.previous_request = ebus.state.request.copy();
        fbus.state.previous_request = fbus.state.request.copy();
    };
    
    get_current_value( config : CPUBUS_CONFIGURATIONS, paths : CPU_SIGNAL_PATH ) is {
        var preq           : uint (bits: 01);
        var pabus          : uint (bits: 24);
        var pdbus          : uint (bits: 32);
        var pready         : uint (bits: 01);
        var pdismiss       : uint (bits: 01);
        var breq           : uint (bits: 01);
        var babus          : uint (bits: 23);
        var bdbus          : uint (bits: 32);
        var bready         : uint (bits: 01);
        var creq           : uint (bits: 01);
        var cabus          : uint (bits: 23);
        var cdbus          : uint (bits: 16);
        var cready         : uint (bits: 01);
        var dreq           : uint (bits: 01);
        var iodreq         : uint (bits: 01);
        var dabus          : uint (bits: 23);
        var ddbus          : uint (bits: 16);
        var dready         : uint (bits: 01);
        var ereq           : uint (bits: 01);
        var ioereq         : uint (bits: 01);
        var eabus          : uint (bits: 23);
        var edbus          : uint (bits: 16);
        var eready         : uint (bits: 01);
        var freq           : uint (bits: 01);
        var fabus          : uint (bits: 23);
        var fdbus          : uint (bits: 16);
        var fready         : uint (bits: 01);
        var wrbyte         : uint (bits: 01);
        var wrlsbe         : uint (bits: 01);
        var rdbyte         : uint (bits: 01);
        var rdlsbe         : uint (bits: 01);
        var doubler        : uint (bits: 01);
        var doublew        : uint (bits: 01);
        var dblcoef        : uint (bits: 01);
        var lock           : uint (bits: 01);
        var dmapr          : uint (bits: 01);
        var dmapw          : uint (bits: 01);
        var dbugd          : uint (bits: 01);
        var dbuge          : uint (bits: 01);
        var debugenable    : uint (bits: 01);
        var abortready     : uint (bits: 01);
        var suspend        : uint (bits: 01);
        
        var cachefreeze    : uint (bits: 01);
        var cacheenable    : uint (bits: 01);
        var cacheclr       : uint (bits: 01);
        var cacheendclr    : uint (bits: 01);
        
        var progdismiss    : uint (bits: 01);
        var gl_resetn      : uint (bits: 01);
        var cpu_resetn     : uint (bits: 01);
        var gl_mpnmc       : uint (bits: 01);
        var cpu_mpnmc      : uint (bits: 01);
        var gl_hompnsamp_init : uint (bits: 01);
        var gl_homrnsamr_init : uint (bits: 01);
        var gl_homxnsamx_init : uint (bits: 01);
        var gl_homynsamy_init : uint (bits: 01);
        var gl_homnsam     : uint (bits: 04);
        var cpu_homnsam    : uint (bits: 04);
        var avis           : uint (bits: 01);
        var idlemaster     : uint (bits: 01);
        var idlewakeup     : uint (bits:  1);
        var idlecpu        : uint (bits:  1);
        var idlecpud       : uint (bits:  1);
        var iack           : uint (bits:  1);
        var int            : uint (bits: 22);
        var nmi            : uint (bits:  1);
        var buserror       : uint (bits:  1);
        var debugerror     : uint (bits:  1);
        var perrack        : uint (bits:  1);
        
        var gl_privenable  : uint (bits:  1);
        var mpcxviol       : uint (bits:  1);
        var mpcbviol       : uint (bits:  1);
        var mpccviol       : uint (bits:  1);
        var mpcdviol       : uint (bits:  1);
        var mpceviol       : uint (bits:  1);
        var mpcfviol       : uint (bits:  1);
        var cpu_privenable : uint (bits:  1);
        var ppm            : uint (bits:  1);
        var bpm            : uint (bits:  1);
        var cpm            : uint (bits:  1);
        var dpm            : uint (bits:  1);
        var epm            : uint (bits:  1);
        var fpm            : uint (bits:  1);
        var rptpcvio       : uint (bits: 24);
        var rptintr        : uint (bits:  1);
        var rptpvio        : uint (bits:  1);
        var rptbvio        : uint (bits:  1);
        var rptcvio        : uint (bits:  1);
        var rptdvio        : uint (bits:  1);
        var rptiodvio      : uint (bits:  1);
        var rptevio        : uint (bits:  1);
        var rptioevio      : uint (bits:  1);
        var rptfvio        : uint (bits:  1);
        var rptpm          : uint (bits:  1);
        var dbgmpviol      : uint (bits:  1);
        var xdmadmpovr     : uint (bits:  1);
        var xdmaempovr     : uint (bits:  1);
        var emiodready     : uint (bits:  1);
        var emioeready     : uint (bits:  1);
        
        -----------------------------
        -- for getting top signals --
        -----------------------------
        if( paths.preq.hdl_path()               != "" ) { preq              = paths.preq$               } else { preq              = 1 };
        if( paths.pready.hdl_path()             != "" ) { pready            = paths.pready$             } else { pready            = 1 };
        if( paths.pdismiss.hdl_path()           != "" ) { pdismiss          = paths.pdismiss$           } else { pdismiss          = 0 };
        
        if( paths.breq.hdl_path()               != "" ) { breq              = paths.breq$               } else { breq              = 1 };
        if( paths.bready.hdl_path()             != "" ) { bready            = paths.bready$             } else { bready            = 1 };
        
        if( paths.creq.hdl_path()               != "" ) { creq              = paths.creq$               } else { creq              = 1 };
        if( paths.cready.hdl_path()             != "" ) { cready            = paths.cready$             } else { cready            = 1 };
        
        if( paths.dreq.hdl_path()               != "" ) { dreq              = paths.dreq$               } else { dreq              = 1 };
        if( paths.iodreq.hdl_path()             != "" ) { iodreq            = paths.iodreq$             } else { iodreq            = 1 };
        if( paths.dready.hdl_path()             != "" ) { dready            = paths.dready$             } else { dready            = 1 };
        
        if( paths.ereq.hdl_path()               != "" ) { ereq              = paths.ereq$               } else { ereq              = 1 };
        if( paths.ioereq.hdl_path()             != "" ) { ioereq            = paths.ioereq$             } else { ioereq            = 1 };
        if( paths.eready.hdl_path()             != "" ) { eready            = paths.eready$             } else { eready            = 1 };
        
        if( paths.freq.hdl_path()               != "" ) { freq              = paths.freq$               } else { freq              = 1 };
        if( paths.fready.hdl_path()             != "" ) { fready            = paths.fready$             } else { fready            = 1 };
        
        if( paths.wrbyte.hdl_path()             != "" ) { wrbyte            = paths.wrbyte$             } else { wrbyte            = 0 };
        if( paths.wrlsbe.hdl_path()             != "" ) { wrlsbe            = paths.wrlsbe$             } else { wrlsbe            = 0 };
        if( paths.rdbyte.hdl_path()             != "" ) { rdbyte            = paths.rdbyte$             } else { rdbyte            = 0 };
        if( paths.rdlsbe.hdl_path()             != "" ) { rdlsbe            = paths.rdlsbe$             } else { rdlsbe            = 0 };
        if( paths.doubler.hdl_path()            != "" ) { doubler           = paths.doubler$            } else { doubler           = 0 };
        if( paths.doublew.hdl_path()            != "" ) { doublew           = paths.doublew$            } else { doublew           = 0 };
        if( paths.dblcoef.hdl_path()            != "" ) { dblcoef           = paths.dblcoef$            } else { dblcoef           = 0 };
        if( paths.lock.hdl_path()               != "" ) { lock              = paths.lock$               } else { lock              = 0 };
        
        if( paths.dmapr.hdl_path()              != "" ) { dmapr             = paths.dmapr$              } else { dmapr             = 0 };
        if( paths.dmapw.hdl_path()              != "" ) { dmapw             = paths.dmapw$              } else { dmapw             = 0 };
        if( paths.dbugd.hdl_path()              != "" ) { dbugd             = paths.dbugd$              } else { dbugd             = 0 };
        if( paths.dbuge.hdl_path()              != "" ) { dbuge             = paths.dbuge$              } else { dbuge             = 0 };
        if( paths.debugenable.hdl_path()        != "" ) { debugenable       = paths.debugenable$        } else { debugenable       = 0 };
        if( paths.abortready.hdl_path()         != "" ) { abortready        = paths.abortready$         } else { abortready        = 0 };
        if( paths.suspend.hdl_path()            != "" ) { suspend           = paths.suspend$            } else { suspend           = 1 };
        
        if( paths.progdismiss.hdl_path()        != "" ) { progdismiss       = paths.progdismiss$        } else { progdismiss       = 1 };
        if( paths.gl_resetn.hdl_path()          != "" ) { gl_resetn         = paths.gl_resetn$          } else { gl_resetn         = 0 };
        if( paths.cpu_resetn.hdl_path()         != "" ) { cpu_resetn        = paths.cpu_resetn$         } else { cpu_resetn        = 1 };
        if( paths.gl_mpnmc.hdl_path()           != "" ) { gl_mpnmc          = paths.gl_mpnmc$           } else { gl_mpnmc          = 1 };
        if( paths.gl_hompnsamp_init.hdl_path()  != "" ) { gl_hompnsamp_init = paths.gl_hompnsamp_init$  } else { gl_hompnsamp_init = 1 };
        if( paths.gl_homrnsamr_init.hdl_path()  != "" ) { gl_homrnsamr_init = paths.gl_homrnsamr_init$  } else { gl_homrnsamr_init = 1 };
        if( paths.gl_homxnsamx_init.hdl_path()  != "" ) { gl_homxnsamx_init = paths.gl_homxnsamx_init$  } else { gl_homxnsamx_init = 1 };
        if( paths.gl_homynsamy_init.hdl_path()  != "" ) { gl_homynsamy_init = paths.gl_homynsamy_init$  } else { gl_homynsamy_init = 1 };
        if( paths.gl_homnsam.hdl_path()         != "" ) { gl_homnsam        = paths.gl_homnsam$         } else { gl_homnsam        = 1 };
        if( paths.idlecpu.hdl_path()            != "" ) { idlecpu           = paths.idlecpu$            } else { idlecpu           = 0 };
        if( paths.idlecpud.hdl_path()           != "" ) { idlecpud          = paths.idlecpu$            } else { idlecpud          = 0 };
        if( paths.int.hdl_path()                != "" ) { int               = paths.int$                } else { int               = 0x3fffff };
        if( paths.nmi.hdl_path()                != "" ) { nmi               = paths.nmi$                } else { nmi               = 1 };
        if( paths.buserror.hdl_path()           != "" ) { buserror          = paths.buserror$           } else { buserror          = 1 };
        if( paths.debugerror.hdl_path()         != "" ) { debugerror        = paths.debugerror$         } else { debugerror        = 1 };
        
        if( paths.gl_privenable.hdl_path()      != "" ) { gl_privenable     = paths.gl_privenable$      } else { gl_privenable     = 0 };
        if( paths.mpcxviol.hdl_path()           != "" ) { mpcxviol          = paths.mpcxviol$           } else { mpcxviol          = 0 };
        if( paths.mpcbviol.hdl_path()           != "" ) { mpcbviol          = paths.mpcbviol$           } else { mpcbviol          = 0 };
        if( paths.mpccviol.hdl_path()           != "" ) { mpccviol          = paths.mpccviol$           } else { mpccviol          = 0 };
        if( paths.mpcdviol.hdl_path()           != "" ) { mpcdviol          = paths.mpcdviol$           } else { mpcdviol          = 0 };
        if( paths.mpceviol.hdl_path()           != "" ) { mpceviol          = paths.mpceviol$           } else { mpceviol          = 0 };
        if( paths.mpcfviol.hdl_path()           != "" ) { mpcfviol          = paths.mpcfviol$           } else { mpcfviol          = 0 };
        if( paths.dbgmpviol.hdl_path()          != "" ) { dbgmpviol         = paths.dbgmpviol$          } else { dbgmpviol         = 0 };
        if( paths.xdmadmpovr.hdl_path()         != "" ) { xdmadmpovr        = paths.xdmadmpovr$         } else { xdmadmpovr        = 0 };
        if( paths.xdmaempovr.hdl_path()         != "" ) { xdmaempovr        = paths.xdmaempovr$         } else { xdmaempovr        = 0 };
        if( paths.emiodready.hdl_path()         != "" ) { emiodready        = paths.emiodready$         } else { emiodready        = 1 };
        if( paths.emioeready.hdl_path()         != "" ) { emioeready        = paths.emioeready$         } else { emioeready        = 1 };

        --------------------------------
        -- speed improvement          --
        -- remove unused data capture --
        --------------------------------
        if( config.module_type     != MASTER ) { -- Master driver does't use own driving value 
           if( paths.pabus.hdl_path()              != "" ) { pabus          = paths.pabus$          } else { pabus          = 0 };
           if( paths.pdbus.hdl_path()              != "" ) { pdbus          = paths.pdbus$          } else { pdbus          = 0 };
           if( paths.babus.hdl_path()              != "" ) { babus          = paths.babus$          } else { babus          = 0 };
           if( paths.bdbus.hdl_path()              != "" ) { bdbus          = paths.bdbus$          } else { bdbus          = 0 };
           if( paths.cabus.hdl_path()              != "" ) { cabus          = paths.cabus$          } else { cabus          = 0 };
           if( paths.cdbus.hdl_path()              != "" ) { cdbus          = paths.cdbus $         } else { cdbus          = 0 };
           if( paths.dabus.hdl_path()              != "" ) { dabus          = paths.dabus$          } else { dabus          = 0 };
           if( paths.ddbus.hdl_path()              != "" ) { ddbus          = paths.ddbus$          } else { ddbus          = 0 };
           if( paths.eabus.hdl_path()              != "" ) { eabus          = paths.eabus$          } else { eabus          = 0 };
           if( paths.edbus.hdl_path()              != "" ) { edbus          = paths.edbus$          } else { edbus          = 0 };
           if( paths.fabus.hdl_path()              != "" ) { fabus          = paths.fabus$          } else { fabus          = 0 };
           if( paths.fdbus.hdl_path()              != "" ) { fdbus          = paths.fdbus$          } else { fdbus          = 0 };
           if( paths.cachefreeze.hdl_path()        != "" ) { cachefreeze    = paths.cachefreeze$    } else { cachefreeze    = 0 };
           if( paths.cacheenable.hdl_path()        != "" ) { cacheenable    = paths.cacheenable$    } else { cacheenable    = 0 };
           if( paths.cacheclr.hdl_path()           != "" ) { cacheclr       = paths.cacheclr$       } else { cacheclr       = 0 };
           if( paths.cacheendclr.hdl_path()        != "" ) { cacheendclr    = paths.cacheendclr$    } else { cacheendclr    = 0 };
           if( paths.cpu_mpnmc.hdl_path()          != "" ) { cpu_mpnmc      = paths.cpu_mpnmc$      } else { cpu_mpnmc      = 0 };
           if( paths.cpu_homnsam.hdl_path()        != "" ) { cpu_homnsam    = paths.cpu_homnsam$    } else { cpu_homnsam    = 1 };
           if( paths.avis.hdl_path()               != "" ) { avis           = paths.avis$           } else { avis           = 0 };
           if( paths.idlemaster.hdl_path()         != "" ) { idlemaster     = paths.idlemaster$     } else { idlemaster     = 0 };
           if( paths.idlewakeup.hdl_path()         != "" ) { idlewakeup     = paths.idlewakeup$     } else { idlewakeup     = 0 };
           if( paths.iack.hdl_path()               != "" ) { iack           = paths.iack$           } else { iack           = 1 };
           if( paths.perrack.hdl_path()            != "" ) { perrack        = paths.perrack$        } else { perrack        = 0 };
           if( paths.cpu_privenable.hdl_path()     != "" ) { cpu_privenable = paths.cpu_privenable$ } else { cpu_privenable = 0 };
           if( paths.ppm.hdl_path()                != "" ) { ppm            = paths.ppm$            } else { ppm            = 0 };
           if( paths.bpm.hdl_path()                != "" ) { bpm            = paths.bpm$            } else { bpm            = 0 };
           if( paths.cpm.hdl_path()                != "" ) { cpm            = paths.cpm$            } else { cpm            = 0 };
           if( paths.dpm.hdl_path()                != "" ) { dpm            = paths.dpm$            } else { dpm            = 0 };
           if( paths.epm.hdl_path()                != "" ) { epm            = paths.epm$            } else { epm            = 0 };
           if( paths.fpm.hdl_path()                != "" ) { fpm            = paths.fpm$            } else { fpm            = 0 };
           if( paths.rptpcvio.hdl_path()           != "" ) { rptpcvio       = paths.rptpcvio$       } else { rptpcvio       = 0 };
           if( paths.rptintr.hdl_path()            != "" ) { rptintr        = paths.rptintr$        } else { rptintr        = 0 };
           if( paths.rptpvio.hdl_path()            != "" ) { rptpvio        = paths.rptpvio$        } else { rptpvio        = 0 };
           if( paths.rptbvio.hdl_path()            != "" ) { rptbvio        = paths.rptbvio$        } else { rptbvio        = 0 };
           if( paths.rptcvio.hdl_path()            != "" ) { rptcvio        = paths.rptcvio$        } else { rptcvio        = 0 };
           if( paths.rptdvio.hdl_path()            != "" ) { rptdvio        = paths.rptdvio$        } else { rptdvio        = 0 };
           if( paths.rptiodvio.hdl_path()          != "" ) { rptiodvio      = paths.rptiodvio$      } else { rptiodvio      = 0 };
           if( paths.rptevio.hdl_path()            != "" ) { rptevio        = paths.rptevio$        } else { rptevio        = 0 };
           if( paths.rptioevio.hdl_path()          != "" ) { rptioevio      = paths.rptioevio$      } else { rptioevio      = 0 };
           if( paths.rptfvio.hdl_path()            != "" ) { rptfvio        = paths.rptfvio$        } else { rptfvio        = 0 };
           if( paths.rptpm.hdl_path()              != "" ) { rptpm          = paths.rptpm$          } else { rptpm          = 0 };
        };

        ------------------------
        -- ready for icemaker --
        ------------------------
        -- 
        -- ice_ready : icemaker's ready
        -- ready     : real ready signal that not include icemaker's ready
        -- 
        dbus.current.ice_ready = emiodready;
        ebus.current.ice_ready = emioeready;
        dbus.current.raw_ready = dready;
        ebus.current.raw_ready = eready;
        
        ------------------------
        -- set current values --
        ------------------------
        pbus.current.req              = preq;
        pbus.current.abus             = pabus;
        pbus.current.dbus             = pdbus;
        pbus.current.ready            = pready;
        pbus.current.dismiss          = pdismiss;
        
        bbus.current.req              = breq;
        bbus.current.abus             = babus;
        bbus.current.dbus             = bdbus;
        bbus.current.ready            = bready;
        
        cbus.current.req              = creq;
        cbus.current.abus             = cabus;
        cbus.current.dbus             = cdbus;
        cbus.current.ready            = cready;
        
        dbus.current.req              = dreq;
        dbus.current.ioreq            = iodreq;
        dbus.current.abus             = dabus;
        dbus.current.dbus             = ddbus;
        dbus.current.ready            = dready & dbus.current.ice_ready;
        
        ebus.current.req              = ereq;
        ebus.current.ioreq            = ioereq;
        ebus.current.abus             = eabus;
        ebus.current.dbus             = edbus;
        ebus.current.ready            = eready & ebus.current.ice_ready;
        
        fbus.current.req              = freq;
        fbus.current.abus             = fabus;
        fbus.current.dbus             = fdbus;
        fbus.current.ready            = fready;
        
        qualifier.current.wrbyte      = wrbyte;
        qualifier.current.wrlsbe      = wrlsbe;
        qualifier.current.rdbyte      = rdbyte;
        qualifier.current.rdlsbe      = rdlsbe;
        qualifier.current.doubler     = doubler;
        qualifier.current.doublew     = doublew;
        qualifier.current.dblcoef     = dblcoef;
        qualifier.current.lock        = lock;
        
        emulation.current.dmapr       = dmapr;
        emulation.current.dmapw       = dmapw;
        emulation.current.dbugd       = dbugd;
        emulation.current.dbuge       = dbuge;
        emulation.current.debugenable = debugenable;
        emulation.current.abortready  = abortready;
        emulation.current.suspend     = suspend;
        
        cache.current.cachefreeze     = cachefreeze;
        cache.current.cacheenable     = cacheenable;
        cache.current.cacheclr        = cacheclr;
        cache.current.cacheendclr     = cacheendclr;
        
        system.current.progdismiss    = progdismiss;
        system.current.gl_resetn      = gl_resetn;
        system.current.cpu_resetn     = cpu_resetn & gl_resetn;
        system.current.gl_mpnmc       = gl_mpnmc;
        system.current.cpu_mpnmc      = cpu_mpnmc;
        system.current.gl_hompnsamp_init = gl_hompnsamp_init;
        system.current.gl_homrnsamr_init = gl_homrnsamr_init;
        system.current.gl_homxnsamx_init = gl_homxnsamx_init;
        system.current.gl_homynsamy_init = gl_homynsamy_init;
        system.current.gl_homnsam     = gl_homnsam;
        system.current.cpu_homnsam    = cpu_homnsam;
        system.current.avis           = avis;
        system.current.idlemaster     = idlemaster;
        system.current.idlewakeup     = idlewakeup;
        system.current.idlecpu        = idlecpu;
        system.current.idlecpud       = idlecpud;
        system.current.int            = int;
        system.current.nmi            = nmi;
        system.current.iack           = iack;
        system.current.buserror       = buserror;
        system.current.debugerror     = debugerror;
        system.current.perrack        = perrack;
        
        mp.current.gl_privenable      = gl_privenable;
        mp.current.mpcxviol           = mpcxviol;
        mp.current.mpcbviol           = mpcbviol;
        mp.current.mpccviol           = mpccviol;
        mp.current.mpcdviol           = mpcdviol;
        mp.current.mpceviol           = mpceviol;
        mp.current.mpcfviol           = mpcfviol;
        mp.current.cpu_privenable     = cpu_privenable;
        mp.current.ppm                = ppm;
        mp.current.bpm                = bpm;
        mp.current.cpm                = cpm;
        mp.current.dpm                = dpm;
        mp.current.epm                = epm;
        mp.current.fpm                = fpm;
        mp.current.rptpcvio           = rptpcvio;
        mp.current.rptintr            = rptintr;
        mp.current.rptpvio            = rptpvio;
        mp.current.rptbvio            = rptbvio;
        mp.current.rptcvio            = rptcvio;
        mp.current.rptdvio            = rptdvio;
        mp.current.rptiodvio          = rptiodvio;
        mp.current.rptevio            = rptevio;
        mp.current.rptioevio          = rptioevio;
        mp.current.rptfvio            = rptfvio;
        mp.current.rptpm              = rptpm;
        mp.current.dbgmpviol          = dbgmpviol;
        mp.current.xdmadmpovr         = xdmadmpovr;
        mp.current.xdmaempovr         = xdmaempovr;
        mp.current.emiodready         = emiodready;
        mp.current.emioeready         = emioeready;
        
        ---------------------------
        -- get for unknown info. --
        ---------------------------
        if( config.unknown_drive     != TRUE and not ( config.module_type == MASTER or config.module_type == SLAVE )) {
            if( paths.preq.hdl_path()               != "" ) { pbus.x.req                 = paths.preq.has_x().as_a(uint(bits:1))               } else { pbus.x.req                 = 0 };
            if( paths.pabus.hdl_path()              != "" ) { pbus.x.abus                = paths.pabus.has_x().as_a(uint(bits:24))             } else { pbus.x.abus                = 0 };
            if( paths.pdbus.hdl_path()              != "" ) { pbus.x.dbus                = paths.pdbus.has_x().as_a(uint(bits:32))             } else { pbus.x.dbus                = 0 };
            if( paths.pready.hdl_path()             != "" ) { pbus.x.ready               = paths.pready.has_x().as_a(uint(bits:1))             } else { pbus.x.ready               = 0 };
            if( paths.pdismiss.hdl_path()           != "" ) { pbus.x.dismiss             = paths.pdismiss.has_x().as_a(uint(bits:1))           } else { pbus.x.dismiss             = 0 };
            
            if( paths.breq.hdl_path()               != "" ) { bbus.x.req                 = paths.breq.has_x().as_a(uint(bits:1))               } else { bbus.x.req                 = 0 };
            if( paths.babus.hdl_path()              != "" ) { bbus.x.abus                = paths.babus.has_x().as_a(uint(bits:23))             } else { bbus.x.abus                = 0 };
            if( paths.bdbus.hdl_path()              != "" ) { bbus.x.dbus                = paths.bdbus.has_x().as_a(uint(bits:16))             } else { bbus.x.dbus                = 0 };
            if( paths.bready.hdl_path()             != "" ) { bbus.x.ready               = paths.bready.has_x().as_a(uint(bits:1))             } else { bbus.x.ready               = 0 };
            
            if( paths.creq.hdl_path()               != "" ) { cbus.x.req                 = paths.creq.has_x().as_a(uint(bits:1))               } else { cbus.x.req                 = 0 };
            if( paths.cabus.hdl_path()              != "" ) { cbus.x.abus                = paths.cabus.has_x().as_a(uint(bits:23))             } else { cbus.x.abus                = 0 };
            if( paths.cdbus.hdl_path()              != "" ) { cbus.x.dbus                = paths.cdbus.has_x().as_a(uint(bits:16))             } else { cbus.x.dbus                = 0 };
            if( paths.cready.hdl_path()             != "" ) { cbus.x.ready               = paths.cready.has_x().as_a(uint(bits:1))             } else { cbus.x.ready               = 0 };
            
            if( paths.dreq.hdl_path()               != "" ) { dbus.x.req                 = paths.dreq.has_x().as_a(uint(bits:1))               } else { dbus.x.req                 = 0 };
            if( paths.iodreq.hdl_path()             != "" ) { dbus.x.ioreq               = paths.iodreq.has_x().as_a(uint(bits:1))             } else { dbus.x.ioreq               = 0 };
            if( paths.dabus.hdl_path()              != "" ) { dbus.x.abus                = paths.dabus.has_x().as_a(uint(bits:23))             } else { dbus.x.abus                = 0 };
            if( paths.ddbus.hdl_path()              != "" ) { dbus.x.dbus                = paths.ddbus.has_x().as_a(uint(bits:16))             } else { dbus.x.dbus                = 0 };
            if( paths.dready.hdl_path()             != "" ) { dbus.x.ready               = paths.dready.has_x().as_a(uint(bits:1))             } else { dbus.x.ready               = 0 };
            
            if( paths.ereq.hdl_path()               != "" ) { ebus.x.req                 = paths.ereq.has_x().as_a(uint(bits:1))               } else { ebus.x.req                 = 0 };
            if( paths.ioereq.hdl_path()             != "" ) { ebus.x.ioreq               = paths.ioereq.has_x().as_a(uint(bits:1))             } else { ebus.x.ioreq               = 0 };
            if( paths.eabus.hdl_path()              != "" ) { ebus.x.abus                = paths.eabus.has_x().as_a(uint(bits:23))             } else { ebus.x.abus                = 0 };
            if( paths.edbus.hdl_path()              != "" ) { ebus.x.dbus                = paths.edbus.has_x().as_a(uint(bits:16))             } else { ebus.x.dbus                = 0 };
            if( paths.eready.hdl_path()             != "" ) { ebus.x.ready               = paths.eready.has_x().as_a(uint(bits:1))             } else { ebus.x.ready               = 0 };
            
            if( paths.freq.hdl_path()               != "" ) { fbus.x.req                 = paths.freq.has_x().as_a(uint(bits:1))               } else { fbus.x.req                 = 0 };
            if( paths.fabus.hdl_path()              != "" ) { fbus.x.abus                = paths.fabus.has_x().as_a(uint(bits:23))             } else { fbus.x.abus                = 0 };
            if( paths.fdbus.hdl_path()              != "" ) { fbus.x.dbus                = paths.fdbus.has_x().as_a(uint(bits:16))             } else { fbus.x.dbus                = 0 };
            if( paths.fready.hdl_path()             != "" ) { fbus.x.ready               = paths.fready.has_x().as_a(uint(bits:1))             } else { fbus.x.ready               = 0 };
            
            if( paths.wrbyte.hdl_path()             != "" ) { qualifier.x.wrbyte         = paths.wrbyte.has_x().as_a(uint(bits:1))             } else { qualifier.x.wrbyte         = 0 };
            if( paths.wrlsbe.hdl_path()             != "" ) { qualifier.x.wrlsbe         = paths.wrlsbe.has_x().as_a(uint(bits:1))             } else { qualifier.x.wrlsbe         = 0 };
            if( paths.rdbyte.hdl_path()             != "" ) { qualifier.x.rdbyte         = paths.rdbyte.has_x().as_a(uint(bits:1))             } else { qualifier.x.rdbyte         = 0 };
            if( paths.rdlsbe.hdl_path()             != "" ) { qualifier.x.rdlsbe         = paths.rdlsbe.has_x().as_a(uint(bits:1))             } else { qualifier.x.rdlsbe         = 0 };
            if( paths.doubler.hdl_path()            != "" ) { qualifier.x.doubler        = paths.doubler.has_x().as_a(uint(bits:1))            } else { qualifier.x.doubler        = 0 };
            if( paths.doublew.hdl_path()            != "" ) { qualifier.x.doublew        = paths.doublew.has_x().as_a(uint(bits:1))            } else { qualifier.x.doublew        = 0 };
            if( paths.dblcoef.hdl_path()            != "" ) { qualifier.x.dblcoef        = paths.dblcoef.has_x().as_a(uint(bits:1))            } else { qualifier.x.dblcoef        = 0 };
            if( paths.lock.hdl_path()               != "" ) { qualifier.x.lock           = paths.lock.has_x().as_a(uint(bits:1))               } else { qualifier.x.lock           = 0 };
                
            if( paths.dmapr.hdl_path()              != "" ) { emulation.x.dmapr          = paths.dmapr.has_x().as_a(uint(bits:1))              } else { emulation.x.dmapr          = 0 };
            if( paths.dmapw.hdl_path()              != "" ) { emulation.x.dmapw          = paths.dmapw.has_x().as_a(uint(bits:1))              } else { emulation.x.dmapw          = 0 };
            if( paths.dbugd.hdl_path()              != "" ) { emulation.x.dbugd          = paths.dbugd.has_x().as_a(uint(bits:1))              } else { emulation.x.dbugd          = 0 };
            if( paths.dbuge.hdl_path()              != "" ) { emulation.x.dbuge          = paths.dbuge.has_x().as_a(uint(bits:1))              } else { emulation.x.dbuge          = 0 };
            
            if( paths.debugenable.hdl_path()        != "" ) { emulation.x.debugenable    = paths.debugenable.has_x().as_a(uint(bits:1))        } else { emulation.x.debugenable    = 0 };
            if( paths.abortready.hdl_path()         != "" ) { emulation.x.abortready     = paths.abortready.has_x().as_a(uint(bits:1))         } else { emulation.x.abortready     = 0 };
            if( paths.suspend.hdl_path()            != "" ) { emulation.x.suspend        = paths.suspend.has_x().as_a(uint(bits:1))            } else { emulation.x.suspend        = 0 };
            
            if( paths.cachefreeze.hdl_path()        != "" ) { cache.x.cachefreeze        = paths.cachefreeze.has_x().as_a(uint(bits:1))        } else { cache.x.cachefreeze        = 0 };
            if( paths.cacheenable.hdl_path()        != "" ) { cache.x.cacheenable        = paths.cacheenable.has_x().as_a(uint(bits:1))        } else { cache.x.cacheenable        = 0 };
            if( paths.cacheclr.hdl_path()           != "" ) { cache.x.cacheclr           = paths.cacheclr.has_x().as_a(uint(bits:1))           } else { cache.x.cacheclr           = 0 };
            if( paths.cacheendclr.hdl_path()        != "" ) { cache.x.cacheendclr        = paths.cacheendclr.has_x().as_a(uint(bits:1))        } else { cache.x.cacheendclr        = 0 };
            
            if( paths.progdismiss.hdl_path()        != "" ) { system.x.progdismiss       = paths.progdismiss.has_x().as_a(uint(bits:1))        } else { system.x.progdismiss       = 0 };
            if( paths.gl_resetn.hdl_path()          != "" ) { system.x.gl_resetn         = paths.gl_resetn.has_x().as_a(uint(bits:1))          } else { system.x.gl_resetn         = 0 };
            if( paths.cpu_resetn.hdl_path()         != "" ) { system.x.cpu_resetn        = paths.cpu_resetn.has_x().as_a(uint(bits:1))         } else { system.x.cpu_resetn        = 0 };
            if( paths.gl_mpnmc.hdl_path()           != "" ) { system.x.gl_mpnmc          = paths.gl_mpnmc.has_x().as_a(uint(bits:1))           } else { system.x.gl_mpnmc          = 0 };
            if( paths.cpu_mpnmc.hdl_path()          != "" ) { system.x.cpu_mpnmc         = paths.gl_mpnmc.has_x().as_a(uint(bits:1))           } else { system.x.cpu_mpnmc         = 0 };
            if( paths.gl_hompnsamp_init.hdl_path()  != "" ) { system.x.gl_hompnsamp_init = paths.gl_hompnsamp_init.has_x().as_a(uint(bits:1))  } else { system.x.gl_hompnsamp_init = 1 };
            if( paths.gl_homrnsamr_init.hdl_path()  != "" ) { system.x.gl_homrnsamr_init = paths.gl_homrnsamr_init.has_x().as_a(uint(bits:1))  } else { system.x.gl_homrnsamr_init = 1 };
            if( paths.gl_homxnsamx_init.hdl_path()  != "" ) { system.x.gl_homxnsamx_init = paths.gl_homxnsamx_init.has_x().as_a(uint(bits:1))  } else { system.x.gl_homxnsamx_init = 1 };
            if( paths.gl_homynsamy_init.hdl_path()  != "" ) { system.x.gl_homynsamy_init = paths.gl_homynsamy_init.has_x().as_a(uint(bits:1))  } else { system.x.gl_homynsamy_init = 1 };
            if( paths.gl_homnsam.hdl_path()         != "" ) { system.x.gl_homnsam        = paths.gl_homnsam.has_x().as_a(uint(bits:1))         } else { system.x.gl_homnsam        = 1 };
            if( paths.cpu_homnsam.hdl_path()        != "" ) { system.x.cpu_homnsam       = paths.cpu_homnsam.has_x().as_a(uint(bits:1))        } else { system.x.cpu_homnsam       = 1 };
            if( paths.avis.hdl_path()               != "" ) { system.x.avis              = paths.avis.has_x().as_a(uint(bits:1))               } else { system.x.avis              = 0 };
            if( paths.idlemaster.hdl_path()         != "" ) { system.x.idlemaster        = paths.idlemaster.has_x().as_a(uint(bits:1))         } else { system.x.idlemaster        = 0 };
            if( paths.idlewakeup.hdl_path()         != "" ) { system.x.idlewakeup        = paths.idlewakeup.has_x().as_a(uint(bits:1))         } else { system.x.idlewakeup        = 0 };
            if( paths.idlecpu.hdl_path()            != "" ) { system.x.idlecpu           = paths.idlecpu.has_x().as_a(uint(bits:1))            } else { system.x.idlecpu           = 0 };
            if( paths.idlecpud.hdl_path()           != "" ) { system.x.idlecpud          = paths.idlecpud.has_x().as_a(uint(bits:1))           } else { system.x.idlecpud          = 0 };
            if( paths.int.hdl_path()                != "" ) { system.x.int               = paths.int.has_x().as_a(uint(bits:1))                } else { system.x.int               = 0 };
            if( paths.nmi.hdl_path()                != "" ) { system.x.nmi               = paths.nmi.has_x().as_a(uint(bits:1))                } else { system.x.nmi               = 0 };
            if( paths.iack.hdl_path()               != "" ) { system.x.iack              = paths.iack.has_x().as_a(uint(bits:1))               } else { system.x.iack              = 0 };
            if( paths.buserror.hdl_path()           != "" ) { system.x.buserror          = paths.buserror.has_x().as_a(uint(bits:1))           } else { system.x.buserror          = 0 };
            if( paths.debugerror.hdl_path()         != "" ) { system.x.debugerror        = paths.debugerror.has_x().as_a(uint(bits:1))         } else { system.x.debugerror        = 0 };
            if( paths.perrack.hdl_path()            != "" ) { system.x.perrack           = paths.perrack.has_x().as_a(uint(bits:1))            } else { system.x.perrack           = 0 };
            
            if( paths.gl_privenable.hdl_path()      != "" ) { mp.x.gl_privenable         = paths.gl_privenable.has_x().as_a(uint(bits:1))      } else { mp.x.gl_privenable         = 0 };
            if( paths.mpcxviol.hdl_path()           != "" ) { mp.x.mpcxviol              = paths.mpcxviol.has_x().as_a(uint(bits:1))           } else { mp.x.mpcxviol              = 0 };
            if( paths.mpcbviol.hdl_path()           != "" ) { mp.x.mpcbviol              = paths.mpcbviol.has_x().as_a(uint(bits:1))           } else { mp.x.mpcbviol              = 0 };
            if( paths.mpccviol.hdl_path()           != "" ) { mp.x.mpccviol              = paths.mpccviol.has_x().as_a(uint(bits:1))           } else { mp.x.mpccviol              = 0 };
            if( paths.mpcdviol.hdl_path()           != "" ) { mp.x.mpcdviol              = paths.mpcdviol.has_x().as_a(uint(bits:1))           } else { mp.x.mpcdviol              = 0 };
            if( paths.mpceviol.hdl_path()           != "" ) { mp.x.mpceviol              = paths.mpceviol.has_x().as_a(uint(bits:1))           } else { mp.x.mpceviol              = 0 };
            if( paths.mpcfviol.hdl_path()           != "" ) { mp.x.mpcfviol              = paths.mpcfviol.has_x().as_a(uint(bits:1))           } else { mp.x.mpcfviol              = 0 };
            if( paths.cpu_privenable.hdl_path()     != "" ) { mp.x.cpu_privenable        = paths.cpu_privenable.has_x().as_a(uint(bits:1))     } else { mp.x.cpu_privenable        = 0 };
            if( paths.ppm.hdl_path()                != "" ) { mp.x.ppm                   = paths.ppm.has_x().as_a(uint(bits:1))                } else { mp.x.ppm                   = 0 };
            if( paths.bpm.hdl_path()                != "" ) { mp.x.bpm                   = paths.bpm.has_x().as_a(uint(bits:1))                } else { mp.x.bpm                   = 0 };
            if( paths.cpm.hdl_path()                != "" ) { mp.x.cpm                   = paths.cpm.has_x().as_a(uint(bits:1))                } else { mp.x.cpm                   = 0 };
            if( paths.dpm.hdl_path()                != "" ) { mp.x.dpm                   = paths.dpm.has_x().as_a(uint(bits:1))                } else { mp.x.dpm                   = 0 };
            if( paths.epm.hdl_path()                != "" ) { mp.x.epm                   = paths.epm.has_x().as_a(uint(bits:1))                } else { mp.x.epm                   = 0 };
            if( paths.fpm.hdl_path()                != "" ) { mp.x.fpm                   = paths.fpm.has_x().as_a(uint(bits:1))                } else { mp.x.fpm                   = 0 };
            if( paths.rptpcvio.hdl_path()           != "" ) { mp.x.rptpcvio              = paths.rptpcvio.has_x().as_a(uint(bits:1))           } else { mp.x.rptpcvio              = 0 };
            if( paths.rptintr.hdl_path()            != "" ) { mp.x.rptintr               = paths.rptintr.has_x().as_a(uint(bits:1))            } else { mp.x.rptintr               = 0 };
            if( paths.rptpvio.hdl_path()            != "" ) { mp.x.rptpvio               = paths.rptpvio.has_x().as_a(uint(bits:1))            } else { mp.x.rptpvio               = 0 };
            if( paths.rptbvio.hdl_path()            != "" ) { mp.x.rptbvio               = paths.rptbvio.has_x().as_a(uint(bits:1))            } else { mp.x.rptbvio               = 0 };
            if( paths.rptcvio.hdl_path()            != "" ) { mp.x.rptcvio               = paths.rptcvio.has_x().as_a(uint(bits:1))            } else { mp.x.rptcvio               = 0 };
            if( paths.rptdvio.hdl_path()            != "" ) { mp.x.rptdvio               = paths.rptdvio.has_x().as_a(uint(bits:1))            } else { mp.x.rptdvio               = 0 };
            if( paths.rptiodvio.hdl_path()          != "" ) { mp.x.rptiodvio             = paths.rptiodvio.has_x().as_a(uint(bits:1))          } else { mp.x.rptiodvio             = 0 };
            if( paths.rptevio.hdl_path()            != "" ) { mp.x.rptevio               = paths.rptevio.has_x().as_a(uint(bits:1))            } else { mp.x.rptevio               = 0 };
            if( paths.rptioevio.hdl_path()          != "" ) { mp.x.rptioevio             = paths.rptioevio.has_x().as_a(uint(bits:1))          } else { mp.x.rptioevio             = 0 };
            if( paths.rptfvio.hdl_path()            != "" ) { mp.x.rptfvio               = paths.rptfvio.has_x().as_a(uint(bits:1))            } else { mp.x.rptfvio               = 0 };
            if( paths.rptpm.hdl_path()              != "" ) { mp.x.rptpm                 = paths.rptpm.has_x().as_a(uint(bits:1))              } else { mp.x.rptpm                 = 0 };
            if( paths.dbgmpviol.hdl_path()          != "" ) { mp.x.dbgmpviol             = paths.dbgmpviol.has_x().as_a(uint(bits:1))          } else { mp.x.dbgmpviol             = 0 };
            if( paths.xdmadmpovr.hdl_path()         != "" ) { mp.x.xdmadmpovr            = paths.xdmadmpovr.has_x().as_a(uint(bits:1))         } else { mp.x.xdmadmpovr            = 0 };
            if( paths.xdmaempovr.hdl_path()         != "" ) { mp.x.xdmaempovr            = paths.xdmaempovr.has_x().as_a(uint(bits:1))         } else { mp.x.xdmaempovr            = 0 };
            if( paths.emiodready.hdl_path()         != "" ) { mp.x.emiodready            = paths.emiodready.has_x().as_a(uint(bits:1))         } else { mp.x.emiodready            = 0 };
            if( paths.emioeready.hdl_path()         != "" ) { mp.x.emioeready            = paths.emioeready.has_x().as_a(uint(bits:1))         } else { mp.x.emioeready            = 0 };
            
            -----------------------
            -- get for hiz info. --
            -----------------------
            if( paths.preq.hdl_path()               != "" ) { pbus.z.req                 = paths.preq.has_z().as_a(uint(bits:1))               } else { pbus.z.req                 = 0 };
            if( paths.pabus.hdl_path()              != "" ) { pbus.z.abus                = paths.pabus.has_z().as_a(uint(bits:24))             } else { pbus.z.abus                = 0 };
            if( paths.pdbus.hdl_path()              != "" ) { pbus.z.dbus                = paths.pdbus.has_z().as_a(uint(bits:32))             } else { pbus.z.dbus                = 0 };
            if( paths.pready.hdl_path()             != "" ) { pbus.z.ready               = paths.pready.has_z().as_a(uint(bits:1))             } else { pbus.z.ready               = 0 };
            if( paths.pdismiss.hdl_path()           != "" ) { pbus.z.dismiss             = paths.pdismiss.has_z().as_a(uint(bits:1))           } else { pbus.z.dismiss             = 0 };
            
            if( paths.breq.hdl_path()               != "" ) { bbus.z.req                 = paths.breq.has_z().as_a(uint(bits:1))               } else { bbus.z.req                 = 0 };
            if( paths.babus.hdl_path()              != "" ) { bbus.z.abus                = paths.babus.has_z().as_a(uint(bits:23))             } else { bbus.z.abus                = 0 };
            if( paths.bdbus.hdl_path()              != "" ) { bbus.z.dbus                = paths.bdbus.has_z().as_a(uint(bits:16))             } else { bbus.z.dbus                = 0 };
            if( paths.bready.hdl_path()             != "" ) { bbus.z.ready               = paths.bready.has_z().as_a(uint(bits:1))             } else { bbus.z.ready               = 0 };
            
            if( paths.creq.hdl_path()               != "" ) { cbus.z.req                 = paths.creq.has_z().as_a(uint(bits:1))               } else { cbus.z.req                 = 0 };
            if( paths.cabus.hdl_path()              != "" ) { cbus.z.abus                = paths.cabus.has_z().as_a(uint(bits:23))             } else { cbus.z.abus                = 0 };
            if( paths.cdbus.hdl_path()              != "" ) { cbus.z.dbus                = paths.cdbus.has_z().as_a(uint(bits:16))             } else { cbus.z.dbus                = 0 };
            if( paths.cready.hdl_path()             != "" ) { cbus.z.ready               = paths.cready.has_z().as_a(uint(bits:1))             } else { cbus.z.ready               = 0 };
            
            if( paths.dreq.hdl_path()               != "" ) { dbus.z.req                 = paths.dreq.has_z().as_a(uint(bits:1))               } else { dbus.z.req                 = 0 };
            if( paths.iodreq.hdl_path()             != "" ) { dbus.z.ioreq               = paths.iodreq.has_z().as_a(uint(bits:1))             } else { dbus.z.ioreq               = 0 };
            if( paths.dabus.hdl_path()              != "" ) { dbus.z.abus                = paths.dabus.has_z().as_a(uint(bits:23))             } else { dbus.z.abus                = 0 };
            if( paths.ddbus.hdl_path()              != "" ) { dbus.z.dbus                = paths.ddbus.has_z().as_a(uint(bits:16))             } else { dbus.z.dbus                = 0 };
            if( paths.dready.hdl_path()             != "" ) { dbus.z.ready               = paths.dready.has_z().as_a(uint(bits:1))             } else { dbus.z.ready               = 0 };
            
            if( paths.ereq.hdl_path()               != "" ) { ebus.z.req                 = paths.ereq.has_z().as_a(uint(bits:1))               } else { ebus.z.req                 = 0 };
            if( paths.ioereq.hdl_path()             != "" ) { ebus.z.ioreq               = paths.ioereq.has_z().as_a(uint(bits:1))             } else { ebus.z.ioreq               = 0 };
            if( paths.eabus.hdl_path()              != "" ) { ebus.z.abus                = paths.eabus.has_z().as_a(uint(bits:23))             } else { ebus.z.abus                = 0 };
            if( paths.edbus.hdl_path()              != "" ) { ebus.z.dbus                = paths.edbus.has_z().as_a(uint(bits:16))             } else { ebus.z.dbus                = 0 };
            if( paths.eready.hdl_path()             != "" ) { ebus.z.ready               = paths.eready.has_z().as_a(uint(bits:1))             } else { ebus.z.ready               = 0 };
            
            if( paths.freq.hdl_path()               != "" ) { fbus.z.req                 = paths.freq.has_z().as_a(uint(bits:1))               } else { fbus.z.req                 = 0 };
            if( paths.fabus.hdl_path()              != "" ) { fbus.z.abus                = paths.fabus.has_z().as_a(uint(bits:23))             } else { fbus.z.abus                = 0 };
            if( paths.fdbus.hdl_path()              != "" ) { fbus.z.dbus                = paths.fdbus.has_z().as_a(uint(bits:1))              } else { fbus.z.dbus                = 0 };
            if( paths.fready.hdl_path()             != "" ) { fbus.z.ready               = paths.fready.has_z().as_a(uint(bits:1))             } else { fbus.z.ready               = 0 };
            
            if( paths.wrbyte.hdl_path()             != "" ) { qualifier.z.wrbyte         = paths.wrbyte.has_z().as_a(uint(bits:1))             } else { qualifier.z.wrbyte         = 0 };
            if( paths.wrlsbe.hdl_path()             != "" ) { qualifier.z.wrlsbe         = paths.wrlsbe.has_z().as_a(uint(bits:1))             } else { qualifier.z.wrlsbe         = 0 };
            if( paths.rdbyte.hdl_path()             != "" ) { qualifier.z.rdbyte         = paths.rdbyte.has_z().as_a(uint(bits:1))             } else { qualifier.z.rdbyte         = 0 };
            if( paths.rdlsbe.hdl_path()             != "" ) { qualifier.z.rdlsbe         = paths.rdlsbe.has_z().as_a(uint(bits:1))             } else { qualifier.z.rdlsbe         = 0 };
            if( paths.doubler.hdl_path()            != "" ) { qualifier.z.doubler        = paths.doubler.has_z().as_a(uint(bits:1))            } else { qualifier.z.doubler        = 0 };
            if( paths.doublew.hdl_path()            != "" ) { qualifier.z.doublew        = paths.doublew.has_z().as_a(uint(bits:1))            } else { qualifier.z.doublew        = 0 };
            if( paths.dblcoef.hdl_path()            != "" ) { qualifier.z.dblcoef        = paths.dblcoef.has_z().as_a(uint(bits:1))            } else { qualifier.z.dblcoef        = 0 };
            if( paths.lock.hdl_path()               != "" ) { qualifier.z.lock           = paths.lock.has_z().as_a(uint(bits:1))               } else { qualifier.z.lock           = 0 };
            
            if( paths.dmapr.hdl_path()              != "" ) { emulation.z.dmapr          = paths.dmapr.has_z().as_a(uint(bits:1))              } else { emulation.z.dmapr          = 0 };
            if( paths.dmapw.hdl_path()              != "" ) { emulation.z.dmapw          = paths.dmapw.has_z().as_a(uint(bits:1))              } else { emulation.z.dmapw          = 0 };
            if( paths.dbugd.hdl_path()              != "" ) { emulation.z.dbugd          = paths.dbugd.has_z().as_a(uint(bits:1))              } else { emulation.z.dbugd          = 0 };
            if( paths.dbuge.hdl_path()              != "" ) { emulation.z.dbuge          = paths.dbuge.has_z().as_a(uint(bits:1))              } else { emulation.z.dbuge          = 0 };
            
            if( paths.debugenable.hdl_path()        != "" ) { emulation.z.debugenable    = paths.debugenable.has_z().as_a(uint(bits:1))        } else { emulation.z.debugenable    = 0 };
            if( paths.abortready.hdl_path()         != "" ) { emulation.z.abortready     = paths.abortready.has_z().as_a(uint(bits:1))         } else { emulation.z.abortready     = 0 };
            if( paths.suspend.hdl_path()            != "" ) { emulation.z.suspend        = paths.suspend.has_z().as_a(uint(bits:1))            } else { emulation.z.suspend        = 0 };
            
            if( paths.cachefreeze.hdl_path()        != "" ) { cache.z.cachefreeze        = paths.cachefreeze.has_z().as_a(uint(bits:1))        } else { cache.z.cachefreeze        = 0 };
            if( paths.cacheenable.hdl_path()        != "" ) { cache.z.cacheenable        = paths.cacheenable.has_z().as_a(uint(bits:1))        } else { cache.z.cacheenable        = 0 };
            if( paths.cacheclr.hdl_path()           != "" ) { cache.z.cacheclr           = paths.cacheclr.has_z().as_a(uint(bits:1))           } else { cache.z.cacheclr           = 0 };
            if( paths.cacheendclr.hdl_path()        != "" ) { cache.z.cacheendclr        = paths.cacheendclr.has_z().as_a(uint(bits:1))        } else { cache.z.cacheendclr        = 0 };
            
            if( paths.progdismiss.hdl_path()        != "" ) { system.z.progdismiss       = paths.progdismiss.has_z().as_a(uint(bits:1))        } else { system.z.progdismiss       = 0 };
            if( paths.gl_resetn.hdl_path()          != "" ) { system.z.gl_resetn         = paths.gl_resetn.has_z().as_a(uint(bits:1))          } else { system.z.gl_resetn         = 0 };
            if( paths.cpu_resetn.hdl_path()         != "" ) { system.z.cpu_resetn        = paths.cpu_resetn.has_z().as_a(uint(bits:1))         } else { system.z.cpu_resetn        = 0 };
            if( paths.gl_mpnmc.hdl_path()           != "" ) { system.z.gl_mpnmc          = paths.gl_mpnmc.has_z().as_a(uint(bits:1))           } else { system.z.gl_mpnmc          = 0 };
            if( paths.cpu_mpnmc.hdl_path()          != "" ) { system.z.cpu_mpnmc         = paths.gl_mpnmc.has_z().as_a(uint(bits:1))           } else { system.z.cpu_mpnmc         = 0 };
            if( paths.gl_hompnsamp_init.hdl_path()  != "" ) { system.z.gl_hompnsamp_init = paths.gl_hompnsamp_init.has_z().as_a(uint(bits:1))  } else { system.z.gl_hompnsamp_init = 1 };
            if( paths.gl_homrnsamr_init.hdl_path()  != "" ) { system.z.gl_homrnsamr_init = paths.gl_homrnsamr_init.has_z().as_a(uint(bits:1))  } else { system.z.gl_homrnsamr_init = 1 };
            if( paths.gl_homxnsamx_init.hdl_path()  != "" ) { system.z.gl_homxnsamx_init = paths.gl_homxnsamx_init.has_z().as_a(uint(bits:1))  } else { system.z.gl_homxnsamx_init = 1 };
            if( paths.gl_homynsamy_init.hdl_path()  != "" ) { system.z.gl_homynsamy_init = paths.gl_homynsamy_init.has_z().as_a(uint(bits:1))  } else { system.z.gl_homynsamy_init = 1 };
            if( paths.gl_homnsam.hdl_path()         != "" ) { system.z.gl_homnsam        = paths.gl_homnsam.has_z().as_a(uint(bits:1))         } else { system.z.gl_homnsam        = 1 };
            if( paths.cpu_homnsam.hdl_path()        != "" ) { system.z.cpu_homnsam       = paths.cpu_homnsam.has_z().as_a(uint(bits:1))        } else { system.z.cpu_homnsam       = 1 };
            if( paths.avis.hdl_path()               != "" ) { system.z.avis              = paths.avis.has_z().as_a(uint(bits:1))               } else { system.z.avis              = 0 };
            if( paths.idlemaster.hdl_path()         != "" ) { system.z.idlemaster        = paths.idlemaster.has_z().as_a(uint(bits:1))         } else { system.z.idlemaster        = 0 };
            if( paths.idlewakeup.hdl_path()         != "" ) { system.z.idlewakeup        = paths.idlewakeup.has_z().as_a(uint(bits:1))         } else { system.z.idlewakeup        = 0 };
            if( paths.idlecpu.hdl_path()            != "" ) { system.z.idlecpu           = paths.idlecpu.has_z().as_a(uint(bits:1))            } else { system.z.idlecpu           = 0 };
            if( paths.idlecpud.hdl_path()           != "" ) { system.z.idlecpud          = paths.idlecpud.has_z().as_a(uint(bits:1))           } else { system.z.idlecpud          = 0 };
            if( paths.int.hdl_path()                != "" ) { system.z.int               = paths.int.has_z().as_a(uint(bits:1))                } else { system.z.int               = 0 };
            if( paths.nmi.hdl_path()                != "" ) { system.z.nmi               = paths.nmi.has_z().as_a(uint(bits:1))                } else { system.z.nmi               = 0 };
            if( paths.iack.hdl_path()               != "" ) { system.z.iack              = paths.iack.has_z().as_a(uint(bits:1))               } else { system.z.iack              = 0 };
            if( paths.buserror.hdl_path()           != "" ) { system.z.buserror          = paths.buserror.has_z().as_a(uint(bits:1))           } else { system.z.buserror          = 0 };
            if( paths.debugerror.hdl_path()         != "" ) { system.z.debugerror        = paths.debugerror.has_z().as_a(uint(bits:1))         } else { system.z.debugerror        = 0 };
            if( paths.perrack.hdl_path()            != "" ) { system.z.perrack           = paths.perrack.has_z().as_a(uint(bits:1))            } else { system.z.perrack           = 0 };
            
            if( paths.gl_privenable.hdl_path()      != "" ) { mp.z.gl_privenable         = paths.gl_privenable.has_z().as_a(uint(bits:1))      } else { mp.z.gl_privenable         = 0 };
            if( paths.mpcxviol.hdl_path()           != "" ) { mp.z.mpcxviol              = paths.mpcxviol.has_z().as_a(uint(bits:1))           } else { mp.z.mpcxviol              = 0 };
            if( paths.mpcbviol.hdl_path()           != "" ) { mp.z.mpcbviol              = paths.mpcbviol.has_z().as_a(uint(bits:1))           } else { mp.z.mpcbviol              = 0 };
            if( paths.mpccviol.hdl_path()           != "" ) { mp.z.mpccviol              = paths.mpccviol.has_z().as_a(uint(bits:1))           } else { mp.z.mpccviol              = 0 };
            if( paths.mpcdviol.hdl_path()           != "" ) { mp.z.mpcdviol              = paths.mpcdviol.has_z().as_a(uint(bits:1))           } else { mp.z.mpcdviol              = 0 };
            if( paths.mpceviol.hdl_path()           != "" ) { mp.z.mpceviol              = paths.mpceviol.has_z().as_a(uint(bits:1))           } else { mp.z.mpceviol              = 0 };
            if( paths.mpcfviol.hdl_path()           != "" ) { mp.z.mpcfviol              = paths.mpcfviol.has_z().as_a(uint(bits:1))           } else { mp.z.mpcfviol              = 0 };
            if( paths.cpu_privenable.hdl_path()     != "" ) { mp.z.cpu_privenable        = paths.cpu_privenable.has_z().as_a(uint(bits:1))     } else { mp.z.cpu_privenable        = 0 };
            if( paths.ppm.hdl_path()                != "" ) { mp.z.ppm                   = paths.ppm.has_z().as_a(uint(bits:1))                } else { mp.z.ppm                   = 0 };
            if( paths.bpm.hdl_path()                != "" ) { mp.z.bpm                   = paths.bpm.has_z().as_a(uint(bits:1))                } else { mp.z.bpm                   = 0 };
            if( paths.cpm.hdl_path()                != "" ) { mp.z.cpm                   = paths.cpm.has_z().as_a(uint(bits:1))                } else { mp.z.cpm                   = 0 };
            if( paths.dpm.hdl_path()                != "" ) { mp.z.dpm                   = paths.dpm.has_z().as_a(uint(bits:1))                } else { mp.z.dpm                   = 0 };
            if( paths.epm.hdl_path()                != "" ) { mp.z.epm                   = paths.epm.has_z().as_a(uint(bits:1))                } else { mp.z.epm                   = 0 };
            if( paths.fpm.hdl_path()                != "" ) { mp.z.fpm                   = paths.fpm.has_z().as_a(uint(bits:1))                } else { mp.z.fpm                   = 0 };
            if( paths.rptpcvio.hdl_path()           != "" ) { mp.z.rptpcvio              = paths.rptpcvio.has_z().as_a(uint(bits:1))           } else { mp.z.rptpcvio              = 0 };
            if( paths.rptintr.hdl_path()            != "" ) { mp.z.rptintr               = paths.rptintr.has_z().as_a(uint(bits:1))            } else { mp.z.rptintr               = 0 };
            if( paths.rptpvio.hdl_path()            != "" ) { mp.z.rptpvio               = paths.rptpvio.has_z().as_a(uint(bits:1))            } else { mp.z.rptpvio               = 0 };
            if( paths.rptbvio.hdl_path()            != "" ) { mp.z.rptbvio               = paths.rptbvio.has_z().as_a(uint(bits:1))            } else { mp.z.rptbvio               = 0 };
            if( paths.rptcvio.hdl_path()            != "" ) { mp.z.rptcvio               = paths.rptcvio.has_z().as_a(uint(bits:1))            } else { mp.z.rptcvio               = 0 };
            if( paths.rptdvio.hdl_path()            != "" ) { mp.z.rptdvio               = paths.rptdvio.has_z().as_a(uint(bits:1))            } else { mp.z.rptdvio               = 0 };
            if( paths.rptiodvio.hdl_path()          != "" ) { mp.z.rptiodvio             = paths.rptiodvio.has_z().as_a(uint(bits:1))          } else { mp.z.rptiodvio             = 0 };
            if( paths.rptevio.hdl_path()            != "" ) { mp.z.rptevio               = paths.rptevio.has_z().as_a(uint(bits:1))            } else { mp.z.rptevio               = 0 };
            if( paths.rptioevio.hdl_path()          != "" ) { mp.z.rptioevio             = paths.rptioevio.has_z().as_a(uint(bits:1))          } else { mp.z.rptioevio             = 0 };
            if( paths.rptfvio.hdl_path()            != "" ) { mp.z.rptfvio               = paths.rptfvio.has_z().as_a(uint(bits:1))            } else { mp.z.rptfvio               = 0 };
            if( paths.rptpm.hdl_path()              != "" ) { mp.z.rptpm                 = paths.rptpm.has_z().as_a(uint(bits:1))              } else { mp.z.rptpm                 = 0 };
            if( paths.dbgmpviol.hdl_path()          != "" ) { mp.z.dbgmpviol             = paths.dbgmpviol.has_z().as_a(uint(bits:1))          } else { mp.z.dbgmpviol             = 0 };
            if( paths.xdmadmpovr.hdl_path()         != "" ) { mp.z.xdmadmpovr            = paths.xdmadmpovr.has_z().as_a(uint(bits:1))         } else { mp.z.xdmadmpovr            = 0 };
            if( paths.xdmaempovr.hdl_path()         != "" ) { mp.z.xdmaempovr            = paths.xdmaempovr.has_z().as_a(uint(bits:1))         } else { mp.z.xdmaempovr            = 0 };
            if( paths.emiodready.hdl_path()         != "" ) { mp.z.emiodready            = paths.emiodready.has_z().as_a(uint(bits:1))         } else { mp.z.emiodready            = 0 };
            if( paths.emioeready.hdl_path()         != "" ) { mp.z.emioeready            = paths.emioeready.has_z().as_a(uint(bits:1))         } else { mp.z.emioeready            = 0 };
            
        };
        
    };
    
    -- detect request state --
    ocuur_data_write_requests() : bool is {
        if( ebus.state.request == request or fbus.state.request == request ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    ocuur_data_read_requests() : bool is {
        if( bbus.state.request == request or cbus.state.request == request or dbus.state.request == request ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    ocuur_program_request() : bool is {
        if( pbus.state.request == request ) { result = TRUE;  }
                                       else { result = FALSE; };
    };
    -- detect current/previous ready state --
    ocuur_current_program_ready() : bool is {
        if( pbus.state.ready == ready ) { result = TRUE;  }
                                   else { result = FALSE; };
    };
    ocuur_previous_program_ready() : bool is {
        if( pbus.state.previous_ready == ready ) { result = TRUE;  }
                                            else { result = FALSE; };
    };
    ocuur_current_data_ready() : bool is {
        if( bbus.state.ready == ready or cbus.state.ready == ready or dbus.state.ready == ready or
            ebus.state.ready == ready or fbus.state.ready == ready ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    ocuur_previous_data_ready() : bool is {
        if( bbus.state.previous_ready == ready or cbus.state.previous_ready == ready or dbus.state.previous_ready == ready or
            ebus.state.previous_ready == ready or fbus.state.previous_ready == ready ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    
    ----------------------------------
    -- for pending program requests --
    ----------------------------------
    pending_prog_requests( prog_attributes : PROGRAM_ACCESS_ATTRIBUTES ) is {
        
        -- when pdismiss occur --
        if( pbus.current.dismiss == 1 ) {
            pbus.pending.dismiss_rdy_index   = pbus.pending.rdy_index;
            pbus.pending.dismiss_counter     = pbus.pending.counter;
            if ( pbus.pending.progress[0] != NULL ) { pbus.pending.dismiss_progress[0] = pbus.pending.progress[0].copy(); };
            if ( pbus.pending.progress[1] != NULL ) { pbus.pending.dismiss_progress[1] = pbus.pending.progress[1].copy(); };
            pbus.pending.rdy_index  = 0;
            pbus.pending.req_index  = 0;
            pbus.pending.counter    = 0;
        };
        
        -- update pending requests --
        if( pbus.state.previous_ready == ready ) { 
            if( pbus.pending.detect_ready.pbus_active != finished ) {
                pbus.pending.detect_ready.pbus_active    = finished;
                pbus.pending.detect_ready.pdbus          = pbus.current.dbus;
            };
            pbus.pending.finish = pbus.pending.detect_ready;
        };
        if( pbus.state.ready == ready ) { 
            pbus.pending.detect_ready = pbus.pending.progress[pbus.pending.rdy_index];
            pbus.pending.rdy_index = pbus.pending.rdy_index + 1;
            pbus.pending.counter = pbus.pending.counter - 1;
        };
        
        -- keep pending requests --
        if( pbus.state.request == request ) { 
            pbus.pending.progress[pbus.pending.req_index] = prog_attributes;
            pbus.pending.req_index = pbus.pending.req_index + 1;
            pbus.pending.counter = pbus.pending.counter + 1;
        };
        
    };
    
    -------------------------------
    -- for pending data requests --
    -------------------------------
    pending_data_requests( read_attributes : DATA_ACCESS_ATTRIBUTES, write_attributes : DATA_ACCESS_ATTRIBUTES ) is {
        -- update pending requests --
        update_pending_data_requests();
        -- keep pending requests --
        keep_pending_data_read_requests( read_attributes );
        keep_pending_data_write_requests( write_attributes );
    };
    -- keep pending requests --
    update_pending_data_requests() is {
        -- detect previous ready --
        if( bbus.state.previous_ready == ready ) { 
            if( bbus.pending.detect_ready.bbus_active != finished ) {
                bbus.pending.detect_ready.bbus_active = finished;
                bbus.pending.detect_ready.bdbus       = bbus.current.dbus;
            };
            bbus.pending.finish = bbus.pending.detect_ready;
        };
        if( cbus.state.previous_ready == ready ) {
            if( cbus.pending.detect_ready.cbus_active != finished ) {
                cbus.pending.detect_ready.cbus_active = finished;
                cbus.pending.detect_ready.cdbus       = cbus.current.dbus;
            };
            cbus.pending.finish = cbus.pending.detect_ready;
        };
        if( dbus.state.previous_ready == ready and dbus.pending.detect_ready.doubler != double ) {
            if( dbus.pending.detect_ready.dbus_active != finished ) {
                dbus.pending.detect_ready.dbus_active = finished;
                dbus.pending.detect_ready.ddbus = dbus.current.dbus;
            };
            dbus.pending.finish = dbus.pending.detect_ready;
        };
        if( dbus.state.previous_ready == ready and dbus.pending.detect_ready.doubler == double ) {
            if( dbus.pending.detect_ready.dbus_active != finished ) {
                dbus.pending.detect_ready.dbus_active = finished;
                dbus.pending.detect_ready.cbus_active = finished;
                dbus.pending.detect_ready.ddbus = cbus.current.dbus;   -- only when double access
                dbus.pending.detect_ready.cdbus = dbus.current.dbus;   -- only when double access
            };
            dbus.pending.finish = dbus.pending.detect_ready;
        };
        if( ebus.state.previous_ready == ready and ebus.pending.detect_ready.doublew != double ) {
            ebus.pending.detect_ready.ebus_active = finished;
            ebus.pending.detect_ready.edbus = ebus.previous.dbus;
            ebus.pending.finish = ebus.pending.detect_ready;
        };
        if( ebus.state.previous_ready == ready and ebus.pending.detect_ready.doublew == double ) {
            ebus.pending.detect_ready.ebus_active = finished;
            ebus.pending.detect_ready.fbus_active = finished;
            ebus.pending.detect_ready.fdbus = ebus.previous.dbus;     -- only when double access
            ebus.pending.detect_ready.edbus = fbus.previous.dbus;     -- only when double access
            ebus.pending.finish = ebus.pending.detect_ready;
        };
        if( fbus.state.previous_ready == ready ) {
            fbus.pending.detect_ready.fbus_active = finished;
            fbus.pending.detect_ready.fdbus = fbus.previous.dbus;
            fbus.pending.finish = fbus.pending.detect_ready;
        };
        -- detect ready --
        if( bbus.state.ready == ready ) { 
            bbus.pending.detect_ready = bbus.pending.progress[bbus.pending.rdy_index];
            bbus.pending.rdy_index = bbus.pending.rdy_index + 1;
            bbus.pending.counter = bbus.pending.counter - 1;
            bbus.pending.detect_ready.mp_viol_bbus = (mp.current.mpcbviol == 1);
        };
        if( cbus.state.ready == ready ) {
            cbus.pending.detect_ready = cbus.pending.progress[cbus.pending.rdy_index];
            cbus.pending.rdy_index = cbus.pending.rdy_index + 1;
            cbus.pending.counter = cbus.pending.counter - 1;
            cbus.pending.detect_ready.mp_viol_cbus = (mp.current.mpccviol == 1);
        };
        if( dbus.state.ready == ready ) {
            dbus.pending.detect_ready = dbus.pending.progress[dbus.pending.rdy_index];
            dbus.pending.rdy_index = dbus.pending.rdy_index + 1;
            dbus.pending.counter = dbus.pending.counter - 1;
            dbus.pending.detect_ready.mp_viol_dbus = (mp.current.mpcdviol == 1);
        };
        if( ebus.state.ready == ready ) {
            ebus.pending.detect_ready = ebus.pending.progress[ebus.pending.rdy_index];
            ebus.pending.rdy_index = ebus.pending.rdy_index + 1;
            ebus.pending.counter = ebus.pending.counter - 1;
            ebus.pending.detect_ready.mp_viol_ebus = (mp.current.mpceviol == 1);
        };
        if( fbus.state.ready == ready ) {
            fbus.pending.detect_ready = fbus.pending.progress[fbus.pending.rdy_index];
            fbus.pending.rdy_index = fbus.pending.rdy_index + 1;
            fbus.pending.counter = fbus.pending.counter - 1;
            fbus.pending.detect_ready.mp_viol_fbus = (mp.current.mpcfviol == 1);
        };
    };
    
    -- detect request --
    !read_inst_slot[2]  : list of DATA_ACCESS_ATTRIBUTES;
    !read_inst_index    : uint(bits:  1);
    
    keep_pending_data_read_requests( data_attributes : DATA_ACCESS_ATTRIBUTES ) is {
        if( bbus.state.request == request ) { 
            bbus.pending.progress[bbus.pending.req_index] = data_attributes;
            bbus.pending.req_index = bbus.pending.req_index + 1;
            bbus.pending.counter = bbus.pending.counter + 1;
        };
        if( cbus.state.request == request ) { 
            cbus.pending.progress[cbus.pending.req_index] = data_attributes;
            cbus.pending.req_index = cbus.pending.req_index + 1;
            cbus.pending.counter = cbus.pending.counter + 1;
        };
        if( dbus.state.request == request ) { 
            dbus.pending.progress[dbus.pending.req_index] = data_attributes;
            dbus.pending.req_index = dbus.pending.req_index + 1;
            dbus.pending.counter = dbus.pending.counter + 1;
        };
        
        -- update instruction buffer --
        if( bbus.state.request == request or cbus.state.request == request or dbus.state.request == request ) {
            read_inst_slot[ read_inst_index ] = data_attributes;
            read_inst_index = ~read_inst_index;
        };
        
    };
    
    !write_inst_slot[2] : list of DATA_ACCESS_ATTRIBUTES;
    !write_inst_index   : uint(bits:  1);
    
    keep_pending_data_write_requests( data_attributes : DATA_ACCESS_ATTRIBUTES ) is {
        if( ebus.state.request == request ) { 
            ebus.pending.progress[ebus.pending.req_index] = data_attributes;
            ebus.pending.req_index = ebus.pending.req_index + 1;
            ebus.pending.counter = ebus.pending.counter + 1;
        };
        if( fbus.state.request == request ) { 
            fbus.pending.progress[fbus.pending.req_index] = data_attributes;
            fbus.pending.req_index = fbus.pending.req_index + 1;
            fbus.pending.counter = fbus.pending.counter + 1;
        };
        
        -- update instruction buffer --
        if( ebus.state.request == request or fbus.state.request == request ) {
            write_inst_slot[ write_inst_index ] = data_attributes;
            write_inst_index = ~write_inst_index;
        };
    };
    
    init() is also {
        read_inst_slot[0]  = NULL;
        read_inst_slot[1]  = NULL;
        read_inst_index      = 0;
        write_inst_slot[0] = NULL;
        write_inst_slot[1] = NULL;
        write_inst_index     = 0;
    };
    
    -------------------------------------
    -- functions of stall informations --
    -------------------------------------
    -- ---------------------------------------------------------------------------------------------------------------
    -- Xbus_current_rdystall()       : rdy_stall_two
    -- Xbus_current_bothrdystall()   : rdy_stall_one or rdy_stall_two
    -- data_read_current_rdystall()  : bbus_current_rdystall() or cbus_current_rdystall() or dbus_current_rdystall()
    -- data_write_current_rdystall() : ebus_current_rdystall() or fbus_current_rdystall()
    -- data_current_rdystall()       : data_read_current_rdystall()  or data_write_current_rdystall()
    -- data_read_current_stall()     : data_read_current_rdystall()
    -- data_write_current_stall()    : data_write_current_rdystall()
    -- ---------------------------------------------------------------------------------------------------------------
    
    pbus_current_rdystall() : bool is {
        result = ( pbus.state.current == rdy_stall_two );
    };
    bbus_current_rdystall() : bool is {
        result = ( bbus.state.current == rdy_stall_two );
    };
    cbus_current_rdystall() : bool is {
        result = ( cbus.state.current == rdy_stall_two );
    };
    dbus_current_rdystall() : bool is {
        result = ( dbus.state.current == rdy_stall_two );
    };
    ebus_current_rdystall() : bool is {
        result = ( ebus.state.current == rdy_stall_two );
    };
    fbus_current_rdystall() : bool is {
        result = ( fbus.state.current == rdy_stall_two );
    };
    
    data_read_current_rdystall() : bool is {
        result = ( bbus_current_rdystall() or cbus_current_rdystall() or dbus_current_rdystall() );
    };
    
    data_write_current_rdystall() : bool is {
        result = ( ebus_current_rdystall() or fbus_current_rdystall() );
    };
    
    data_current_rdystall() : bool is {
        result = ( data_read_current_rdystall() or data_write_current_rdystall() );
    };
    
    data_read_current_stall() : bool is {
        result = data_read_current_rdystall();
    };
    data_write_current_stall() : bool is {
        result = data_write_current_rdystall();
    };
    
    -- check for lock stall --
    check_lock_stall( read_phase_attributes : DATA_ACCESS_ATTRIBUTES ) : bool is {
        if( read_phase_attributes != NULL and read_phase_attributes.lock == enable and
            (dbus.pending.counter == 1 and dbus.pending.progress[ dbus.pending.rdy_index ] == read_phase_attributes) or 
            (dbus.pending.counter == 2 and (dbus.pending.progress[ 0 ] == read_phase_attributes or
                                            dbus.pending.progress[ 1 ] == read_phase_attributes    )) ) { result = TRUE;  }
                                                                                                   else { result = FALSE; };
    };
    
    -----------------------
    -- bus status output --
    -----------------------
    debug_output() is {
        -- output data transition informations --
        if( bbus.state.previous_ready == ready and bbus.pending.finish != NULL ) { 
            out("(debug)ready   ",sys.time," bbus finish = ",bbus.pending.finish );
        };
        if( cbus.state.previous_ready == ready and cbus.pending.finish != NULL ) { 
            out("(debug)ready   ",sys.time," cbus finish = ",cbus.pending.finish );
        };
        if( dbus.state.previous_ready == ready and dbus.pending.finish != NULL ) { 
            out("(debug)ready   ",sys.time," dbus finish = ",dbus.pending.finish );
        };
        if( ebus.state.previous_ready == ready and ebus.pending.finish != NULL ) { 
            out("(debug)ready   ",sys.time," ebus finish = ",ebus.pending.finish );
        };
        if( fbus.state.previous_ready == ready and fbus.pending.finish != NULL ) { 
            out("(debug)ready   ",sys.time," fbus finish = ",fbus.pending.finish );
        };
    };
    
    set_reset_values() is {
        pbus.state.init();
        bbus.state.init();
        cbus.state.init();
        dbus.state.init();
        ebus.state.init();
        fbus.state.init();
        pbus.pending.init();
        bbus.pending.init();
        cbus.pending.init();
        dbus.pending.init();
        ebus.pending.init();
        fbus.pending.init();
    };
    
};

struct PROGRAM_ACCESS_ATTRIBUTES {
   !pbus_active : activate_bus;
    pbus_type   : pbus_access_type;
    
   !pabus : uint(bits : 24);
   !pdbus : uint(bits : 32);
    
    set_current_attributes( bus_states : CPU_BUSES ) is {
        pabus = bus_states.pbus.current.abus;
        pdbus = bus_states.pbus.current.dbus;
        if( bus_states.pbus.state.request == request ) { pbus_active = active; } else { pbus_active = no_active; };
        if( bus_states.emulation.current.dmapr == 1  ) { pbus_type   = debug;  } else { pbus_type   = no_debug;  };
    };
    
    -- default constraint --
    keep pabus[1:0]     == 0;
    keep soft pbus_type == no_debug;
    
    init() is also {
        pbus_active = no_active;
        pabus       = 0;
        pdbus       = 0;
    };
    
};

struct DATA_ACCESS_ATTRIBUTES {
    
    write_width : access_width;   -- for Ebus
    read_width  : access_width;   -- for Dbus
    space       : access_space;   -- for Dbus/Ebus
    doubler     : access_double;  -- for D request with Dbus/Cbus
    doublew     : access_double;  -- for E request with Ebus/Fbus
    dblcoef     : access_double;  -- for B request with Bbus only
    lock        : access_lock;    -- for Dbus/Ebus or Dbus/Fbus
    
   !bbus_active : activate_bus;
   !cbus_active : activate_bus;
   !dbus_active : activate_bus;
   !ebus_active : activate_bus;
   !fbus_active : activate_bus;
    
    dbus_type : dbus_access_type;
    ebus_type : ebus_access_type;
    
    -- address for data buses
    babus : uint(bits : 23);
    cabus : uint(bits : 23);
    dabus : uint(bits : 23);
    eabus : uint(bits : 23);
    fabus : uint(bits : 23);
    -- data for data buses
   !bdbus : uint(bits : 32);
   !cdbus : uint(bits : 16);
   !ddbus : uint(bits : 16);
    edbus : uint(bits : 16);
    fdbus : uint(bits : 16);
    
    -- for slave driver --
    wspace : access_space;
    rspace : access_space;
    
    -- violation flag for memory protection --
    !mp_viol_bbus : bool;
    !mp_viol_cbus : bool;
    !mp_viol_dbus : bool;
    !mp_viol_ebus : bool;
    !mp_viol_fbus : bool;
    
    set_current_attributes( bus_states : CPU_BUSES ) is {
        babus       = bus_states.bbus.current.abus;
        cabus       = bus_states.cbus.current.abus;
        dabus       = bus_states.dbus.current.abus;
        eabus       = bus_states.ebus.current.abus;
        fabus       = bus_states.fbus.current.abus;
        
        bdbus       = bus_states.bbus.current.dbus;
        cdbus       = bus_states.cbus.current.dbus;
        ddbus       = bus_states.dbus.current.dbus;
        edbus       = bus_states.ebus.current.dbus;
        fdbus       = bus_states.fbus.current.dbus;
        
        if( bus_states.bbus.state.request == request ) { bbus_active = active; } else { bbus_active = no_active; };
        if( bus_states.cbus.state.request == request ) { cbus_active = active; } else { cbus_active = no_active; };
        if( bus_states.dbus.state.request == request ) { dbus_active = active; } else { dbus_active = no_active; };
        if( bus_states.ebus.state.request == request ) { ebus_active = active; } else { ebus_active = no_active; };
        if( bus_states.fbus.state.request == request ) { fbus_active = active; } else { fbus_active = no_active; };
        
        -- space --
        if( bus_states.dbus.state.request == request and bus_states.dbus.current.ioreq == 0 ) { rspace = io; } else { rspace = memory; };
        if( bus_states.ebus.state.request == request and bus_states.ebus.current.ioreq == 0 ) { wspace = io; } else { wspace = memory; };
        
        -- width --
        if( bus_states.dbus.state.request == request and bus_states.qualifier.current.rdbyte == 1 ) {
            if( bus_states.qualifier.current.rdlsbe == 1 ) { read_width = low_byte; } else { read_width = high_byte; };
        } else { read_width = word; };
        if( bus_states.ebus.state.request == request and bus_states.qualifier.current.wrbyte == 1 ) {
            if( bus_states.qualifier.current.wrlsbe == 1 ) { write_width = low_byte; } else { write_width = high_byte; };
        } else { write_width = word; };
        
        -- doubler --
        if( bus_states.dbus.state.request == request and bus_states.qualifier.current.doubler == 1 ) {
            doubler  = double;
            cabus    = %{bus_states.dbus.current.abus[22:1], (~bus_states.dbus.current.abus)[0:0]};
        } else { doubler = word; };
        
        -- doublew --
        if( bus_states.ebus.state.request == request and bus_states.qualifier.current.doublew == 1 ) {
            doublew = double;
            fabus   = %{bus_states.ebus.current.abus[22:1], (~bus_states.ebus.current.abus)[0:0]};
        } else { doublew = word; };
        
        -- dblcoef --
        if( bus_states.bbus.state.request == request and bus_states.qualifier.current.dblcoef == 1 ) {
            dblcoef = double;
        } else { dblcoef = word; };
        
        -- lock --
        if( (bus_states.dbus.state.request == request or bus_states.ebus.state.request == request or
             bus_states.fbus.state.request == request) and bus_states.qualifier.current.lock == 1 ) {
            lock = enable;
        } else { lock = disable; };
        
        -- dbus_type --
        if( bus_states.emulation.current.dbugd == 1 ) { dbus_type = debug;  } else { dbus_type = no_debug;  };
        
        -- ebus_type --
        if( bus_states.emulation.current.dbuge == 1 and bus_states.emulation.current.dmapw == 1 ) { ebus_type = debug_prog_write; } else { ebus_type = no_debug; };
        if( bus_states.emulation.current.dbuge == 1 and bus_states.emulation.current.dmapw == 0 ) { ebus_type = debug_data_write; } else { ebus_type = no_debug; };
        
    };
    
    -- default constraint --
    keep soft dbus_type == no_debug;
    keep soft ebus_type == no_debug;
    
    -- initialize --
    init() is also {
        bbus_active  = no_active;
        cbus_active  = no_active;
        dbus_active  = no_active;
        ebus_active  = no_active;
        fbus_active  = no_active;
        bdbus        = 0;
        cdbus        = 0;
        ddbus        = 0;
        mp_viol_bbus = FALSE;
        mp_viol_cbus = FALSE;
        mp_viol_dbus = FALSE;
        mp_viol_ebus = FALSE;
        mp_viol_fbus = FALSE;
    };
    
    -- output for debug --
    debug_output_read( bus_states : CPU_BUSES ) is {
        if( bus_states.bbus.state.request == request ) {
            out("(debug)request ",sys.time," bbus : address = ",babus, " data = ", bdbus);
        };
        if( bus_states.cbus.state.request == request ) {
            out("(debug)request ",sys.time," cbus : address = ",cabus, " data = ", cdbus);
        };
        if( bus_states.dbus.state.request == request ) {
            out("(debug)request ",sys.time," dbus : address = ",dabus, " data = ", ddbus, " lock = ",lock);
        };
    };
    debug_output_write( bus_states : CPU_BUSES ) is {
        if( bus_states.ebus.state.request == request ) {
            out("(debug)request ",sys.time," ebus : address = ",eabus, " data = ", edbus, " lock = ",lock);
        };
        if( bus_states.fbus.state.request == request ) {
            out("(debug)request ",sys.time," fbus : address = ",fabus, " data = ", fdbus, " lock = ",lock);
        };
    };
};

'>
