//********************************************************************************
//
//  Filename: cpubus_data_extractor.e
//
//  Description: cpu bus data extractor
//
//  Major Revision Descriptions:
//  oishi          09/08/02         copied from UMA1.0 branch
//  oishi          09/09/02         removed mdp related functions
//  oishi          09/11/02         changed bdbus width to 32bits
//  oishi          09/11/02         added cpu_dblcoef_tr qualifier
//                                  and byte read information
//  oishi          09/13/02         modified resetn to cpu_resetn and gl_resetn
//  oishi          10/04/02         added checking function of signal connection.
//
//*********************************************************************************
<'
unit CPUBUS_SCOREBOARD_STRUCT {
    -- time
    !s_time      : uint;
    
    -- ready detect
    !pbus_active : activate_bus;
    !bbus_active : activate_bus;
    !cbus_active : activate_bus;
    !dbus_active : activate_bus;
    !ebus_active : activate_bus;
    !fbus_active : activate_bus;
    
    -- emulation qualifiers --
    pbus_type : pbus_access_type;
    dbus_type : dbus_access_type;
    ebus_type : ebus_access_type;
    
    -- address in program space
    !pabus       : uint(bits : 24);
    
    -- address in data space
    !babus       : uint(bits : 23);
    !cabus       : uint(bits : 23);
    !dabus       : uint(bits : 23);
    !eabus       : uint(bits : 23);
    !fabus       : uint(bits : 23);
    
    -- data
    !pdbus       : uint(bits : 32);
    !bdbus       : uint(bits : 32);
    !cdbus       : uint(bits : 16);
    !ddbus       : uint(bits : 16);
    !edbus       : uint(bits : 16);
    !fdbus       : uint(bits : 16);
    
    -- qualifiers
    !ebus_wspace : access_space;   -- indicate io or memory access in Ebus
    !dbus_rspace : access_space;   -- indicate io or memory access in Dbus
    !ebus_width  : access_width;   -- indicate word or byte access in Ebus
    !dbus_width  : access_width;   -- indicate word or byte access in Dbus
    !doubler     : access_double;  -- indicate double read access in Dbus/Cbus
    !doublew     : access_double;  -- indicate double write access in Ebus/Fbus
    !dblcoef     : access_double;  -- indicate double coef read in Bbus
    
    init() is also {
        s_time      = 0;
        pbus_active = no_active;
        bbus_active = no_active;
        cbus_active = no_active;
        dbus_active = no_active;
        ebus_active = no_active;
        fbus_active = no_active;
        pbus_type   = no_debug;
        dbus_type   = no_debug;
        ebus_type   = no_debug;
        pabus       = 0;
        babus       = 0;
        cabus       = 0;
        dabus       = 0;
        eabus       = 0;
        fabus       = 0;
        pdbus       = 0;
        bdbus       = 0;
        cdbus       = 0;
        ddbus       = 0;
        edbus       = 0;
        fdbus       = 0;
        ebus_wspace = memory;
        dbus_rspace = memory;
        ebus_width  = word;
        dbus_width  = word;
        doubler     = word;
        doublew     = word;
        dblcoef     = word;
    };
    
};

unit CPUBUS_DATA_EXTRACTOR_STRUCT {
    -- enable/disable flag --
    disable  : bool; 
    keep soft disable == FALSE;
    
    -- global configurations --
    config   : CPUBUS_CONFIGURATIONS;
    paths    : CPU_SIGNAL_PATH;
    
    -- ready detect signal --
    readys_detect : bool;
    
    -- cpubus states --
    bus_states : CPU_BUSES;
    
    -- access attributes --
    !prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
    !data_attributes : DATA_ACCESS_ATTRIBUTES;
    
    -- scoreboard --
    !cpubus_scoreboard : list of CPUBUS_SCOREBOARD_STRUCT;
    
    -- set clock event for state transitions
    event raw_rise_clock is rise(paths.clock$)    @sim;
    event rise_clock     is true( disable == FALSE ) @raw_rise_clock;
    
    -- CPUBUS data extractors will emit this event for all detected accesses
    event CPUBUS_access_detected;
    
    post_generate() is also {
        
        -- check signal connection --
        paths.check_signal_connection( config.id );
        
    };
    
    ------------------------------
    -- cpu bus state transition --
    ------------------------------
    on rise_clock
    {
        -- keep current values
        bus_states.keep_current_values();
        
        -- get current value
        bus_states.get_current_value( config, paths );
        
        -- state transition
        bus_states.state_transition();
        
        -- set current request attributes --
        if( bus_states.ocuur_program_request() ) {
            gen prog_attributes; 
            prog_attributes.set_current_attributes( bus_states );
        };
        if( bus_states.ocuur_data_read_requests() or bus_states.ocuur_data_write_requests() ) {
            gen data_attributes; 
            data_attributes.set_current_attributes( bus_states );
        };
        
        -- for keeping pending requests --
        bus_states.pending_prog_requests( prog_attributes );
        bus_states.pending_data_requests( data_attributes, data_attributes );
        
        -- state initialize when reset 
        if( bus_states.system.current.cpu_resetn == 0 ) {
            bus_states.set_reset_values()
        };
        
    };
    
   temp_scoreboard : CPUBUS_SCOREBOARD_STRUCT;
    TCM_cpubus_data_extranctor() @rise_clock is {
        
        while( TRUE ) {
            
            temp_scoreboard.init();
            
            if( bus_states.pbus.state.previous_ready == ready or bus_states.bbus.state.previous_ready == ready or
                bus_states.cbus.state.previous_ready == ready or bus_states.dbus.state.previous_ready == ready or
                bus_states.ebus.state.previous_ready == ready or bus_states.fbus.state.previous_ready == ready    ) {
                temp_scoreboard.s_time = sys.time;
                readys_detect          = TRUE;
            } else {
                readys_detect = FALSE;
            };
            
            -- for single access --
            if( bus_states.pbus.state.previous_ready == ready ) {
                temp_scoreboard.pbus_active = finished;
                temp_scoreboard.pabus       = bus_states.pbus.pending.finish.pabus;
                temp_scoreboard.pdbus       = bus_states.pbus.pending.finish.pdbus;
                temp_scoreboard.pbus_type   = bus_states.pbus.pending.finish.pbus_type;
            };
            if( bus_states.bbus.state.previous_ready == ready ) {
                temp_scoreboard.bbus_active = finished;
                temp_scoreboard.babus       = bus_states.bbus.pending.finish.babus;
                temp_scoreboard.bdbus       = bus_states.bbus.pending.finish.bdbus;
                temp_scoreboard.dblcoef     = bus_states.bbus.pending.finish.dblcoef;
            };
            if( bus_states.cbus.state.previous_ready == ready ) {
                temp_scoreboard.cbus_active = finished;
                temp_scoreboard.cabus       = bus_states.cbus.pending.finish.cabus;
                temp_scoreboard.cdbus       = bus_states.cbus.pending.finish.cdbus;
            };
            if( bus_states.dbus.state.previous_ready == ready and bus_states.dbus.pending.finish.doubler != double ) {
                temp_scoreboard.dbus_active = finished;
                temp_scoreboard.dabus       = bus_states.dbus.pending.finish.dabus;
                temp_scoreboard.ddbus       = bus_states.dbus.pending.finish.ddbus;
                temp_scoreboard.dbus_rspace = bus_states.dbus.pending.finish.rspace;
                temp_scoreboard.dbus_width  = bus_states.ebus.pending.finish.read_width;
                temp_scoreboard.doubler     = bus_states.dbus.pending.finish.doubler;
                temp_scoreboard.dbus_type   = bus_states.dbus.pending.finish.dbus_type;
            };
            if( bus_states.ebus.state.previous_ready == ready and bus_states.ebus.pending.finish.doublew != double ) {
                temp_scoreboard.ebus_active = finished;
                temp_scoreboard.eabus       = bus_states.ebus.pending.finish.eabus;
                temp_scoreboard.edbus       = bus_states.ebus.pending.finish.edbus;
                temp_scoreboard.ebus_wspace = bus_states.ebus.pending.finish.wspace;
                temp_scoreboard.ebus_width  = bus_states.ebus.pending.finish.write_width;
                temp_scoreboard.doublew     = bus_states.ebus.pending.finish.doublew;
                temp_scoreboard.ebus_type   = bus_states.ebus.pending.finish.ebus_type;
            };
            if( bus_states.fbus.state.previous_ready == ready ) {
                temp_scoreboard.fbus_active = finished;
                temp_scoreboard.fabus       = bus_states.fbus.pending.finish.fabus;
                temp_scoreboard.fdbus       = bus_states.fbus.pending.finish.fdbus;
            };
            -- for dobule access --
            if( bus_states.dbus.state.previous_ready == ready and bus_states.dbus.pending.finish.doubler == double ) {
                -- set cbus attributes --
                temp_scoreboard.cbus_active = finished;
                temp_scoreboard.cabus       = bus_states.dbus.pending.finish.cabus;
                -- set dbus attributes --
                temp_scoreboard.dbus_active = finished;
                temp_scoreboard.dabus       = bus_states.dbus.pending.finish.dabus;
                -- set qualifiers attributes --
                temp_scoreboard.dbus_rspace = bus_states.dbus.pending.finish.rspace;
                temp_scoreboard.dbus_width  = bus_states.ebus.pending.finish.read_width;
                temp_scoreboard.doubler     = bus_states.dbus.pending.finish.doubler;
                -- set data --
                temp_scoreboard.cdbus       = bus_states.dbus.pending.finish.cdbus;
                temp_scoreboard.ddbus       = bus_states.dbus.pending.finish.ddbus;
            };
            if( bus_states.ebus.state.previous_ready == ready and bus_states.ebus.pending.finish.doublew == double ) {
                -- set ebus attributes --
                temp_scoreboard.ebus_active = finished;
                temp_scoreboard.eabus       = bus_states.ebus.pending.finish.eabus;
                -- set fbus attributes --
                temp_scoreboard.fbus_active = finished;
                temp_scoreboard.fabus       = bus_states.ebus.pending.finish.fabus;
                -- set qualifiers attributes --
                temp_scoreboard.ebus_wspace = bus_states.ebus.pending.finish.wspace;
                temp_scoreboard.ebus_width  = bus_states.ebus.pending.finish.write_width;
                temp_scoreboard.doublew     = bus_states.ebus.pending.finish.doublew;
                -- set data --
                temp_scoreboard.edbus       = bus_states.ebus.pending.finish.edbus;
                temp_scoreboard.fdbus       = bus_states.ebus.pending.finish.fdbus;
            };
            
            if( readys_detect ) {
                cpubus_scoreboard.add( temp_scoreboard );
                emit CPUBUS_access_detected;
            };
            
            wait cycle;
        };
    };
    
  --run() is also {
  --   start TCM_cpubus_data_extranctor();
  --};
    
};

'>
