//********************************************************************************
//
//  Filename: cpubus_master_driver.e
//
//  Description: cpubus master driver
//
//  Major Revision Descriptions:
//  oishi          09/08/02         copied from UMA1.0 branch
//  oishi          09/09/02         removed bottom signals
//                                  removed mdp related functions
//  oishi          09/11/02         supported following qualifiers;
//                                    cpu_dblcoef_tr/cpu_rdbyte_tr/cpu_rdlsbe_tr
//  oishi          09/13/02         modified resetn to cpu_resetn and gl_resetn
//  oishi          09/27/02         removed old function as DismissBranch() and Dismiss()
//  oishi          09/30/02         supports simple interrupt behaviors
//  oishi          10/01/02         removed LockAccess_SingleWrite/LockAccess_SingleRead
//  oishi          10/04/02         added checking function of signal connection.
//  oishi          10/07/02         modified interrupt function for supporting specman version 3.x and 4.x
//                                  modified reset function
//                                  revive old functions as Dismiss() and DismissBranch()
//  oishi          10/08/02         added Suspend() function
//  oishi          10/21/02         backed old functions like a uma1.0.
//                                  and added following functions for supports byte read and double bbus read function
//                                    SetSingleRead_width(data_attributes,space,width,dabus)
//                                    SetDualRead_width(data_attributes,space,width,cabus,dabus)
//                                    SetCoefRead_width(data_attributes,width,cabus)
//                                    DebugDataSingleRead_width(space,width,dabus)
//  oishi          11/01/02         applied MikeP suggesion improvement for Nop functions
//                                  called get_current_value_for_master() instead of get_current_value()
//  oishi          11/25/02         added switch for interrupt behavior is enable/disable
//  oishi          11/29/02         added memory protection behaviors
//  oishi          12/13/02         supports interrupt functions
//  oishi          01/23/03         fixed address inclement bug
//  thiraoka       05/13/03         blocking new transaction after calling ProgramAccessFinish/DataAccessFinish
//                                  take care of "Finish vs Intr"
//  thiraoka       05/19/03         blocking condition change,  prog_access_complete/data_access_complete
//  thiraoka       07/09/03         speed improvement
//                                    -- update int_check()
//                                    -- update drive_master_signals
//*********************************************************************************
<'
import ./cpubus_master_structures;

--------------------------
-- cpubus master driver --
--------------------------
unit CPUBUS_MASTER_DRIVER {
    
    -- enable/disable flag --
    disable  : bool; 
    keep soft disable == FALSE;
    
    -- configuration for memory protection behaviors --
    mp_enable           : bool;
    mp_supervisor       : bool;
    mp_viol_pdelay      : uint;
    mp_viol_pclear      : bool;
    mp_viol_address     : uint(bits: 24);
    mp_viol_contextsave : bool;
    -- default values --
    keep soft mp_enable           == TRUE;
    keep soft mp_supervisor       == TRUE;
    keep soft mp_viol_pdelay      <  10;
    keep soft mp_viol_pclear      == select { 80 : FALSE; 20 : TRUE; };
    keep soft mp_viol_contextsave == select { 90 : FALSE; 10 : TRUE; };
    
    -- global configurations --
    config   : CPUBUS_CONFIGURATIONS  is instance;
    paths    : CPU_SIGNAL_PATH        is instance;
    
    -- control structures --
    bus_states : CPU_BUSES            is instance;
    bus_driver : CPU_SIGNALS          is instance;
    
    !prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
    !data_attributes : DATA_ACCESS_ATTRIBUTES;
    
    !prog_attributes_forRandom : PROGRAM_ACCESS_ATTRIBUTES;
    !data_attributes_forRandom : DATA_ACCESS_ATTRIBUTES;
    
    !next_prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
    
    pipeline : CPU_PIPELINE           is instance;
    buffer   : PROGRAM_BUFFER         is instance;
    
    -- set clock event for state transitions
    event raw_rise_clock is rise(paths.clock$)    @sim;
    event rise_clock     is true( disable == FALSE ) @raw_rise_clock;
    
    -- event for getting attributes from stimulus --
    event request_program_attributes;
    event request_data_attributes;
    
    -- event queue --
    !data_queue:    list of DATA_ACCESS_ATTRIBUTES;
    !program_queue: list of PROGRAM_ACCESS_ATTRIBUTES;
    
    -- access complete flag
    prog_access_complete : bool;
    data_access_complete : bool;
    
    -- for user defined functions as MyDataAccess/MyProgramAccess
    my_pcount : uint;
    my_dcount : uint;
    
    keep soft my_pcount == 0;
    keep soft my_dcount == 0;
    
    post_generate() is also {
        -- check signal connection --
        paths.check_signal_connection( config.id );
        
        -- update instruction queue --
        update_program_instructions_buffer( get_prog_nop() );
        update_data_instructions_buffer( get_data_nop() );
        
        -- initialize --
        prog_attributes = get_prog_nop();
        prog_attributes.mp_supervisor = mp_supervisor;
        
        data_attributes = get_data_nop();
        data_attributes.mp_supervisor = mp_supervisor;
        
        prog_access_complete = FALSE;
        data_access_complete = FALSE;
    };
    
    -- for program bus access --
    -- 
    --  function name                        | configuration of program_attributes
    --                                       | instr_group   | branch_occur | dismiss_occur | action
    -- --------------------------------------+---------------+--------------+---------------+-----------------------------------------------
    --  ProgramRandomAccess( cycles )        | (random)      | (random)     | (random)      | random access
    --  ProgramRead( cycles )                | ProgramRead   | FALSE        | -             | Program read(address is continuous with provious one)
    --  Branch( address )                    | ProgramRead   | TRUE         | FALSE         | branch without dismiss
    --  WaitDismissBranch( cycles, address ) | ProgramRead   | TRUE         | TRUE          | branch with dismiss(*1)
    --  WaitDismiss( cycles, address )       | None          | TRUE         | TRUE          | generate dismiss without read request(*1)(*2)
    --  ProgramNop( cycles )                 | None          | -            | FALSE         | no action
    --  ProgramAccessFinish()                | None          | FALSE        | FALSE         | (flash buffer instructions)
    -- 
    -- == for emulation pbus access ==
    --  DebugProgramRead( address )
    -- 
    -- == for override by user ==
    --  MyProgramAccess( my_pcount )
    -- 
    -- (*1) Note that pdismiss signal can generate only when two or one requests are pending.
    --      If pending requests doesn't exist, pdismiss signal can't generate even if 
    --      dismiss function as WaitDismiss()/WaitDismissBranch() use.
    -- (*2) If pdismiss occur without read request, next read request is branch access.
    --
    -- --------------------------------------------------------------- 
    -- Note that MyProgramAccess is supposed to override with user function,
    -- but if you use this function, need to takecare.
    -- This MyDataAccess function is random access in default.
    -- ---------------------------------------------------------------
    
    ProgramRandomAccess( count : int ) @request_program_attributes is {
        var flag : bool;
        var i    : uint;
        
        flag = (count < 0);
        for{ i = 0; (i < count) or flag ; i+=1 } do {
            gen prog_attributes_forRandom;
            update_program_instructions_buffer( prog_attributes_forRandom );
            wait until @request_program_attributes;
        };
        
    };
    
    ProgramRead( count : int ) @request_program_attributes is {
        var prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
        var i : int;
        var flag : bool;
        
        flag = (count < 0);
        for{ i = 0; (i < count) or flag ; i+=1 } do {
            gen prog_attributes keeping { 
                it.instr_group   == ProgramRead;
                it.branch_occur  == FALSE;
                it.dismiss_occur == FALSE;
            };
            update_program_instructions_buffer( prog_attributes );
            wait until @request_program_attributes;
        };
    };
    
    Branch( address : uint(bits : 24) ) @request_program_attributes is {
        var prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
        
        gen prog_attributes keeping { 
            it.instr_group    == ProgramRead;
            it.branch_occur   == TRUE;
            it.dismiss_occur  == FALSE;
            it.branch_address == address;
        };
        update_program_instructions_buffer( prog_attributes );
        wait until @request_program_attributes;
    };
    
    -- DismissBranch function --
    WaitDismissBranch( count : int, address : uint(bits : 24) ) @request_program_attributes is {
        var prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
        
        gen prog_attributes keeping { 
            it.instr_group     == ProgramRead;
            it.branch_occur    == TRUE;
            it.dismiss_occur   == TRUE;
            it.branch_address  == address;
        };
        prog_attributes.dismiss_waiting = count;
        
        update_program_instructions_buffer( prog_attributes );
        wait until @request_program_attributes;
    };
    
    DismissBranch( address : uint(bits : 24) ) @request_program_attributes is {
        var prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
        
        gen prog_attributes keeping { 
            it.instr_group    == ProgramRead;
            it.branch_occur   == TRUE;
            it.dismiss_occur  == TRUE;
            it.branch_address == address;
        };
        update_program_instructions_buffer( prog_attributes );
        wait until @request_program_attributes;
    };
    
    -- Dismiss function --
    WaitDismiss( count : int, address : uint(bits : 24) ) @request_program_attributes is {
        var prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
        
        gen prog_attributes keeping { 
            it.instr_group    == None;
            it.branch_occur   == TRUE;
            it.dismiss_occur  == TRUE;
            it.branch_address == address;
        };
        prog_attributes.dismiss_waiting = count;
        
        update_program_instructions_buffer( prog_attributes );
        wait until @request_program_attributes;
    };
    
    Dismiss( address : uint(bits : 24) ) @request_program_attributes is {
        var prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
        
        gen prog_attributes keeping { 
            it.instr_group    == None;
            it.branch_occur   == TRUE;
            it.dismiss_occur  == TRUE;
            it.branch_address == address;
        };
        update_program_instructions_buffer( prog_attributes );
        wait until @request_program_attributes;
    };
    
    ProgramNop( count : int ) @request_program_attributes is {
        var i : int;
        var flag : bool;
        
        flag = (count < 0);
        for{ i = 0; (i < count) or flag ; i+=1 } do {
            update_program_instructions_buffer( get_prog_nop() );
            wait until @request_program_attributes;
        };
    };
    
    ProgramAccessFinish() @request_program_attributes is {
        var i : uint;
        
        for{ i = 0; i < 10; i+=1 } do {
            update_program_instructions_buffer( get_prog_nop() );
            wait until @request_program_attributes;
        };
        
        prog_access_complete = TRUE;
        
    };
    
    DebugProgramRead( address : uint(bits : 24) ) @request_program_attributes is {
        var prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
        
        gen prog_attributes keeping {
            it.pbus_type      == debug;
            it.instr_group    == ProgramRead;
            it.branch_occur   == TRUE;
            it.dismiss_occur  == FALSE;
            it.branch_address == address;
        };
        update_program_instructions_buffer( prog_attributes );
        wait until @request_program_attributes;
    };
    
    update_program_instructions_buffer( prog_attributes : PROGRAM_ACCESS_ATTRIBUTES ) is {
        -- set mp attributes --
        prog_attributes.mp_supervisor = mp_supervisor;;
        
        -- update the queue --
        if ( prog_access_complete == FALSE or data_access_complete == FALSE ) {
           program_queue.push( prog_attributes );
        };
    };
    update_program_instructions_buffer0( prog_attributes : PROGRAM_ACCESS_ATTRIBUTES ) is {
        -- set mp attributes --
        prog_attributes.mp_supervisor = mp_supervisor;
        
        -- update the queue --
        if ( prog_access_complete == FALSE or data_access_complete == FALSE ) {
           program_queue.push0( prog_attributes );
        };
    };
    
    -- for data buses access --
    --
    -- direct generation functions
    -- 
    -- DataRandomAccess( cycles )       : random access
    -- DataNop( cycles )                : no access
    -- PutPipeline( data_attributes )   : buses drive with data_attributes
    -- 
    -- == attributes set for read access ==                                       b  c  d  e  f
    -- SetSingleRead(data_attributes,space,dabus)                                 -  -  o  -  -     iodreq
    -- SetSingleRead_width( data_attributes, space, width, dabus )                -  -  o  -  -     iodreq, rdbyte, rdlsbe
    -- SetDualRead(data_attributes,space,cabus,dabus)                             -  o  o  -  -     iodreq
    -- SetDualRead_width( data_attributes, space, width, cabus, dabus )           -  o  o  -  -     iodreq, rdbyte, rdlsbe
    -- SetDoubleRead(data_attributes,dabus)                                       -  o  o  -  -     doubler
    -- 
    -- == attributes set for write access ==
    -- SetSingleWrite(data_attributes,space,width,eabus,edbus)                    -  -  -  o  -     ioereq, wrbyte, wrlsbe
    -- SetDualWrite(data_attributes,space,width,eabus,edbus,fabus,fdbus)          -  -  -  o  o     ioereq, wrbyte, wrlsbe
    -- SetDoubleWrite(data_attributes,eabus,edbus,fdbus)                          -  -  -  o  o     doublew
    -- 
    -- == attributes set for lock access ==
    -- SetLockAccess(data_attributes,space,abus,dbus)                             -  -  o  o  -     lock(D/Ebus is locked),iodreq,ioereq
    -- SetLockAccess_SingleWrite(data_attributes,space,dfabus,eabus,edbus)        -  -  -  -  -     ** illegal bus access in Laijin3.x **
    -- SetLockAccess_SingleRead(data_attributes,space,deabus,cabus)               -  -  -  -  -     ** illegal bus access in Laijin3.x **
    -- 
    -- == attributes set for coef access ==
    -- SetCoefRead(data_attributes,babus)                                         o  -  -  -  -     dblcoef(word)
    -- SetCoefRead_width( data_attributes, width, cabus )                         o  -  -  -  -     dblcoef(word/double)
    --
    -- == for MMR access ==
    -- MMRReadAccess( abus )                                                      -  -  o  -  -
    -- MMRWriteAccess( abus )                                                     -  -  -  o  -
    --
    -- == for emulation access ==
    -- DebugDataSingleRead(space,dabus)                                           -  -  o  -  -     dbugd
    -- DebugDataSingleRead_width(space,width,dabus)                               -  -  o  -  -     dbugd, rdbyte, rdlsbe
    -- DebugDataDoubleRead(dabus)                                                 -  o  o  -  -     dbugd
    -- DebugDataSingleWrite(space,width,eabus,edbus)                              -  -  -  o  -     dbuge
    -- DebugDataDoubleWrite(eabus,edbus,fdbus)                                    -  -  -  o  o     dbuge
    -- DebugProgramSingleWrite(space,width,eabus,edbus)                           -  -  -  o  -     dbuge, dmapw
    -- DebugProgramDoubleWrite(eabus,edbus,fdbus)                                 -  -  -  o  o     dbuge, dmapw
    -- 
    -- == for override by user ==
    -- MyConfigAccess()
    -- MyDataAccess( my_dcount )
    --
    -- --------------------------------------------------------------- 
    -- Note that MMRReadAccess/MMRWriteAccess functions is same as 
    -- SingleRead/SignleWrite function. Just diffrence is;
    --   access address is only 0x00~0x5f,
    --   no pipeline protection.
    -- And MyDataAccess is supposed to override with user function,
    -- but if you use this function, need to takecare.
    -- This MyDataAccess function is random access in default.
    -- ---------------------------------------------------------------

    temp_flag : bool;
    post_generate() is also { temp_flag = TRUE; };
    
    DataRandomAccess( count : int ) @request_data_attributes is {
        var i : int;
        var flag : bool;
        
        flag = (count < 0);
        for{ i = 0; (i < count) or flag ; i+=1 } do {
            gen data_attributes_forRandom;
            update_data_instructions_buffer( data_attributes_forRandom );
            wait until @request_data_attributes;
        };
        temp_flag = TRUE;
    };
    
    DataNop( count : int ) @request_data_attributes is {
        var i : int;
        var flag : bool;
        
        flag = (count < 0);
        for{ i = 0; (i < count) or flag ; i+=1 } do {
            update_data_instructions_buffer( get_data_nop() );
            wait until @request_data_attributes;
        };
        temp_flag = TRUE;
    };
    
    Idle() @request_data_attributes is {
        var data_attributes : DATA_ACCESS_ATTRIBUTES;
        var i : int;
        var flag : bool;
        
        gen data_attributes;
        data_attributes.set_noaccess_instruction();
        
        data_attributes.idle_flag = TRUE;
        
        update_data_instructions_buffer( data_attributes );
        wait until @request_data_attributes;
        
        temp_flag = TRUE;
    };
    
    PutPipeline( data_attributes : DATA_ACCESS_ATTRIBUTES ) @request_data_attributes is {
        if( data_attributes == NULL ) {
            out("Warring : data_attributes is NULL!");
        };
        update_data_instructions_buffer( data_attributes );
        wait until @request_data_attributes;
        temp_flag = TRUE;
    };
    
    PutPipeline0( data_attributes : DATA_ACCESS_ATTRIBUTES ) is {
        if( data_attributes == NULL ) {
            out("Warring : data_attributes is NULL!");
        };
        data_attributes.isr_flag = TRUE;
        update_data_instructions_buffer0( data_attributes );
        temp_flag = TRUE;
    };
    
    SetSingleRead( data_attributes : DATA_ACCESS_ATTRIBUTES,             -- data_attributes
                   space           : access_space,                       -- select io/memory
                   dabus           : uint(bits : 23)                     -- dbus address
                  ) : DATA_ACCESS_ATTRIBUTES is {
        
        if( data_attributes == NULL or temp_flag ) {
            gen data_attributes keeping { it.lock == disable; };
            data_attributes.set_noaccess_instruction();
        };
        
        data_attributes.read_width = word;
        data_attributes.space      = space;
        data_attributes.dabus      = dabus;
        data_attributes.set_instruction( SingleRead );
        result = data_attributes;
        temp_flag = FALSE;
    };
    
    SetSingleRead_width( data_attributes : DATA_ACCESS_ATTRIBUTES,             -- data_attributes
                         space           : access_space,                       -- select io/memory
                         width           : access_width,                       -- byte/word access
                         dabus           : uint(bits : 23)                     -- dbus address
                       ) : DATA_ACCESS_ATTRIBUTES is {
        
        if( data_attributes == NULL or temp_flag ) {
            gen data_attributes keeping { it.lock == disable; };
            data_attributes.set_noaccess_instruction();
        };
        
        data_attributes.read_width = width;
        data_attributes.space      = space;
        data_attributes.dabus      = dabus;
        data_attributes.set_instruction( SingleRead );
        result = data_attributes;
        temp_flag = FALSE;
    };
    
    SetDualRead( data_attributes : DATA_ACCESS_ATTRIBUTES,             -- data_attributes
                 space           : access_space,                       -- select io/memory
                 cabus           : uint(bits : 23),                    -- cbus address
                 dabus           : uint(bits : 23)                     -- dbus address
               ) : DATA_ACCESS_ATTRIBUTES is {
        
        if( data_attributes == NULL or temp_flag ) {
            gen data_attributes keeping { it.lock == disable; };
            data_attributes.set_noaccess_instruction();
        };
        
        data_attributes.read_width = word;
        data_attributes.space      = space;
        data_attributes.cabus      = cabus;
        data_attributes.dabus      = dabus;
        data_attributes.set_instruction( DualRead );
        result = data_attributes;
        temp_flag = FALSE;
    };
    
    SetDualRead_width( data_attributes : DATA_ACCESS_ATTRIBUTES,             -- data_attributes
                       space           : access_space,                       -- select io/memory
                       width           : access_width,                       -- byte/word access
                       cabus           : uint(bits : 23),                    -- cbus address
                       dabus           : uint(bits : 23)                     -- dbus address
                     ) : DATA_ACCESS_ATTRIBUTES is {
        
        if( data_attributes == NULL or temp_flag ) {
            gen data_attributes keeping { it.lock == disable; };
            data_attributes.set_noaccess_instruction();
        };
        
        data_attributes.read_width = width;
        data_attributes.space      = space;
        data_attributes.cabus      = cabus;
        data_attributes.dabus      = dabus;
        data_attributes.set_instruction( DualRead );
        result = data_attributes;
        temp_flag = FALSE;
    };
    
    SetDoubleRead( data_attributes : DATA_ACCESS_ATTRIBUTES,             -- data_attributes
                   dabus           : uint(bits : 23)                     -- dbus address(cbus address)
                 ) : DATA_ACCESS_ATTRIBUTES is {
        
        if( data_attributes == NULL or temp_flag ) {
            gen data_attributes keeping { it.lock == disable; };
            data_attributes.set_noaccess_instruction();
        };
        
        data_attributes.dabus = dabus;
        data_attributes.set_instruction( DoubleRead );
        result = data_attributes;
        temp_flag = FALSE;
    };
    
    SetSingleWrite( data_attributes : DATA_ACCESS_ATTRIBUTES,             -- data_attributes
                    space           : access_space,                       -- select io/memory
                    width           : access_width,                       -- byte/word access
                    eabus           : uint(bits : 23),                    -- address
                    edbus           : uint(bits : 16)                     -- write data
                  ) : DATA_ACCESS_ATTRIBUTES is {
        
        if( data_attributes == NULL or temp_flag ) {
            gen data_attributes keeping { it.lock == disable; };
            data_attributes.set_noaccess_instruction();
        };
        
        data_attributes.write_width = width;
        data_attributes.space       = space;
        data_attributes.eabus       = eabus;
        data_attributes.edbus       = edbus;
        data_attributes.set_instruction( SingleWrite );
        result = data_attributes;
        temp_flag = FALSE;
    };
    
    SetDualWrite( data_attributes : DATA_ACCESS_ATTRIBUTES,             -- data_attributes
                  space           : access_space,                       -- select io/memory
                  width           : access_width,                       -- byte/word access
                  eabus           : uint(bits : 23),                    -- ebus address
                  edbus           : uint(bits : 16),                    -- write data
                  fabus           : uint(bits : 23),                    -- fbus address
                  fdbus           : uint(bits : 16)                     -- write data
                ) : DATA_ACCESS_ATTRIBUTES is {
        
        if( data_attributes == NULL or temp_flag ) {
            gen data_attributes keeping { it.lock == disable; };
            data_attributes.set_noaccess_instruction();
        };
        
        data_attributes.write_width = width;
        data_attributes.space       = space;
        data_attributes.eabus       = eabus;
        data_attributes.edbus       = edbus;
        data_attributes.fabus       = fabus;
        data_attributes.fdbus       = fdbus;
        data_attributes.set_instruction( DualWrite );
        result = data_attributes;
        temp_flag = FALSE;
    };
    
    SetDoubleWrite( data_attributes : DATA_ACCESS_ATTRIBUTES,             -- data_attributes
                    eabus           : uint(bits : 23),                    -- ebus address(fbus address)
                    edbus           : uint(bits : 16),                    -- write data
                    fdbus           : uint(bits : 16)                     -- write data
                  ) : DATA_ACCESS_ATTRIBUTES is {
        
        if( data_attributes == NULL or temp_flag ) {
            gen data_attributes keeping { it.lock == disable; };
            data_attributes.set_noaccess_instruction();
        };
        
        data_attributes.eabus = eabus;
        data_attributes.edbus = edbus;
        data_attributes.fdbus = fdbus;
        data_attributes.set_instruction( DoubleWrite );
        result = data_attributes;
        temp_flag = FALSE;
    };
    
    SetLockAccess( data_attributes : DATA_ACCESS_ATTRIBUTES,             -- data_attributes
                   space           : access_space,                       -- select io/memory
                   abus            : uint(bits : 23),                    -- dbus/ebus address
                   dbus            : uint(bits : 16)                     -- ebus write data
                  ) : DATA_ACCESS_ATTRIBUTES is {
        
        if( data_attributes == NULL or temp_flag ) {
            gen data_attributes keeping { it.lock == enable; };
            data_attributes.set_noaccess_instruction();
        };
        
        data_attributes.lock_incomplete = FALSE;
        
        data_attributes.space = space;
        data_attributes.dabus = abus;
        data_attributes.eabus = abus;
        data_attributes.edbus = dbus;
        data_attributes.set_instruction( LockAccess );
        result = data_attributes;
        temp_flag = FALSE;
    };
    
    SetLockAccess_SingleWrite( data_attributes : DATA_ACCESS_ATTRIBUTES,             -- data_attributes
                               space          : access_space,                       -- select io/memory
                               abus           : uint(bits : 23),                    -- dbus/fbus address
                               dbus           : uint(bits : 16),                    -- fbus write data
                               eabus          : uint(bits : 23),                    -- ebus address
                               edbus          : uint(bits : 16)                     -- ebus write data
                             ) : DATA_ACCESS_ATTRIBUTES is {
        
        if( data_attributes == NULL or temp_flag ) {
            gen data_attributes keeping { it.lock == enable; };
            data_attributes.set_noaccess_instruction();
        };
        
        data_attributes.set_noaccess_instruction();
        result = data_attributes;
        temp_flag = FALSE;
    };
    
    SetLockAccess_SingleRead( data_attributes : DATA_ACCESS_ATTRIBUTES,             -- data_attributes
                              space          : access_space,                       -- select io/memory
                              abus           : uint(bits : 23),                    -- dbus/ebus address
                              dbus           : uint(bits : 16),                    -- ebus write data
                              cabus          : uint(bits : 23)                     -- cbus address
                            ) : DATA_ACCESS_ATTRIBUTES is {
        
        if( data_attributes == NULL or temp_flag ) {
            gen data_attributes keeping { it.lock == enable; };
            data_attributes.set_noaccess_instruction();
        };
        
        if( data_attributes == NULL or temp_flag ) {
            gen data_attributes keeping { it.lock == enable; };
            data_attributes.set_noaccess_instruction();
        };
        
        data_attributes.set_noaccess_instruction();
        result = data_attributes;
        temp_flag = FALSE;
    };
    
    SetCoefRead( data_attributes : DATA_ACCESS_ATTRIBUTES,             -- data_attributes
                 babus           : uint(bits : 23)                     -- address
               ) : DATA_ACCESS_ATTRIBUTES is {
        
        if( data_attributes == NULL or temp_flag ) {
            gen data_attributes keeping { it.lock == disable; };   -- coef access doesn't allow with lock access
            data_attributes.set_noaccess_instruction();
        };
        
        data_attributes.babus   = babus;
        data_attributes.dblcoef = word;
        
        data_attributes.set_instruction( CoefRead );
        
        result = data_attributes;
        temp_flag = FALSE;
    };
    
    SetCoefRead_width( data_attributes : DATA_ACCESS_ATTRIBUTES,             -- data_attributes
                       width           : access_double,                      -- word/double
                       babus           : uint(bits : 23)                     -- address
                     ) : DATA_ACCESS_ATTRIBUTES is {
        
        if( data_attributes == NULL or temp_flag ) {
            gen data_attributes keeping { it.lock == disable; };   -- coef access doesn't allow with lock access
            data_attributes.set_noaccess_instruction();
        };
        
        data_attributes.babus   = babus;
        data_attributes.dblcoef = width;
        
        data_attributes.set_instruction( CoefRead );
        
        result = data_attributes;
        temp_flag = FALSE;
    };
    
    DataAccessFinish() @request_data_attributes is {
        var i : uint;
        
        for{ i = 0; i < 10; i+=1 } do {
            update_data_instructions_buffer( get_data_nop() );
            wait until @request_data_attributes;
        };
        temp_flag = FALSE;
        
        data_access_complete = TRUE;
        
    };
    
    -- for MMR access functions --
    MMRReadAccess( abus : uint(bits:16) ) : uint(bits:16) @request_data_attributes is {
        var data_attributes : DATA_ACCESS_ATTRIBUTES;
        
        gen data_attributes keeping { it.lock == disable; };
        data_attributes.set_noaccess_instruction();
        
        -- single read for mmr --
        if( abus < 0x5f ) {
            data_attributes = SetSingleRead( data_attributes, memory, abus );
        };
        
        update_data_instructions_buffer( data_attributes );
        wait until @request_data_attributes;
        
        DataNop( 3 );
        
        result = bus_states.mmr.rdata;
    };
    
    MMRWriteAccess( abus : uint(bits:16), dbus : uint(bits:16) ) @request_data_attributes is {
        var data_attributes : DATA_ACCESS_ATTRIBUTES;
        
        gen data_attributes keeping { it.lock == disable; };
        data_attributes.set_noaccess_instruction();
        
        -- single write for mmr --
        if( abus < 0x5f ) {
            data_attributes = SetSingleWrite( data_attributes, memory, word, abus, dbus );
        };
        
        update_data_instructions_buffer( data_attributes );
        wait until @request_data_attributes;
        
        DataNop( 6 );
    };
    
    -- emulation access functions --
    DebugDataSingleRead( space : access_space, dabus : uint(bits : 23) ) @request_data_attributes is {
        var data_attributes : DATA_ACCESS_ATTRIBUTES;
        
        gen data_attributes keeping { it.lock == disable; };
        data_attributes.set_noaccess_instruction();
        
        data_attributes.dbus_type  = debug;
        data_attributes.read_width = word;
        data_attributes.space      = space;
        data_attributes.dabus      = dabus;
        data_attributes.set_instruction( SingleRead );
        
        update_data_instructions_buffer( data_attributes );
        wait until @request_data_attributes;
    };
    
    DebugDataSingleRead_width( space : access_space, width : access_width, dabus : uint(bits : 23) ) @request_data_attributes is {
        var data_attributes : DATA_ACCESS_ATTRIBUTES;
        
        gen data_attributes keeping { it.lock == disable; };
        data_attributes.set_noaccess_instruction();
        
        data_attributes.dbus_type  = debug;
        data_attributes.read_width = width;
        data_attributes.space      = space;
        data_attributes.dabus      = dabus;
        data_attributes.set_instruction( SingleRead );
        
        update_data_instructions_buffer( data_attributes );
        wait until @request_data_attributes;
    };
    
    DebugDataDoubleRead( dabus : uint(bits : 23) ) @request_data_attributes is {
        var data_attributes : DATA_ACCESS_ATTRIBUTES;
        
        gen data_attributes keeping { it.lock == disable; };
        data_attributes.set_noaccess_instruction();
        
        data_attributes.dbus_type = debug;
        data_attributes.dabus     = dabus;
        data_attributes.set_instruction( DoubleRead );
        
        update_data_instructions_buffer( data_attributes );
        wait until @request_data_attributes;
    };
    
    DebugDataSingleWrite( space : access_space, width : access_width, eabus : uint(bits : 23), edbus : uint(bits : 16) ) @request_data_attributes is {
        var data_attributes : DATA_ACCESS_ATTRIBUTES;
        
        gen data_attributes keeping { it.lock == disable; };
        data_attributes.set_noaccess_instruction();
        
        data_attributes.ebus_type   = debug_data_write;
        data_attributes.write_width = width;
        data_attributes.space       = space;
        data_attributes.eabus       = eabus;
        data_attributes.edbus       = edbus;
        data_attributes.set_instruction( SingleWrite );
        
        update_data_instructions_buffer( data_attributes );
        wait until @request_data_attributes;
    };
    
    DebugDataDoubleWrite( eabus : uint(bits : 23), edbus : uint(bits : 16), fdbus : uint(bits : 16) ) @request_data_attributes is {
        var data_attributes : DATA_ACCESS_ATTRIBUTES;
        
        gen data_attributes keeping { it.lock == disable; };
        data_attributes.set_noaccess_instruction();
        
        data_attributes.ebus_type = debug_data_write;
        data_attributes.eabus     = eabus;
        data_attributes.edbus     = edbus;
        data_attributes.fdbus     = fdbus;
        data_attributes.set_instruction( DoubleWrite );
        
        update_data_instructions_buffer( data_attributes );
        wait until @request_data_attributes;
    };
    
    DebugProgramSingleWrite( width : access_width, eabus : uint(bits : 23), edbus : uint(bits : 16) ) @request_data_attributes is {
        var data_attributes : DATA_ACCESS_ATTRIBUTES;
        
        gen data_attributes keeping { it.lock == disable; };
        data_attributes.set_noaccess_instruction();
        
        data_attributes.ebus_type   = debug_prog_write;
        data_attributes.write_width = width;
        data_attributes.space       = memory;
        data_attributes.eabus       = eabus;
        data_attributes.edbus       = edbus;
        data_attributes.set_instruction( SingleWrite );
        
        update_data_instructions_buffer( data_attributes );
        wait until @request_data_attributes;
    };
    
    DebugProgramDoubleWrite( eabus : uint(bits : 23), edbus : uint(bits : 16), fdbus : uint(bits : 16) ) @request_data_attributes is {
        var data_attributes : DATA_ACCESS_ATTRIBUTES;
        
        gen data_attributes keeping { it.lock == disable; };
        data_attributes.set_noaccess_instruction();
        
        data_attributes.ebus_type = debug_prog_write;
        data_attributes.eabus     = eabus;
        data_attributes.edbus     = edbus;
        data_attributes.fdbus     = fdbus;
        data_attributes.set_instruction( DoubleWrite );
        
        update_data_instructions_buffer( data_attributes );
        wait until @request_data_attributes;
    };
    
    gen_mp_attributes( data_attributes : DATA_ACCESS_ATTRIBUTES ) is {
        
        data_attributes.mp_supervisor = mp_supervisor;
        
        gen mp_viol_address;
        data_attributes.mp_viol_address = mp_viol_address;
        
        if( data_attributes.ebus_active == active and data_attributes.doublew == word ) {
          gen mp_viol_contextsave;
          data_attributes.mp_viol_contextsave = mp_viol_contextsave;
        } else {
          data_attributes.mp_viol_contextsave = FALSE;
        };
        
    };
    
    update_data_instructions_buffer( data_attributes : DATA_ACCESS_ATTRIBUTES ) is {
        -- set mp attributes --
        gen_mp_attributes( data_attributes );
        
        -- update the queue --
        if ( prog_access_complete == FALSE or data_access_complete == FALSE ) {
           data_queue.push(data_attributes);
        };
    };
    update_data_instructions_buffer0( data_attributes : DATA_ACCESS_ATTRIBUTES ) is {
        -- set mp attributes --
        gen_mp_attributes( data_attributes );
        
        -- update the queue --
        if ( prog_access_complete == FALSE or data_access_complete == FALSE ) {
           data_queue.push0(data_attributes);
        };
    };
    
    -- emulation enable functions --
    DebugEnable() is {
        bus_driver.emulations.debugenable = 1;
    };
    
    DebugDisable() is {
        bus_driver.emulations.debugenable = 0;
    };
    
    -- abortready assertion --
    !abortready_cycle : uint;
    Abortready( cycle : uint ) is {
        abortready_cycle = cycle;
    };
    TCM_abortready_counter() @rise_clock is {
        while( TRUE ) {
            if( abortready_cycle != 0 ) {
                abortready_cycle = abortready_cycle - 1;
            };
            
            if( abortready_cycle != 0 ) {
                bus_driver.emulations.abortready = 1;
            } else {
                bus_driver.emulations.abortready = 0;
            };
            
            wait cycle;
        };
    };
    
    -- suspend assertion --
    !suspend_cycle : uint;
    Suspend( cycle : uint ) is {
        suspend_cycle = cycle;
    };
    TCM_suspend_counter() @rise_clock is {
        while( TRUE ) {
            if( suspend_cycle != 0 ) {
                suspend_cycle = suspend_cycle - 1;
            };
            
            if( suspend_cycle != 0 ) {
                bus_driver.emulations.suspend = 0;
            } else {
                bus_driver.emulations.suspend = 1;
            };
            
            wait cycle;
        };
    };
    
    init() is also {
        abortready_cycle = 0;
        suspend_cycle    = 0;
    };
    
    -- override function --
    MyProgramAccess() @request_program_attributes is {
        var i : uint;
        
        for{ i = 0; i < my_pcount; i += 1 } do {
            gen prog_attributes_forRandom;
            update_program_instructions_buffer( prog_attributes_forRandom );
            wait until @request_program_attributes;
        };
        
        ProgramAccessFinish();
    };

    MyDataAccess() @request_data_attributes is {
        var i : uint;
        
        for{ i = 0; i < my_dcount; i += 1 } do {
            gen data_attributes_forRandom;
            update_data_instructions_buffer( data_attributes_forRandom );
            wait until @request_data_attributes;
        };
        
        DataAccessFinish();
    };
    
    MyConfigAccess() @request_data_attributes is {
    };
    
    -- nop generate --
    get_prog_nop() : PROGRAM_ACCESS_ATTRIBUTES is {
        var prog_attributes : PROGRAM_ACCESS_ATTRIBUTES = new;
        
        prog_attributes.set_noaccess_instruction();
        prog_attributes.pbus_active = no_active;
        
        result = prog_attributes;
    };
    
    get_data_nop() : DATA_ACCESS_ATTRIBUTES is {
        var data_attributes : DATA_ACCESS_ATTRIBUTES = new;
        
        data_attributes.set_noaccess_instruction();
        data_attributes.set_activate_buses();
        
        result = data_attributes;
    };
    
    -----------------------------------------
    -- Memory Protection Control functions --
    -----------------------------------------
    MpPrivilegeEnable() is {
        mp_enable = TRUE;
    };
    MpPrivilegeDisable() is {
        mp_enable = FALSE;
    };
    MpPrivilegeUpdate() is {
        if( bus_states.mp.current.gl_privenable == 1 ) { 
            mp_enable = TRUE;
        };
        
        if( mp_enable ) {
            bus_driver.mp.cpu_privenable = 1;
        } else {
            bus_driver.mp.cpu_privenable = 0;
        };
        
    };
    
    MpPrivilegeSupervisor() is {
        mp_supervisor = TRUE;
    };
    MpPrivilegeUser() is {
        mp_supervisor = FALSE;
    };
    MpPrivilegeRandom() is {
        gen mp_supervisor;
    };
    
    --------------------------------
    -- interrupt checking routine --
    --------------------------------
    int_branch( int_index : uint ) is {
        var prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
        
        gen prog_attributes keeping { 
            it.instr_group     == ProgramRead;
            it.branch_occur    == TRUE;
            it.dismiss_occur   == TRUE;
            it.isr_flag        == TRUE;
            it.isr_index       == int_index;
        };
        update_program_instructions_buffer0( prog_attributes );
    };
    
    int_check() is {
        var int_value : uint;
        var var_ifr0  : uint(bits : 16);
        var var_ifr1  : uint(bits : 16);
        
        -- when buserror happen --
        if( bus_states.system.current.nmi        == 0                                              ) { buffer.isr_queue.push(30); int_branch(30); };
        if( bus_states.system.current.int        != 0x3fffff ) {
            if( bus_states.system.current.int[00:00] == 0 and bus_states.mmr.reg[MMR_IMR0][02:02] != 1 ) { buffer.isr_queue.push(00); int_branch(00); };
            if( bus_states.system.current.int[01:01] == 0 and bus_states.mmr.reg[MMR_IMR0][03:03] != 1 ) { buffer.isr_queue.push(01); int_branch(01); };
            if( bus_states.system.current.int[02:02] == 0 and bus_states.mmr.reg[MMR_IMR0][04:04] != 1 ) { buffer.isr_queue.push(02); int_branch(02); };
            if( bus_states.system.current.int[03:03] == 0 and bus_states.mmr.reg[MMR_IMR0][05:05] != 1 ) { buffer.isr_queue.push(03); int_branch(03); };
            if( bus_states.system.current.int[04:04] == 0 and bus_states.mmr.reg[MMR_IMR0][06:06] != 1 ) { buffer.isr_queue.push(04); int_branch(04); };
            if( bus_states.system.current.int[05:05] == 0 and bus_states.mmr.reg[MMR_IMR0][07:07] != 1 ) { buffer.isr_queue.push(05); int_branch(05); };
            if( bus_states.system.current.int[06:06] == 0 and bus_states.mmr.reg[MMR_IMR0][08:08] != 1 ) { buffer.isr_queue.push(06); int_branch(06); };
            if( bus_states.system.current.int[07:07] == 0 and bus_states.mmr.reg[MMR_IMR0][09:09] != 1 ) { buffer.isr_queue.push(07); int_branch(07); };
            if( bus_states.system.current.int[08:08] == 0 and bus_states.mmr.reg[MMR_IMR0][10:10] != 1 ) { buffer.isr_queue.push(08); int_branch(08); };
            if( bus_states.system.current.int[09:09] == 0 and bus_states.mmr.reg[MMR_IMR0][11:11] != 1 ) { buffer.isr_queue.push(09); int_branch(09); };
            if( bus_states.system.current.int[10:10] == 0 and bus_states.mmr.reg[MMR_IMR0][12:12] != 1 ) { buffer.isr_queue.push(10); int_branch(10); };
            if( bus_states.system.current.int[11:11] == 0 and bus_states.mmr.reg[MMR_IMR0][13:13] != 1 ) { buffer.isr_queue.push(11); int_branch(11); };
            if( bus_states.system.current.int[12:12] == 0 and bus_states.mmr.reg[MMR_IMR0][14:14] != 1 ) { buffer.isr_queue.push(12); int_branch(12); };
            if( bus_states.system.current.int[13:13] == 0 and bus_states.mmr.reg[MMR_IMR0][15:15] != 1 ) { buffer.isr_queue.push(13); int_branch(13); };
            if( bus_states.system.current.int[14:14] == 0 and bus_states.mmr.reg[MMR_IMR1][00:00] != 1 ) { buffer.isr_queue.push(14); int_branch(14); };
            if( bus_states.system.current.int[15:15] == 0 and bus_states.mmr.reg[MMR_IMR1][01:01] != 1 ) { buffer.isr_queue.push(15); int_branch(15); };
            if( bus_states.system.current.int[16:16] == 0 and bus_states.mmr.reg[MMR_IMR1][02:02] != 1 ) { buffer.isr_queue.push(16); int_branch(16); };
            if( bus_states.system.current.int[17:17] == 0 and bus_states.mmr.reg[MMR_IMR1][03:03] != 1 ) { buffer.isr_queue.push(17); int_branch(17); };
            if( bus_states.system.current.int[18:18] == 0 and bus_states.mmr.reg[MMR_IMR1][04:04] != 1 ) { buffer.isr_queue.push(18); int_branch(18); };
            if( bus_states.system.current.int[19:19] == 0 and bus_states.mmr.reg[MMR_IMR1][05:05] != 1 ) { buffer.isr_queue.push(19); int_branch(19); };
            if( bus_states.system.current.int[20:20] == 0 and bus_states.mmr.reg[MMR_IMR1][06:06] != 1 ) { buffer.isr_queue.push(20); int_branch(20); };
            if( bus_states.system.current.int[21:21] == 0 and bus_states.mmr.reg[MMR_IMR1][07:07] != 1 ) { buffer.isr_queue.push(21); int_branch(21); };
            if( bus_states.system.current.buserror   == 0 and bus_states.mmr.reg[MMR_IMR1][08:08] != 1 ) { buffer.isr_queue.push(31); int_branch(31); };
            if( bus_states.system.current.debugerror == 0 and bus_states.mmr.reg[MMR_IMR1][08:08] != 1 ) { buffer.isr_queue.push(32); int_branch(32); };
        };

        var_ifr0 = bus_states.mmr.reg[MMR_IFR0];
        var_ifr1 = bus_states.mmr.reg[MMR_IFR1];
        if( bus_states.system.current.int        != 0x3fffff ) {
            if( bus_states.system.current.int[00:00] == 0 and bus_states.mmr.reg[MMR_IMR0][02:02] != 1 ) { var_ifr0 = var_ifr0 | 16'b0000000000000100; };
            if( bus_states.system.current.int[01:01] == 0 and bus_states.mmr.reg[MMR_IMR0][03:03] != 1 ) { var_ifr0 = var_ifr0 | 16'b0000000000001000; };
            if( bus_states.system.current.int[02:02] == 0 and bus_states.mmr.reg[MMR_IMR0][04:04] != 1 ) { var_ifr0 = var_ifr0 | 16'b0000000000010000; };
            if( bus_states.system.current.int[03:03] == 0 and bus_states.mmr.reg[MMR_IMR0][05:05] != 1 ) { var_ifr0 = var_ifr0 | 16'b0000000000100000; };
            if( bus_states.system.current.int[04:04] == 0 and bus_states.mmr.reg[MMR_IMR0][06:06] != 1 ) { var_ifr0 = var_ifr0 | 16'b0000000001000000; };
            if( bus_states.system.current.int[05:05] == 0 and bus_states.mmr.reg[MMR_IMR0][07:07] != 1 ) { var_ifr0 = var_ifr0 | 16'b0000000010000000; };
            if( bus_states.system.current.int[06:06] == 0 and bus_states.mmr.reg[MMR_IMR0][08:08] != 1 ) { var_ifr0 = var_ifr0 | 16'b0000000100000000; };
            if( bus_states.system.current.int[07:07] == 0 and bus_states.mmr.reg[MMR_IMR0][09:09] != 1 ) { var_ifr0 = var_ifr0 | 16'b0000001000000000; };
            if( bus_states.system.current.int[08:08] == 0 and bus_states.mmr.reg[MMR_IMR0][10:10] != 1 ) { var_ifr0 = var_ifr0 | 16'b0000010000000000; };
            if( bus_states.system.current.int[09:09] == 0 and bus_states.mmr.reg[MMR_IMR0][11:11] != 1 ) { var_ifr0 = var_ifr0 | 16'b0000100000000000; };
            if( bus_states.system.current.int[10:10] == 0 and bus_states.mmr.reg[MMR_IMR0][12:12] != 1 ) { var_ifr0 = var_ifr0 | 16'b0001000000000000; };
            if( bus_states.system.current.int[11:11] == 0 and bus_states.mmr.reg[MMR_IMR0][13:13] != 1 ) { var_ifr0 = var_ifr0 | 16'b0010000000000000; };
            if( bus_states.system.current.int[12:12] == 0 and bus_states.mmr.reg[MMR_IMR0][14:14] != 1 ) { var_ifr0 = var_ifr0 | 16'b0100000000000000; };
            if( bus_states.system.current.int[13:13] == 0 and bus_states.mmr.reg[MMR_IMR0][15:15] != 1 ) { var_ifr0 = var_ifr0 | 16'b1000000000000000; };
            if( bus_states.system.current.int[14:14] == 0 and bus_states.mmr.reg[MMR_IMR1][00:00] != 1 ) { var_ifr1 = var_ifr1 | 16'b0000000000000001; };
            if( bus_states.system.current.int[15:15] == 0 and bus_states.mmr.reg[MMR_IMR1][01:01] != 1 ) { var_ifr1 = var_ifr1 | 16'b0000000000000010; };
            if( bus_states.system.current.int[16:16] == 0 and bus_states.mmr.reg[MMR_IMR1][02:02] != 1 ) { var_ifr1 = var_ifr1 | 16'b0000000000000100; };
            if( bus_states.system.current.int[17:17] == 0 and bus_states.mmr.reg[MMR_IMR1][03:03] != 1 ) { var_ifr1 = var_ifr1 | 16'b0000000000001000; };
            if( bus_states.system.current.int[18:18] == 0 and bus_states.mmr.reg[MMR_IMR1][04:04] != 1 ) { var_ifr1 = var_ifr1 | 16'b0000000000010000; };
            if( bus_states.system.current.int[19:19] == 0 and bus_states.mmr.reg[MMR_IMR1][05:05] != 1 ) { var_ifr1 = var_ifr1 | 16'b0000000000100000; };
            if( bus_states.system.current.int[20:20] == 0 and bus_states.mmr.reg[MMR_IMR1][06:06] != 1 ) { var_ifr1 = var_ifr1 | 16'b0000000001000000; };
            if( bus_states.system.current.int[21:21] == 0 and bus_states.mmr.reg[MMR_IMR1][07:07] != 1 ) { var_ifr1 = var_ifr1 | 16'b0000000010000000; };
            if( bus_states.system.current.buserror   == 0 and bus_states.mmr.reg[MMR_IMR1][08:08] != 1 ) { var_ifr1 = var_ifr1 | 16'b0000000100000000; };
            if( bus_states.system.current.debugerror == 0 and bus_states.mmr.reg[MMR_IMR1][08:08] != 1 ) { var_ifr1 = var_ifr1 | 16'b0000000100000000; };
        };
    
        if( bus_states.system.current.iack == 0 and !buffer.isr_queue.is_empty() ) {
            int_value = buffer.isr_queue.pop0();
            if( int_value == 31 ) { var_ifr1 = var_ifr1 & 16'b1111111011111111; };
            if( int_value == 32 ) { var_ifr1 = var_ifr1 & 16'b1111111011111111; };
            if( int_value == 00 ) { var_ifr0 = var_ifr0 & 16'b1111111111111011; };
            if( int_value == 01 ) { var_ifr0 = var_ifr0 & 16'b1111111111110111; };
            if( int_value == 02 ) { var_ifr0 = var_ifr0 & 16'b1111111111101111; };
            if( int_value == 03 ) { var_ifr0 = var_ifr0 & 16'b1111111111011111; };
            if( int_value == 04 ) { var_ifr0 = var_ifr0 & 16'b1111111110111111; };
            if( int_value == 05 ) { var_ifr0 = var_ifr0 & 16'b1111111101111111; };
            if( int_value == 06 ) { var_ifr0 = var_ifr0 & 16'b1111111011111111; };
            if( int_value == 07 ) { var_ifr0 = var_ifr0 & 16'b1111110111111111; };
            if( int_value == 08 ) { var_ifr0 = var_ifr0 & 16'b1111101111111111; };
            if( int_value == 09 ) { var_ifr0 = var_ifr0 & 16'b1111011111111111; };
            if( int_value == 10 ) { var_ifr0 = var_ifr0 & 16'b1110111111111111; };
            if( int_value == 11 ) { var_ifr0 = var_ifr0 & 16'b1101111111111111; };
            if( int_value == 12 ) { var_ifr0 = var_ifr0 & 16'b1011111111111111; };
            if( int_value == 13 ) { var_ifr0 = var_ifr0 & 16'b0111111111111111; };
            if( int_value == 14 ) { var_ifr1 = var_ifr1 & 16'b1111111111111110; };
            if( int_value == 15 ) { var_ifr1 = var_ifr1 & 16'b1111111111111101; };
            if( int_value == 16 ) { var_ifr1 = var_ifr1 & 16'b1111111111111011; };
            if( int_value == 17 ) { var_ifr1 = var_ifr1 & 16'b1111111111110111; };
            if( int_value == 18 ) { var_ifr1 = var_ifr1 & 16'b1111111111101111; };
            if( int_value == 19 ) { var_ifr1 = var_ifr1 & 16'b1111111111011111; };
            if( int_value == 20 ) { var_ifr1 = var_ifr1 & 16'b1111111110111111; };
            if( int_value == 21 ) { var_ifr1 = var_ifr1 & 16'b1111111101111111; };
        };
        
        bus_states.mmr.reg[MMR_IFR0] = var_ifr0;
        bus_states.mmr.reg[MMR_IFR1] = var_ifr1 ;
        
    };
    
    call_isr_functions(  prog_attributes : PROGRAM_ACCESS_ATTRIBUTES ) is {
        
        if( prog_attributes.isr_flag == TRUE ) {
            if( prog_attributes.isr_index == 30 ) { isr_nmi();        };
            if( prog_attributes.isr_index == 00 ) { isr_int0();       };
            if( prog_attributes.isr_index == 01 ) { isr_int1();       };
            if( prog_attributes.isr_index == 02 ) { isr_int2();       };
            if( prog_attributes.isr_index == 03 ) { isr_int3();       };
            if( prog_attributes.isr_index == 04 ) { isr_int4();       };
            if( prog_attributes.isr_index == 05 ) { isr_int5();       };
            if( prog_attributes.isr_index == 06 ) { isr_int6();       };
            if( prog_attributes.isr_index == 07 ) { isr_int7();       };
            if( prog_attributes.isr_index == 08 ) { isr_int8();       };
            if( prog_attributes.isr_index == 09 ) { isr_int9();       };
            if( prog_attributes.isr_index == 10 ) { isr_int10();      };
            if( prog_attributes.isr_index == 11 ) { isr_int11();      };
            if( prog_attributes.isr_index == 12 ) { isr_int12();      };
            if( prog_attributes.isr_index == 13 ) { isr_int13();      };
            if( prog_attributes.isr_index == 14 ) { isr_int14();      };
            if( prog_attributes.isr_index == 15 ) { isr_int15();      };
            if( prog_attributes.isr_index == 16 ) { isr_int16();      };
            if( prog_attributes.isr_index == 17 ) { isr_int17();      };
            if( prog_attributes.isr_index == 18 ) { isr_int18();      };
            if( prog_attributes.isr_index == 19 ) { isr_int19();      };
            if( prog_attributes.isr_index == 20 ) { isr_int20();      };
            if( prog_attributes.isr_index == 21 ) { isr_int21();      };
            if( prog_attributes.isr_index == 31 ) { isr_buserror();   };
            if( prog_attributes.isr_index == 32 ) { isr_debugerror(); };
        };
        
    };
    
    -- isr functions for override by user --
    --
    -- Note : only following functions are allowed to use for overriding these isr functions;
    --          SetSingleRead(data_attributes,space,dabus)
    --          SetSingleWrite(data_attributes,space,width,eabus,edbus)
    --          PutPipeline0( data_attributes )
    --        And these function should be set with inverted order.
    -- 
    -- Example : 
    --   
    --   isr_nmi() is only {
    --     var data_attributes : DATA_ACCESS_ATTRIBUTES;
    --     
    --     data_attributes = SetSingleWrite( data_attributes, io , word, 0x0200, 0xffff );  PutPipeline0( data_attributes );
    --     data_attributes = SetSingleWrite( data_attributes, io , word, 0x0201, 0xffff );  PutPipeline0( data_attributes );
    --     data_attributes = SetSingleWrite( data_attributes, io , word, 0x0202, 0xffff );  PutPipeline0( data_attributes );
    --     data_attributes = SetSingleWrite( data_attributes, io , word, 0x0203, 0xffff );  PutPipeline0( data_attributes );
    --   };
    --   
    --   this overrided function will execute following sequence when nmi is happned;
    --     1. write to 0x0203 in io space
    --     2. write to 0x0202 in io space
    --     3. write to 0x0201 in io space
    --     4. write to 0x0200 in io space
    --   
    
    isr_nmi()        is { };
    isr_int0()       is { };
    isr_int1()       is { };
    isr_int2()       is { };
    isr_int3()       is { };
    isr_int4()       is { };
    isr_int5()       is { };
    isr_int6()       is { };
    isr_int7()       is { };
    isr_int8()       is { };
    isr_int9()       is { };
    isr_int10()      is { };
    isr_int11()      is { };
    isr_int12()      is { };
    isr_int13()      is { };
    isr_int14()      is { };
    isr_int15()      is { };
    isr_int16()      is { };
    isr_int17()      is { };
    isr_int18()      is { };
    isr_int19()      is { };
    isr_int20()      is { };
    isr_int21()      is { };
    isr_buserror()   is { };
    isr_debugerror() is { };
    
    ------------------------------
    -- for mp control functions --
    ------------------------------
    !mp_viol_stall : bool;
    !mp_viol_flag  : bool;
    init() is also { mp_viol_stall = FALSE; mp_viol_flag = FALSE; };
    
    mp_violation_check() is {
        var i : uint;
        var prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
        
        if( bus_states.system.current.nmi == 0 ) {
            mp_viol_stall = FALSE;
            mp_viol_flag  = FALSE;
        };
        
        if( mp_viol_flag == FALSE and mp_viol_stall == FALSE and
            bus_states.pbus.state.previous_ready == ready and 
            bus_states.mp.current.mpcxviol       == 1     and 
            bus_states.pbus.previous.dismiss     != 1         ) {
            
            gen mp_viol_pdelay;
            gen mp_viol_pclear;
            
            -- if buffer is flush, brunch occur --
            if( mp_viol_pclear ) {
                gen prog_attributes keeping { 
                    it.instr_group   == ProgramRead;
                    it.branch_occur  == TRUE;
                };
                update_program_instructions_buffer0( prog_attributes );
            };
            
            -- dummy nop --
            update_program_instructions_buffer0( get_prog_nop() );
            
            -- set violation flag --
            prog_attributes = get_prog_nop();
            prog_attributes.mp_viol_pbus   = TRUE;
            prog_attributes.mp_viol_pclear = mp_viol_pclear;
            update_program_instructions_buffer0( prog_attributes );
            
            -- set nop for delay --
            for{ i = 0; i < mp_viol_pdelay; i = i + 1; } do {
                update_program_instructions_buffer0( get_prog_nop() );
            };
            
            -- set violation flag --
            if( bus_states.mp.current.cpu_privenable == 1 ) { mp_viol_flag = TRUE; };
            
        };
        
    };
    
    set_mp_attributes( prog_attributes : PROGRAM_ACCESS_ATTRIBUTES, data_attributes : DATA_ACCESS_ATTRIBUTES ) is {
        
        -- set mp violation flag --
        if( prog_attributes.mp_viol_pbus and !prog_attributes.mp_viol_pclear ) {
            data_attributes.mp_viol_pbus = TRUE;
            if( bus_states.mp.current.cpu_privenable == 0 ) { int_branch( 31 ); };
        };
        
        -- update control flags --
        if( mp_viol_flag and prog_attributes.mp_viol_pbus ) {
            
            mp_viol_flag  = FALSE;
            mp_viol_stall = TRUE;
            
            if( bus_states.mp.current.cpu_privenable == 0 or prog_attributes.mp_viol_pclear ) {
                mp_viol_stall = FALSE;
            };
            
        };
        
    };
    
    -----------------------------
    -- for async-reset driving --
    -----------------------------
    event reset_true is true( paths.gl_resetn$ == 1'b0 );
    
    on reset_true {
        
        -- driving reset output --
        bus_states.system.current.cpu_resetn = 0;
        bus_driver.systems.cpu_resetn        = 0;
        if( paths.cpu_resetn.hdl_path() != "" ) { paths.cpu_resetn$ = 0; };
        
        -- update privenable output --
        MpPrivilegeUpdate();
        
    };
    
    ------------------------------
    -- cpu bus state transition --
    ------------------------------
    on rise_clock
    {
        -- keep current values
        bus_states.keep_current_values();
        
        -- get current value
      --bus_states.get_current_value( config, paths );
        bus_states.get_current_value_for_master( config, paths, bus_driver );
        
        -- state transition
        bus_states.state_transition();
        
        -- for keeping pending requests --
        bus_states.pending_prog_requests( prog_attributes );
        bus_states.pending_data_requests( pipeline.access1, pipeline.execute );
        
        -- state initialize when reset 
        if( bus_states.system.current.cpu_resetn == 0 ) {
            bus_states.set_reset_values();
            bus_states.init();
            pipeline.init();
            buffer.init();
        };
        
        -- inserting dummy nops for program space                --
        -- for avoiding immediately prequest issuing after reset --
        if( bus_states.system.previous.gl_resetn == 0 and bus_states.system.current.gl_resetn == 1 ) {
            update_program_instructions_buffer0( get_prog_nop() );
            update_program_instructions_buffer0( get_prog_nop() );
            update_program_instructions_buffer0( get_prog_nop() );
            update_program_instructions_buffer0( get_prog_nop() );
            update_program_instructions_buffer0( get_prog_nop() );
        };
        
        -- DEBUG OUTPUT --
      --bus_states.debug_output();                                                              -- DEBUG OUT 
      --if( pipeline.access1 != NULL ) { pipeline.access1.debug_output_read( bus_states );  };  -- DEBUG OUT 
      --if( pipeline.execute != NULL ) { pipeline.execute.debug_output_write( bus_states ); };  -- DEBUG OUT 
      --pipeline.debug_output( bus_states );                                                    -- DEBUG OUT 
    };
    
    ------------------
    -- driving data --
    ------------------
    TCM_master_driver() @rise_clock is {
        var isr_flag : bool;
        
        while( TRUE ) {
            
            -----------------------------
            -- check pbus mp violation --
            -----------------------------
            mp_violation_check();
            
            ---------------------
            -- check interrupt --
            ---------------------
            if( config.interrupt_enable ) { int_check(); };
            
            -------------------------------------
            -- generate program bus attributes --
            -------------------------------------
            if( program_queue.size() != 0 ) { next_prog_attributes = program_queue[0]; } 
                                       else { next_prog_attributes = get_prog_nop();   };
            
            if( !buffer.buffer_stall( bus_states, next_prog_attributes ) ) {
                
                -- set attributes --
                if( program_queue.size() != 0 and !mp_viol_stall ) { prog_attributes = program_queue.pop0(); } 
                                                              else { prog_attributes = get_prog_nop();       };
                
                if( bus_states.system.current.progdismiss == 0 ) { prog_attributes.dismiss_occur  = FALSE; };
                
                -- call event --
                emit request_program_attributes;
            };
            
            ----------------------------------
            -- generate data bus attributes --
            ----------------------------------
            -- check isr instruction or not --
            if( !data_queue.is_empty() and data_queue[0].isr_flag ) { 
                isr_flag = TRUE;
            } else { 
                isr_flag = FALSE;
            };
            
            -- update access attributes --
            if( !pipeline.pipeline_stall( bus_states ) ) {
                
                -- set attributes --
                if( (data_queue.size() != 0 and !data_attributes.idle_flag and !mp_viol_stall) or isr_flag ) { data_attributes = data_queue.pop0(); } 
                                                                                                        else { data_attributes = get_data_nop();    };
                
                -- call event --
                emit request_data_attributes;
            };
            
            ------------------------
            -- set activate buses --
            ------------------------
            -- set activate buses --
            if( !buffer.buffer_stall( bus_states, next_prog_attributes ) ) {
                prog_attributes.set_activate_buses( buffer, bus_states, config.pbus_start_address );
            };
            data_attributes.set_activate_buses();
            
            -- set mp attributes --
            if( !buffer.buffer_stall( bus_states, next_prog_attributes ) ) { set_mp_attributes( prog_attributes, data_attributes ); };
            
            -- for calling isr functions --
            if( config.interrupt_enable ) { call_isr_functions( prog_attributes ); };
            
            ----------------------------
            -- pipeline/buffer update --
            ----------------------------
            buffer.transition( next_prog_attributes, prog_attributes, bus_states );
            pipeline.transition( data_attributes, bus_states );
            
            ------------------------
            -- set data/prog data --
            ------------------------
            prog_attributes.set_driving_data( config, buffer,   bus_states, bus_driver );
            data_attributes.set_driving_data( config, pipeline, bus_states, bus_driver );
            
            ---------------------
            -- driving signals --
            ---------------------
            wait delay(config.driver_delay);
            bus_driver.drive_master_signals( paths, config, bus_states );
            
            -- wait --
            wait cycle;
            
            -- output current statue --
            if( config.logger_on ) {
                bus_states.output_bus_informations( pipeline, buffer );
            };
            
        };
    };
    
    ---------------------------------
    -- check for access completion --
    ---------------------------------
    TCM_access_complete() @rise_clock is {
        while( !(prog_access_complete and data_access_complete and
                 data_queue.size() == 0 and program_queue.size() == 0 and
                 bus_states.pbus.pending.counter == 0 and bus_states.bbus.pending.counter == 0 and
                 bus_states.cbus.pending.counter == 0 and bus_states.dbus.pending.counter == 0 and
                 bus_states.ebus.pending.counter == 0 and bus_states.fbus.pending.counter == 0     ) ) {
            wait cycle;
        };
        stop_run();
    };
    
    run() is also {
        start TCM_access_complete();
        start TCM_abortready_counter();
        start TCM_suspend_counter();
        start TCM_master_driver();
    };
    
};

'>
