//********************************************************************************
//
//  Filename: cpubus_master_structures.e
//
//  Description: structures for master driver
//
//  Major Revision Descriptions:
//  oishi          09/08/02         copied from UMA1.0 branch
//  oishi          09/09/02         removed bottom signals
//                                  removed mdp related functions
//                                  removed ack related functions
//                                  removed data dismiss functions except bypass stall function
//  oishi          09/11/02         added following qualifiers;
//                                    cpu_dblcoef_tr/cpu_rdbyte_tr/cpu_rdlsbe_tr
//  oishi          09/12/02         added MemoryProtection related signals
//                                  added cpu_avis_tr siganl
//  oishi          09/13/02         modified resetn to cpu_resetn and gl_resetn
//                                  And added cpu_reset driving function
//                                  And added branch function when buserr is asserted.
//  oishi          09/17/02         added lock stall function and
//                                  modified lock driving function
//  oishi          09/26/02         added iodready_stall()
//  oishi          09/27/02         modified information output function
//                                  added interrupt branch behaviors
//  oishi          09/30/02         supports simple interrupt behaviors
//  oishi          10/01/02         updated lock behavior;
//                                  . The lock signal is activated only during 
//                                    the corresponding read request is activate.
//                                  . for lock access, lock() qualifier is used,
//                                    so b/c/f bus are NOT driven when lock access.
//  oishi          10/02/02         removed check_lock_access()
//  oishi          10/04/02         supports cpu_suspend_nr signal
//  oishi          10/07/02         excepted dual access when emulation access
//  oishi          10/08/02         added suspend_stall() function
//  oishi          10/21/02         modified unknown driving for address buses
//  oishi          10/31/02         commented out unknown driving for mpc signals
//  oishi          11/01/02         added get_current_value_for_master() 
//                                  for performance enhancement
//  oishi          11/11/02         updated unknown driving for address bus
//  oishi          11/13/02         applied MikeP suggesion code for several functions
//  oishi          11/14/02         corrected address holding function
//  oishi          11/22/02         supported xdmadmpovr/xdmaempovr signal as IceMaker XDMA access
//                                  generated icemaker's ready as emiodready and emioeready
//                                  added write2 phase for pipeline
//  oishi          11/29/02         supported driving mp related signals
//  oishi          12/13/02         added isr index for PROGRAM_ACCESS_ATTRIBUTES
//                                  added perrack signal
//  oishi          01/21/03         added limitation that is avoid to generate byte qualifier
//                                  when dual access.(UMA_2X.BTS_UMA_CORE.50)
//  hiraoka        03/10/03         following Laijin R3.0 spec, write data on write byte access
//  hiraoka        03/25/03         add following signal
//                                    gl_homnsam_tr(3 DOWNTO 0)
//                                    cpu_homnsam_tr(3 DOWNTO 0)
//                                    gl_hompnsamp_init_tr
//                                    gl_homrnsamr_init_tr
//                                    gl_homxnsamx_init_tr
//                                    gl_homynsamy_init_tr
//  hiraoka        05/21/03         Pdismiss doesn't occur when P bus DT-DMA exist.
//                                  add condition detection on buffer_stall()
//  hiraoka        07/09/03         speed improvement, around drive_master_signals() of CPU_SIGNALS
//  hiraoka        07/22/03         drive qualifier on C req without D req at drive_master_signals
//  hiraoka        08/26/03         bug fix:E/F bus data drive on drive_master_signals()
//*********************************************************************************
<'
extend CPU_BUSES {
    
    -- for control idle instruction --
    !curr_idle_stall : bool;
    !prev_idle_stall : bool;
    
    -- for performance enhancement of get_current_value() --
    get_current_value_for_master( config : CPUBUS_CONFIGURATIONS, paths : CPU_SIGNAL_PATH, driver : CPU_SIGNALS ) is {
        var preq           : uint (bits: 01);
        var pabus          : uint (bits: 24);
        var pdbus          : uint (bits: 32);
        var pready         : uint (bits: 01);
        var pdismiss       : uint (bits: 01);
        var breq           : uint (bits: 01);
        var babus          : uint (bits: 23);
        var bdbus          : uint (bits: 32);
        var bready         : uint (bits: 01);
        var creq           : uint (bits: 01);
        var cabus          : uint (bits: 23);
        var cdbus          : uint (bits: 16);
        var cready         : uint (bits: 01);
        var dreq           : uint (bits: 01);
        var iodreq         : uint (bits: 01);
        var dabus          : uint (bits: 23);
        var ddbus          : uint (bits: 16);
        var dready         : uint (bits: 01);
        var ereq           : uint (bits: 01);
        var ioereq         : uint (bits: 01);
        var eabus          : uint (bits: 23);
        var edbus          : uint (bits: 16);
        var eready         : uint (bits: 01);
        var freq           : uint (bits: 01);
        var fabus          : uint (bits: 23);
        var fdbus          : uint (bits: 16);
        var fready         : uint (bits: 01);
        var wrbyte         : uint (bits: 01);
        var wrlsbe         : uint (bits: 01);
        var rdbyte         : uint (bits: 01);
        var rdlsbe         : uint (bits: 01);
        var doubler        : uint (bits: 01);
        var doublew        : uint (bits: 01);
        var dblcoef        : uint (bits: 01);
        var lock           : uint (bits: 01);
        var dmapr          : uint (bits: 01);
        var dmapw          : uint (bits: 01);
        var dbugd          : uint (bits: 01);
        var dbuge          : uint (bits: 01);
        var debugenable    : uint (bits: 01);
        var abortready     : uint (bits: 01);
        var suspend        : uint (bits: 01);
        
        var cachefreeze    : uint (bits: 01);
        var cacheenable    : uint (bits: 01);
        var cacheclr       : uint (bits: 01);
        var cacheendclr    : uint (bits: 01);
        
        var progdismiss    : uint (bits: 01);
        var gl_resetn      : uint (bits: 01);
        var cpu_resetn     : uint (bits: 01);
        var gl_mpnmc       : uint (bits: 01);
        var cpu_mpnmc      : uint (bits: 01);
        var gl_hompnsamp_init : uint (bits: 01);
        var gl_homrnsamr_init : uint (bits: 01);
        var gl_homxnsamx_init : uint (bits: 01);
        var gl_homynsamy_init : uint (bits: 01);
        var gl_homnsam     : uint (bits: 04);
        var cpu_homnsam    : uint (bits: 04);
        var avis           : uint (bits: 01);
        var idlemaster     : uint (bits: 01);
        var idlewakeup     : uint (bits:  1);
        var idlecpu        : uint (bits:  1);
        var idlecpud       : uint (bits:  1);
        var iack           : uint (bits:  1);
        var int            : uint (bits: 22);
        var nmi            : uint (bits:  1);
        var buserror       : uint (bits:  1);
        var debugerror     : uint (bits:  1);
        var perrack        : uint (bits:  1);
        
        var gl_privenable  : uint (bits:  1);
        var mpcxviol       : uint (bits:  1);
        var mpcbviol       : uint (bits:  1);
        var mpccviol       : uint (bits:  1);
        var mpcdviol       : uint (bits:  1);
        var mpceviol       : uint (bits:  1);
        var mpcfviol       : uint (bits:  1);
        var cpu_privenable : uint (bits:  1);
        var ppm            : uint (bits:  1);
        var bpm            : uint (bits:  1);
        var cpm            : uint (bits:  1);
        var dpm            : uint (bits:  1);
        var epm            : uint (bits:  1);
        var fpm            : uint (bits:  1);
        var rptpcvio       : uint (bits: 24);
        var rptintr        : uint (bits:  1);
        var rptpvio        : uint (bits:  1);
        var rptbvio        : uint (bits:  1);
        var rptcvio        : uint (bits:  1);
        var rptdvio        : uint (bits:  1);
        var rptiodvio      : uint (bits:  1);
        var rptevio        : uint (bits:  1);
        var rptioevio      : uint (bits:  1);
        var rptfvio        : uint (bits:  1);
        var rptpm          : uint (bits:  1);
        var dbgmpviol      : uint (bits:  1);
        var xdmadmpovr     : uint (bits:  1);
        var xdmaempovr     : uint (bits:  1);
        var emiodready     : uint (bits:  1);
        var emioeready     : uint (bits:  1);
        
        -----------------------------
        -- for getting top signals --
        -----------------------------
        preq           = driver.pbus.req;
        pabus          = driver.pbus.abus;
        pdismiss       = driver.pbus.dismiss;
        
        breq           = driver.bbus.req;
        babus          = driver.bbus.abus;
        
        creq           = driver.cbus.req;
        cabus          = driver.cbus.abus;
        
        dreq           = driver.dbus.req;
        iodreq         = driver.dbus.ioreq;
        dabus          = driver.dbus.abus;
        
        ereq           = driver.ebus.req;
        ioereq         = driver.ebus.ioreq;
        eabus          = driver.ebus.abus;
        edbus          = driver.ebus.dbus;
        
        freq           = driver.fbus.req;
        fabus          = driver.fbus.abus;
        fdbus          = driver.fbus.dbus;
        
        wrbyte         = driver.qualifiers.wrbyte;
        wrlsbe         = driver.qualifiers.wrlsbe;
        rdbyte         = driver.qualifiers.rdbyte;
        rdlsbe         = driver.qualifiers.rdlsbe;
        doubler        = driver.qualifiers.doubler;
        doublew        = driver.qualifiers.doublew;
        dblcoef        = driver.qualifiers.dblcoef;
        lock           = driver.qualifiers.lock;
        
        dmapr          = driver.emulations.dmapr;
        dmapw          = driver.emulations.dmapw;
        dbugd          = driver.emulations.dbugd;
        dbuge          = driver.emulations.dbuge;
        debugenable    = driver.emulations.debugenable;
        abortready     = driver.emulations.abortready;
        suspend        = driver.emulations.suspend;
        
        cachefreeze    = driver.cache.cachefreeze;
        cacheenable    = driver.cache.cacheenable;
        cacheclr       = driver.cache.cacheclr;
        
        cpu_resetn     = driver.systems.cpu_resetn;
        cpu_mpnmc      = driver.systems.cpu_mpnmc;
        cpu_homnsam    = driver.systems.cpu_homnsam;
        avis           = driver.systems.avis;
        idlemaster     = driver.systems.idlemaster;
        idlewakeup     = driver.systems.idlewakeup;
        iack           = driver.systems.iack;
        perrack        = driver.systems.perrack;
        
        cpu_privenable = driver.mp.cpu_privenable;
        ppm            = driver.mp.ppm;
        bpm            = driver.mp.bpm;
        cpm            = driver.mp.cpm;
        dpm            = driver.mp.dpm;
        epm            = driver.mp.epm;
        fpm            = driver.mp.fpm;
        rptpcvio       = driver.mp.rptpcvio;
        rptintr        = driver.mp.rptintr;
        rptpvio        = driver.mp.rptpvio;
        rptbvio        = driver.mp.rptbvio;
        rptcvio        = driver.mp.rptcvio;
        rptdvio        = driver.mp.rptdvio;
        rptiodvio      = driver.mp.rptiodvio;
        rptevio        = driver.mp.rptevio;
        rptioevio      = driver.mp.rptioevio;
        rptfvio        = driver.mp.rptfvio;
        rptpm          = driver.mp.rptpm;
        xdmadmpovr     = driver.mp.xdmadmpovr;
        xdmaempovr     = driver.mp.xdmaempovr;
        emiodready     = driver.mp.emiodready;
        emioeready     = driver.mp.emioeready;
        
        if( paths.pdbus.hdl_path()          != "" ) { pdbus          = paths.pdbus$          } else { pdbus          = 0 };
        if( paths.pready.hdl_path()         != "" ) { pready         = paths.pready$         } else { pready         = 1 };
        
        if( paths.bdbus.hdl_path()          != "" ) { bdbus          = paths.bdbus$          } else { bdbus          = 0 };
        if( paths.bready.hdl_path()         != "" ) { bready         = paths.bready$         } else { bready         = 1 };
        
        if( paths.cdbus.hdl_path()          != "" ) { cdbus          = paths.cdbus $         } else { cdbus          = 0 };
        if( paths.cready.hdl_path()         != "" ) { cready         = paths.cready$         } else { cready         = 1 };
        
        if( paths.ddbus.hdl_path()          != "" ) { ddbus          = paths.ddbus$          } else { ddbus          = 0 };
        if( paths.dready.hdl_path()         != "" ) { dready         = paths.dready$         } else { dready         = 1 };
        
        if( paths.eready.hdl_path()         != "" ) { eready         = paths.eready$         } else { eready         = 1 };
        
        if( paths.fready.hdl_path()         != "" ) { fready         = paths.fready$         } else { fready         = 1 };
        
        if( paths.cacheendclr.hdl_path()    != "" ) { cacheendclr    = paths.cacheendclr$    } else { cacheendclr    = 0 };
        
        if( paths.progdismiss.hdl_path()    != "" ) { progdismiss    = paths.progdismiss$    } else { progdismiss    = 1 };
        if( paths.gl_resetn.hdl_path()      != "" ) { gl_resetn      = paths.gl_resetn$      } else { gl_resetn      = 0 };
        if( paths.gl_mpnmc.hdl_path()       != "" ) { gl_mpnmc       = paths.gl_mpnmc$       } else { gl_mpnmc       = 1 };
        if( paths.gl_hompnsamp_init.hdl_path() != "" ) { gl_hompnsamp_init = paths.gl_hompnsamp_init$ } else { gl_hompnsamp_init = 1 };
        if( paths.gl_homrnsamr_init.hdl_path() != "" ) { gl_homrnsamr_init = paths.gl_homrnsamr_init$ } else { gl_homrnsamr_init = 1 };
        if( paths.gl_homxnsamx_init.hdl_path() != "" ) { gl_homxnsamx_init = paths.gl_homxnsamx_init$ } else { gl_homxnsamx_init = 1 };
        if( paths.gl_homynsamy_init.hdl_path() != "" ) { gl_homynsamy_init = paths.gl_homynsamy_init$ } else { gl_homynsamy_init = 1 };
        if( paths.gl_homnsam.hdl_path()     != "" ) { gl_homnsam     = paths.gl_homnsam$     } else { gl_homnsam     = 1 };
        if( paths.idlecpu.hdl_path()        != "" ) { idlecpu        = paths.idlecpu$        } else { idlecpu        = 0 };
        if( paths.idlecpud.hdl_path()       != "" ) { idlecpud       = paths.idlecpu$        } else { idlecpud       = 0 };
        if( paths.int.hdl_path()            != "" ) { int            = paths.int$            } else { int            = 0x3fffff };
        if( paths.nmi.hdl_path()            != "" ) { nmi            = paths.nmi$            } else { nmi            = 1 };
        if( paths.buserror.hdl_path()       != "" ) { buserror       = paths.buserror$       } else { buserror       = 1 };
        if( paths.debugerror.hdl_path()     != "" ) { debugerror     = paths.debugerror$     } else { debugerror     = 1 };
        
        if( paths.gl_privenable.hdl_path()  != "" ) { gl_privenable  = paths.gl_privenable$  } else { gl_privenable  = 0 };
        if( paths.mpcxviol.hdl_path()       != "" ) { mpcxviol       = paths.mpcxviol$       } else { mpcxviol       = 0 };
        if( paths.mpcbviol.hdl_path()       != "" ) { mpcbviol       = paths.mpcbviol$       } else { mpcbviol       = 0 };
        if( paths.mpccviol.hdl_path()       != "" ) { mpccviol       = paths.mpccviol$       } else { mpccviol       = 0 };
        if( paths.mpcdviol.hdl_path()       != "" ) { mpcdviol       = paths.mpcdviol$       } else { mpcdviol       = 0 };
        if( paths.mpceviol.hdl_path()       != "" ) { mpceviol       = paths.mpceviol$       } else { mpceviol       = 0 };
        if( paths.mpcfviol.hdl_path()       != "" ) { mpcfviol       = paths.mpcfviol$       } else { mpcfviol       = 0 };
        if( paths.dbgmpviol.hdl_path()      != "" ) { dbgmpviol      = paths.dbgmpviol$      } else { dbgmpviol      = 0 };
        
        ------------------------
        -- ready for icemaker --
        ------------------------
        -- ice_ready : icemaker's ready
        -- ready     : include icemaker ready and real ready signal
        
        if( dbus.pending.progress[dbus.pending.rdy_index] != NULL and
            dbus.pending.progress[dbus.pending.rdy_index].dbus_active == active and
            dbus.pending.progress[dbus.pending.rdy_index].rspace == io and
            dbus.pending.progress[dbus.pending.rdy_index].dabus is in ICEMAKER_IOMMR ) { 
            dbus.current.ice_ready = 0;
        } else {
            dbus.current.ice_ready = 1;
        };
        
        if( ebus.pending.progress[ebus.pending.rdy_index] != NULL and
            ebus.pending.progress[ebus.pending.rdy_index].ebus_active == active and
            ebus.pending.progress[ebus.pending.rdy_index].wspace == io and
            ebus.pending.progress[ebus.pending.rdy_index].eabus is in ICEMAKER_IOMMR ) { 
            ebus.current.ice_ready = 0;
        } else {
            ebus.current.ice_ready = 1;
        };
        
        dbus.current.raw_ready = dready;
        ebus.current.raw_ready = eready;
        
        ------------------------
        -- set current values --
        ------------------------
        pbus.current.req              = preq;
        pbus.current.abus             = pabus;
        pbus.current.dbus             = pdbus;
        pbus.current.ready            = pready;
        pbus.current.dismiss          = pdismiss;
        
        bbus.current.req              = breq;
        bbus.current.abus             = babus;
        bbus.current.dbus             = bdbus;
        bbus.current.ready            = bready;
        
        cbus.current.req              = creq;
        cbus.current.abus             = cabus;
        cbus.current.dbus             = cdbus;
        cbus.current.ready            = cready;
        
        dbus.current.req              = dreq;
        dbus.current.ioreq            = iodreq;
        dbus.current.abus             = dabus;
        dbus.current.dbus             = ddbus;
        dbus.current.ready            = dready & dbus.current.ice_ready;
        
        ebus.current.req              = ereq;
        ebus.current.ioreq            = ioereq;
        ebus.current.abus             = eabus;
        ebus.current.dbus             = edbus;
        ebus.current.ready            = eready & ebus.current.ice_ready;
        
        fbus.current.req              = freq;
        fbus.current.abus             = fabus;
        fbus.current.dbus             = fdbus;
        fbus.current.ready            = fready;
        
        qualifier.current.wrbyte      = wrbyte;
        qualifier.current.wrlsbe      = wrlsbe;
        qualifier.current.rdbyte      = rdbyte;
        qualifier.current.rdlsbe      = rdlsbe;
        qualifier.current.doubler     = doubler;
        qualifier.current.doublew     = doublew;
        qualifier.current.dblcoef     = dblcoef;
        qualifier.current.lock        = lock;
        
        emulation.current.dmapr       = dmapr;
        emulation.current.dmapw       = dmapw;
        emulation.current.dbugd       = dbugd;
        emulation.current.dbuge       = dbuge;
        emulation.current.debugenable = debugenable;
        emulation.current.abortready  = abortready;
        emulation.current.suspend     = suspend;
        
        cache.current.cachefreeze     = cachefreeze;
        cache.current.cacheenable     = cacheenable;
        cache.current.cacheclr        = cacheclr;
        cache.current.cacheendclr     = cacheendclr;
        
        system.current.progdismiss    = progdismiss;
        system.current.gl_resetn      = gl_resetn;
        system.current.cpu_resetn     = cpu_resetn & gl_resetn;
        system.current.gl_mpnmc       = gl_mpnmc;
        system.current.cpu_mpnmc      = cpu_mpnmc;
        system.current.gl_hompnsamp_init = gl_hompnsamp_init;
        system.current.gl_homrnsamr_init = gl_homrnsamr_init;
        system.current.gl_homxnsamx_init = gl_homxnsamx_init;
        system.current.gl_homynsamy_init = gl_homynsamy_init;
        system.current.gl_homnsam     = gl_homnsam;
        system.current.cpu_homnsam    = cpu_homnsam;
        system.current.avis           = avis;
        system.current.idlemaster     = idlemaster;
        system.current.idlewakeup     = idlewakeup;
        system.current.idlecpu        = idlecpu;
        system.current.idlecpud       = idlecpud;
        system.current.int            = int;
        system.current.nmi            = nmi;
        system.current.iack           = iack;
        system.current.buserror       = buserror;
        system.current.debugerror     = debugerror;
        system.current.perrack        = perrack;
        
        mp.current.gl_privenable      = gl_privenable;
        mp.current.mpcxviol           = mpcxviol;
        mp.current.mpcbviol           = mpcbviol;
        mp.current.mpccviol           = mpccviol;
        mp.current.mpcdviol           = mpcdviol;
        mp.current.mpceviol           = mpceviol;
        mp.current.mpcfviol           = mpcfviol;
        mp.current.cpu_privenable     = cpu_privenable;
        mp.current.ppm                = ppm;
        mp.current.bpm                = bpm;
        mp.current.cpm                = cpm;
        mp.current.dpm                = dpm;
        mp.current.epm                = epm;
        mp.current.fpm                = fpm;
        mp.current.rptpcvio           = rptpcvio;
        mp.current.rptintr            = rptintr;
        mp.current.rptpvio            = rptpvio;
        mp.current.rptbvio            = rptbvio;
        mp.current.rptcvio            = rptcvio;
        mp.current.rptdvio            = rptdvio;
        mp.current.rptiodvio          = rptiodvio;
        mp.current.rptevio            = rptevio;
        mp.current.rptioevio          = rptioevio;
        mp.current.rptfvio            = rptfvio;
        mp.current.rptpm              = rptpm;
        mp.current.dbgmpviol          = dbgmpviol;
        mp.current.xdmadmpovr         = xdmadmpovr;
        mp.current.xdmaempovr         = xdmaempovr;
        mp.current.emiodready         = emiodready;
        mp.current.emioeready         = emioeready;
        
    };
    
};

--------------------------
-- update signal struct --
--------------------------
extend CPU_PBUS_SIGNALS {
    
    !valid_abus : bool;
    
    set_init_values() is also {
        valid_abus = FALSE;
    };
    
    unknown_drive( paths : CPU_SIGNAL_PATH ) is {
        if( paths.pabus.hdl_path()   != "" and !valid_abus ) { paths.pabus$ = 24'h000000; }; // 24'hxxxxxx
    };
    
    drive_master_signals( paths : CPU_SIGNAL_PATH ) is {
        if( paths.preq.hdl_path()       != "" ) { paths.preq$       = req;     };
        if( paths.pabus.hdl_path()      != "" ) { paths.pabus$      = abus;    };
        if( paths.pdismiss.hdl_path()   != "" ) { paths.pdismiss$   = dismiss; };
    };
};

extend CPU_BBUS_SIGNALS {
    
    !valid_abus    : bool;
    
    set_init_values() is also {
        valid_abus    = FALSE;
    };
    
    unknown_drive( paths : CPU_SIGNAL_PATH ) is {
        if( paths.babus.hdl_path()      != "" and !valid_abus    ) { paths.babus$      = 23'h000000; }; // 23'hxxxxxx
    };
    
    drive_master_signals( paths : CPU_SIGNAL_PATH ) is {
        if( paths.breq.hdl_path()       != "" ) { paths.breq$       = req;     };
        if( paths.babus.hdl_path()      != "" ) { paths.babus$      = abus;    };
    };
    
};

extend CPU_CBUS_SIGNALS {
    
    !valid_abus    : bool;
    
    set_init_values() is also {
        valid_abus    = FALSE;
    };
    
    unknown_drive( paths : CPU_SIGNAL_PATH ) is {
        if( paths.cabus.hdl_path()      != "" and !valid_abus    ) { paths.cabus$      = 23'h000000; }; // 23'hxxxxxx
    };
    
    drive_master_signals( paths : CPU_SIGNAL_PATH ) is {
        if( paths.creq.hdl_path()       != "" ) { paths.creq$     = req;     };
        if( paths.cabus.hdl_path()      != "" ) { paths.cabus$    = abus;    };
    };
    
};

extend CPU_DBUS_SIGNALS {
    
    !valid_abus    : bool;
    
    set_init_values() is also {
        valid_abus    = FALSE;
    };
    
    unknown_drive( paths : CPU_SIGNAL_PATH ) is {
        if( paths.dabus.hdl_path()      != "" and !valid_abus    ) { paths.dabus$      = 23'h000000; }; // 23'hxxxxxx
    };
    
    drive_master_signals( paths : CPU_SIGNAL_PATH ) is {
        if( paths.iodreq.hdl_path()     != "" ) { paths.iodreq$     = ioreq;   };
        if( paths.dreq.hdl_path()       != "" ) { paths.dreq$       = req;     };
        if( paths.dabus.hdl_path()      != "" ) { paths.dabus$      = abus;    };
    };
    
};

extend CPU_EBUS_SIGNALS {
    
    !valid_abus    : bool;
    !valid_dbus    : bool;
    
    set_init_values() is also {
        valid_abus    = FALSE;
        valid_dbus    = FALSE;
    };
    
    unknown_drive( paths : CPU_SIGNAL_PATH ) is {
        if( paths.eabus.hdl_path()      != "" and !valid_abus    ) { paths.eabus$      = 23'h000000; }; // 23'hxxxxxx
        if( paths.edbus.hdl_path()      != "" and !valid_dbus    ) { paths.edbus$      = 16'h0000;   }; // 16'hxxxx
    };
    
    drive_master_signals( paths : CPU_SIGNAL_PATH ) is {
        if( paths.ioereq.hdl_path()     != "" ) { paths.ioereq$     = ioreq;   };
        if( paths.ereq.hdl_path()       != "" ) { paths.ereq$       = req;     };
        if( paths.eabus.hdl_path()      != "" ) { paths.eabus$      = abus;    };
        if( paths.edbus.hdl_path()      != "" ) { paths.edbus$      = dbus;    };
    };
    
};

extend CPU_FBUS_SIGNALS {
    
    !valid_abus    : bool;
    !valid_dbus    : bool;
    
    set_init_values() is also {
        valid_abus    = FALSE;
        valid_dbus    = FALSE;
    };
    
    unknown_drive( paths : CPU_SIGNAL_PATH ) is {
        if( paths.fabus.hdl_path()      != "" and !valid_abus    ) { paths.fabus$      = 23'h000000; }; // 23'hxxxxxx
        if( paths.fdbus.hdl_path()      != "" and !valid_dbus    ) { paths.fdbus$      = 16'h0000;   }; // 16'hxxxx
    };
    
    drive_master_signals( paths : CPU_SIGNAL_PATH ) is {
        if( paths.freq.hdl_path()       != "" ) { paths.freq$       = req;     };
        if( paths.fabus.hdl_path()      != "" ) { paths.fabus$      = abus;    };
        if( paths.fdbus.hdl_path()      != "" ) { paths.fdbus$      = dbus;    };
    };
    
};

extend CPU_QUALIFIERS_SIGNALS {
    !valid_wrbyte  : bool;
    !valid_wrlsbe  : bool;
    !valid_rdbyte  : bool;
    !valid_rdlsbe  : bool;
    !valid_doubler : bool;
    !valid_doublew : bool;
    !valid_dblcoef : bool;
    
    set_init_values() is also {
        valid_wrbyte  = FALSE;
        valid_wrlsbe  = FALSE;
        valid_rdbyte  = FALSE;
        valid_rdlsbe  = FALSE;
        valid_doubler = FALSE;
        valid_doublew = FALSE;
        valid_dblcoef = FALSE;
    };
    
    unknown_drive( paths : CPU_SIGNAL_PATH ) is {
        if( paths.wrbyte.hdl_path()    != "" and !valid_wrbyte  ) { paths.wrbyte$    = 1'b0; }; // 1'bx
        if( paths.wrlsbe.hdl_path()    != "" and !valid_wrlsbe  ) { paths.wrlsbe$    = 1'b0; }; // 1'bx
        if( paths.rdbyte.hdl_path()    != "" and !valid_rdbyte  ) { paths.rdbyte$    = 1'b0; }; // 1'bx
        if( paths.rdlsbe.hdl_path()    != "" and !valid_rdlsbe  ) { paths.rdlsbe$    = 1'b0; }; // 1'bx
        if( paths.doubler.hdl_path()   != "" and !valid_doubler ) { paths.doubler$   = 1'b0; }; // 1'bx
        if( paths.doublew.hdl_path()   != "" and !valid_doublew ) { paths.doublew$   = 1'b0; }; // 1'bx
        if( paths.dblcoef.hdl_path()   != "" and !valid_dblcoef ) { paths.dblcoef$   = 1'b0; }; // 1'bx
    };
    
    drive_master_signals( paths : CPU_SIGNAL_PATH ) is {
        if( paths.wrbyte.hdl_path()    != "" ) { paths.wrbyte$    = wrbyte;  };
        if( paths.wrlsbe.hdl_path()    != "" ) { paths.wrlsbe$    = wrlsbe;  };
        if( paths.rdbyte.hdl_path()    != "" ) { paths.rdbyte$    = rdbyte;  };
        if( paths.rdlsbe.hdl_path()    != "" ) { paths.rdlsbe$    = rdlsbe;  };
        if( paths.doubler.hdl_path()   != "" ) { paths.doubler$   = doubler; };
        if( paths.doublew.hdl_path()   != "" ) { paths.doublew$   = doublew; };
        if( paths.dblcoef.hdl_path()   != "" ) { paths.dblcoef$   = dblcoef; };
        if( paths.lock.hdl_path()      != "" ) { paths.lock$      = lock;    };
    };
    
};

extend CPU_EMULATION_SIGNALS {
    !valid_dmapr : bool;
    !valid_dmapw : bool;
    !valid_dbugd : bool;
    !valid_dbuge : bool;
    
    set_init_values() is also {
        valid_dmapr = FALSE;
        valid_dmapw = FALSE;
        valid_dbugd = FALSE;
        valid_dbuge = FALSE;
    };
    
    unknown_drive( paths : CPU_SIGNAL_PATH ) is {
        if( paths.dmapr.hdl_path()   != "" and !valid_dmapr ) { paths.dmapr$   = 1'b0; }; // 1'bx
        if( paths.dmapw.hdl_path()   != "" and !valid_dmapw ) { paths.dmapw$   = 1'b0; }; // 1'bx
        if( paths.dbugd.hdl_path()   != "" and !valid_dbugd ) { paths.dbugd$   = 1'b0; }; // 1'bx
        if( paths.dbuge.hdl_path()   != "" and !valid_dbuge ) { paths.dbuge$   = 1'b0; }; // 1'bx
    };
    
    drive_master_signals( paths : CPU_SIGNAL_PATH ) is {
        if( paths.dmapr.hdl_path()       != "" ) { paths.dmapr$       = dmapr; };
        if( paths.dmapw.hdl_path()       != "" ) { paths.dmapw$       = dmapw; };
        if( paths.dbugd.hdl_path()       != "" ) { paths.dbugd$       = dbugd; };
        if( paths.dbuge.hdl_path()       != "" ) { paths.dbuge$       = dbuge; };
        if( paths.debugenable.hdl_path() != "" ) { paths.debugenable$ = debugenable; };
        if( paths.abortready.hdl_path()  != "" ) { paths.abortready$  = abortready;  };
        if( paths.suspend.hdl_path()     != "" ) { paths.suspend$     = suspend;     };
    };
    
};

extend CPU_MP_SIGNALS {
    !valid_ppm        : bool;
    !valid_bpm        : bool;
    !valid_cpm        : bool;
    !valid_dpm        : bool;
    !valid_epm        : bool;
    !valid_fpm        : bool;
    !valid_xdmadmpovr : bool;
    !valid_xdmaempovr : bool;
    
    set_init_values() is also {
        valid_ppm        = FALSE;
        valid_bpm        = FALSE;
        valid_cpm        = FALSE;
        valid_dpm        = FALSE;
        valid_epm        = FALSE;
        valid_fpm        = FALSE;
        valid_xdmadmpovr = FALSE;
        valid_xdmaempovr = FALSE;
    };
    
    unknown_drive( paths : CPU_SIGNAL_PATH ) is {
        if( paths.ppm.hdl_path()        != "" and !valid_ppm        ) { paths.ppm$          = 1'b0; }; // 1'bx
        if( paths.bpm.hdl_path()        != "" and !valid_bpm        ) { paths.bpm$          = 1'b0; }; // 1'bx
        if( paths.cpm.hdl_path()        != "" and !valid_cpm        ) { paths.cpm$          = 1'b0; }; // 1'bx
        if( paths.dpm.hdl_path()        != "" and !valid_dpm        ) { paths.dpm$          = 1'b0; }; // 1'bx
        if( paths.epm.hdl_path()        != "" and !valid_epm        ) { paths.epm$          = 1'b0; }; // 1'bx
        if( paths.fpm.hdl_path()        != "" and !valid_fpm        ) { paths.fpm$          = 1'b0; }; // 1'bx
        if( paths.xdmadmpovr.hdl_path() != "" and !valid_xdmadmpovr ) { paths.xdmadmpovr$   = 1'b0; }; // 1'bx
        if( paths.xdmaempovr.hdl_path() != "" and !valid_xdmaempovr ) { paths.xdmaempovr$   = 1'b0; }; // 1'bx
    };
    
    drive_master_signals( paths : CPU_SIGNAL_PATH ) is {
        if( paths.cpu_privenable.hdl_path() != "" ) { paths.cpu_privenable$ = cpu_privenable; };
        if( paths.ppm.hdl_path()            != "" ) { paths.ppm$            = ppm;            };
        if( paths.bpm.hdl_path()            != "" ) { paths.bpm$            = bpm;            };
        if( paths.cpm.hdl_path()            != "" ) { paths.cpm$            = cpm;            };
        if( paths.dpm.hdl_path()            != "" ) { paths.dpm$            = dpm;            };
        if( paths.epm.hdl_path()            != "" ) { paths.epm$            = epm;            };
        if( paths.fpm.hdl_path()            != "" ) { paths.fpm$            = fpm;            };
        if( paths.rptpcvio.hdl_path()       != "" ) { paths.rptpcvio$       = rptpcvio;       };
        if( paths.rptintr.hdl_path()        != "" ) { paths.rptintr$        = rptintr;        };
        if( paths.rptpvio.hdl_path()        != "" ) { paths.rptpvio$        = rptpvio;        };
        if( paths.rptbvio.hdl_path()        != "" ) { paths.rptbvio$        = rptbvio;        };
        if( paths.rptcvio.hdl_path()        != "" ) { paths.rptcvio$        = rptcvio;        };
        if( paths.rptdvio.hdl_path()        != "" ) { paths.rptdvio$        = rptdvio;        };
        if( paths.rptiodvio.hdl_path()      != "" ) { paths.rptiodvio$      = rptiodvio;      };
        if( paths.rptevio.hdl_path()        != "" ) { paths.rptevio$        = rptevio;        };
        if( paths.rptioevio.hdl_path()      != "" ) { paths.rptioevio$      = rptioevio;      };
        if( paths.rptfvio.hdl_path()        != "" ) { paths.rptfvio$        = rptfvio;        };
        if( paths.rptpm.hdl_path()          != "" ) { paths.rptpm$          = rptpm;          };
        if( paths.xdmadmpovr.hdl_path()     != "" ) { paths.xdmadmpovr$     = xdmadmpovr;     };
        if( paths.xdmaempovr.hdl_path()     != "" ) { paths.xdmaempovr$     = xdmaempovr;     };
        if( paths.emiodready.hdl_path()     != "" ) { paths.emiodready$     = emiodready;     };
        if( paths.emioeready.hdl_path()     != "" ) { paths.emioeready$     = emioeready;     };
    };
    
};

extend CPU_SYSTEM_SIGNALS {
    
    drive_master_signals( paths : CPU_SIGNAL_PATH ) is {
        if( paths.cpu_resetn.hdl_path() != "" ) { paths.cpu_resetn$ = cpu_resetn; };
        if( paths.idlemaster.hdl_path() != "" ) { paths.idlemaster$ = idlemaster; };
        if( paths.cpu_mpnmc.hdl_path()  != "" ) { paths.cpu_mpnmc$  = cpu_mpnmc;  };
        if( paths.cpu_homnsam.hdl_path()!= "" ) { paths.cpu_homnsam$= cpu_homnsam;};
        if( paths.avis.hdl_path()       != "" ) { paths.avis$       = avis;       };
        if( paths.iack.hdl_path()       != "" ) { paths.iack$       = iack;       };
        if( paths.perrack.hdl_path()    != "" ) { paths.perrack$    = perrack;    };
      --if( paths.idlewakeup.hdl_path() != "" ) { paths.idlewakeup$ = idlewakeup; };
    };
    
};

extend CPU_CACHE_SIGNALS {
    
    drive_master_signals( paths : CPU_SIGNAL_PATH ) is {
        if( paths.cachefreeze.hdl_path() != "" ) { paths.cachefreeze$ = cachefreeze; };
        if( paths.cacheenable.hdl_path() != "" ) { paths.cacheenable$ = cacheenable; };
        if( paths.cacheclr.hdl_path()    != "" ) { paths.cacheclr$    = cacheclr;    };
    };
    
};

extend CPU_SIGNALS {
    
    drive_master_signals( paths : CPU_SIGNAL_PATH, config : CPUBUS_CONFIGURATIONS, bus_states : CPU_BUSES ) is {
        
        if( !config.disable_drive ) {
            if ( (bus_states.system.current.cpu_resetn     == 0   ) or
                 (bus_states.pbus.current.req     != pbus.req     ) or
                 (bus_states.pbus.current.abus    != pbus.abus    ) or
                 (bus_states.pbus.current.ready   != pbus.ready   ) or
                 (bus_states.pbus.current.dismiss != pbus.dismiss )     )  { pbus.drive_master_signals( paths ); };
            if ( (bus_states.system.current.cpu_resetn     == 0   ) or
                 (bus_states.bbus.current.req     != bbus.req     ) or
                 (bus_states.bbus.current.abus    != bbus.abus    ) or
                 (bus_states.bbus.current.ready   != bbus.ready   )     ) { bbus.drive_master_signals( paths ); };
            if ( (bus_states.system.current.cpu_resetn     == 0   ) or
                 (bus_states.cbus.current.req     != cbus.req     ) or
                 (bus_states.cbus.current.abus    != cbus.abus    ) or
                 (bus_states.cbus.current.ready   != cbus.ready   )     ) { cbus.drive_master_signals( paths ); };
            if ( (bus_states.system.current.cpu_resetn     == 0   ) or
                 (bus_states.dbus.current.req     != dbus.req     ) or
                 (bus_states.dbus.current.ioreq   != dbus.ioreq   ) or
                 (bus_states.dbus.current.abus    != dbus.abus    ) or
                 (bus_states.dbus.current.ready   != dbus.ready   )     ) { dbus.drive_master_signals( paths ); };
            if ( (bus_states.system.current.cpu_resetn     == 0   ) or
                 (bus_states.ebus.pending.counter != 0            ) or // when two pending req of same addr/data..
                 (bus_states.ebus.current.req     != ebus.req     ) or
                 (bus_states.ebus.current.ioreq   != ebus.ioreq   ) or
                 (bus_states.ebus.current.abus    != ebus.abus    ) or
                 (bus_states.ebus.current.ready   != ebus.ready   ) or
                 (bus_states.ebus.current.dbus    != ebus.dbus    )     ) { ebus.drive_master_signals( paths ); };
            if ( (bus_states.system.current.cpu_resetn     == 0   ) or
                 (bus_states.fbus.pending.counter != 0            ) or // when two pending req of same addr/data..
                 (bus_states.fbus.current.req     != fbus.req     ) or
                 (bus_states.fbus.current.abus    != fbus.abus    ) or
                 (bus_states.fbus.current.ready   != fbus.ready   ) or
                 (bus_states.fbus.current.dbus    != fbus.dbus    )     ) { fbus.drive_master_signals( paths ); };
            if ( (bus_states.system.current.cpu_resetn     == 0   ) or
                 (bus_states.qualifier.current.wrbyte  != qualifiers.wrbyte)  or
                 (bus_states.qualifier.current.wrlsbe  != qualifiers.wrlsbe)  or
                 (bus_states.qualifier.current.rdbyte  != qualifiers.rdbyte)  or
                 (bus_states.qualifier.current.rdlsbe  != qualifiers.rdlsbe)  or
                 (bus_states.qualifier.current.doubler != qualifiers.doubler) or
                 (bus_states.qualifier.current.doublew != qualifiers.doublew) or
                 (bus_states.qualifier.current.dblcoef != qualifiers.dblcoef) or
                 (bus_states.qualifier.current.lock    != qualifiers.lock)    or
                 (bus_states.bbus.current.req          != bbus.req     ) or -- for dblcoef
                 (bus_states.cbus.current.req          != cbus.req     ) or -- Only Creq access
                 (bus_states.dbus.current.req          != dbus.req     ) or -- for rdbyte, rdlsbe, doubler, lock
                 (bus_states.dbus.current.ioreq        != dbus.ioreq   ) or
                 (bus_states.ebus.current.req          != ebus.req     ) or -- for wrbyte, wrlsbe, doublew
                 (bus_states.ebus.current.ioreq        != ebus.ioreq   ) ) { qualifiers.drive_master_signals( paths ); };
            if ( (bus_states.system.current.cpu_resetn     == 0   ) or
                 (bus_states.emulation.current.dmapr       != emulations.dmapr)       or
                 (bus_states.emulation.current.dmapw       != emulations.dmapw)       or
                 (bus_states.emulation.current.dbugd       != emulations.dbugd)       or
                 (bus_states.emulation.current.dbuge       != emulations.dbuge)       or
                 (bus_states.emulation.current.debugenable != emulations.debugenable) or
                 (bus_states.emulation.current.abortready  != emulations.abortready)  or
                 (bus_states.emulation.current.suspend     != emulations.suspend)     or
                 (bus_states.pbus.current.req              != pbus.req     ) or -- for dmapr
                 (bus_states.dbus.current.req              != dbus.req     ) or -- dbugd
                 (bus_states.dbus.current.ioreq            != dbus.ioreq   ) or 
                 (bus_states.ebus.current.req              != ebus.req     ) or -- dbuge, dmapw
                 (bus_states.ebus.current.ioreq            != ebus.ioreq   ) ) { emulations.drive_master_signals( paths ); };
            if ( (bus_states.system.current.cpu_resetn     == 0   ) or
                 (bus_states.cache.current.cachefreeze     != cache.cachefreeze) or
                 (bus_states.cache.current.cacheenable     != cache.cacheenable) or
                 (bus_states.cache.current.cacheclr        != cache.cacheclr)    or
                 (bus_states.cache.current.cacheendclr     != cache.cacheendclr) ) { cache.drive_master_signals( paths ); };
            if ( (bus_states.system.current.cpu_resetn     == 0   ) or
                 (bus_states.mp.current.emiodready         != mp.emiodready) or -- need to driver emiodready because protocol checker uses this signal
                 (bus_states.mp.current.emioeready         != mp.emioeready) or -- need to driver emiodready because protocol checker uses this signal
                  ( (bus_states.mp.current.gl_privenable   == 1 ) and 
                    ( (bus_states.mp.current.cpu_privenable     != mp.cpu_privenable)  or
                      (bus_states.mp.current.ppm                != mp.ppm)       or
                      (bus_states.mp.current.bpm                != mp.bpm)       or
                      (bus_states.mp.current.cpm                != mp.cpm)       or
                      (bus_states.mp.current.dpm                != mp.dpm)       or
                      (bus_states.mp.current.epm                != mp.epm)       or
                      (bus_states.mp.current.fpm                != mp.fpm)       or
                      (bus_states.mp.current.rptpcvio           != mp.rptpcvio)  or
                      (bus_states.mp.current.rptintr            != mp.rptintr)   or
                      (bus_states.mp.current.rptpvio            != mp.rptpvio)   or
                      (bus_states.mp.current.rptbvio            != mp.rptbvio)   or
                      (bus_states.mp.current.rptcvio            != mp.rptcvio)   or
                      (bus_states.mp.current.rptdvio            != mp.rptdvio)   or
                      (bus_states.mp.current.rptiodvio          != mp.rptiodvio) or
                      (bus_states.mp.current.rptevio            != mp.rptevio)   or
                      (bus_states.mp.current.rptioevio          != mp.rptioevio) or
                      (bus_states.mp.current.rptfvio            != mp.rptfvio)   or
                      (bus_states.mp.current.rptpm              != mp.rptpm)     or
                      (bus_states.mp.current.xdmadmpovr         != mp.xdmadmpovr)or
                      (bus_states.mp.current.xdmaempovr         != mp.xdmaempovr)))     ) { mp.drive_master_signals( paths ); };
            if ( (bus_states.system.current.cpu_resetn     != systems.cpu_resetn)  or
                 (bus_states.system.current.idlemaster     != systems.idlemaster)  or
                 (bus_states.system.current.idlewakeup     != systems.idlewakeup)  or
                 (bus_states.system.current.cpu_mpnmc      != systems.cpu_mpnmc)   or
                 (bus_states.system.current.avis           != systems.avis)        or
                 (bus_states.system.current.cpu_homnsam    != systems.cpu_homnsam) or
                 (bus_states.system.current.iack           != systems.iack)        or
                 (bus_states.system.current.perrack        != systems.perrack) ) { systems.drive_master_signals( paths ); };
            
            if( config.unknown_drive ) {
                pbus.unknown_drive( paths );
                bbus.unknown_drive( paths );
                cbus.unknown_drive( paths );
                dbus.unknown_drive( paths );
                ebus.unknown_drive( paths );
                fbus.unknown_drive( paths );
                qualifiers.unknown_drive( paths );
                emulations.unknown_drive( paths );
                if (bus_states.mp.current.gl_privenable   == 1 ) { mp.unknown_drive( paths );};
            };
            
        };
        
    };
    
};

-----------------------------
-- for updating attributes --
-----------------------------
extend DATA_ACCESS_ATTRIBUTES {
    
    !bbus_mmrflag : bool;
    !cbus_mmrflag : bool;
    !dbus_mmrflag : bool;
    !ebus_mmrflag : bool;
    !fbus_mmrflag : bool;
    
    -- for mp control --
    mp_supervisor       : bool;           keep soft mp_supervisor       == TRUE;
    mp_viol_pbus        : bool;           keep soft mp_viol_pbus        == FALSE;
    mp_viol_contextsave : bool;           keep soft mp_viol_contextsave == FALSE;
    mp_viol_address     : uint(bits: 24); keep soft mp_viol_address     == 0;
    ---------------------
    
    idle_flag : bool;
    keep soft idle_flag == FALSE;
    
    lock_incomplete : bool;
    keep soft lock_incomplete == FALSE;
    
    isr_flag : bool;
    keep soft isr_flag == FALSE;
    
    when disable'lock DATA_ACCESS_ATTRIBUTES {
        
        instr_coef_group  : data_access_instruction;
        keep instr_coef_group  in [ None, CoefRead ];
        
        instr_read_group  : data_access_instruction;
        keep instr_read_group  in [ None, SingleRead, DualRead, DoubleRead ];
        
        instr_write_group : data_access_instruction;
        keep instr_write_group in [ None, SingleWrite, DualWrite, DoubleWrite ];
        
        Select_set_activate_buses() is {
            
            -- for idle control --
            if( idle_flag == TRUE ) {
                instr_coef_group  = None;
                instr_read_group  = None;
                instr_write_group = None;
            };
            
            -- set activate buses
            case instr_coef_group {
                [ None     ] : { bbus_active = no_active; };
                [ CoefRead ] : { bbus_active = active;    };
            };
            case instr_read_group {
                [ None       ] : { cbus_active = no_active; dbus_active = no_active; };
                [ SingleRead ] : { cbus_active = no_active; dbus_active =    active; };
                [ DualRead   ] : { cbus_active =    active; dbus_active =    active; };
                [ DoubleRead ] : { cbus_active =    active; dbus_active =    active; };
            };
            case instr_write_group {
                [ None        ] : { ebus_active = no_active; fbus_active = no_active; };
                [ SingleWrite ] : { ebus_active =    active; fbus_active = no_active; };
                [ DualWrite   ] : { ebus_active =    active; fbus_active =    active; };
                [ DoubleWrite ] : { ebus_active =    active; fbus_active =    active; };
            };
            
            -- byte access is not allowed when dual access
            if( instr_read_group  == DualRead  ) { read_width  = word; };
            if( instr_write_group == DualWrite ) { write_width = word; };
            
            -- emulation access is not allowed for dual access 
            if( dbus_type != no_debug and instr_coef_group  == CoefRead  ) { dbus_type = no_debug };
            if( dbus_type != no_debug and instr_read_group  == DualRead  ) { dbus_type = no_debug };
            if( ebus_type != no_debug and instr_write_group == DualWrite ) { ebus_type = no_debug };
            
            -- emulation program write is not allowed for io space
            if( space == io and ebus_active == active and ebus_type == debug_prog_write ) { ebus_type = debug_data_write; };
            
            -- limitations for io space access
            if( !(instr_read_group  == SingleRead and instr_write_group == None       ) and 
                !(instr_read_group  == None       and instr_write_group == SingleWrite)     ){ space = memory; };
            
            -- when double access, io space access and byte access can't allow
            if( instr_read_group  == DoubleRead  ) { doubler = double; space = memory; read_width  = word; } else { doubler = word; };
            if( instr_write_group == DoubleWrite ) { doublew = double; space = memory; write_width = word; } else { doublew = word; };
            
            -- when io space access, mdp address should be set '0'
            if( dbus_active == active and space == io ) { dabus[22:16] = 0; };
            if( ebus_active == active and space == io ) { eabus[22:16] = 0; };
            
            -- address set when double access
            if( instr_read_group == DoubleRead ) {
                cabus    = %{dabus[22:1], (~dabus)[0:0]};
            };
            if( instr_write_group == DoubleWrite ) {
                fabus    = %{eabus[22:1], (~eabus)[0:0]};
            };

            -- byte data manage --
            if( write_width == high_byte or write_width == low_byte ) {
                edbus[15:08] = edbus[07:00];
            };
            
            -- mmr access check --
            if( babus < 0x60 and space == memory ) {
                bbus_mmrflag = TRUE;
                bbus_active  = no_active;
            } else {
                bbus_mmrflag = FALSE;
            };
            if( cabus < 0x60 and space == memory ) {
                cbus_mmrflag = TRUE;
                cbus_active  = no_active;
            } else {
                cbus_mmrflag = FALSE;
            };
            if( dabus < 0x60 and space == memory ) {
                dbus_mmrflag = TRUE;
                dbus_active  = no_active;
            } else {
                dbus_mmrflag = FALSE;
            };
            if( eabus < 0x60 and space == memory ) {
                ebus_mmrflag = TRUE;
                ebus_active  = no_active;
            } else {
                ebus_mmrflag = FALSE;
            };
            if( fabus < 0x60 and space == memory ) {
                fbus_mmrflag = TRUE;
                fbus_active  = no_active;
            } else {
                fbus_mmrflag = FALSE;
            };
            
            -- set rspace/wspace flags --
            rspace = space;
            wspace = space;
            
            -- this is not lock access --
            lock_incomplete = FALSE;
            
            -- when e/fbus access to same address, ebus does not activate --
            if( ebus_active == active and fbus_active == active and eabus == fabus ) {
                ebus_active = no_active;
            };
            
        };
        
        Select_set_instruction( instruction : data_access_instruction ) is {
            case instruction {
                [ None, CoefRead ]                            : { instr_coef_group  = instruction; };
                [ None, SingleRead,  DualRead,  DoubleRead  ] : { instr_read_group  = instruction; };
                [ None, SingleWrite, DualWrite, DoubleWrite ] : { instr_write_group = instruction; };
                default                                       : { outf("Warning : can't set instruction as %s \n",instruction); };
            };
        };
        
        Select_set_noaccess_instruction() is {
            instr_coef_group  = None;
            instr_read_group  = None;
            instr_write_group = None;
            idle_flag         = FALSE;
        };
        Select_get_read_instruction() : data_access_instruction is {
            result = instr_read_group;
        };
        Select_get_write_instruction() : data_access_instruction is {
            result = instr_write_group;
        };
        Select_get_lock_instruction() : data_access_instruction is {
            result = None;
        };
        Select_get_coef_instruction() : data_access_instruction is {
            result = instr_coef_group;
        };
        
    };
    when enable'lock DATA_ACCESS_ATTRIBUTES {
        
        instr_lock_group  : data_access_instruction;
        keep instr_lock_group in [ None, LockAccess ];
        
        Select_set_activate_buses() is {
            
            -- for idle control --
            if( idle_flag == TRUE ) {
                instr_lock_group  = None;
            };
            
            -- set activate buses
            
            if( !lock_incomplete ) {
                case instr_lock_group {
                    [ None                   ] : { bbus_active = no_active; cbus_active = no_active; dbus_active = no_active; ebus_active = no_active; fbus_active = no_active; };
                    [ LockAccess             ] : { bbus_active = no_active; cbus_active = no_active; dbus_active =    active; ebus_active =    active; fbus_active = no_active; };
                };
            } else {
                case instr_lock_group {
                    [ None                   ] : { bbus_active = no_active; cbus_active = no_active; dbus_active = no_active; ebus_active = no_active; fbus_active = no_active; };
                    [ LockAccess             ] : { bbus_active = no_active; cbus_active = no_active; dbus_active =    active; ebus_active = no_active; fbus_active = no_active; };
                };
            };
            doubler     = word;
            doublew     = word;
            read_width  = word;
            write_width = word;
            space       = memory;
            
            -- emulation access is not allowed for lock access
            dbus_type = no_debug;
            ebus_type = no_debug;
            
            -- when io space access, mdp address should be set '0'
            if( dbus_active == active and space == io ) { dabus[22:16] = 0; };
            if( ebus_active == active and space == io ) { eabus[22:16] = 0; };
            
            -- same address set for lock read/write --
            if( instr_lock_group is in [ LockAccess ] ) {
                eabus = dabus;
            };
            
            -- mmr access check --
            if( babus < 0x60 and space == memory ) {
                bbus_mmrflag = TRUE;
                bbus_active  = no_active;
            } else {
                bbus_mmrflag = FALSE;
            };
            if( cabus < 0x60 and space == memory ) {
                cbus_mmrflag = TRUE;
                cbus_active  = no_active;
            } else {
                cbus_mmrflag = FALSE;
            };
            if( dabus < 0x60 and space == memory ) {
                dbus_mmrflag = TRUE;
                dbus_active  = no_active;
            } else {
                dbus_mmrflag = FALSE;
            };
            if( eabus < 0x60 and space == memory ) {
                ebus_mmrflag = TRUE;
                ebus_active  = no_active;
            } else {
                ebus_mmrflag = FALSE;
            };
            if( fabus < 0x60 and space == memory ) {
                fbus_mmrflag = TRUE;
                fbus_active  = no_active;
            } else {
                fbus_mmrflag = FALSE;
            };
            
            -- set rspace/wspace flags --
            rspace = space;
            wspace = space;
            
        };
        
        Select_set_instruction( instruction : data_access_instruction ) is {
            case instruction {
                [ None, LockAccess ] : { instr_lock_group  = instruction; };
                default              : { outf("Warning : can't set instruction as %s \n",instruction); };
            };
        };
        
        Select_set_noaccess_instruction() is {
            instr_lock_group = None;
            idle_flag        = FALSE;
        };
        Select_get_read_instruction() : data_access_instruction is {
            result = None;
        };
        Select_get_write_instruction() : data_access_instruction is {
            result = None;
        };
        Select_get_lock_instruction() : data_access_instruction is {
            result = instr_lock_group;
        };
        Select_get_coef_instruction() : data_access_instruction is {
            result = None;
        };
        
    };
    
    -- duplicate function call that depend on lock enable/disable --
    set_activate_buses() is {
        if( me is a disable'lock DATA_ACCESS_ATTRIBUTES (attributes) ) { attributes.Select_set_activate_buses(); };
        if( me is a enable'lock  DATA_ACCESS_ATTRIBUTES (attributes) ) { attributes.Select_set_activate_buses(); };
    };
    set_instruction( instruction : data_access_instruction ) is {
        if( me is a disable'lock DATA_ACCESS_ATTRIBUTES (attributes) ) { attributes.Select_set_instruction( instruction ); };
        if( me is a enable'lock  DATA_ACCESS_ATTRIBUTES (attributes) ) { attributes.Select_set_instruction( instruction );  };
    };
    set_noaccess_instruction() is {
        if( me is a disable'lock DATA_ACCESS_ATTRIBUTES (attributes) ) { attributes.Select_set_noaccess_instruction(); };
        if( me is a enable'lock  DATA_ACCESS_ATTRIBUTES (attributes) ) { attributes.Select_set_noaccess_instruction(); };
    };
    get_read_instruction() : data_access_instruction is {
        if( me is a disable'lock DATA_ACCESS_ATTRIBUTES (attributes) ) { result = attributes.Select_get_read_instruction(); };
        if( me is a enable'lock  DATA_ACCESS_ATTRIBUTES (attributes) ) { result = attributes.Select_get_read_instruction();  };
    };
    get_write_instruction() : data_access_instruction is {
        if( me is a disable'lock DATA_ACCESS_ATTRIBUTES (attributes) ) { result = attributes.Select_get_write_instruction(); };
        if( me is a enable'lock  DATA_ACCESS_ATTRIBUTES (attributes) ) { result = attributes.Select_get_write_instruction();  };
    };
    get_lock_instruction() : data_access_instruction is {
        if( me is a disable'lock DATA_ACCESS_ATTRIBUTES (attributes) ) { result = attributes.Select_get_lock_instruction(); };
        if( me is a enable'lock  DATA_ACCESS_ATTRIBUTES (attributes) ) { result = attributes.Select_get_lock_instruction();  };
    };
    get_coef_instruction() : data_access_instruction is {
        if( me is a disable'lock DATA_ACCESS_ATTRIBUTES (attributes) ) { result = attributes.Select_get_coef_instruction(); };
        if( me is a enable'lock  DATA_ACCESS_ATTRIBUTES (attributes) ) { result = attributes.Select_get_coef_instruction();  };
    };
    
    -- set driving data --
    set_driving_data( config : CPUBUS_CONFIGURATIONS, pipeline : CPU_PIPELINE, bus_states : CPU_BUSES, driver : CPU_SIGNALS ) is {
        
        -- Bbus : request/address --
        if( pipeline.access1 != NULL and pipeline.access1.bbus_active == active and pipeline.pipeline_update_read ) {
            
            -- set request and address --
            driver.bbus.req  = 0;
            driver.bbus.abus = pipeline.access1.babus;
            
            -- privilege mode --
            if( pipeline.access1.mp_supervisor ) {
                driver.mp.bpm = 0;
            } else {
                driver.mp.bpm = 1;
            };
            
            -- set double qualifier --
            if( pipeline.access1.dblcoef == double ) {
                driver.qualifiers.dblcoef = 1;
            } else {
                driver.qualifiers.dblcoef = 0;
            };
            
        } else {
            driver.bbus.req           = 1;
            driver.mp.bpm             = 0;
            driver.qualifiers.dblcoef = 0;
        };
        
        -- Cbus : request/address --
        if( pipeline.access1 != NULL and pipeline.access1.cbus_active == active and pipeline.access1.doubler != double and pipeline.pipeline_update_read ) {
            driver.cbus.req  = 0;
            driver.cbus.abus = pipeline.access1.cabus;
            
            -- privilege mode --
            if( pipeline.access1.mp_supervisor ) {
                driver.mp.cpm = 0;
            } else {
                driver.mp.cpm = 1;
            };
            
        } else {
            driver.cbus.req  = 1;
            driver.mp.cpm = 0;
        };
        
        -- Dbus : request/address/double qualifier --
        if( pipeline.access1 != NULL and pipeline.access1.dbus_active == active and pipeline.pipeline_update_read ) {
            
            -- set io request when io access --
            if( pipeline.access1.space == io ) { 
                driver.dbus.ioreq  = 0;
                driver.dbus.req    = 1;
            } else {
                driver.dbus.ioreq  = 1;
                driver.dbus.req    = 0;
            };
            
            -- privilege mode --
            if( pipeline.access1.mp_supervisor ) {
                driver.mp.dpm = 0;
            } else {
                driver.mp.dpm = 1;
            };
            
            -- set double qualifier when double access --
            if( pipeline.access1.doubler == double ){
                driver.qualifiers.doubler = 1;
            } else {
                driver.qualifiers.doubler = 0;
            };
            
            -- set byte qualifiers when byte access --
            if( pipeline.access1.read_width != word ){
                driver.qualifiers.rdbyte = 1;
                if( pipeline.access1.read_width == high_byte ) { driver.qualifiers.rdlsbe = 0; }
                                                          else { driver.qualifiers.rdlsbe = 1; };
            } else {
                driver.qualifiers.rdbyte = 0;
                driver.qualifiers.rdlsbe = 0;
            };
            
            -- set address value --
            driver.dbus.abus = pipeline.access1.dabus;
            
            -- set emulation qualifiers --
            if( pipeline.access1.dbus_type == debug and bus_states.emulation.current.debugenable == 1 ) {
                driver.emulations.dbugd = 1;
            } else {
                driver.emulations.dbugd = 0;
            };
            
            -- set icemaker xdma access qualifiers --
            if( bus_states.emulation.current.debugenable == 1 and pipeline.access1.dbus_type == xdma ) {
                driver.mp.xdmadmpovr = 1;
            } else {
                driver.mp.xdmadmpovr = 0;
            };
            
        } else {
            driver.dbus.req           = 1;
            driver.dbus.ioreq         = 1;
            driver.mp.dpm             = 0;
            driver.qualifiers.doubler = 0;
            driver.qualifiers.rdbyte  = 0;
            driver.qualifiers.rdlsbe  = 0;
            driver.emulations.dbugd   = 0;
            driver.mp.xdmadmpovr      = 0;
        };
        
        -- Ebus : request/address/double qualifier --
        if( pipeline.execute != NULL and pipeline.execute.ebus_active == active and pipeline.pipeline_update_write ) {
            
            -- set io request when io access --
            if( pipeline.execute.space == io ) { 
                driver.ebus.ioreq  = 0;
                driver.ebus.req    = 1;
            } else {
                driver.ebus.ioreq  = 1;
                driver.ebus.req    = 0;
            };
            
            -- privilege mode --
            if( pipeline.execute.mp_supervisor ) {
                driver.mp.epm = 0;
            } else {
                driver.mp.epm = 1;
            };
            
            -- set double qualifier when double access --
            if( pipeline.execute.doublew == double ){
                driver.qualifiers.doublew = 1;
            } else {
                driver.qualifiers.doublew = 0;
            };
            
            -- set byte qualifiers when byte access --
            if( pipeline.execute.write_width != word ){
                driver.qualifiers.wrbyte = 1;
                if( pipeline.execute.write_width == high_byte ) { driver.qualifiers.wrlsbe = 0; }
                                                           else { driver.qualifiers.wrlsbe = 1; };
            } else {
                driver.qualifiers.wrbyte = 0;
                driver.qualifiers.wrlsbe = 0;
            };
            
            -- set address value --
            driver.ebus.abus = pipeline.execute.eabus;
            
            -- set emulation qualifiers --
            if( bus_states.emulation.current.debugenable == 1 and
                (pipeline.execute.ebus_type == debug_prog_write or pipeline.execute.ebus_type == debug_data_write) ) {
                driver.emulations.dbuge = 1;
                if( pipeline.execute.ebus_type == debug_prog_write ) {
                    driver.emulations.dmapw = 1;
                } else {
                    driver.emulations.dmapw = 0;
                };
            } else {
                driver.emulations.dbuge = 0;
                driver.emulations.dmapw = 0;
            };
            
            -- set icemaker xdma access qualifiers --
            if( bus_states.emulation.current.debugenable == 1 and pipeline.execute.ebus_type == xdma ) {
                driver.mp.xdmaempovr = 1;
            } else {
                driver.mp.xdmaempovr = 0;
            };
            
        } else {
            driver.ebus.req           = 1;
            driver.ebus.ioreq         = 1;
            driver.mp.epm             = 0;
            driver.qualifiers.doublew = 0;
            driver.qualifiers.wrbyte  = 0;
            driver.qualifiers.wrlsbe  = 0;
            driver.emulations.dbuge   = 0;
            driver.emulations.dmapw   = 0;
            driver.mp.xdmaempovr      = 0;
        };
        
        -- Fbus : request/address --
        if( pipeline.execute != NULL and pipeline.execute.fbus_active == active and pipeline.execute.doublew != double and pipeline.pipeline_update_write ) {
            driver.fbus.req  = 0;
            driver.fbus.abus = pipeline.execute.fabus;
            
            -- privilege mode --
            if( pipeline.execute.mp_supervisor ) {
                driver.mp.fpm = 0;
            } else {
                driver.mp.fpm = 1;
            };
            
        } else {
            driver.fbus.req  = 1;
            driver.mp.fpm    = 0;
        };
        
        -- address update when mmr access --
        if( pipeline.access1 != NULL and pipeline.access1.bbus_mmrflag and pipeline.pipeline_update_read ) {
            driver.bbus.abus = pipeline.access1.babus;
        };
        if( pipeline.access1 != NULL and pipeline.access1.cbus_mmrflag and pipeline.pipeline_update_read ) {
            driver.cbus.abus = pipeline.access1.cabus;
        };
        if( pipeline.access1 != NULL and pipeline.access1.dbus_mmrflag and pipeline.pipeline_update_read ) {
            driver.dbus.abus = pipeline.access1.dabus;
        };
        if( pipeline.execute != NULL and pipeline.execute.ebus_mmrflag and pipeline.pipeline_update_write ) {
            driver.ebus.abus = pipeline.execute.eabus;
        };
        if( pipeline.execute != NULL and pipeline.execute.fbus_mmrflag and pipeline.pipeline_update_write ) {
            driver.fbus.abus = pipeline.execute.fabus;
        };
        
        -- drive for write data in Ebus --
        if( bus_states.ebus.pending.counter == 1 ) {
            driver.ebus.dbus = bus_states.ebus.pending.progress[bus_states.ebus.pending.rdy_index].edbus;
        } else {
            if( bus_states.ebus.pending.counter == 0 and 
                pipeline.previous_execute != NULL and pipeline.previous_execute.ebus_mmrflag == TRUE ) {
                driver.ebus.dbus = pipeline.previous_execute.edbus;
            };
        };
        -- drive for write data in Fbus --
        var double_write_flag : bool;
        if( bus_states.ebus.pending.progress[bus_states.ebus.pending.rdy_index] != NULL and
            bus_states.ebus.pending.progress[bus_states.ebus.pending.rdy_index].ebus_active == active and
            bus_states.ebus.pending.progress[bus_states.ebus.pending.rdy_index].doublew == double ) { 
            double_write_flag = TRUE;
        } else { 
            double_write_flag = FALSE;
        };
        
        if( bus_states.fbus.pending.counter == 1 and double_write_flag == FALSE ) {
            driver.fbus.dbus = bus_states.fbus.pending.progress[bus_states.fbus.pending.rdy_index].fdbus;
        };
        if( bus_states.fbus.pending.counter == 0 and double_write_flag == TRUE and bus_states.ebus.pending.counter != 0 ) {
            driver.fbus.dbus = bus_states.ebus.pending.progress[bus_states.ebus.pending.rdy_index].fdbus;
        };
        if( bus_states.fbus.pending.counter == 0 and double_write_flag == FALSE and 
            pipeline.previous_execute != NULL and pipeline.previous_execute.fbus_mmrflag == TRUE ) {
            driver.fbus.dbus = pipeline.previous_execute.fdbus;
        };
        
        -- set lock qualifier(for Laijin3.x)--
        if( pipeline.access1              != NULL   and 
            pipeline.access1.lock         == enable and 
            pipeline.access1.dbus_active  == active and 
            pipeline.pipeline_update_read ) {
            driver.qualifiers.lock = 1;
        } else {
            driver.qualifiers.lock = 0;
        };
        
        -- drive for idlemaster signal --
        if( pipeline.prev_idle_stall == TRUE and pipeline.curr_idle_stall == FALSE ) {
            driver.systems.idlemaster = 1;
        } else {
            driver.systems.idlemaster = 0;
        };
        
        -- mmr update --
        if( pipeline.access1 != NULL and pipeline.access1.dbus_mmrflag ) {
          bus_states.mmr.rdata = bus_states.mmr.reg[ pipeline.access1.dabus ];
        };
        if( pipeline.execute != NULL and pipeline.execute.ebus_mmrflag ) {
          bus_states.mmr.reg[ pipeline.execute.eabus ] = pipeline.execute.edbus;
        };
        
        -- driving cache related signals --
        driver.cache.cachefreeze = bus_states.mmr.reg[MMR_ST3_55][15:15];
        driver.cache.cacheenable = bus_states.mmr.reg[MMR_ST3_55][14:14];
        driver.cache.cacheclr    = bus_states.mmr.reg[MMR_ST3_55][13:13];
        if( bus_states.cache.current.cacheendclr == 1 ) {
            bus_states.mmr.reg[MMR_ST3_55] = bus_states.mmr.reg[MMR_ST3_55] & 0xdfff;
        };

        -- HOM/SAM related signals --
        if ( bus_states.system.current.cpu_resetn == 0 ) {
             bus_states.mmr.shadow_reg[SHADOW_HOMSAM][08:08] = bus_states.system.current.gl_hompnsamp_init;
             bus_states.mmr.shadow_reg[SHADOW_HOMSAM][09:09] = bus_states.system.current.gl_homrnsamr_init;
             bus_states.mmr.shadow_reg[SHADOW_HOMSAM][10:10] = bus_states.system.current.gl_homxnsamx_init;
             bus_states.mmr.shadow_reg[SHADOW_HOMSAM][11:11] = bus_states.system.current.gl_homynsamy_init;
        } else if( pipeline.execute != NULL and pipeline.execute.ebus_mmrflag and pipeline.execute.eabus == MMR_ST3_55 ) {
             bus_states.mmr.shadow_reg[SHADOW_HOMSAM]        = pipeline.execute.edbus;
        };
        if ( bus_states.system.current.cpu_resetn == 0 ) { 
             driver.systems.cpu_homnsam = 0xf;
        } else {
             driver.systems.cpu_homnsam = bus_states.mmr.shadow_reg[SHADOW_HOMSAM][11:08];
        };
        -- MMR read value reflects gl_homnsam_tr
        bus_states.mmr.reg[MMR_ST3_55] = (bus_states.mmr.reg[MMR_ST3_55] & 0xf0ff) | (bus_states.system.current.gl_homnsam << 8);

        -- driving cpu_mpnmc_tr signal --
        driver.systems.cpu_mpnmc = bus_states.mmr.reg[MMR_ST3_55][06:06];

        -- driveing cpu_avis_tr signal --
        driver.systems.avis = bus_states.mmr.reg[MMR_ST3_55][04:04];
        
        -- driving icemaker's ready as emiodready and emioeready --
        if( bus_states.dbus.pending.progress[bus_states.dbus.pending.rdy_index] != NULL and
            bus_states.dbus.pending.progress[bus_states.dbus.pending.rdy_index].dbus_active == active and
            bus_states.dbus.pending.progress[bus_states.dbus.pending.rdy_index].rspace == io and
            bus_states.dbus.pending.progress[bus_states.dbus.pending.rdy_index].dabus is in ICEMAKER_IOMMR ) { 
            driver.mp.emiodready = 0;
        } else {
            driver.mp.emiodready = 1;
        };
        
        if( bus_states.ebus.pending.progress[bus_states.ebus.pending.rdy_index] != NULL and
            bus_states.ebus.pending.progress[bus_states.ebus.pending.rdy_index].ebus_active == active and
            bus_states.ebus.pending.progress[bus_states.ebus.pending.rdy_index].wspace == io and
            bus_states.ebus.pending.progress[bus_states.ebus.pending.rdy_index].eabus is in ICEMAKER_IOMMR ) { 
            driver.mp.emioeready = 0;
        } else {
            driver.mp.emioeready = 1;
        };
        
        -- driving violation signals for memory protection --
        if( bus_states.mp.current.cpu_privenable == 1    and
            pipeline.mp_viol_report              != NULL and 
            pipeline.mp_viol_report.mp_viol_pbus == TRUE     ) {
            driver.mp.rptpvio = 1;
        } else {
            driver.mp.rptpvio = 0;
        };
        
        if( bus_states.mp.current.cpu_privenable == 1    and
            pipeline.mp_viol_report              != NULL     and 
            pipeline.mp_viol_report.bbus_active  == finished and 
            pipeline.mp_viol_report.mp_viol_bbus == TRUE         ) {
            driver.mp.rptbvio = 1;
        } else {
            driver.mp.rptbvio = 0;
        };
        
        if( bus_states.mp.current.cpu_privenable  == 1        and
            pipeline.mp_viol_report               != NULL     and 
            pipeline.mp_viol_report.cbus_active   == finished and 
            pipeline.mp_viol_report.mp_viol_cbus == TRUE         ) {
            driver.mp.rptcvio = 1;
        } else {
            driver.mp.rptcvio = 0;
        };
        
        if( bus_states.mp.current.cpu_privenable == 1        and
            pipeline.mp_viol_report              != NULL     and 
            pipeline.mp_viol_report.dbus_active  == finished and 
            pipeline.mp_viol_report.mp_viol_dbus == TRUE         ) {
            if( pipeline.mp_viol_report.space == io ) {
                driver.mp.rptdvio   = 0;
                driver.mp.rptiodvio = 1;
            } else {
                driver.mp.rptdvio   = 1;
                driver.mp.rptiodvio = 0;
            };
        } else {
            driver.mp.rptdvio   = 0;
            driver.mp.rptiodvio = 0;
        };
        
        if( bus_states.mp.current.cpu_privenable == 1        and
            pipeline.mp_viol_report              != NULL     and 
            pipeline.mp_viol_report.ebus_active  == finished and 
            pipeline.mp_viol_report.mp_viol_ebus == TRUE         ) {
            
            if( pipeline.mp_viol_report.space == io ) {
                driver.mp.rptevio   = 0;
                driver.mp.rptioevio = 1;
            } else {
                driver.mp.rptevio   = 1;
                driver.mp.rptioevio = 0;
            };
            
            if( pipeline.mp_viol_report.space == space and pipeline.mp_viol_report.mp_viol_contextsave ) {
                driver.mp.rptintr = 1;
            } else {
                driver.mp.rptintr = 0;
            };
            
        } else {
            driver.mp.rptevio   = 0;
            driver.mp.rptioevio = 0;
            driver.mp.rptintr   = 0;
        };
        
        if( bus_states.mp.current.cpu_privenable == 1        and
            pipeline.mp_viol_report              != NULL     and 
            pipeline.mp_viol_report.fbus_active  == finished and 
            pipeline.mp_viol_report.mp_viol_fbus == TRUE         ) {
            driver.mp.rptfvio = 1;
        } else {
            driver.mp.rptfvio = 0;
        };
        
        if( bus_states.mp.current.cpu_privenable  == 1    and
            pipeline.mp_viol_report               != NULL and
            ( pipeline.mp_viol_report.mp_viol_pbus == TRUE or
             (pipeline.mp_viol_report.mp_viol_bbus == TRUE and pipeline.mp_viol_report.bbus_active == finished) or
             (pipeline.mp_viol_report.mp_viol_cbus == TRUE and pipeline.mp_viol_report.cbus_active == finished) or
             (pipeline.mp_viol_report.mp_viol_dbus == TRUE and pipeline.mp_viol_report.dbus_active == finished) or
             (pipeline.mp_viol_report.mp_viol_ebus == TRUE and pipeline.mp_viol_report.ebus_active == finished) or
             (pipeline.mp_viol_report.mp_viol_fbus == TRUE and pipeline.mp_viol_report.fbus_active == finished)   ) ) {
            driver.mp.rptpcvio = pipeline.mp_viol_report.mp_viol_address;
            
        } else {
            driver.mp.rptpcvio  = 0;
        };
        
        if( pipeline.mp_viol_report != NULL ) {
          if(  pipeline.mp_viol_report.mp_supervisor ) { driver.mp.rptpm = 0; };
          if( !pipeline.mp_viol_report.mp_supervisor ) { driver.mp.rptpm = 1; };
        };
        
        -------------------------
        -- for unknown driving --
        -------------------------
        driver.bbus.valid_abus = FALSE;
        driver.cbus.valid_abus = FALSE;
        driver.dbus.valid_abus = FALSE;
        driver.ebus.valid_abus = FALSE;
        driver.fbus.valid_abus = FALSE;
        if( bus_states.bbus.pending.counter == 2 or driver.bbus.req == 0                            ) { driver.bbus.valid_abus = TRUE; };
        if( bus_states.cbus.pending.counter == 2 or driver.cbus.req == 0                            ) { driver.cbus.valid_abus = TRUE; };
        if( bus_states.dbus.pending.counter == 2 or driver.dbus.req == 0 or driver.dbus.ioreq  == 0 ) { driver.dbus.valid_abus = TRUE; };
        if( bus_states.ebus.pending.counter == 2 or driver.ebus.req == 0 or driver.ebus.ioreq  == 0 ) { driver.ebus.valid_abus = TRUE; };
        if( bus_states.fbus.pending.counter == 2 or driver.fbus.req == 0                            ) { driver.fbus.valid_abus = TRUE; };
        
        var rslot0 : DATA_ACCESS_ATTRIBUTES; rslot0 = bus_states.read_inst_slot[  bus_states.read_inst_index ];
        var rslot1 : DATA_ACCESS_ATTRIBUTES; rslot1 = bus_states.read_inst_slot[ ~bus_states.read_inst_index ];
        var wslot0 : DATA_ACCESS_ATTRIBUTES; wslot0 = bus_states.write_inst_slot[  bus_states.write_inst_index ];
        var wslot1 : DATA_ACCESS_ATTRIBUTES; wslot1 = bus_states.write_inst_slot[ ~bus_states.write_inst_index ];
        
        if( rslot0 != NULL and rslot1 != NULL ) {
            if( rslot0.bbus_active == active and bus_states.bbus.state.ready != ready ) {
                if( rslot1.cbus_active != no_active and rslot1.doubler == word ) { driver.cbus.valid_abus = TRUE; };
                if( rslot1.dbus_active != no_active                            ) { driver.dbus.valid_abus = TRUE; };
            };
            if( rslot0.cbus_active == active and rslot0.doubler == word and bus_states.cbus.state.ready != ready ) {
                if( rslot1.bbus_active != no_active                            ) { driver.bbus.valid_abus = TRUE; };
                if( rslot1.dbus_active != no_active                            ) { driver.dbus.valid_abus = TRUE; };
            };
            if( rslot0.dbus_active == active and bus_states.dbus.state.ready != ready ) {
                if( rslot1.cbus_active != no_active and rslot1.doubler == word ) { driver.cbus.valid_abus = TRUE; };
                if( rslot1.bbus_active != no_active                            ) { driver.bbus.valid_abus = TRUE; };
            };
        };
        
        if( wslot0 != NULL and wslot1 != NULL ) {
            if( wslot0.ebus_active == active and bus_states.ebus.state.ready != ready ) {
                if( wslot1.fbus_active != no_active and wslot1.doublew == word ) { driver.fbus.valid_abus = TRUE; };
            };
            if( wslot0.fbus_active == active and wslot0.doublew == word and bus_states.fbus.state.ready != ready ) {
                if( wslot1.ebus_active != no_active                            ) { driver.ebus.valid_abus = TRUE; };
            };
        };
        
        if( driver.bbus.req == 0 or driver.cbus.req == 0 or driver.dbus.req == 0 or driver.dbus.ioreq  == 0 ) {
            driver.qualifiers.valid_doubler = TRUE;
            driver.qualifiers.valid_rdbyte  = TRUE;
            driver.qualifiers.valid_rdlsbe  = TRUE;
            driver.qualifiers.valid_dblcoef = TRUE;
            driver.emulations.valid_dbugd   = TRUE;
            driver.mp.valid_bpm             = TRUE;
            driver.mp.valid_cpm             = TRUE;
            driver.mp.valid_dpm             = TRUE;
            driver.mp.valid_xdmadmpovr      = TRUE;
        } else {
            driver.qualifiers.valid_doubler = FALSE;
            driver.qualifiers.valid_rdbyte  = FALSE;
            driver.qualifiers.valid_rdlsbe  = FALSE;
            driver.qualifiers.valid_dblcoef = FALSE;
            driver.emulations.valid_dbugd   = FALSE;
            driver.mp.valid_bpm             = FALSE;
            driver.mp.valid_cpm             = FALSE;
            driver.mp.valid_dpm             = FALSE;
            driver.mp.valid_xdmadmpovr      = FALSE;
        };
        if( driver.ebus.req == 0 or driver.fbus.req == 0 or driver.ebus.ioreq  == 0 ) {
            driver.qualifiers.valid_doublew = TRUE;
            driver.qualifiers.valid_wrbyte  = TRUE;
            driver.qualifiers.valid_wrlsbe  = TRUE;
            driver.emulations.valid_dmapw   = TRUE;
            driver.emulations.valid_dbuge   = TRUE;
            driver.mp.valid_epm             = TRUE;
            driver.mp.valid_fpm             = TRUE;
            driver.mp.valid_xdmaempovr      = TRUE;
        } else {
            driver.qualifiers.valid_doublew = FALSE;
            driver.qualifiers.valid_wrbyte  = FALSE;
            driver.qualifiers.valid_wrlsbe  = FALSE;
            driver.emulations.valid_dmapw   = FALSE;
            driver.emulations.valid_dbuge   = FALSE;
            driver.mp.valid_epm             = FALSE;
            driver.mp.valid_fpm             = FALSE;
            driver.mp.valid_xdmaempovr      = FALSE;
        };
        
        if( bus_states.ebus.pending.counter != 0 ) {
            driver.ebus.valid_dbus = TRUE;
        } else {
            driver.ebus.valid_dbus = FALSE;
        };
        if( bus_states.fbus.pending.counter != 0 or 
            (bus_states.ebus.pending.progress[bus_states.ebus.pending.rdy_index] != NULL               and
             bus_states.ebus.pending.progress[bus_states.ebus.pending.rdy_index].ebus_active == active and
             bus_states.ebus.pending.progress[bus_states.ebus.pending.rdy_index].doublew     == double    ) ) {
            driver.fbus.valid_dbus = TRUE;
        } else {
            driver.fbus.valid_dbus = FALSE;
        };
        
        -- for cpu_reset driving --
        if( bus_states.system.current.gl_resetn == 1 and bus_states.system.current.cpu_resetn == 0 ) {
            reset_counter = reset_counter + 1;
        } else {
            reset_counter = 0;
        };
        if( bus_states.system.current.gl_resetn == 0 or (reset_counter > 0 and reset_counter < 9) ) {
            driver.systems.cpu_resetn = 0;
        } else {
            driver.systems.cpu_resetn = 1;
        };
        
    };

    !reset_counter : uint;
    
    init() is also {
        reset_counter = 0;
    }
    
};

extend PROGRAM_ACCESS_ATTRIBUTES {
    
    --
    --  instr_group    branch_occur    dismiss_occur   action
    -- --------------+---------------+---------------+----------
    --  ProgramRead    TRUE            FALSE           Program read with branch
    --  ProgramRead    TRUE            TRUE            Program read with branch/dismiss
    --  ProgramRead    FALSE           --              Program read(address is continuous with previous one)
    --  None           --              FALSE           no action
    --  None           TRUE            TRUE            generate dismiss without read request
    --  
    
    instr_group    : program_access_instruction;
    
    branch_occur   : bool;
    dismiss_occur  : bool;
    branch_address : uint(bits : 24);
    
    keep branch_address[1:0] == 0;
    
    dismiss_waiting : uint;
    keep soft dismiss_waiting >= 0;
    keep soft dismiss_waiting <= 5;
    
    -- for interrupt control --
    isr_flag       : bool; keep soft isr_flag  == FALSE;
    isr_index      : uint; keep soft isr_index == 0;
    ---------------------------
    
    -- for mp control --
    mp_supervisor  : bool; keep soft mp_supervisor  == TRUE;
    mp_viol_pbus   : bool; keep soft mp_viol_pbus   == FALSE;
    mp_viol_pclear : bool; keep soft mp_viol_pclear == FALSE;
    --------------------
    
    set_activate_buses( buffer : PROGRAM_BUFFER, bus_states : CPU_BUSES, start_address : uint(bits : 24) ) is {
        if( instr_group == ProgramRead ) { pbus_active =    active; }
                                    else { pbus_active = no_active; };
        
        if( branch_occur and dismiss_occur and bus_states.pbus.pending.counter > 0 ) {
               buffer.keep_dismiss_flag    = TRUE;
               buffer.keep_dismiss_address = branch_address;
        };
        
        if( buffer.previous_prog_attributes != NULL ) { 
            if( instr_group == ProgramRead ) {
                if( !buffer.keep_dismiss_flag ) { 
                  if( branch_occur ) { pabus = branch_address; } 
                                else { pabus = buffer.current_prog_attributes.pabus + 4; };
                } else { pabus = buffer.keep_dismiss_address; };
                buffer.keep_dismiss_flag = FALSE;
            } else { pabus = buffer.current_prog_attributes.pabus; };
        } else { 
            instr_group = None;
            pbus_active = no_active;
            pabus = start_address - 4;
        };
    };
    
    set_noaccess_instruction() is {
        instr_group   = None;
        branch_occur  = FALSE;
        dismiss_occur = FALSE;
        isr_flag      = FALSE;
    };
    
    set_driving_data( config : CPUBUS_CONFIGURATIONS, buffer: PROGRAM_BUFFER, bus_states : CPU_BUSES, driver : CPU_SIGNALS ) is {
        
        -- pbus request/address --
        if( buffer.current_prog_attributes != NULL and 
            buffer.current_prog_attributes.instr_group == ProgramRead and
            buffer.current_prog_attributes.pbus_active == active and buffer.buffer_update ) {
            driver.pbus.req = 0;
            driver.pbus.abus = pabus;
            
            -- privilege mode --
            if( buffer.current_prog_attributes.mp_supervisor ) {
                driver.mp.ppm = 0;
            } else {
                driver.mp.ppm = 1;
            };
            
            -- iack --
            if( buffer.current_prog_attributes.isr_flag ) { 
                driver.systems.iack = 0;
            } else {
                driver.systems.iack = 1;
            };
            
            -- emulation qualifier
            if( buffer.current_prog_attributes.pbus_type == debug and bus_states.emulation.current.debugenable == 1 ) {
                driver.emulations.dmapr = 1;
            } else {
                driver.emulations.dmapr = 0;
            };
            
        } else {
            driver.pbus.req         = 1;
            driver.mp.ppm           = 0;
            driver.systems.iack     = 1;
            driver.emulations.dmapr = 0;
        };
        
        -- pdismiss --
        if( branch_occur and dismiss_occur and bus_states.pbus.pending.counter > 0 and buffer.buffer_update ) {
            driver.pbus.dismiss = 1;
        } else {
            driver.pbus.dismiss = 0;
        };
        
        -- perrack --
        if( buffer.current_prog_attributes != NULL and buffer.buffer_update and 
            bus_states.mp.current.cpu_privenable == 0 and 
            buffer.current_prog_attributes.mp_viol_pbus and !buffer.current_prog_attributes.mp_viol_pclear ) {
            driver.systems.perrack = 1;
        } else {
            driver.systems.perrack = 0;
        };
        
        -------------------------
        -- for unknown driving --
        -------------------------
        if( driver.pbus.req == 0 or bus_states.pbus.pending.counter > 0 ) {
            driver.pbus.valid_abus        = TRUE; 
        } else {
            driver.pbus.valid_abus        = FALSE; 
        };
        
        if( driver.pbus.req == 0 ) { 
            driver.emulations.valid_dmapr = TRUE; 
            driver.mp.valid_ppm           = TRUE;
        } else { 
            driver.emulations.valid_dmapr = FALSE; 
            driver.mp.valid_ppm           = FALSE;
        };
        
    };
    
};

------------------------------
-- pipeline for data access --
------------------------------
unit CPU_PIPELINE {
    !decode  : DATA_ACCESS_ATTRIBUTES;
    !address : DATA_ACCESS_ATTRIBUTES;   -- read mdp req
    !access1 : DATA_ACCESS_ATTRIBUTES;   -- read req
    !access2 : DATA_ACCESS_ATTRIBUTES;   -- data dismiss when bypass occur
    !read    : DATA_ACCESS_ATTRIBUTES;   -- write mdp req
    !execute : DATA_ACCESS_ATTRIBUTES;   -- write req
    !write   : DATA_ACCESS_ATTRIBUTES;
    !write2  : DATA_ACCESS_ATTRIBUTES;   -- for reporting mp violations
    
    d0_bypass : BYPASS_ATTRIBUTES;
    d1_bypass : BYPASS_ATTRIBUTES;
    d2_bypass : BYPASS_ATTRIBUTES;
    
    !previous_access1 : DATA_ACCESS_ATTRIBUTES;
    !previous_access2 : DATA_ACCESS_ATTRIBUTES;
    !previous_read    : DATA_ACCESS_ATTRIBUTES;
    !previous_execute : DATA_ACCESS_ATTRIBUTES;
    !previous_write2  : DATA_ACCESS_ATTRIBUTES;
    
    !prev_idle_stall : bool;
    !curr_idle_stall : bool;
    
    pipeline_update_read   : bool;
    pipeline_update_write  : bool;
    pipeline_update_write2 : bool;
    
    !write2_queue   : list of DATA_ACCESS_ATTRIBUTES;
    !mp_viol_report : DATA_ACCESS_ATTRIBUTES;
    
    --  ------------------------------------------------------------
    --     <-------------------- insert stall when idle inst is in pipeline or abortready or suspend
    --  decode
    --  address
    --     <-------------------- insert stall when 'io access stall' or 'lock stall'
    --  access1(gen read req)
    --     <-------------------- insert stall when ('read stall' and not 'write stall') or 'bypass ocuur' or ('read ready stall' and not 'write ready stall')
    --  access2
    --  read
    --     <-------------------- insert stall when 'lock stall'
    --  execute(gen write req)
    --     <-------------------- insert stall when 'write stall' or 'write ready stall'
    --  write
    --  write2
    --  ------------------------------------------------------------
    
    transition( attributes : DATA_ACCESS_ATTRIBUTES, bus_states : CPU_BUSES ) is {
        
        -- keeping pipeline status --
        previous_access1 = access1;
        previous_access2 = access2;
        previous_read    = read;
        previous_execute = execute;
        previous_write2  = write2;
        
        -- keeping idle stall flag --
        prev_idle_stall = curr_idle_stall;
        curr_idle_stall = idle_stall(bus_states);
        bus_states.curr_idle_stall = curr_idle_stall;
        bus_states.prev_idle_stall = prev_idle_stall;
        
        -- pipeline transitions --
        if( bus_states.data_write_current_stall() or ready_stall_for_write(bus_states) ) {
            write2  = write;
            write   = NULL;
        } else if( bus_states.check_lock_stall(read) or write_stall(bus_states) or iodready_stall(bus_states) ) {
            write2  = write;
            write   = execute;
            execute = NULL;
        } else if( bus_states.data_read_current_stall() or d0_bypass.current_bypass_occur() or ready_stall_for_read(bus_states) ) {
            write2  = write;
            write   = execute;
            execute = read;
            read    = access2;
            access2 = NULL;
        } else if( io_stall() or read_stall(bus_states) or lock_stall(bus_states) ) {
            write2  = write;
            write   = execute;
            execute = read;
            read    = access2;
            access2 = access1;
            access1 = NULL;
        } else if( idle_stall(bus_states) or abortready_stall(bus_states) or suspend_stall(bus_states) ) {
            write2  = write;
            write   = execute;
            execute = read;
            read    = access2;
            access2 = access1;
            access1 = address;
            address = decode;
            decode  = NULL;
        } else {
            write2  = write;
            write   = execute;
            execute = read;
            read    = access2;
            access2 = access1;
            access1 = address;
            address = decode;
            decode  = attributes;
        };
        
        -- check update --
        if( previous_access1 != NULL and access1 != NULL ) {
            if( previous_access1 != access1 ) { pipeline_update_read  = TRUE;  } 
                                         else { pipeline_update_read  = FALSE; };
        } else { 
            pipeline_update_read  = TRUE;
        };
        
        if( previous_execute != NULL and execute != NULL ) {
            if( previous_execute != execute ) { pipeline_update_write = TRUE;  }
                                         else { pipeline_update_write = FALSE; };
        } else { 
            pipeline_update_write = TRUE;
        };
        
        if( previous_write2 != NULL and write2 != NULL ) {
            if( previous_write2 != write2 ) { pipeline_update_write2 = TRUE;  }
                                       else { pipeline_update_write2 = FALSE; };
        } else { 
            pipeline_update_write2 = TRUE;
        };
        
        -- bypass check --
        d0_bypass.check_bypass_condition( bus_states, access1, access2 );
        d1_bypass.check_bypass_condition( bus_states, access1, read    );
        d2_bypass.check_bypass_condition( bus_states, access1, execute );
        
        -- update write2 queue --
        if( write2 != NULL and pipeline_update_write2 ) { write2_queue.push( write2 ); };
        
        if( write2_queue.size() != 0 and
            (write2_queue[0].bbus_active == no_active or write2_queue[0].bbus_active == finished) and
            (write2_queue[0].cbus_active == no_active or write2_queue[0].cbus_active == finished) and
            (write2_queue[0].dbus_active == no_active or write2_queue[0].dbus_active == finished) and
            (write2_queue[0].ebus_active == no_active or write2_queue[0].ebus_active == finished) and
            (write2_queue[0].fbus_active == no_active or write2_queue[0].fbus_active == finished)     ) {
            mp_viol_report = write2_queue.pop0();
        } else {
            mp_viol_report = NULL;
        };
        
    };
    
    -- ready stall detection --
    ready_stall_for_read(bus_states : CPU_BUSES) : bool is {
        if( (bus_states.bbus.state.previous_request == request and bus_states.bbus.state.ready == no_ready) or 
            (bus_states.cbus.state.previous_request == request and bus_states.cbus.state.ready == no_ready) or
            (bus_states.dbus.state.previous_request == request and bus_states.dbus.state.ready == no_ready)    ) { 
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    ready_stall_for_write(bus_states : CPU_BUSES) : bool is {
        if( (bus_states.ebus.state.previous_request == request and bus_states.ebus.state.ready == no_ready) or 
            (bus_states.fbus.state.previous_request == request and bus_states.fbus.state.ready == no_ready)    ) { 
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    
    -- io stall detection --
    io_stall() : bool is {
        if( (address != NULL and address.space == io and address.dbus_active == active) and 
            ( (access1 != NULL and access1.space == io and access1.ebus_active == active) or
              (access2 != NULL and access2.space == io and access2.ebus_active == active) or
              (read    != NULL and read.space    == io and read.ebus_active    == active)   ) ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    
    -- read/write stall --
    read_stall( bus_states : CPU_BUSES ) : bool is {
        if( bus_states.read_inst_slot[ bus_states.read_inst_index ] != NULL and 
            ( (bus_states.bbus.pending.counter != 0 and
               bus_states.read_inst_slot[ bus_states.read_inst_index ] == bus_states.bbus.pending.progress[bus_states.bbus.pending.rdy_index]) or
              (bus_states.cbus.pending.counter != 0 and
               bus_states.read_inst_slot[ bus_states.read_inst_index ] == bus_states.cbus.pending.progress[bus_states.cbus.pending.rdy_index]) or
              (bus_states.dbus.pending.counter != 0 and
               bus_states.read_inst_slot[ bus_states.read_inst_index ] == bus_states.dbus.pending.progress[bus_states.dbus.pending.rdy_index])      ) ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    
    write_stall( bus_states : CPU_BUSES ) : bool is {
        if( bus_states.write_inst_slot[ bus_states.write_inst_index ] != NULL and 
            ( (bus_states.ebus.pending.counter != 0 and 
               bus_states.write_inst_slot[ bus_states.write_inst_index ] == bus_states.ebus.pending.progress[bus_states.ebus.pending.rdy_index]) or
              (bus_states.fbus.pending.counter != 0 and 
               bus_states.write_inst_slot[ bus_states.write_inst_index ] == bus_states.fbus.pending.progress[bus_states.fbus.pending.rdy_index])    ) ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    
    iodready_stall( bus_states : CPU_BUSES ) : bool is {
        if( (read != NULL and read.space == io and read.dbus_active == active) and 
            (read == bus_states.dbus.pending.progress[bus_states.dbus.pending.rdy_index]) ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    
    -- idle stall --
    !idle_stall_flag : bool;
    idle_stall( bus_states : CPU_BUSES ) : bool is {
        if( write != NULL and write.idle_flag == TRUE ) {
            idle_stall_flag = TRUE;
        };
        if( idle_stall_flag == TRUE and 
            bus_states.pbus.pending.counter == 0 and bus_states.bbus.pending.counter == 0 and
            bus_states.cbus.pending.counter == 0 and bus_states.dbus.pending.counter == 0 and
            bus_states.ebus.pending.counter == 0 and bus_states.fbus.pending.counter == 0     ) {
            idle_stall_flag = FALSE;
        };
        if( idle_stall_flag or 
            (decode  != NULL and decode.idle_flag  == TRUE) or
            (address != NULL and address.idle_flag == TRUE) or
            (access1 != NULL and access1.idle_flag == TRUE) or
            (access2 != NULL and access2.idle_flag == TRUE) or
            (read    != NULL and read.idle_flag    == TRUE) or
            (execute != NULL and execute.idle_flag == TRUE) or
            (write   != NULL and write.idle_flag   == TRUE)    ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    
    -- lock stall --
    lock_stall( bus_states : CPU_BUSES ) : bool is {
        if( -- clear the pipeline --
            ((address != NULL and address.lock == enable and address.dbus_active == active ) and 
             ((access1 != NULL and (access1.ebus_active == active or access1.fbus_active == active)) or
              (access2 != NULL and (access2.ebus_active == active or access2.fbus_active == active)) or
              (read    != NULL and (read.ebus_active    == active or read.fbus_active    == active))   )) or
            -- stall for next request --
            ((access1 != NULL and access1.lock == enable and access1.dbus_active != no_active) or
             (access2 != NULL and access2.lock == enable and access2.dbus_active != no_active) or
             (read    != NULL and read.lock    == enable and read.dbus_active    != no_active)    ) ) { 
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    
    -- abortready stall --
    abortready_stall( bus_states : CPU_BUSES ) : bool is {
        if( bus_states.emulation.current.abortready == 1 ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    
    -- suspend stall --
    suspend_stall( bus_states : CPU_BUSES ) : bool is {
        if( bus_states.emulation.current.suspend == 0 ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    
    -- indicate pipeline stall --
    pipeline_stall( bus_states : CPU_BUSES ) : bool is {
        if( bus_states.data_write_current_stall() or ready_stall_for_write(bus_states) or write_stall(bus_states) or iodready_stall(bus_states) or
            bus_states.check_lock_stall(read) or
            bus_states.data_read_current_stall() or d0_bypass.current_bypass_occur() or ready_stall_for_read(bus_states) or read_stall(bus_states) or
            io_stall() or lock_stall(bus_states) or
            idle_stall(bus_states) or
            bus_states.system.current.cpu_resetn == 0 or
            abortready_stall(bus_states) or suspend_stall(bus_states) ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    
    -- for bypass logic --
    current_bbus_bypass_occur() : bool is {
        if( d0_bypass.current_bbus_bypass_occur() or d1_bypass.current_bbus_bypass_occur() or d2_bypass.current_bbus_bypass_occur() ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    current_cbus_bypass_occur() : bool is {
        if( d0_bypass.current_cbus_bypass_occur() or d1_bypass.current_cbus_bypass_occur() or d2_bypass.current_cbus_bypass_occur() ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    current_dbus_bypass_occur() : bool is {
        if( d0_bypass.current_dbus_bypass_occur() or d1_bypass.current_dbus_bypass_occur() or d2_bypass.current_dbus_bypass_occur() ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    previous_bbus_bypass_occur() : bool is {
        if( d0_bypass.previous_bbus_bypass_occur() or d1_bypass.previous_bbus_bypass_occur() or d2_bypass.previous_bbus_bypass_occur() ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    previous_cbus_bypass_occur() : bool is {
        if( d0_bypass.previous_cbus_bypass_occur() or d1_bypass.previous_cbus_bypass_occur() or d2_bypass.previous_cbus_bypass_occur() ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    previous_dbus_bypass_occur() : bool is {
        if( d0_bypass.previous_dbus_bypass_occur() or d1_bypass.previous_dbus_bypass_occur() or d2_bypass.previous_dbus_bypass_occur() ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    bypass_data_treatment( bus_states : CPU_BUSES )  is {
        if( d0_bypass.previous_bbus_bypass_occur() )      { d0_bypass.read_data_treatment( bus_states, previous_access1, previous_access2 ); } 
        else if( d1_bypass.previous_bbus_bypass_occur() ) { d1_bypass.read_data_treatment( bus_states, previous_access1, previous_read    ); } 
        else if( d2_bypass.previous_bbus_bypass_occur() ) { d2_bypass.read_data_treatment( bus_states, previous_access1, previous_execute ); };
            
        if( d0_bypass.previous_cbus_bypass_occur() )      { d0_bypass.read_data_treatment( bus_states, previous_access1, previous_access2 ); }
        else if( d1_bypass.previous_cbus_bypass_occur() ) { d1_bypass.read_data_treatment( bus_states, previous_access1, previous_read    ); }
        else if( d2_bypass.previous_cbus_bypass_occur() ) { d2_bypass.read_data_treatment( bus_states, previous_access1, previous_execute ); };
        
        if( d0_bypass.previous_dbus_bypass_occur() )      { d0_bypass.read_data_treatment( bus_states, previous_access1, previous_access2 ); }
        else if( d1_bypass.previous_dbus_bypass_occur() ) { d1_bypass.read_data_treatment( bus_states, previous_access1, previous_read    ); }
        else if( d2_bypass.previous_dbus_bypass_occur() ) { d2_bypass.read_data_treatment( bus_states, previous_access1, previous_execute ); };
    };
    
    debug_output( bus_states : CPU_BUSES ) is {
        if( bus_states.bbus.state.request == request ) {
            if( d0_bypass.current_bbus_bypass_occur() )      { out("<d0_bypass> ",sys.time, " +++ bdismiss +++ ",access1," <-> ", access2); }
            else if( d1_bypass.current_bbus_bypass_occur() ) { out("<d0_bypass> ",sys.time, " +++ bdismiss +++ ",access1," <-> ", read   ); }
            else if( d2_bypass.current_bbus_bypass_occur() ) { out("<d0_bypass> ",sys.time, " +++ bdismiss +++ ",access1," <-> ", execute); }
        };
        if( bus_states.cbus.state.request == request ) {
            if( d0_bypass.current_cbus_bypass_occur() )      { out("<d0_bypass> ",sys.time, " +++ cdismiss +++ ",access1," <-> ", access2); }
            else if( d1_bypass.current_cbus_bypass_occur() ) { out("<d0_bypass> ",sys.time, " +++ cdismiss +++ ",access1," <-> ", read   ); }
            else if( d2_bypass.current_cbus_bypass_occur() ) { out("<d0_bypass> ",sys.time, " +++ cdismiss +++ ",access1," <-> ", execute); }
        };
        if( bus_states.dbus.state.request == request ) {
            if( d0_bypass.current_dbus_bypass_occur() )      { out("<d0_bypass> ",sys.time, " +++ ddismiss +++ ",access1," <-> ", access2); }
            else if( d1_bypass.current_dbus_bypass_occur() ) { out("<d0_bypass> ",sys.time, " +++ ddismiss +++ ",access1," <-> ", read   ); }
            else if( d2_bypass.current_dbus_bypass_occur() ) { out("<d0_bypass> ",sys.time, " +++ ddismiss +++ ",access1," <-> ", execute); }
        };
    };
    
    -- set initial values --
    init() is also {
        
        -- init pipeline --
        decode  = NULL;
        address = NULL;
        access1 = NULL;
        access2 = NULL;
        read    = NULL;
        execute = NULL;
        write   = NULL;
        
    };
    
};

---------------------------------
-- struct for bypass detection --
---------------------------------
struct BYPASS_ATTRIBUTES {
    !current_busFB : bool;
    !current_busFC : bool;
    !current_busFD : bool;
    !current_busEB : bool;
    !current_busEC : bool;
    !current_busED : bool;
    
    !previous_busFB : bool;
    !previous_busFC : bool;
    !previous_busFD : bool;
    !previous_busEB : bool;
    !previous_busEC : bool;
    !previous_busED : bool;
    
    -- bypass doesn't applied when byte write and double read access
    check_bypass_condition( bus_states : CPU_BUSES, read_attributes : DATA_ACCESS_ATTRIBUTES, write_attributes : DATA_ACCESS_ATTRIBUTES ) is {
        
        -- keep current values
        previous_busFB = current_busFB;
        previous_busFC = current_busFC;
        previous_busFD = current_busFD;
        previous_busEB = current_busEB;
        previous_busEC = current_busEC;
        previous_busED = current_busED;
        
        -- compare Fbus and Bbus 
        if( write_attributes             != NULL   and read_attributes             != NULL   and 
            write_attributes.write_width == word   and read_attributes.dblcoef     != double and
            write_attributes.space       == memory and read_attributes.space       == memory and
            write_attributes.fbus_active == active and read_attributes.bbus_active == active and 
            write_attributes.fabus == read_attributes.babus ) {
            current_busFB = TRUE;
        } else {
            current_busFB = FALSE;
        };
        
        -- compare Fbus and Cbus 
        if( write_attributes             != NULL   and read_attributes             != NULL   and 
            write_attributes.write_width == word   and read_attributes.doubler     != double and
            write_attributes.space       == memory and read_attributes.space       == memory and
            write_attributes.fbus_active == active and read_attributes.cbus_active == active and 
            write_attributes.fabus == read_attributes.cabus ) {
            current_busFC = TRUE;
        } else {
            current_busFC = FALSE;
        };
        
        -- compare Fbus and Dbus 
        if( write_attributes             != NULL   and read_attributes             != NULL   and 
            write_attributes.write_width == word   and read_attributes.read_width  == word   and
                                                       read_attributes.doubler     != double and 
            write_attributes.space       == memory and read_attributes.space       == memory and
            write_attributes.fbus_active == active and read_attributes.dbus_active == active and 
            write_attributes.fabus == read_attributes.dabus ) {
            current_busFD = TRUE;
        } else {
            current_busFD = FALSE;
        };
        
        -- compare Ebus and Bbus 
        if( write_attributes             != NULL   and read_attributes             != NULL   and 
            write_attributes.write_width == word   and read_attributes.dblcoef     != double and
            write_attributes.space       == memory and read_attributes.space       == memory and
            write_attributes.ebus_active == active and read_attributes.bbus_active == active and 
            write_attributes.eabus == read_attributes.babus ) {
            current_busEB = TRUE;
        } else {
            current_busEB = FALSE;
        };
        
        -- compare Ebus and Cbus 
        if( write_attributes             != NULL   and read_attributes             != NULL   and 
            write_attributes.write_width == word   and read_attributes.doubler     != double and
            write_attributes.space       == memory and read_attributes.space       == memory and
            write_attributes.ebus_active == active and read_attributes.cbus_active == active and 
            write_attributes.eabus == read_attributes.cabus ) {
            current_busEC = TRUE;
        } else {
            current_busEC = FALSE;
        };
        
        -- compare Ebus and Dbus 
        if( write_attributes             != NULL   and read_attributes             != NULL   and 
            write_attributes.write_width == word   and read_attributes.read_width  == word   and
                                                       read_attributes.doubler     != double and
            write_attributes.space       == memory and read_attributes.space       == memory and
            write_attributes.ebus_active == active and read_attributes.dbus_active == active and 
            write_attributes.eabus == read_attributes.dabus ) {
            current_busED = TRUE;
        } else {
            current_busED = FALSE;
        };
        
    };
    
    current_bbus_bypass_occur() : bool is {
        if( current_busFB or current_busEB ) { result = TRUE;  } 
                                        else { result = FALSE; };
    };
    current_cbus_bypass_occur() : bool is {
        if( current_busFC or current_busEC ) { result = TRUE;  } 
                                        else { result = FALSE; };
    };
    current_dbus_bypass_occur() : bool is {
        if( current_busFD or current_busED ) { result = TRUE;  } 
                                        else { result = FALSE; };
    };
    current_bypass_occur() : bool is {
        if( current_bbus_bypass_occur() or current_cbus_bypass_occur() or current_dbus_bypass_occur() ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    
    previous_bbus_bypass_occur() : bool is {
        if( previous_busFB or previous_busEB ) { result = TRUE;  } 
                                          else { result = FALSE; };
    };
    previous_cbus_bypass_occur() : bool is {
        if( previous_busFC or previous_busEC ) { result = TRUE;  } 
                                          else { result = FALSE; };
    };
    previous_dbus_bypass_occur() : bool is {
        if( previous_busFD or previous_busED ) { result = TRUE;  } 
                                          else { result = FALSE; };
    };
    previous_bypass_occur() : bool is {
        if( previous_bbus_bypass_occur() or previous_cbus_bypass_occur() or previous_dbus_bypass_occur() ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    
    read_data_treatment( bus_states : CPU_BUSES, read_attributes : DATA_ACCESS_ATTRIBUTES, write_attributes : DATA_ACCESS_ATTRIBUTES ) is {
        -- Note : if bypass occur when 2 diffrent data is writen to same address, Fbus has high priority.
        
        if( previous_busEB and bus_states.bbus.state.request == request ) {
            read_attributes.bbus_active = finished;
            read_attributes.bdbus       = write_attributes.edbus;
        };
        if( previous_busEC and bus_states.cbus.state.request == request ) {
            read_attributes.cbus_active = finished;
            read_attributes.cdbus       = write_attributes.edbus;
        };
        if( previous_busED and bus_states.dbus.state.request == request ) {
            read_attributes.dbus_active = finished;
            read_attributes.ddbus       = write_attributes.edbus;
        };
        if( previous_busFB and bus_states.bbus.state.request == request ) {
            read_attributes.bbus_active = finished;
            read_attributes.bdbus       = write_attributes.fdbus;
        };
        if( previous_busFC and bus_states.cbus.state.request == request ) {
            read_attributes.cbus_active = finished;
            read_attributes.cdbus       = write_attributes.fdbus;
        };
        if( previous_busFD and bus_states.dbus.state.request == request ) {
            read_attributes.dbus_active = finished;
            read_attributes.ddbus       = write_attributes.fdbus;
        };
    };
    
    post_generate() is also {
        current_busFB  = FALSE;
        current_busFC  = FALSE;
        current_busFD  = FALSE;
        current_busEB  = FALSE;
        current_busEC  = FALSE;
        current_busED  = FALSE;
        previous_busFB = FALSE;
        previous_busFC = FALSE;
        previous_busFD = FALSE;
        previous_busEB = FALSE;
        previous_busEC = FALSE;
        previous_busED = FALSE;
    };

};

-------------------------------
-- struct for program buffer --
-------------------------------
unit PROGRAM_BUFFER {
    !current_prog_attributes  : PROGRAM_ACCESS_ATTRIBUTES;
    !previous_prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
    
    !keep_dismiss_flag    : bool;
    !keep_dismiss_address : uint(bits : 24);
    
    !dismiss_waiting_counter : uint;
    
    buffer_update : bool;
    
    isr_queue     : list of uint(bits: 5);
    
    transition( next_attributes : PROGRAM_ACCESS_ATTRIBUTES, attributes : PROGRAM_ACCESS_ATTRIBUTES, bus_states : CPU_BUSES ) is {
        
        -- set pbus instruction --
        previous_prog_attributes = current_prog_attributes;
        if( !buffer_stall( bus_states, next_attributes ) ) {
            current_prog_attributes  = attributes;
        };
        
        buffer_update = ( current_prog_attributes != previous_prog_attributes );
        
        -- dismiss waiting counter --
      --if( next_attributes.branch_occur and next_attributes.dismiss_occur and !buffer_update ) {
      --    dismiss_waiting_counter = dismiss_waiting_counter + 1;
      --} else {
      --    dismiss_waiting_counter = 0;
      --};
        dismiss_waiting_counter += ( next_attributes.branch_occur and next_attributes.dismiss_occur and !buffer_update ) ? 1 : 0;
        
    };
    
    buffer_stall( bus_states : CPU_BUSES, next_attributes : PROGRAM_ACCESS_ATTRIBUTES ) : bool is {
        if( bus_states.emulation.current.abortready == 1 or 
            bus_states.emulation.current.suspend == 0 or 
            bus_states.system.current.cpu_resetn == 0 or
            -- Pdismiss doesn't occur when P bus DT-DMA access exists
            ( ( (    bus_states.pbus.pending.progress[0] != NULL and 
                     bus_states.pbus.pending.progress[0].pbus_active == active and bus_states.pbus.pending.progress[0].pbus_type == debug )
                or ( bus_states.pbus.pending.progress[1] != NULL and
                     bus_states.pbus.pending.progress[1].pbus_active == active and bus_states.pbus.pending.progress[1].pbus_type == debug ))
              and bus_states.pbus.pending.counter > 0 and next_attributes.dismiss_occur ) or
            (bus_states.curr_idle_stall or bus_states.system.current.idlemaster == 1 or bus_states.system.previous.idlemaster == 1) or
            ( bus_states.pbus_current_rdystall()  and !(next_attributes.branch_occur and next_attributes.dismiss_occur)) or 
            ( bus_states.pbus.pending.counter > 0 and  (next_attributes.branch_occur and next_attributes.dismiss_occur) and 
              dismiss_waiting_counter < next_attributes.dismiss_waiting ) ) {
            result = TRUE;
        } else {
            result = FALSE;
        };
    };
    
    init() is also {
        current_prog_attributes  = NULL;
        previous_prog_attributes = NULL;
        isr_queue.clear();
    };
    
};

-----------------------------
-- output bus informations --
-----------------------------
extend CPU_BUSES {
    
    output_bus_informations( pipeline : CPU_PIPELINE, buffer : PROGRAM_BUFFER ) is {
        var any_requests_occur       : bool;
        var any_previous_ready_occur : bool;
        
        if( pbus.state.request == request or bbus.state.request == request or cbus.state.request == request or
            dbus.state.request == request or ebus.state.request == request or fbus.state.request == request    ) { 
            any_requests_occur = TRUE;
        } else {
            any_requests_occur = FALSE;
        };
        
        if( pbus.state.previous_ready == ready or bbus.state.previous_ready == ready or cbus.state.previous_ready == ready or
            dbus.state.previous_ready == ready or ebus.state.previous_ready == ready or fbus.state.previous_ready == ready    ) { 
            any_previous_ready_occur = TRUE;
        } else {
            any_previous_ready_occur = FALSE;
        };
        
        if( any_requests_occur or any_previous_ready_occur ) { outf("%s : current CPU buses status \n", sys.time); };
        -- request infomations --
        var temp_abus : uint(bits:23);
        
        if( pbus.state.request == request ) {
            temp_abus = buffer.current_prog_attributes.pabus;
            outf(" - Pbus request : address = 0x%06x \n",temp_abus );
        };
        if( bbus.state.request == request ) {
            temp_abus = pipeline.access1.babus;
            outf(" - Bbus request : address = 0x%06x \n",temp_abus );
        };
        if( cbus.state.request == request ) {
            temp_abus = pipeline.access1.cabus;
            outf(" - Cbus request : address = 0x%06x \n",temp_abus );
        };
        if( dbus.state.request == request ) {
            temp_abus = pipeline.access1.dabus;
            outf(" - Dbus request : address = 0x%06x \n",temp_abus );
        };
        if( ebus.state.request == request ) {
            temp_abus = pipeline.execute.eabus;
            outf(" - Ebus request : address = 0x%06x data = 0x%04x \n", temp_abus, pipeline.execute.edbus );
        };
        if( fbus.state.request == request ) {
            temp_abus = pipeline.execute.fabus;
            outf(" - Fbus request : address = 0x%06x data = 0x%04x \n", temp_abus, pipeline.execute.fdbus );
        };
        
        if( bbus.state.request == request ) {
            if( pipeline.access1 != NULL and pipeline.access1.bbus_active == active and system.current.cpu_resetn != 0 ) {
                if( pipeline.access1.dblcoef     == double ) { outf(" - (qualifier)dblcoef read     \n"); };
            };
        };
        if( dbus.state.request == request ) {
            if( pipeline.access1 != NULL and pipeline.access1.dbus_active == active and system.current.cpu_resetn != 0 ) {
                if( pipeline.access1.doubler     == double    ) { outf(" - (qualifier)double read     \n"); };
                if( pipeline.access1.read_width  == high_byte ) { outf(" - (qualifier)high byte read  \n"); };
                if( pipeline.access1.read_width  == low_byte  ) { outf(" - (qualifier)low byte read   \n"); };
                if( pipeline.access1.space       == io        ) { outf(" - (qualifier)io space read   \n"); };
            };
        };
        if( ebus.state.request == request ) {
            if( pipeline.execute != NULL and pipeline.execute.ebus_active == active and system.current.cpu_resetn != 0 ) {
                if( pipeline.execute.doublew     == double    ) { outf(" - (qualifier)double write    \n"); };
                if( pipeline.execute.write_width == high_byte ) { outf(" - (qualifier)high byte write \n"); };
                if( pipeline.execute.write_width == low_byte  ) { outf(" - (qualifier)low byte write  \n"); };
                if( pipeline.execute.space       == io        ) { outf(" - (qualifier)io space write  \n"); };
            };
        };
        if( any_requests_occur ) {
            if( (pipeline.access1 != NULL and pipeline.access1.lock == enable) or 
                (pipeline.execute != NULL and pipeline.execute.lock == enable)    ) { outf(" - (qualifier)lock access     \n"); };
        };
        
        -- reset informations --
        if( system.current.cpu_resetn == 0 ) {
            outf("%s : current CPU buses status \n",sys.time);
            outf( " - Reset assertion \n" );
        };
        
        -- ready informations --
        if( pbus.state.previous_ready == ready ) {
            temp_abus = pbus.pending.finish.pabus;
            outf(" - Pbus ready/data : address = 0x%06x data = 0x%08x \n", temp_abus, pbus.pending.finish.pdbus );
        };
        if( bbus.state.previous_ready == ready ) {
            temp_abus = bbus.pending.finish.babus;
            outf(" - Bbus ready/data : address = 0x%06x data = 0x%08x \n", temp_abus, bbus.pending.finish.bdbus );
        };
        if( cbus.state.previous_ready == ready ) {
            temp_abus = cbus.pending.finish.cabus;
            outf(" - Cbus ready/data : address = 0x%06x data = 0x%04x \n", temp_abus, cbus.pending.finish.cdbus );
        };
        if( dbus.state.previous_ready == ready ) {
            temp_abus = dbus.pending.finish.dabus;
            outf(" - Dbus ready/data : address = 0x%06x data = 0x%04x \n", temp_abus, dbus.pending.finish.ddbus );
        };
        if( ebus.state.previous_ready == ready ) {
            temp_abus = ebus.pending.finish.eabus;
            outf(" - Ebus ready : address = 0x%06x data = 0x%04x \n", temp_abus, ebus.pending.finish.edbus );
        };
        if( fbus.state.previous_ready == ready ) {
            temp_abus = fbus.pending.finish.fabus;
            outf(" - Fbus ready : address = 0x%06x data = 0x%04x \n", temp_abus, fbus.pending.finish.fdbus );
        };
        
        -- bypass informations --
        var bypass_message : string;
        if( bbus.state.request == request and pipeline.current_bbus_bypass_occur() ) {
            if( pipeline.d2_bypass.current_busEB ) { bypass_message = " - (bypass) Bbus ddd-bypass : Ebus write data --> Bbus read data \n"; };
            if( pipeline.d2_bypass.current_busFB ) { bypass_message = " - (bypass) Bbus ddd-bypass : Fbus write data --> Bbus read data \n"; };
            if( pipeline.d1_bypass.current_busEB ) { bypass_message = " - (bypass) Bbus  dd-bypass : Ebus write data --> Bbus read data \n"; };
            if( pipeline.d1_bypass.current_busFB ) { bypass_message = " - (bypass) Bbus  dd-bypass : Fbus write data --> Bbus read data \n"; };
            if( pipeline.d0_bypass.current_busEB ) { bypass_message = " - (bypass) Bbus   d-bypass : Ebus write data --> Bbus read data \n"; };
            if( pipeline.d0_bypass.current_busFB ) { bypass_message = " - (bypass) Bbus   d-bypass : Fbus write data --> Bbus read data \n"; };
            outf( bypass_message );
        };
        if( cbus.state.request == request and pipeline.current_cbus_bypass_occur() ) {
            if( pipeline.d2_bypass.current_busEC ) { bypass_message = " - (bypass) Cbus ddd-bypass : Ebus write data --> Cbus read data \n"; };
            if( pipeline.d2_bypass.current_busFC ) { bypass_message = " - (bypass) Cbus ddd-bypass : Fbus write data --> Cbus read data \n"; };
            if( pipeline.d1_bypass.current_busEC ) { bypass_message = " - (bypass) Cbus  dd-bypass : Ebus write data --> Cbus read data \n"; };
            if( pipeline.d1_bypass.current_busFC ) { bypass_message = " - (bypass) Cbus  dd-bypass : Fbus write data --> Cbus read data \n"; };
            if( pipeline.d0_bypass.current_busEC ) { bypass_message = " - (bypass) Cbus   d-bypass : Ebus write data --> Cbus read data \n"; };
            if( pipeline.d0_bypass.current_busFC ) { bypass_message = " - (bypass) Cbus   d-bypass : Fbus write data --> Cbus read data \n"; };
            outf( bypass_message );
        };
        if( dbus.state.request == request and pipeline.current_dbus_bypass_occur() ) {
            if( pipeline.d2_bypass.current_busED ) { bypass_message = " - (bypass) Dbus ddd-bypass : Ebus write data --> Dbus read data \n"; };
            if( pipeline.d2_bypass.current_busFD ) { bypass_message = " - (bypass) Dbus ddd-bypass : Fbus write data --> Dbus read data \n"; };
            if( pipeline.d1_bypass.current_busED ) { bypass_message = " - (bypass) Dbus  dd-bypass : Ebus write data --> Dbus read data \n"; };
            if( pipeline.d1_bypass.current_busFD ) { bypass_message = " - (bypass) Dbus  dd-bypass : Fbus write data --> Dbus read data \n"; };
            if( pipeline.d0_bypass.current_busED ) { bypass_message = " - (bypass) Dbus   d-bypass : Ebus write data --> Dbus read data \n"; };
            if( pipeline.d0_bypass.current_busFD ) { bypass_message = " - (bypass) Dbus   d-bypass : Fbus write data --> Dbus read data \n"; };
            outf( bypass_message );
        };
        
        -- pdismiss informations --
        if( pbus.state.request == request and 
            (buffer.current_prog_attributes.branch_occur or buffer.keep_dismiss_flag) ) {
            outf(" - (Branch occur) pbus address : %x \n",pbus.current.abus );
        };
        if( pbus.current.dismiss == 1 ) {
            outf(" - (Pdismiss occur)\n");
        };
        
    };
};
'>
