//********************************************************************************
//
//  Filename: cpubus_slave_driver.e
//
//  Description: cpubus slave driver
//
//  Major Revision Descriptions:
//  oishi          09/08/02         copied from UMA1.0 branch
//  oishi          09/09/02         removed mdp related functions
//                                  removed ack related functions
//                                  And major function change
//  oishi          09/11/02         moved bfm extend part to cpubus_slave_structures.e
//  oishi          09/11/02         changed bdbus width to 32bits
//  oishi          09/13/02         modified resetn to cpu_resetn and gl_resetn
//  oishi          10/02/02         added address decoder function for override
//  oishi          10/03/02         modified address decoder
//  oishi          10/04/02         added ready synchronizer for synchronized requests
//  oishi          10/04/02         added checking function of signal connection.
//
//*********************************************************************************
<'
import ./cpubus_slave_structures;

-----------------------------
-- simple memory structure --
-----------------------------
struct MEMORY_8k {
    memory[ 8*1024 ] : list of uint(bits: 16);
    
    put_memory( address : uint(bits : 32), data : uint(bits: 16) ) is {
        memory[ address[12:0] ] = data;
    };
    
    get_memory( address : uint(bits : 32) ) : uint(bits : 16 ) is {
        result = memory[ address[12:0] ]
    };
    
};

-------------------------
-- cpubus slave driver --
-------------------------
unit CPUBUS_SLAVE_DRIVER {
    -- enable/disable flag --
    disable  : bool; 
    keep soft disable == FALSE;
    
    -- global configurations --
    config   : CPUBUS_CONFIGURATIONS;
    paths    : CPU_SIGNAL_PATH;
    
    -- simple memory structure configurations --
    memory_enable : bool;        keep soft memory_enable == TRUE;
    memory8k      : MEMORY_8k;
    
    -- local structures --
    bus_states : CPU_BUSES;
    bus_driver : CPU_SIGNALS;
    
    !prog_attributes : PROGRAM_ACCESS_ATTRIBUTES;
    !data_attributes : DATA_ACCESS_ATTRIBUTES;
    
    -- set clock event for state transitions
    event raw_rise_clock is rise(paths.clock$)    @sim;
    event rise_clock     is true( disable == FALSE ) @raw_rise_clock;
    
    -- xbus_wait_ready(uint)      : indicate ready wait cycles
    -- 
    -- < Example - ready base protocol - >
    --   xbus_wait_ready     = 3
    --                __    __    __    __    __    __    __ 
    --   clk       __|  |__|  |__|  |__|  |__|  |__|  |__|  |
    --             ___       _____       ____________________
    --   requset      |_____|     |_____|                    
    --             _____________________       _____       __
    --   ready                          |_____|     |_____|  
    --
    -- 
    
    pbus_data : uint (bits: 32);
    bbus_data : uint (bits: 32);
    cbus_data : uint (bits: 16);
    dbus_data : uint (bits: 16);
    
    pbus_wait_ready : uint;  keep soft pbus_wait_ready < 5; keep pbus_wait_ready >= 0;
    bbus_wait_ready : uint;  keep soft bbus_wait_ready < 5; keep bbus_wait_ready >= 0;
    cbus_wait_ready : uint;  keep soft cbus_wait_ready < 5; keep cbus_wait_ready >= 0;
    dbus_wait_ready : uint;  keep soft dbus_wait_ready < 5; keep dbus_wait_ready >= 0;
    ebus_wait_ready : uint;  keep soft ebus_wait_ready < 5; keep ebus_wait_ready >= 0;
    fbus_wait_ready : uint;  keep soft fbus_wait_ready < 5; keep fbus_wait_ready >= 0;
    
    pready_queue : list of uint(bits: 1); pdbus_queue : list of uint(bits: 32);
    bready_queue : list of uint(bits: 1); bdbus_queue : list of uint(bits: 32);
    cready_queue : list of uint(bits: 1); cdbus_queue : list of uint(bits: 16);
    dready_queue : list of uint(bits: 1); ddbus_queue : list of uint(bits: 16);
    eready_queue : list of uint(bits: 1);
    fready_queue : list of uint(bits: 1);
    
    preq_index : uint(bits: 2); preq_queue : list of uint(bits: 2);
    breq_index : uint(bits: 2); breq_queue : list of uint(bits: 2);
    creq_index : uint(bits: 2); creq_queue : list of uint(bits: 2);
    dreq_index : uint(bits: 2); dreq_queue : list of uint(bits: 2);
    ereq_index : uint(bits: 2); ereq_queue : list of uint(bits: 2);
    freq_index : uint(bits: 2); freq_queue : list of uint(bits: 2);
    
    local_pstall : uint(bits: 1);
    local_rstall : uint(bits: 1);
    local_wstall : uint(bits: 1);
    
    post_generate() is also {

        -- check signal connection --
        paths.check_signal_connection( config.id );
        
        -- initialize --
        pready_queue.clear();
        bready_queue.clear();
        cready_queue.clear();
        dready_queue.clear();
        eready_queue.clear();
        fready_queue.clear();
        
        pdbus_queue.clear();
        bdbus_queue.clear();
        cdbus_queue.clear();
        ddbus_queue.clear();
        
        preq_queue.clear();
        breq_queue.clear();
        creq_queue.clear();
        dreq_queue.clear();
        ereq_queue.clear();
        freq_queue.clear();
        
        preq_index = 0;
        breq_index = 0;
        creq_index = 0;
        dreq_index = 0;
        ereq_index = 0;
        freq_index = 0;
        
        local_pstall = 0;
        local_rstall = 0;
        local_wstall = 0;
    };
    
    ------------------------------
    -- cpu bus state transition --
    ------------------------------
    on rise_clock
    {   
        -- keep current values
        bus_states.keep_current_values();
        
        -- get current value
        bus_states.get_current_value( config, paths );
        
        -- state transition
        bus_states.state_transition();
        
        -- set current request attributes --
        if( bus_states.ocuur_program_request() ) {
            gen prog_attributes; 
            prog_attributes.set_current_attributes( bus_states );
        };
        if( bus_states.ocuur_data_read_requests() or bus_states.ocuur_data_write_requests() ) {
            gen data_attributes; 
            data_attributes.set_current_attributes( bus_states );
        };
        
        -- for keeping pending requests --
        bus_states.pending_prog_requests( prog_attributes );
        bus_states.pending_data_requests( data_attributes, data_attributes );
        
        -- state initialize when reset 
        if( bus_states.system.current.cpu_resetn == 0 ) {
            bus_states.set_reset_values()
        };
        
    };
    
    ---------------------
    -- address decoder --
    ---------------------
    -- 
    -- user should extend these function if need.
    -- 
    decoder_prog( is : string, address : uint(bits : 24) ) : bool is {
        result = TRUE;
    };
    decoder_data( id : string, address : uint(bits : 23) ) : bool is {
        result = TRUE;
    };
    decoder_io( id : string, address : uint(bits : 23) ) : bool is {
        result = TRUE;
    };
    
    -----------------------------------
    -- ready/data driveing functions --
    -----------------------------------
    TCM_pbus_driving( cycles : uint ) @rise_clock is {
        var index      : uint(bits: 2);
        var attributes : PROGRAM_ACCESS_ATTRIBUTES;
        
        -- keep attributes --
        attributes = prog_attributes;
        
        -- push current index --
        preq_queue.push( preq_index );
        index = preq_index;
        preq_index = preq_index + 1;
        
        -- wait cycles --
        wait [ cycles ]*cycle;
        
        -- waiting pending ready/stall --
        while( attributes != bus_states.pbus.pending.progress[bus_states.pbus.pending.rdy_index] ) { wait cycle; };
        
        -- set read data --
        gen pbus_data;
        
        -- set driving queues --
        pready_queue.push(0);        wait cycle; pready_queue.push(1);
        pdbus_queue.push(pbus_data); wait cycle; pdbus_queue.push(0);
    };
    TCM_bbus_driving( cycles : uint ) @rise_clock is {
        var index      : uint(bits: 2);
        var attributes : DATA_ACCESS_ATTRIBUTES;
        var bbus_flag  : bool;
        var cbus_flag  : bool;
        var dbus_flag  : bool;
        
        bbus_flag = FALSE;
        cbus_flag = FALSE;
        dbus_flag = FALSE;
        if( bus_states.bbus.state.request == request ) { bbus_flag = decoder_data(config.id, bus_states.bbus.current.abus); };
        if( bus_states.cbus.state.request == request ) { cbus_flag = decoder_data(config.id, bus_states.cbus.current.abus); };
        if( bus_states.dbus.state.request == request )
            { if( bus_states.dbus.current.ioreq != 0 ) { dbus_flag = decoder_data(config.id, bus_states.dbus.current.abus); }
                                                  else { dbus_flag = decoder_io(config.id, bus_states.dbus.current.abus); }; };
        
        -- keep attributes --
        attributes = data_attributes;
        
        -- push current index --
        breq_queue.push( breq_index );
        index = breq_index;
        breq_index = breq_index + 1;
        
        -- wait cycles --
        wait [ cycles ]*cycle;
        
        -- waiting pending ready/stall --
        while( attributes != bus_states.bbus.pending.progress[bus_states.bbus.pending.rdy_index] ) { wait cycle; };
        attributes.bbus_active = finished;
        
        -- synchronize another ready --
        wait delay(0); -- dummy delay --
        
        while( (bbus_flag and attributes.bbus_active == active) or 
               (cbus_flag and attributes.cbus_active == active) or
               (              attributes.dbus_active == active)    ) { wait cycle; };
        
        -- set read data --
        gen bbus_data;
        
        -- set driving queues --
        bready_queue.push(0);        wait cycle; bready_queue.push(1);
        bdbus_queue.push(bbus_data); wait cycle; bdbus_queue.push(0);
    };
    TCM_cbus_driving( cycles : uint ) @rise_clock is {
        var index      : uint(bits: 2);
        var attributes : DATA_ACCESS_ATTRIBUTES;
        var bbus_flag  : bool;
        var cbus_flag  : bool;
        var dbus_flag  : bool;
        
        bbus_flag = FALSE;
        cbus_flag = FALSE;
        dbus_flag = FALSE;
        if( bus_states.bbus.state.request == request ) { bbus_flag = decoder_data(config.id, bus_states.bbus.current.abus); };
        if( bus_states.cbus.state.request == request ) { cbus_flag = decoder_data(config.id, bus_states.cbus.current.abus); };
        if( bus_states.dbus.state.request == request )
            { if( bus_states.dbus.current.ioreq != 0 ) { dbus_flag = decoder_data(config.id, bus_states.dbus.current.abus); }
                                                  else { dbus_flag = decoder_io(config.id, bus_states.dbus.current.abus); }; };
        
        -- keep attributes --
        attributes = data_attributes;
        
        -- push current index --
        creq_queue.push( creq_index );
        index = creq_index;
        creq_index = creq_index + 1;
        
        -- wait cycles --
        wait [ cycles ]*cycle;
        
        -- waiting pending ready/stall --
        while( attributes != bus_states.cbus.pending.progress[bus_states.cbus.pending.rdy_index] ) { wait cycle; };
        attributes.cbus_active = finished;
        
        -- synchronize another ready --
        wait delay(0); -- dummy delay --
        
        while( (bbus_flag and attributes.bbus_active == active) or 
               (cbus_flag and attributes.cbus_active == active) or
               (              attributes.dbus_active == active)    ) { wait cycle; };
        
        -- set read data --
        gen cbus_data;
        
        -- set driving queues --
        cready_queue.push(0);        wait cycle; cready_queue.push(1);
        cdbus_queue.push(cbus_data); wait cycle; cdbus_queue.push(0);
    };
    TCM_dbus_driving( cycles : uint ) @rise_clock is {
        var index      : uint(bits: 2);
        var attributes : DATA_ACCESS_ATTRIBUTES;
        var bbus_flag  : bool;
        var cbus_flag  : bool;
        var dbus_flag  : bool;
        
        bbus_flag = FALSE;
        cbus_flag = FALSE;
        dbus_flag = FALSE;
        if( bus_states.bbus.state.request == request ) { bbus_flag = decoder_data(config.id, bus_states.bbus.current.abus); };
        if( bus_states.cbus.state.request == request ) { cbus_flag = decoder_data(config.id, bus_states.cbus.current.abus); };
        if( bus_states.dbus.state.request == request )
            { if( bus_states.dbus.current.ioreq != 0 ) { dbus_flag = decoder_data(config.id, bus_states.dbus.current.abus); }
                                                  else { dbus_flag = decoder_io(config.id, bus_states.dbus.current.abus); }; };
        
        -- keep attributes --
        attributes = data_attributes;
        
        -- push current index --
        dreq_queue.push( dreq_index );
        index = dreq_index;
        dreq_index = dreq_index + 1;
        
        -- wait cycles --
        wait [ cycles ]*cycle;
        
        -- waiting pending ready/stall --
        while( attributes != bus_states.dbus.pending.progress[bus_states.dbus.pending.rdy_index] ) { wait cycle; };
        attributes.dbus_active = finished;
        
        -- synchronize another ready --
        wait delay(0); -- dummy delay --
        while( (bbus_flag and attributes.bbus_active == active) or 
               (cbus_flag and attributes.cbus_active == active) or
               (              attributes.dbus_active == active)    ) { wait cycle; };
        
        -- set read data --
        gen dbus_data;
        if( attributes.doubler == double ) { gen cbus_data; };
        
        -- set driving queues --
        dready_queue.push(0);        wait cycle; dready_queue.push(1);
        ddbus_queue.push(dbus_data); if( attributes.doubler == double ) { cdbus_queue.push(cbus_data); };
        wait cycle; 
        ddbus_queue.push(0);         if( attributes.doubler == double ) { cdbus_queue.push(0); };
    };
    TCM_ebus_driving( cycles : uint ) @rise_clock is {
        var index      : uint(bits: 2);
        var attributes : DATA_ACCESS_ATTRIBUTES;
        var ebus_flag  : bool;
        var fbus_flag  : bool;
        
        ebus_flag = FALSE;
        fbus_flag = FALSE;
        if( bus_states.fbus.state.request == request ) { fbus_flag = decoder_data(config.id, bus_states.fbus.current.abus); };
        if( bus_states.ebus.state.request == request ) 
            { if( bus_states.ebus.current.ioreq != 0 ) { ebus_flag = decoder_data(config.id, bus_states.ebus.current.abus); }
                                                  else { ebus_flag = decoder_io(config.id, bus_states.ebus.current.abus); }; };
        
        -- keep attributes --
        attributes = data_attributes;
        
        -- push current index --
        ereq_queue.push( ereq_index );
        index = ereq_index;
        ereq_index = ereq_index + 1;
        
        -- wait cycles --
        wait [ cycles ]*cycle;
        
        -- waiting pending ready/stall --
        while( attributes != bus_states.ebus.pending.progress[bus_states.ebus.pending.rdy_index] ) { wait cycle; };
        attributes.ebus_active = finished;
        
        -- synchronize another ready --
        wait delay(0); -- dummy delay --
        while( (ebus_flag and attributes.ebus_active == active) or 
               (fbus_flag and attributes.fbus_active == active)    ) { wait cycle; };
        
        -- set driving queues --
        eready_queue.push(0); wait cycle; eready_queue.push(1);
    };
    TCM_fbus_driving( cycles : uint ) @rise_clock is {
        var index      : uint(bits: 2);
        var attributes : DATA_ACCESS_ATTRIBUTES;
        var ebus_flag  : bool;
        var fbus_flag  : bool;
        
        ebus_flag = FALSE;
        fbus_flag = FALSE;
        if( bus_states.fbus.state.request == request ) { fbus_flag = decoder_data(config.id, bus_states.fbus.current.abus); };
        if( bus_states.ebus.state.request == request ) 
            { if( bus_states.ebus.current.ioreq != 0 ) { ebus_flag = decoder_data(config.id, bus_states.ebus.current.abus); }
                                                  else { ebus_flag = decoder_io(config.id, bus_states.ebus.current.abus); }; };
        
        -- keep attributes --
        attributes = data_attributes;
        
        -- push current index --
        freq_queue.push( freq_index );
        index = freq_index;
        freq_index = freq_index + 1;
        
        -- wait cycles --
        wait [ cycles ]*cycle;
        
        -- waiting pending ready/stall --
        while( attributes != bus_states.fbus.pending.progress[bus_states.fbus.pending.rdy_index] ) { wait cycle; };
        attributes.fbus_active = finished;
        
        -- synchronize another ready --
        wait delay(0); -- dummy delay --
        
        while( (ebus_flag and attributes.ebus_active == active) or 
               (fbus_flag and attributes.fbus_active == active)    ) { wait cycle; };
        
        -- set driving queues --
        fready_queue.push(0); wait cycle; fready_queue.push(1);
    };
    
    ------------------
    -- driving data --
    ------------------
    TCM_slave_driver() @rise_clock is {
        var local_pready : uint(bits: 1);
        var local_bready : uint(bits: 1);
        var local_cready : uint(bits: 1);
        var local_dready : uint(bits: 1);
        var local_eready : uint(bits: 1);
        var local_fready : uint(bits: 1);
        
        var local_pdbus : uint(bits: 32);
        var local_bdbus : uint(bits: 32);
        var local_cdbus : uint(bits: 16);
        var local_ddbus : uint(bits: 16);
        
        var icemaker_read  : bool;
        var icemaker_write : bool;
        
        var pbus_decode    : bool;
        var bbus_decode    : bool;
        var cbus_decode    : bool;
        var dbus_decode    : bool;
        var ebus_decode    : bool;
        var fbus_decode    : bool;
        
        var index          : uint(bits: 2);
        
        while( TRUE ) {
            
            -- pdismiss handling --
            if( bus_states.pbus.current.dismiss == 1 ) {
                var hs := scheduler.get_handles_by_type("CPUBUS_SLAVE_DRIVER","TCM_pbus_driving");
                for each (h) in hs { scheduler.kill(h); };
                preq_queue.clear();
                pready_queue.clear();
                pdbus_queue.clear();
            };
            wait delay(0); -- dummy delay --
            
            -- check for icemaker access --
            if( bus_states.dbus.current.ioreq == 0 and bus_states.dbus.current.abus is in ICEMAKER_IOMMR ) 
                { icemaker_read  = TRUE; } else { icemaker_read  = FALSE; };
            if( bus_states.ebus.current.ioreq == 0 and bus_states.ebus.current.abus is in ICEMAKER_IOMMR ) 
                { icemaker_write = TRUE; } else { icemaker_write = FALSE; };
            
            -- check for address decoder --
            pbus_decode = TRUE;
            bbus_decode = TRUE;
            cbus_decode = TRUE;
            dbus_decode = TRUE;
            ebus_decode = TRUE;
            fbus_decode = TRUE;
            if( bus_states.pbus.state.request == request ) { pbus_decode = decoder_prog(config.id, bus_states.pbus.current.abus); };
            if( bus_states.bbus.state.request == request ) { bbus_decode = decoder_data(config.id, bus_states.bbus.current.abus); };
            if( bus_states.cbus.state.request == request ) { cbus_decode = decoder_data(config.id, bus_states.cbus.current.abus); };
            if( bus_states.fbus.state.request == request ) { fbus_decode = decoder_data(config.id, bus_states.fbus.current.abus); };
            if( bus_states.dbus.state.request == request )
                { if( bus_states.dbus.current.ioreq != 0 ) { dbus_decode = decoder_data(config.id, bus_states.dbus.current.abus); }
                                                      else { dbus_decode = decoder_io(config.id, bus_states.dbus.current.abus); }; };
            if( bus_states.ebus.state.request == request ) 
                { if( bus_states.ebus.current.ioreq != 0 ) { ebus_decode = decoder_data(config.id, bus_states.ebus.current.abus); }
                                                      else { ebus_decode = decoder_io(config.id, bus_states.ebus.current.abus); }; };
            
            -- call TCM function for returning ready and data --
            if( bus_states.pbus.state.request == request and pbus_decode                     ) { gen pbus_wait_ready; start TCM_pbus_driving( pbus_wait_ready ); };
            if( bus_states.bbus.state.request == request and bbus_decode                     ) { gen bbus_wait_ready; start TCM_bbus_driving( bbus_wait_ready ); };
            if( bus_states.cbus.state.request == request and cbus_decode                     ) { gen cbus_wait_ready; start TCM_cbus_driving( cbus_wait_ready ); };
            if( bus_states.dbus.state.request == request and dbus_decode and !icemaker_read  ) { gen dbus_wait_ready; start TCM_dbus_driving( dbus_wait_ready ); };
            if( bus_states.ebus.state.request == request and ebus_decode and !icemaker_write ) { gen ebus_wait_ready; start TCM_ebus_driving( ebus_wait_ready ); };
            if( bus_states.fbus.state.request == request and fbus_decode                     ) { gen fbus_wait_ready; start TCM_fbus_driving( fbus_wait_ready ); };
            
            -- delta delay --
            wait delay(config.driver_delay);
            
            ------------------------------
            -- get and set driving data --
            ------------------------------
            -- set ready signals --
            local_pready = 1; while( pready_queue.size() != 0 ) { local_pready = local_pready & pready_queue.pop(); };
            local_bready = 1; while( bready_queue.size() != 0 ) { local_bready = local_bready & bready_queue.pop(); };
            local_cready = 1; while( cready_queue.size() != 0 ) { local_cready = local_cready & cready_queue.pop(); };
            local_dready = 1; while( dready_queue.size() != 0 ) { local_dready = local_dready & dready_queue.pop(); };
            local_eready = 1; while( eready_queue.size() != 0 ) { local_eready = local_eready & eready_queue.pop(); };
            local_fready = 1; while( fready_queue.size() != 0 ) { local_fready = local_fready & fready_queue.pop(); };
            
            if( local_pready == 0 ) { index = preq_queue.pop0(); };
            if( local_bready == 0 ) { index = breq_queue.pop0(); };
            if( local_cready == 0 ) { index = creq_queue.pop0(); };
            if( local_dready == 0 ) { index = dreq_queue.pop0(); };
            if( local_eready == 0 ) { index = ereq_queue.pop0(); };
            if( local_fready == 0 ) { index = freq_queue.pop0(); };
            
            -- set data signals --
            local_pdbus = 0; while( pdbus_queue.size() != 0 ) { local_pdbus = local_pdbus | pdbus_queue.pop(); };
            local_bdbus = 0; while( bdbus_queue.size() != 0 ) { local_bdbus = local_bdbus | bdbus_queue.pop(); };
            local_cdbus = 0; while( cdbus_queue.size() != 0 ) { local_cdbus = local_cdbus | cdbus_queue.pop(); };
            local_ddbus = 0; while( ddbus_queue.size() != 0 ) { local_ddbus = local_ddbus | ddbus_queue.pop(); };
            
            -- set stall signals --
            if( (bus_states.stall.current.pstall == 0 or bus_states.pbus.current.dismiss == 1) and !preq_queue.is_empty() )
                { local_pstall = 1; };
            if( bus_states.stall.current.rstall == 0 and (!breq_queue.is_empty() or !creq_queue.is_empty() or !dreq_queue.is_empty()) ) 
                { local_rstall = 1; };
            if( bus_states.stall.current.wstall == 0 and (!ereq_queue.is_empty() or !freq_queue.is_empty() ) ) 
                { local_wstall = 1; };
            
            -- reset stall signals --
            if( local_pready == 0 or 
                (bus_states.pbus.current.dismiss == 1 and bus_states.pbus.state.request != request) or
                (bus_states.pbus.current.dismiss == 1 and bus_states.pbus.state.request == request and !decoder_prog(config.id, bus_states.pbus.current.abus) ) )
                { local_pstall = 0; };
            if( local_bready == 0 or local_cready == 0 or local_dready == 0 ) 
                { local_rstall = 0; };
            if( local_eready == 0 or local_fready == 0 ) 
                { local_wstall = 0; };
            
            ---------------------
            -- driving signals --
            ---------------------
            bus_driver.pbus.ready       = local_pready;
            bus_driver.bbus.ready       = local_bready;
            bus_driver.cbus.ready       = local_cready;
            bus_driver.dbus.ready       = local_dready;
            bus_driver.ebus.ready       = local_eready;
            bus_driver.fbus.ready       = local_fready;
            
            bus_driver.pbus.dbus        = local_pdbus;
            bus_driver.bbus.dbus        = local_bdbus;
            bus_driver.cbus.dbus        = local_cdbus;
            bus_driver.dbus.dbus        = local_ddbus;
            
            bus_driver.stall.slv_pstall = local_pstall;
            bus_driver.stall.slv_rstall = local_rstall;
            bus_driver.stall.slv_wstall = local_wstall;
            
            bus_driver.drive_slave_signals( paths, config );
            
            wait cycle;

        };
    };
    
    init() is also {
    };
    
    -------------------------
    -- start TCM functions --
    -------------------------
    run() is also {
        start TCM_slave_driver();
    };
    
};

'>
